/*
VODKA-INDUCED ENTERTAINMENT ADVANCED SOUND MANAGER v0.9.1
COMPATIBILITY LAYER FOR WOLF4SDL v1.6
BY GERARD 'ALUMIUN' WATSON

This file replaces all external ID_SD functions and audio-based ID_CA
functions in Wolf4SDL v1.6 for use with the Advanced Sound Manager.

Please read vieasm.txt before using.

PLEASE ALSO NOTE THAT THIS IS STILL A BETA. IT IS LIKELY TO HAVE LOTS OF BUGS.
IF YOU FIND ANY, PLEASE TELL ME SO I CAN FIX THEM!
*/

#include "wl_def.h"
#include "asmcref.h"

extern bool  nosound;

SDMode          SoundMode;
SMMode          MusicMode;

sample *ASM_Audiosegs;

globalsoundpos channelSoundPos[ASM_ABSMAXCHANNELS];

// SD_MusIsOn
// Returns true if music is on, false otherwise

bool SD_MusIsOn(void)
{
    return (MusicMode == smm_AdLib) ? true : false;
}

bool SD_SndIsOn(void)
{
    return (SoundMode == sdm_AdLib) ? true : false;
}

void SD_ChannelDone(int channel)
{
    ASM_ChannelDone(channel);
    channelSoundPos[channel].valid = 0;
}

void SD_InitAudioSegs(void)
{
    int dataSize;

    if (ASM_Audiosegs != NULL)
    {
        return;
    }

    dataSize = sizeof(sample) * NUMSOUNDS;
	ASM_Audiosegs = (sample *)malloc(dataSize);
    CHECKMALLOCRESULT(ASM_Audiosegs);
    memset (ASM_Audiosegs, 0, dataSize);
}

void SD_Startup(void)
{
    if (nosound)
        return;

    if (!ASM_Open(param_samplerate, MIX_CHANNELS, ASM_ABSMAXCHANNELS/2, param_audiobuffer,
                128, 128, false))
    {        
        printf ("SD_Startup: Unable to open audio device. Trying defaults.\n"); 
        if (!ASM_Open(22050, 16, 128, 2048, 128, 128, false))
        {    
            printf ("SD_Startup: Unable to open audio device with defaults.\n"
                    "            Try closing any other applications that might\n"
                    "            be using the audio driver and try again, or\n"
                    "            start the program with the '--nosound' switch.\n");
            exit(1);
        }
    }
    
    Mix_ChannelFinished(SD_ChannelDone);

    SD_InitAudioSegs();
}

void SD_Shutdown(void)
{
    SD_InitAudioSegs();

    for (int i = 0; i < NUMSOUNDS; i++)
        if (ASM_Audiosegs[i].chunk != NULL)
            UNCACHEAUDIOCHUNK(i);

    ASM_Close();
}

bool SD_MusicPlaying(void)
{
    ASM_AbortIfClosed false;
    SD_AbortIfMusOff false;

    return Mix_PlayingMusic();
}

// This has changed slightly - it returns 0 on an error, 1 on a successful load

int32_t CA_CacheAudioChunk(int chunk)
{
#ifdef USE_SOUNDDIR
    ED_Sound_t *sound;
#endif

    if (chunk >= NUMSOUNDS)
    {
#ifdef VERBOSE_ASMCOMP
        printf("CA_CacheAudioChunk: Invalid sample number %d!", chunk);
#endif
        return 0;
    }

    if (ASM_Audiosegs[chunk].chunk != NULL)
    {
#ifdef VERBOSE_ASMCOMP
        printf("CA_CacheAudioChunk: Already loaded %d!", chunk);
#endif
        return 0;
    }

#ifdef USE_SOUNDDIR
    sound = ED_Sounds(ed) + chunk;
    char *file = (char *) malloc((size_t)(
        strlen(sounddir) + strlen(sound->soundFileName)
        ) + 1);
    strcpy (file, sounddir);
    strcat (file, sound->soundFileName);
#else
    char *file = (char *) malloc((size_t)(strlen(ASM_Soundnames[chunk]))+1);
    strcpy (file, ASM_Soundnames[chunk]);
#endif

    ASM_Audiosegs[chunk] = ASM_Cache(file, sound->soundFileName);

    free (file);
    return 1;
}

void CA_LoadAllSounds(void)
{
    SD_InitAudioSegs();

#ifndef LAZY_LOAD_SOUNDS
    for (int i = 0; i < NUMSOUNDS; i++)
    {
        if (ASM_Audiosegs[i].chunk != NULL)
            continue;
        CA_CacheAudioChunk(i);
    }
#endif
}

void SD_MusicOn(void)
{
    ASM_AbortIfClosed;

    SD_AbortIfMusOff;

    ASM_ResumeMusic();
}

int SD_MusicOff(void)
{
    ASM_AbortIfClosed 0;
    
    ASM_PauseMusic();
    
    return 0;
}

void SD_StartMusic (int chunk)
{
    ASM_AbortIfClosed;

    SD_AbortIfMusOff;

    if (chunk >= NUMMUSICS)
    {
#ifdef VERBOSE_ASMCOMP
        printf("SD_StartMusic: Invalid chunk number %d!", chunk);
#endif
        return;
    }

#ifdef USE_SOUNDDIR
    char *file = (char *) malloc((size_t)(strlen(musicdir)+strlen(ASM_Musicnames[chunk]))+1);
    strcpy (file, musicdir);
    strcat (file, ASM_Musicnames[chunk]);
#else
    char *file = (char *) malloc((size_t)(strlen(ASM_Musicnames[chunk]))+1);
    strcpy (file, ASM_Musicnames[chunk]);
#endif

    ASM_PlayMusic(file);
}

bool SD_SetMusicMode(SMMode mode)
{
    if (mode == MusicMode)
        return true;

    if (mode == smm_Off)
    {
        MusicMode = smm_Off;
        SD_MusicOff();
    }
    else
    {
        MusicMode = smm_AdLib;
        SD_MusicOn();
    }

    return true;
}

bool SD_SetSoundMode(SDMode mode)
{
    if (mode == SoundMode)
        return true;

    ASM_HaltSound();

    if (mode == sdm_Off)
        SoundMode = sdm_Off;
    else
        SoundMode = sdm_AdLib;
    return true;
}

void SD_StopSound(void)
{
    ASM_AbortIfClosed;

#ifdef VERBOSE_ASMCOMP
    printf("SD_StopSound: Stopping all sounds\n");
#endif
    ASM_HaltSound();
}

void SD_SetPosition(int channel, int leftpos, int rightpos, int distance)
{
    if((leftpos < 0) || (leftpos > 15) || (rightpos < 0) || (rightpos > 15) || ((leftpos == 15) && (rightpos == 15)))
    {
#ifdef VERBOSE_ASMCOMP
        printf("SD_SetPosition: Illegal position: leftpos - %d, rightpos - %d", leftpos, rightpos);
#endif
    }
    else
    {
        Mix_SetPanning(channel, ((15 - leftpos) << 4) + 15, ((15 - rightpos) << 4) + 15);
        Mix_SetDistance(channel, distance);
    }
}

int SD_PlaySound(int sound)
{
    ASM_SoundPlay_t soundPlay;

    lwlib_Zero(soundPlay);
    soundPlay.sound = sound;
    soundPlay.vol = Vol_Mono(SD_MaxVolume());

    return SD_PlaySoundEx(soundPlay);
}

int SD_PlaySoundEx(ASM_SoundPlay_t soundPlay)
{
    int sound;
    ED_Sound_t *edSound;

    if (!SD_SndIsOn())
    {
        return -1;
    }

    sound = soundPlay.sound;
    if (sound == -1)
    {
        return -1;
    }
    else if (sound >= NUMSOUNDS)
    {
#ifdef VERBOSE_ASMCOMP
        printf ("SD_PlaySound: Invalid sound number %d!\n", sound);
#endif
        return -1;
    }

    edSound = &ED_Sounds(ed)[sound];

#ifdef VERBOSE_ASMCOMP
    printf("Playing sound %d - id:%s file:%s\n", sound, 
        edSound->soundFileName, edSound->name);
#endif

    if (soundPlay.posOn)
    {
        SetSoundLoc(soundPlay.posGx, soundPlay.posGy,
            soundPlay.noDistAtt ? -1 : edSound->maxTileDist);
        soundPlay.lp = leftchannel;
        soundPlay.rp = rightchannel;
        soundPlay.dist = channeldist;
    }
    else
    {
        soundPlay.lp = soundPlay.rp = soundPlay.dist = 0;
    }

#ifndef LAZY_LOAD_SOUNDS
    if (ASM_Audiosegs[sound].chunk == NULL)
    {
#ifdef VERBOSE_ASMCOMP
        printf ("SD_PlaySound: Sound %d not cached!\n", sound);
#endif
        return -1;
    }
#endif

    int channel = SD_PlayDigitized(soundPlay);
    return channel;
}

int SD_PlaySoundVolume(int sound, int volume)
{
    ASM_SoundPlay_t soundPlay;
    lwlib_Zero(soundPlay);
    soundPlay.sound = sound;
    soundPlay.vol = Vol_Mono(volume);
    return SD_PlaySoundEx(soundPlay);
}

int SD_PlayDigitized(ASM_SoundPlay_t soundPlay)
{
    if (!SD_SndIsOn())
    {
        return 0;
    }

    return ASM_PlaySound(soundPlay);
}

bool SD_ChannelPlaying(int channel)
{
    if (nosound)
    {
        return false;
    }

    if (channel == -1)
    {
        return false;
    }

    return ASM_ChannelPlaying(channel);
}

int SD_MaxVolume(void)
{
    return MIX_MAX_VOLUME;
}

void SD_SetChannelVol(int channel, voltype vol)
{
    ASM_SetChannelVol(channel, vol);
}
