/*************************************************************************
** Wolf3D Legacy
** Copyright (C) 2012 by LinuxWolf - Team RayCAST
**
** This program is free software; you can redistribute it and/or
** modify it under the terms of the GNU General Public License
** as published by the Free Software Foundation; either version 2
** of the License, or (at your option) any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU General Public License for more details.
**************************************************************************
** LinuxWolf Library for Wolf3D Legacy
*************************************************************************/

#ifndef LWLIB_CAMERA_H
#define LWLIB_CAMERA_H

#include "lw_vec.h"
#include "lw_edit.h"
#include "lw_keys.h"
#include "lw_mouse.h"

namespace lwlib
{
	namespace Camera_ProjectionType
	{
		enum e
		{
			Orthographic,
			Perspective,
		};
	}

	typedef struct Camera_KeyToDir_s
	{
		const char *btnStr;
		Vector3f dir;
	} Camera_KeyToDir_t;

	typedef struct Camera_Priv_s
	{
		lwlib_Keys_t keys;
		Mouse_t mouse;
		Camera_ProjectionType::e projectionType;
		bool allowInput;
		Point3f pos;
		Point3f angles;
		double moveSpeed;
		double turnSpeedKeyboard;
		double turnSpeedMouse;
		Edit_Persist_t persist;
	} Camera_Priv_t;

	class Camera
	{
		struct
		{
			lwlib_Keys_t keys;
			Mouse_t mouse;
			Camera_ProjectionType::e projectionType;
			bool allowInput;
			Point3f pos;
			Point3f angles;
			double moveSpeed;
			double turnSpeedKeyboard;
			double turnSpeedMouse;
			Edit_Persist_t persist;
		} priv;

		void processMoveKey(Camera_KeyToDir_t *keyToDir);

	public:
		Camera(lwlib_Keys_t keys, Mouse_t mouse);

		void setProjectionType(Camera_ProjectionType::e projectionType);

		void setAllowInput(bool allowInput);

		void setPosition(Point3f pos);

		Point3f getPosition();

		void setAngles(Point3f angles);

		Point3f getAngles();

		void updateFrame();

		Vector3f getDirection();

		Mat3f rotationMat();

		void addToEdit(Edit *edit);
	};
}

#endif

