/*************************************************************************
** Wolf3D Legacy
** Copyright (C) 2012 by LinuxWolf - Team RayCAST
**
** This program is free software; you can redistribute it and/or
** modify it under the terms of the GNU General Public License
** as published by the Free Software Foundation; either version 2
** of the License, or (at your option) any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU General Public License for more details.
**************************************************************************
** LinuxWolf Library for Wolf3D Legacy
*************************************************************************/

#ifndef LWLIB_COLOR_H
#define LWLIB_COLOR_H

#include "fixedptc.h"

#ifdef SDL_TRUE_COLOR
typedef uint32_t lwlib_Color_Pix_t;
#else
typedef lwlib_Byte_t lwlib_Color_Pix_t;
#endif

#ifdef RGB_LSB_FIRST
#define LWLIB_COLOR_CPT(col, c) \
    (((col) & (0x000000FF << ((c) * 8))) >> ((c) * 8))
#define LWLIB_COLOR_PACK(r, g, b, a) \
    (((r) << 0) + ((g) << 8) + ((b) << 16) + ((a) << 24))
#define LWLIB_COLOR_PALCPT(pal, c) \
    (pal)[c]
#else
#define LWLIB_COLOR_CPT(col, c) \
    (((col) & (0x000000FF << ((2 - (c)) * 8))) >> ((2 - (c)) * 8))
#define LWLIB_COLOR_PACK(r, g, b, a) \
    (((r) << 16) + ((g) << 8) + ((b) << 0) + ((a) << 24))
#define LWLIB_COLOR_PALCPT(pal, c) \
    (pal)[2 - (c)]
#endif

#define LWLIB_COLOR_RED(col) LWLIB_COLOR_CPT(col, 0)
#define LWLIB_COLOR_GREEN(col) LWLIB_COLOR_CPT(col, 1)
#define LWLIB_COLOR_BLUE(col) LWLIB_COLOR_CPT(col, 2)
#define LWLIB_COLOR_ALPHA(col) (((col) & 0xFF000000) >> 24)

#define LWLIB_COLOR_PALRED(pal) LWLIB_COLOR_PALCPT(pal, 0)
#define LWLIB_COLOR_PALGREEN(pal) LWLIB_COLOR_PALCPT(pal, 1)
#define LWLIB_COLOR_PALBLUE(pal) LWLIB_COLOR_PALCPT(pal, 2)
#define LWLIB_COLOR_PALALPHA(pal) (pal)[3]

#define LWLIB_COLOR_RGB_EXPAND(col) LWLIB_COLOR_RED(col), LWLIB_COLOR_GREEN(col), LWLIB_COLOR_BLUE(col)

#define LWLIB_COLOR_TRANS(col) (((col) & 0xFF000000) == 0)

extern int *lwlib_Color_MulTable;

void lwlib_Color_Init(void);

#ifdef SDL_TRUE_COLOR
static inline void lwlib_Color_BlendColor(lwlib_Byte_t *outPix, 
	const int *blend)
{
	const int *table = lwlib_Color_MulTable;
	LWLIB_COLOR_PALRED(outPix) = 
		table[LWLIB_COLOR_PALRED(outPix) + LWLIB_COLOR_PALRED(blend)];
	LWLIB_COLOR_PALGREEN(outPix) = 
		table[LWLIB_COLOR_PALGREEN(outPix) + LWLIB_COLOR_PALGREEN(blend)];
	LWLIB_COLOR_PALBLUE(outPix) = 
		table[LWLIB_COLOR_PALBLUE(outPix) + LWLIB_COLOR_PALBLUE(blend)];
	LWLIB_COLOR_PALALPHA(outPix) = 
		table[LWLIB_COLOR_PALALPHA(outPix) + LWLIB_COLOR_PALALPHA(blend)];
}

static inline void lwlib_Color_BlendAlpha(const lwlib_Byte_t *src,
	lwlib_Byte_t *dst)
{
	const int *table = lwlib_Color_MulTable;
	const int a = LWLIB_COLOR_PALALPHA(src);
	const int as = (a << 8);
	const int at = ((255 - a) << 8);
	LWLIB_COLOR_PALRED(dst) = table[as + LWLIB_COLOR_PALRED(src)] + 
		table[at + LWLIB_COLOR_PALRED(dst)];
	LWLIB_COLOR_PALGREEN(dst) = table[as + LWLIB_COLOR_PALGREEN(src)] + 
		table[at + LWLIB_COLOR_PALGREEN(dst)];
	LWLIB_COLOR_PALBLUE(dst) = table[as + LWLIB_COLOR_PALBLUE(src)] + 
		table[at + LWLIB_COLOR_PALBLUE(dst)];
}
#endif

#endif
