/*************************************************************************
** Wolf3D Legacy
** Copyright (C) 2012 by LinuxWolf - Team RayCAST
**
** This program is free software; you can redistribute it and/or
** modify it under the terms of the GNU General Public License
** as published by the Free Software Foundation; either version 2
** of the License, or (at your option) any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU General Public License for more details.
**************************************************************************
** LinuxWolf Library for Wolf3D Legacy
*************************************************************************/

#include "lw_colorremap.h"
using namespace lwlib;

static const lwlib_Byte_t lwlib_ColorRemap_Start[ColorRemap_Color::Last + 1] =
{
	0x60, // Green
	0x10, // Grey
	0x20, // Red
	0x40, // Yellow
	0x70, // LightBlue
	0x90, // DarkBlue
	0xB0, // Violet
	0xD0, // Brown
};

static const Point4b lwlib_ColorRemap_ColorRgba[
	ColorRemap_Color::Last + 1] =
{
	{ { 0x00, 0xFF, 0x00, 0xFF } }, // Green
	{ { 0xFF, 0xFF, 0xFF, 0xFF } }, // Grey
	{ { 0xFF, 0x00, 0x00, 0xFF } }, // Red
	{ { 0xFF, 0xFF, 0x00, 0xFF } }, // Yellow
	{ { 0x00, 0xFF, 0xFF, 0xFF } }, // LightBlue
	{ { 0x00, 0x00, 0xFF, 0xFF } }, // DarkBlue
	{ { 0xFF, 0x00, 0xFF, 0xFF } }, // Violet
	{ { 0x7F, 0x3F, 0x00, 0xFF } }, // Brown
};

void ColorRemap::initTables()
{
	int i, j;
	int key;

	for (i = 0; i < 256; i++)
	{
		priv.tables[ColorRemap_Color::Default][i] = (lwlib_Byte_t)i;
	}

	key = lwlib_ColorRemap_Start[ColorRemap_Color::Default];

	for (i = 0; i < (int)ColorRemap_Color::Last + 1; i++)
	{
		memcpy(priv.tables[i], 
			priv.tables[ColorRemap_Color::Default], 
			sizeof(ColorRemap_Table_t));
		for (j = 0; j < 16; j++)
		{
			priv.tables[i][j + key] = lwlib_ColorRemap_Start[i] + j;
		}
	}
}

ColorRemap::ColorRemap()
{
	lwlib_Zero(priv);

	priv.activeTable = priv.tables[ColorRemap_Color::Default];
	initTables();
}

void ColorRemap::setActiveColor(ColorRemap_Color::e color)
{
	if (color < 0 || color > ColorRemap_Color::Last)
	{
		return;
	}
	priv.activeTable = priv.tables[color];
}

lwlib_Byte_t *ColorRemap::getActiveTable()
{
	return priv.activeTable;
}

lwlib_Byte_t lwlib::ColorRemap_GetStart(ColorRemap_Color::e color)
{
	return lwlib_ColorRemap_Start[color];
}

Point4b lwlib::ColorRemap_GetColorRgba(ColorRemap_Color::e color)
{
	return lwlib_ColorRemap_ColorRgba[color];
}
