/*************************************************************************
** Wolf3D Legacy
** Copyright (C) 2012 by LinuxWolf - Team RayCAST
**
** This program is free software; you can redistribute it and/or
** modify it under the terms of the GNU General Public License
** as published by the Free Software Foundation; either version 2
** of the License, or (at your option) any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU General Public License for more details.
**************************************************************************
** LinuxWolf Library for Wolf3D Legacy
*************************************************************************/

#ifndef LWLIB_CSG_H
#define LWLIB_CSG_H

#include "lw_vec.h"
#include "lw_misc.h"
#include "lw_edit.h"

#define lwlib_CsgValid(csg) ((csg) != NULL)

namespace lwlib
{
	namespace Csg_NodeType
	{
		enum e
		{
			LeafEmpty,
			ParentUnion,
			ParentIntersection,
			ParentNegate,
			LeafHalfSpace,
			LeafSphere,
			Last = LeafSphere,
		};
	}

	typedef struct Csg_Node_s
	{
		Csg_NodeType::e type;
		Csg_NodeType::e editType;
		Plane3f plane;
		Sphere3f sphere;
		struct Csg_Node_s *childNodes[2];
	} Csg_Node_t;

	class Csg
	{
		struct
		{
			Csg_Node_t *topNode;
			Edit_Persist_t persist;
		} priv;

		explicit Csg(Csg_Node_t *topNode);

	public:
		Csg();

		~Csg();

		Csg *copy();

		static Csg *intersect(Csg *a, Csg *b);

		static Csg *csgUnion(Csg *a, Csg *b);

		Csg *negate();

		static Csg *halfSpace(Plane3f plane);

		static Csg *sphere(Sphere3f sphere);

		void rotate(Mat3f rotMat);

		void translate(Vector3f xlateVec);

		bool testPoint(Point3f point);

		void addToEdit(Edit *edit);

		void updateFrame();

		bool testBox(Box3f box, OverlapTest::e test);
	};
}

#endif

