/*************************************************************************
** Wolf3D Legacy
** Copyright (C) 2012 by LinuxWolf - Team RayCAST
**
** This program is free software; you can redistribute it and/or
** modify it under the terms of the GNU General Public License
** as published by the Free Software Foundation; either version 2
** of the License, or (at your option) any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU General Public License for more details.
**************************************************************************
** LinuxWolf Library for Wolf3D Legacy
*************************************************************************/

#ifndef LWLIB_FONT_H
#define LWLIB_FONT_H

#include <SDL.h>
#include "lw_misc.h"
#include "lw_colorremap.h"
#include "lw_vec.h"
#include "lw_gl.h"
#include "lw_img.h"

namespace lwlib
{
	namespace Font_Color
	{
		enum e
		{
			Green = ColorRemap_Color::Green,
			Grey = ColorRemap_Color::Grey,
			Red = ColorRemap_Color::Red,
			Yellow = ColorRemap_Color::Yellow,
			LightBlue = ColorRemap_Color::LightBlue,
			DarkBlue = ColorRemap_Color::DarkBlue,
			Violet = ColorRemap_Color::Violet,
			Brown = ColorRemap_Color::Brown,
		};
	}

	class Font
	{
		struct
		{
			int width;
			int height;
			int shadow;
			Font_Color::e color;
			Img_t fontImg;
			ColorRemap *colorRemap;
			Gl *gl;
			int indent;
			Img_t charImg;
		} priv;

		void printChar(int c, int cx, int cy, int sw, int sh, 
			Font_Color::e curColor);

		Point2i printPosFitNoNl(Point2i pos, const char *text);

	public:
		Font(const char *fontFile, SDL_Color *palette);

		~Font();

		void setShadow(int shadow);

		void setColor(Font_Color::e color);

		void print(int x, int y, const char *text);

		void printPos(int xl, int yl, const char *text);

		void printPosOff(int xl, int yl, int ox, int oy, const char *text);

		Point2i printPosFit(Point2i pos, const char *text);

		Point2i advancePosFit(Point2i pos, int col);

		void setColorRemap(ColorRemap *colorRemap);

		void setGl(Gl *gl);

		int getIndent();

		void setIndent(int indent);

		int getWidth();

		int getHeight();

		int maxScreenPosY();
	};
}

#endif
