/*************************************************************************
** Wolf3D Legacy
** Copyright (C) 2012 by LinuxWolf - Team RayCAST
**
** This program is free software; you can redistribute it and/or
** modify it under the terms of the GNU General Public License
** as published by the Free Software Foundation; either version 2
** of the License, or (at your option) any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU General Public License for more details.
**************************************************************************
** LinuxWolf Library for Wolf3D Legacy
*************************************************************************/

#ifndef LWLIB_IMG_H
#define LWLIB_IMG_H

#include <SDL.h>
#include "lw_misc.h"
#include "fixedptc.h"
#include "vec.h"

#ifndef LWLIB_GL_PTRTYPE
#define LWLIB_GL_PTRTYPE
namespace lwlib
{
	class Gl;
	typedef Gl *Gl_t;
}
#endif

#define LWLIB_IMG_MIPMAPLEVEL_MAX 16

namespace lwlib
{
	typedef struct Img_MipMapLevel_s
	{
		int w, h;
		int wshift, hshift;
		int pitch;
		lwlib_Byte_t *source;
	} Img_MipMapLevel_t;

	typedef struct Img_MipMap_s
	{
		bool built;
		int texsize;
		int texshift;
		int curLevel;
		Img_MipMapLevel_t levels[LWLIB_IMG_MIPMAPLEVEL_MAX];
		int numLevels;
	} Img_MipMap_t;

	typedef struct Img_Sub_s
	{
		bool built;
		int x, y;
		int w, h;
		int wshift;
		int hshift;
		lwlib_Byte_t *data;
	} Img_Sub_t;

	class Img;
	typedef Img *Img_t;
	typedef vec_t(Img_t) ImgVec_t;

	class Img
	{
		struct
		{
			int w;
			int h;
			int wshift;
			int hshift;
			int bpp;
			int pitch;
			lwlib_Byte_t *data;
			Img_MipMap_t mipmap;
			Img_Sub_t sub;
			bool dataOwned;
			int texId;
			SDL_Color *palette;
		} priv;

		void initDefaultAttr();

	public:
		bool readTGA(char *name);

		bool readPCX(char *name);

		static Img_t load(const char *path);

		static lwlib_Byte_t findClosestColor(SDL_Color *palette, int r, 
			int g, int b);

		static void resample(uint8_t *dst, int outwidth, int outheight,
			uint8_t *data, int width, int height, int bpp);

		static void resamplePaletted(lwlib_Byte_t *dst, int outwidth, 
			int outheight, lwlib_Byte_t *data, int width, int height);

		static void reBpp(uint8_t *dst, int dstbpp, uint8_t *src, 
			int width, int height, int bpp);

		static bool mipMap(uint8_t *src, int *width, int *height, 
			int bpp);

		Img();

		~Img();

		void fromSource(lwlib_Byte_t *source, int w, int h, 
			int bpp, int pitch, bool owned, SDL_Color *palette);

		void palettize(SDL_Color *palette, bool transparent, 
			lwlib_Byte_t transkey);

		int getWidth();

		int getHeight();

		int getWidthShift();

		int getHeightShift();

		int getBpp();

		int getPitch();

		lwlib_Byte_t *getData();

		void applyMonotone(lwlib_Byte_t monosrckey, lwlib_Byte_t monodestkey);

		void buildMipMaps(SDL_Color *palette);

		lwlib_Byte_t getPix(fixedpt x, fixedpt y);

		void setMipLevel(int curLevel);

		void transpose();

		void drawScaled(Gl *gl, int cx, int cy, int sw, int sh,
			bool transparent);

		void subImage(Img_t subImg, int ix, int iy, int iw, int ih);

		void expandPalette(bool transparent);

		void refreshTex(Gl *gl);

		int texId();

		void scaleIntensity(lwlib_Byte_t blend);

		void copyData();

		void expandDimsPow2();

		void expandBits(int bpp);

		void swapRGB();
	};
}

#endif
