/*************************************************************************
** Wolf3D Legacy
** Copyright (C) 2012 by LinuxWolf - Team RayCAST
**
** This program is free software; you can redistribute it and/or
** modify it under the terms of the GNU General Public License
** as published by the Free Software Foundation; either version 2
** of the License, or (at your option) any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU General Public License for more details.
**************************************************************************
** LinuxWolf Library for Wolf3D Legacy
*************************************************************************/

#ifndef LWLIB_KEYS_H
#define LWLIB_KEYS_H

#include <SDL/SDL_keysym.h>
#include <SDL/SDL_events.h>
#include "lw_misc.h"
#include "vec.h"

#define lwlib_Keys_Event(name) KEYS_EVENT_ ## name

#define lwlib_KeyDown(key) lwlib_Keys_CheckEvent(priv.keys, key, lwlib_Keys_Event(Down))
#define lwlib_KeyPressed(key) lwlib_Keys_CheckEvent(priv.keys, key, lwlib_Keys_Event(Pressed))
#define lwlib_KeyReleased(key) lwlib_Keys_CheckEvent(priv.keys, key, lwlib_Keys_Event(Released))
#define lwlib_KeyRepeat(key) lwlib_Keys_CheckEvent(priv.keys, key, lwlib_Keys_Event(Repeat))
#define lwlib_KeyFastRepeat(key) lwlib_Keys_CheckEvent(priv.keys, key, lwlib_Keys_Event(FastRepeat))

#define LWLIB_SDLK_MOUSE1 ((SDLKey)(SDLK_LAST + 0))
#define LWLIB_SDLK_MOUSE3 ((SDLKey)(SDLK_LAST + 1))
#define LWLIB_SDLK_MOUSE2 ((SDLKey)(SDLK_LAST + 2))
#define LWLIB_SDLK_MWHEELUP ((SDLKey)(SDLK_LAST + 3))
#define LWLIB_SDLK_MWHEELDOWN ((SDLKey)(SDLK_LAST + 4))
#define LWLIB_SDLK_MOUSE4 ((SDLKey)(SDLK_LAST + 5))
#define LWLIB_SDLK_MOUSE5 ((SDLKey)(SDLK_LAST + 6))

typedef struct lwlib_Keys_s
{
	void *priv;
} lwlib_Keys_t;

typedef enum lwlib_Keys_Event_e
{
	lwlib_Keys_Event(Down),
	lwlib_Keys_Event(Pressed),
	lwlib_Keys_Event(Released),
	lwlib_Keys_Event(Repeat),
	lwlib_Keys_Event(FastRepeat),
	lwlib_Keys_Event(Last) = lwlib_Keys_Event(FastRepeat),
} lwlib_Keys_Event_t;

lwlib_EnumStrExtern(lwlib_Keys_Event);

// button string can be:
// 1) ctrl, shift, alt
// 2) [a-z`0-9-=[]\;',./]
// 3) shift-[a-z`0-9-=[]\;',./]
// 4) ctrl-[a-z`0-9-=[]\;',./]
// 5) alt-[a-z`0-9-=[]\;',./]
// 6) raw-[a-z`0-9-=[]\;',./]
// 7) f1, f2, f3, f4, f5, f6, f7, f8, f9, f10, f11, f12
// 8) l-ctrl, r-ctrl, l-shift, r-shift, l-alt, r-alt
// 9) enter, space, bksp, del, ins, 
// 10) pgup, pgdn, home, end, tab, esc
// 11) mouse1, mouse2, mouse3, mouse4, mouse5, mwheelup, mwheeldown
// 12) left-arrow, right-arrow, up-arrow, down-arrow

lwlib_Keys_t lwlib_Keys_New(void);

lwlib_Keys_t lwlib_Keys_Free(lwlib_Keys_t keys);

void lwlib_Keys_ResetTransientEvents(lwlib_Keys_t keys);

void lwlib_Keys_ProcessEvent(lwlib_Keys_t keys, SDLKey sdlKeySym, bool isDown);

bool lwlib_Keys_CheckEvent(lwlib_Keys_t keys, const char *btnStr, lwlib_Keys_Event_t event);

bool lwlib_Keys_CheckDown(lwlib_Keys_t keys, const char *btnStr);

void lwlib_Keys_SetIgnore(lwlib_Keys_t keys, bool ignore);

void lwlib_Keys_EventQueueReset(lwlib_Keys_t keys, lwlib_StrSeq_t btnStrs, lwlib_Keys_Event_t event);

int lwlib_Keys_EventQueuePeek(lwlib_Keys_t keys);

bool lwlib_Keys_EventQueuePending(lwlib_Keys_t keys);

char *lwlib_Keys_EventQueueNext(lwlib_Keys_t keys);

char lwlib_Keys_EventQueueNextAscii(lwlib_Keys_t keys);

lwlib_StrSeq_t lwlib_Keys_NumericBtnStrs(lwlib_Keys_t keys);

lwlib_StrSeq_t lwlib_Keys_AsciiBtnStrs(lwlib_Keys_t keys);

int lwlib_Keys_FirstNumericEvent(lwlib_Keys_t keys, lwlib_Keys_Event_t event);

void lwlib_Keys_KeyboardFocusReturned(lwlib_Keys_t keys);

#endif
