/*************************************************************************
** Wolf3D Legacy
** Copyright (C) 2012 by LinuxWolf - Team RayCAST
**
** This program is free software; you can redistribute it and/or
** modify it under the terms of the GNU General Public License
** as published by the Free Software Foundation; either version 2
** of the License, or (at your option) any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU General Public License for more details.
**************************************************************************
** LinuxWolf Library for Wolf3D Legacy
*************************************************************************/

#ifndef LWLIB_MAPTOOL_H
#define LWLIB_MAPTOOL_H

#include "lw_edit.h"
#include "lw_tool.h"
#include "lw_tileviewer.h"
#include "lw_mouse.h"
#include "lw_keys.h"
#include "lw_picker.h"

namespace lwlib
{
	namespace MapTool_FileFormat
	{
		enum e
		{
			Game,
			Raw,
		};
	}

	namespace MapTool_YankMode
	{
		enum e
		{
			Wall,
			Actor,
			FloorCeil,
			WallPatch,
			Picker,
			TileBox,
			Max,
		};
	}

	namespace MapTool_Const
	{
		enum e
		{
			// these must match wl_def.h
			MaxTitleChars = 16, // MAXTITLECHARS
			MapPlanes = 3, // MAPPLANES
			MapArea = 64 * 64, // maparea
			NumMaps = 60, // NUMMAPS
		};
	}

	typedef uint16_t MapTool_Word_t;
	typedef MapTool_Word_t MapTool_Planes_t[MapTool_Const::MapPlanes]
		[MapTool_Const::MapArea];

	typedef struct MapTool_Map_s
	{
		MapTool_Word_t width;
		MapTool_Word_t height;
		char name[MapTool_Const::MaxTitleChars];
		MapTool_Planes_t planes;
		int secretLevelIndex;
	} MapTool_Map_t;

	typedef struct MapTool_CommandSet_s
	{
		Command *yankTile;
		Command *pasteTile;
		Command *zoomIn;
		Command *zoomOut;
		Command *toggleCeilingMode;
		Command *pushWall;
		Command *ambush;
		Command *toggleGrid;
		Command *saveMaps;
		Command *loadMaps;
		Command *picker;
		Command *removeTile;
		Command *nextDir;
		Command *clearMap;
		Command *cutTile;
		Command *tileBox;
		Command *nextMap;
		Command *prevMap;
		Command *moveTileLeft;
		Command *moveTileRight;
		Command *moveTileUp;
		Command *moveTileDown;
		Command *moveViewLeft;
		Command *moveViewRight;
		Command *moveViewUp;
		Command *moveViewDown;
		Command *shiftTileLeft;
		Command *shiftTileRight;
		Command *shiftTileUp;
		Command *shiftTileDown;
		Command *setFogColor;
		Command *panView;
		CommandPtrVec_t all;
	} MapTool_CommandSet_t;

	typedef struct lwlib_MapTool_Yank_s
	{
		MapTool_YankMode::e mode;
		int wall;
		int actor;
		int wallPatch;
		int floor;
		int ceiling;
		bool tileBoxForming;
		Box2i tileBox;
		MapTool_Planes_t planes;
	} lwlib_MapTool_Yank_t;

	typedef struct lwlib_MapTool_DrawCache_s
	{
		MapTool_Planes_t planes;
	} lwlib_MapTool_DrawCache_t;

	typedef struct lwlib_MapTool_TileReplace_s
	{
		int sourceTileFirst;
		int sourceTileLast;
		int destTile;
		int plane;
		bool go; 
	} lwlib_MapTool_TileReplace_t;

	class MapTool : public Tool
	{
		struct
		{
			int mapIndex;
			MapTool_Map_t maps[MapTool_Const::NumMaps];
			bool loadMaps;
			bool saveMaps;
			MapTool_FileFormat::e fileFormat;
			lwlib_String_t rawFileName;
			Gl *gl;
			TileViewer *tileViewer;
			Point2i tilePos;
			Point2f tileMove;
			double tileMoveSensitivity;
			Point3f pan;
			Mouse_t mouse;
			lwlib_Keys_t keys;
			Edit *edit;
			Font *font;
			Picker *picker;
			int grid;
			lwlib_MapTool_Yank_t yank;
			double tileCursorSize;
			bool ceilingMode;
			int planeToManip;
			bool fillArea;
			int areaCode;
			bool areaTileVisit[MapTool_Const::MapArea];
			MapTool_CommandSet_t commandSet;
			bool pickerPending;
			lwlib_MapTool_DrawCache_t drawCache;
			lwlib_MapTool_TileReplace_t tileReplace;
			bool showGrid;
		} priv;

		void map_UpdateEdit(Edit *edit);

		int map_ComputeSecretLevel(MapTool_Map_t *map);

		void loadMaps_FileFormatGame();

		void loadMaps_FileFormatRaw();

		void loadMaps();

		int saveMapHead();

		lwlib_Byte_t *mapPlaneOnDisk(int mapNum, int planeNum,
			int *onDiskSize);

		int mapPlaneSizeOnDisk(int mapNum, int planeNum);

		void saveMapHeaders(int handle, int off);

		void saveMapPlanes(int handle);

		void saveMaps_FileFormatGame();

		void saveMaps_FileFormatRaw();

		void saveMaps();

		void tilePosMotion();

		void panMotion();

		void setAreaTile(Point2i pos);

		void removeSolidWall();

		void setYankMode(MapTool_YankMode::e yankMode);

		void yankTile();

		void insertPicked();

		void pasteTileBox();

		void pasteTile();

		void removeTile();

		void tileManip();

		void drawTileInfo();

		void drawTileCursor();

		void drawTileBox();

		void drawTiles();

		void commandSet_AddToEdit(Edit *edit);

		void tileReplace_AddToEdit(Edit *edit);

		void shiftTile(int x, int y);

		void checkTileReplace();

	public:
		MapTool(
			Gl *gl,
			TileViewer *tileViewer,
			Mouse_t mouse,
			lwlib_Keys_t keys,
			Edit *edit,
			Font *font,
			Picker *picker
			);

		virtual ~MapTool();

		virtual void addToEdit(Edit *edit);

		virtual void updateEdit(Edit *edit);

		virtual void updateFrame();

		virtual void draw(Gl *gl);

		bool pickerPending();

		int getPickerDoorTile(int lock);
	};
}

#endif
