/*************************************************************************
** Wolf3D Legacy
** Copyright (C) 2012 by LinuxWolf - Team RayCAST
**
** This program is free software; you can redistribute it and/or
** modify it under the terms of the GNU General Public License
** as published by the Free Software Foundation; either version 2
** of the License, or (at your option) any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU General Public License for more details.
**************************************************************************
** LinuxWolf Library for Wolf3D Legacy
*************************************************************************/

#ifndef LWLIB_SDLGLEMU_H
#define LWLIB_SDLGLEMU_H

#include <SDL.h>
#include "lw_misc.h"
#include "lw_vec.h"
#include "lw_img.h"
#include "lw_color.h"
#include "lw_raster.h"

#define LWLIB_SDLGLEMU_MATSTACK_MAX 10

namespace lwlib
{
	typedef struct SDLGlEmu_Screen_s
	{
		int width;
		int height;
		int bits;
		bool fullscreen;
		bool useDoubleBuffering;
		SDL_Surface *screenSurf;
		bool isDoubleBuffered;
	#ifndef SDL_TRUE_COLOR
		SDL_Color *palette;
	#endif
	} SDLGlEmu_Screen_t;

	typedef struct SDLGlEmu_Texture_s
	{
		Img_t img;
		bool freed;
	} SDLGlEmu_Texture_t;

	typedef vec_t(SDLGlEmu_Texture_t) SDLGlEmu_TextureVec_t;

	class SDLGlEmu
	{
		struct
		{
			SDLGlEmu_TextureVec_t textureVec;
			SDL_Surface *screenSurf;
#ifndef SDL_TRUE_COLOR
			SDL_Surface *screenBuffer;
			int bufferPitch;
#endif
			lwlib_Color_Pix_t *vbuf;
			fixedpt *depthBuffer;
			int screenWidth, screenHeight, screenPitch;
			bool isDoubleBuffered;
			Mat4f curMat;
			Mat4f matStack[LWLIB_SDLGLEMU_MATSTACK_MAX];
			int matStackSize;
			lwlib_Color_Pix_t curColor;
			int blankTexId;
#ifndef SDL_TRUE_COLOR
			SDL_Color *palette;
#endif
		} priv;

#if 0
		lwlib_Raster_Vertex_t rasterVertex(double x, double y, double z, 
			double u, double v);
#endif

		lwlib_Raster_Vertex_t rasterVertexFast2D(double x, double y, 
			double u, double v);

	public:
		SDLGlEmu();

		~SDLGlEmu();

		void setVideoMode(SDLGlEmu_Screen_t *screen);

		void init();

		void clearScreen();

		void updateScreen();

		int uploadTex(lwlib_Byte_t *data, int w, int h, int bpp, 
			SDL_Color *palette);

		void deleteTex(int texId);

		void lockScreen();

		void unlockScreen();

		void drawScaledImage(int texId, int ix, int iy, int iw, int ih, 
			int w, int h, int cx, int cy, int sw, int sh, 
			bool transparent);

		void setCurColor(Point4b colorRgba);

		void resetCurColor();

		void drawBox(int x, int y, int w, int h, Point4b colorRgba, 
			bool lines);

		void drawBoxArrow(int x, int y, int w, int h, int r, 
			Point4b colorRgba, double dir);

		void pushMatrix();

		void popMatrix();

		void translateMatrix(double tx, double ty, double tz);

		void drawLine(int x1, int y1, int x2, int y2, Point4b colorRgba);

		int getWidth();

		int getHeight();

		Point2i getDims();
	};

#ifndef LWLIB_SDLGLEMU_PTRTYPE
#define LWLIB_SDLGLEMU_PTRTYPE
	typedef SDLGlEmu *SDLGlEmu_t;
#endif
}

#endif
