/*************************************************************************
** Wolf3D Legacy
** Copyright (C) 2012 by LinuxWolf - Team RayCAST
**
** This program is free software; you can redistribute it and/or
** modify it under the terms of the GNU General Public License
** as published by the Free Software Foundation; either version 2
** of the License, or (at your option) any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU General Public License for more details.
**************************************************************************
** LinuxWolf Library for Wolf3D Legacy
*************************************************************************/

#ifndef LWLIB_TILEVIEWER_H
#define LWLIB_TILEVIEWER_H

#include "lw_img.h"
#include "lw_font.h"
#include "lw_keys.h"
#include "lw_edit.h"
#include "lw_tool.h"
#include "wl_ed.h"

namespace lwlib
{
	typedef struct TileViewer_StatInfoMaker_s
	{
		bool doInsert;
		lwlib_String_t name;
		lwlib_String_t picnum;
		ED_Stat_t type;
		ED_BonusDef_t bonus;
		ED_ObjFlag_t specialFlags;
		int animframes;
	} TileViewer_StatInfoMaker_t;

	typedef struct TileViewer_StatInfoReport_s
	{
		bool generateReport;
	} TileViewer_StatInfoReport_t;

	class TileViewer : public Tool
	{
		struct
		{
			lwlib_Keys_t keys;
			Gl *gl;
			Font *font;
			SDL_Color *palette;
			Edit *edit;
			int tileSelector;
			int tilesAcross;
			int tilesDown;
			bool toSpriteStart;
			bool toWallStart;
			ImgVec_t tileImgs;
			lwlib_IntMap_t tileImgMap;
			TileViewer_StatInfoMaker_t statInfoMaker;
			TileViewer_StatInfoReport_t statInfoReport;
		} priv;

		ED_Sprite_t *selectedSprite();

		void drawTileInfo();
	
	public:
		TileViewer();

		~TileViewer();

		void addToEdit(Edit *edit);

		void updateEdit(Edit *edit);

		void updateFrame();

		void draw(Gl *gl);

		void setGl(Gl *gl);

		void setKeys(lwlib_Keys_t keys);

		void setFont(Font *font);

		void setPalette(SDL_Color *palette);

		void setEdit(Edit *edit);

		Img_t getTileImg(int tile);

		void drawTile(int tile, int cx, int cy, int sw, int sh);
	};
}

#endif
