/*
VODKA-INDUCED ENTERTAINMENT ADVANCED SOUND MANAGER v0.9.1
BY GERARD 'ALUMIUN' WATSON

Provides high quality sound and music using SDL_MIXER
Built on SDL_MIXER 1.2.5
Supports all formats supported by SDL_mixer.
For more details, please see the SDL_mixer manual.

Changes
-------

v0.9.1  - Fixed the reverse stereo channel bug
        - Made ASM_ChangeVolume a lot faster and clipped the volume levels
        - Added VERBOSE_VIEASM defines

v0.9    - Initial release

PLEASE ALSO NOTE THAT THIS IS STILL A BETA. IT IS LIKELY TO HAVE LOTS OF BUGS.
IF YOU FIND ANY, PLEASE TELL ME SO I CAN FIX THEM!
*/

// Initialization stuff

#include "vieasm.h"                    // Duh.
#include <SDL.h>
#include "wl_def.h"

const char *ASM_Verstring = "v0.9.1 Beta";  // Version string

Uint8 sndvol, musvol;                       // Volumes for sound
int origchannels, maxchannels, lastchan;    // Channel variables
activechan_t activechans[ASM_ABSMAXCHANNELS];          // Is channel used?
static Mix_Music *music = 0, *switchto = 0; // Music references
int fadetime;                               // Milliseconds to fade in\out
bool deviceopen = false;                    // Is device open?
bool switching = false;                     // Is switching music tracks?
bool reversemode = false;                   // Reverse stereomode

// ASM_IsOpen
// Returns true if the device is open, false otherwise

bool ASM_IsOpen(void)
{
    return (deviceopen) ? true : false;
}

void ASM_AdjustChannels(int reqchan)
{
    if (reqchan <= maxchannels)
    {
        int i = (reqchan > origchannels) ? reqchan : origchannels;
        Mix_AllocateChannels(i + 1);
        lastchan = i;
    }
}

// ASM_GetFreeChannel
// Returns the first free channel, calls ASM_AdjustChannels if another is needed

int ASM_GetFreeChannel(void)
{
    for (int i = 0; i <= maxchannels - 1; i++)
    {
        if (activechans[i].used == false)
        {
            activechans[i].used = true;
            if (i >= lastchan)
                ASM_AdjustChannels(i);
            return i;
        }
    }
    return -1;
}

// ASM_ChannelDone
// Callback for Mix_ChannelFinished
// Turns off channel and adjust if necessary

void ASM_ChannelDone(int channel)
{
    activechans[channel].used = false;

    if (channel == (lastchan - 1) && (lastchan - 1) > origchannels - 1)
        ASM_AdjustChannels(lastchan);
}

// ASM_Open
// Opens audio device at given specs, clears used

bool ASM_Open(int frequency, int channels, int maxchan, int buffersize, Uint8 sndvolume, Uint8 musvolume, bool reverse)
{
    if (ASM_IsOpen())       // Device is already open!
        return false;
#ifdef VERBOSE_VIEASM
    SDL_version compile_version;

    printf("----------\n"
           "Vodka-Induced Entertainment Advanced Sound Manager %s\n"
           "Developed for WolfSDL v1.6\n"
           "By Gerard 'AlumiuN' Watson\n"
           "\n"
           , ASM_Verstring);

    MIX_VERSION(&compile_version);
    printf("Compiled with SDL_mixer version: %d.%d.%d\n", 
            compile_version.major, compile_version.minor, compile_version.patch);

    const SDL_version *link_version=Mix_Linked_Version();
    printf("Running with SDL_mixer version: %d.%d.%d\n"
           "\n" 
           , link_version->major, link_version->minor, link_version->patch);

    printf("Opened with:\n" 
           "      frequency - %d Hz\n"
           "       channels - %d\n"
           "   max channels - %d\n"
           "     buffersize - %d bytes\n"
           "   sound volume - %d\n"
           "   music volume - %d\n"
           " reverse stereo - %s\n"
           "----------\n"
           , frequency, channels, maxchan, buffersize, sndvolume, musvolume, (reverse) ? "true" : "false");
#endif
    if(Mix_OpenAudio(frequency, AUDIO_S16, 2, buffersize) == -1) 
    {
#ifdef VERBOSE_VIEASM
        printf("ASM_Open: %s\n", Mix_GetError());
#endif
        return false;
    }

    if ((maxchan > ASM_ABSMAXCHANNELS) || (channels > maxchan))
    {
#ifdef VERBOSE_VIEASM
        printf("ASM_Open: Invalid maxchan value!\n");
#endif
        return false;
    }

    Mix_AllocateChannels(channels);

    memset(activechans, 0, sizeof(activechans));

    lastchan = origchannels = channels;
    maxchannels = maxchan;

    if (reverse)
        reversemode = true;

    deviceopen = true;

    ASM_ChangeVolume(sndvolume, musvolume);
    Mix_ChannelFinished(ASM_ChannelDone);
    return true;
}

// ASM_Close
// Closes the audio device

void ASM_Close(void)
{
#ifdef VERBOSE_VIEASM
    printf("ASM_Close: Audio device closed.\n");
#endif
    Mix_CloseAudio();
    deviceopen = false;
}

// ASM_HaltSound
// Ends all playing sounds, clears all channels, frees all chunks, etc.

void ASM_HaltSound(void)
{
    ASM_AbortIfClosed;

    Mix_HaltChannel(-1);
    memset(activechans, 0, sizeof(activechans));
    ASM_AdjustChannels(0);
}

// ASM_Pause
// Pauses all sound

void ASM_Pause(void)
{
    ASM_AbortIfClosed;

    ASM_PauseSound();
    ASM_PauseMusic();
}

// ASM_Resume
// Resumes all sound

void ASM_Resume(void)
{
    ASM_AbortIfClosed;

    ASM_ResumeSound();
    ASM_ResumeMusic();
}

// ASM_Halt
// Stops all sound, frees all chunks, etc.

void ASM_Halt(void)
{
    ASM_AbortIfClosed;

    ASM_HaltSound();
    ASM_HaltMusic();
}

// ASM_PlayMusic
// Plays a music sample from a file

bool ASM_PlayMusic(char *musfile)
{
    ASM_AbortIfClosed false;

    if (Mix_PlayingMusic())
    {
        Mix_HaltMusic();
        Mix_FreeMusic(music);
    }

    music = Mix_LoadMUS(musfile);

    if (!music)
    {
#ifdef VERBOSE_VIEASM
        printf("ASM_PlayMusic: %s\n", Mix_GetError());
#endif
        return false;
    }

    Mix_VolumeMusic(musvol);

    if (Mix_PlayMusic (music, -1) == -1)
    {
#ifdef VERBOSE_VIEASM
        printf("ASM_PlayMusic: %s\n", Mix_GetError());
#endif
        return false;
    }

    return true;
}

// ASM_ChangeVolume
// Changes the volume of:
//      Sound - iterates through each channel and sets volume accordingly
//      Music - just calls Mix_VolumeMusic

void ASM_ChangeVolume(Uint8 sndvolume, Uint8 musvolume)
{
    ASM_AbortIfClosed;

    sndvol = (sndvolume > 128) ? 128 : sndvolume;
    musvol = (musvolume > 128) ? 128 : musvolume;
    Mix_Volume(-1, sndvol);
    Mix_VolumeMusic(musvol);
}

// ASM_ReturnVolume
// Takes two byte pointers and returns the internal sound and music volumes using them

void ASM_ReturnVolume(Uint8* retsnd, Uint8* retmus)
{
    *retsnd = 0;
    *retmus = 0;

    ASM_AbortIfClosed;    

    *retsnd = sndvol;
    *retmus = musvol;
}

// ASM_SwitchMus
// Changes music to another track, either directly or via a fade-in/out

bool ASM_SwitchMus(char* loadmus, int fadems, bool fade)
{
    ASM_AbortIfClosed false;

    switchto = Mix_LoadMUS(loadmus);

    if (switchto == NULL)
    {
#ifdef VERBOSE_VIEASM
        printf("ASM_SwitchMus: %s\n", Mix_GetError());
#endif
        return false;
    }

    if (fade != true)
    {
            Mix_HaltMusic();
            Mix_FreeMusic(music);
            music = switchto;
            Mix_VolumeMusic(musvol);
            if (Mix_PlayMusic (music, -1) == -1)
            {
#ifdef VERBOSE_VIEASM
                printf("ASM_SwitchMus: %s\n", Mix_GetError());
#endif
                return false;
            }
            return true;
    }

    Mix_FadeOutMusic(fadems);
    fadetime = fadems;
    switching = true;

    return true;
}

// ASM_FadeInMus
// Fades in a piece of music, stopping whatever was playing before it

bool ASM_FadeInMus(char* loadmus, int fadems)
{
    ASM_AbortIfClosed false;

    if (Mix_PlayingMusic())
        Mix_HaltMusic();
    
    if (music)
        Mix_FreeMusic(music);

    music = Mix_LoadMUS(loadmus);
    
    if (music == NULL)
    {
#ifdef VERBOSE_VIEASM
        printf("ASM_FadeInMus: %s\n", Mix_GetError());
#endif
        return false;
    }
    
    Mix_FadeInMusic (music, -1, fadems);

    return true;
}

// ASM_FadeOutMus
// Fades out and stops the music currently playing

void ASM_FadeOutMus(int fadems)
{
    ASM_AbortIfClosed;

    if (!Mix_PlayingMusic())
        return;

    Mix_FadeOutMusic(fadems);
}

// ASM_Cache
// Loads a sound into memory and returns it as a sample

sample ASM_Cache(char *sndfile, const char *name)
{
    sample sound;
    sound.chunk = NULL;
    sound.name = NULL;

    ASM_AbortIfClosed sound;

    Mix_Chunk *chunk = Mix_LoadWAV(sndfile);

    if (chunk == NULL)
    {
#ifdef VERBOSE_VIEASM
        printf("ASM_Cache: %s on sound %s\n", Mix_GetError(), sndfile);
#endif
        return sound;
    }

    sound.chunk = chunk;
    sound.name = (char *)malloc((size_t)strlen(name)+1);
    strcpy (sound.name, name);

    return sound;
}

// ASM_CacheFromMem
// Loads a sound from a pointer into a sample, returning that sample
// Also frees the original data

sample ASM_CacheFromMem(void *ptr, int size, const char *name)
{
    sample sound;
    sound.chunk = NULL;
    sound.name = NULL;

    ASM_AbortIfClosed sound;

    Mix_Chunk *chunk = Mix_LoadWAV_RW(SDL_RWFromMem(ptr, size), 1);
    if (chunk == NULL)
    {
#ifdef VERBOSE_VIEASM
        printf("ASM_CacheFromMem: %s on sound %s\n", Mix_GetError(), name);
#endif
        return sound;
    }

    sound.chunk = chunk;
    sound.name = (char *)malloc((size_t)strlen(name)+1);
    strcpy (sound.name, name);

    return sound;
}


// ASM_Uncache
// Removes a sound from memory

void ASM_Uncache(sample &sound)
{
    ASM_AbortIfClosed;
    
    if (sound.chunk == NULL)
        return;

    Mix_FreeChunk(sound.chunk);
    sound.chunk = NULL;
    free(sound.name);
    sound.name = NULL;
}

// ASM_PlaySound
// Play a sound in memory loaded with ASM_Cache or Mix_LoadWAV
// Returns -1 on errors or channel number

int ASM_PlaySound(ASM_SoundPlay_t soundPlay)
{
    int chanon;
    ASM_AbortIfClosed -1;
    sample *sound;
    activechan_t *activechan;

    sound = &ASM_Audiosegs[soundPlay.sound];

#ifdef LAZY_LOAD_SOUNDS 
    if (sound->chunk == NULL)
    {
        CA_CacheAudioChunk(soundPlay.sound);
    }
#endif

    if (sound->chunk == NULL)
    {
#ifdef VERBOSE_VIEASM
        printf("ASM_PlaySound: \"%s\" not cached!\n", sound->name);
#endif
        return -1;
    }

    chanon = ASM_GetFreeChannel();

    if (chanon == -1)
    {
#ifdef VERBOSE_VIEASM
        printf("ASM_PlaySound: No free channels!\n");
#endif
        return -1;
    }

    activechan = &activechans[chanon];
    activechan->foreground = soundPlay.foreground;
    activechan->vol = soundPlay.vol;
    activechan->sound = soundPlay.sound;

#ifdef VERBOSE_VIEASM
    printf("Channel %d, dist %d, lp %d, rp %d\n", chanon, 
        soundPlay.dist, soundPlay.lp, soundPlay.rp);
#endif
    Vol_ApplyToChannel(Vol_Adjust(soundPlay.vol, FP(1)), chanon);
    
    if (soundPlay.fadeInTime > 0)
    {
        if (Mix_FadeInChannel(chanon, sound->chunk, soundPlay.loops, 
            soundPlay.fadeInTime) == -1)
        {
#ifdef VERBOSE_VIEASM
            printf("ASM_PlaySound: %s\n", Mix_GetError());
#endif
            return -1;
        }
    }
    else
    {
        if (Mix_PlayChannel(chanon, sound->chunk, soundPlay.loops) == -1)
        {
#ifdef VERBOSE_VIEASM
            printf("ASM_PlaySound: %s\n", Mix_GetError());
#endif
            return -1;
        }
    }

    if (reversemode)
        Mix_SetReverseStereo(chanon, 1);
    else
        Mix_SetReverseStereo(chanon, 0);

    SD_SetPosition(chanon, soundPlay.lp, soundPlay.rp, soundPlay.dist);
#ifdef VERBOSE_VIEASM
    printf("Channel %d has volume %d\n", chanon, Mix_Volume(chanon, -1));
#endif

    return chanon;
}

// ASM_StopChannel
// Stops sound on a given channel

void ASM_StopChannel(int channel)
{
    ASM_AbortIfClosed;

    Mix_HaltChannel(channel);
}

// ASM_FadeOutChannel
// Stops sound on a given channel

void ASM_FadeOutChannel(int channel, int ms)
{
    ASM_AbortIfClosed;

    Mix_FadeOutChannel(channel, ms);
}

// ASM_SwitchStep
// Used each program step for switching music tracks

void ASM_SwitchStep(void)
{
    ASM_AbortIfClosed;

    if (Mix_PlayingMusic())
        return;

    if (switching)       // If music is stopped and we want to switch
    {
        Mix_FreeMusic(music);
        music = switchto;
        switchto = 0;
        Mix_FadeInMusic(music, -1, fadetime);
        switching = false;
    }
}
        
// ASM_ReverseStereo
// Sets reverse stereo mode after initialization

void ASM_ReverseStereo(bool reverse)
{
    ASM_AbortIfClosed;

    Mix_SetReverseStereo(MIX_CHANNEL_POST, (reverse) ? 1 : 0);
}

// ASM_CurChannels
// Returns the number of channels allocated

int ASM_CurChannels(void)
{
    return lastchan;
}

// ASM_GetSndVolume
// Returns the current volume level

int ASM_GetSndVolume(void)
{
    return (int)sndvol;
}

// ASM_SetChannelVol
// Sets the volume of a channel

void ASM_SetChannelVol(int channel, voltype vol)
{
    activechan_t *activechan;

    activechan = &activechans[channel];
    activechan->vol = vol;

    Vol_ApplyToChannel(Vol_Adjust(vol, FP(1)), channel);
}

// ASM_UpdateForeground
// Adjusts the volume of foreground and background sounds

void ASM_UpdateForeground(void)
{
    int i;
    bool found;
    activechan_t *activechan;
    voltype adjVolume;
    fixed scale;

    found = false;

    for (i = 0; i < lastchan; i++)
    {
        if (channelSoundPos[i].valid)
        {
            continue;
        }

        activechan = &activechans[i];
        if (activechan->used && activechan->foreground)
        {
            found = true;
        }
    }

    for (i = 0; i < lastchan; i++)
    {
        if (channelSoundPos[i].valid)
        {
            continue;
        }

        activechan = &activechans[i];
        if (!activechan->used)
        {
            continue;
        }

        scale = FP(1);
        if (found)
        {
            scale = (activechan->foreground ? FP(3) / 2 : FP(1) / 4);
        }

        adjVolume = Vol_Adjust(activechan->vol, scale);
        Vol_ApplyToChannel(adjVolume, i);
    }
}

int ASM_SoundPlaying(void)
{
    int i;
    activechan_t *activechan;

    for (i = 0; i < lastchan; i++)
    {
        activechan = &activechans[i];
        if (activechan->used)
        {
            return activechan->sound;
        }
    }

    return 0;
}

voltype Vol_Zero(bool isStereo)
{
    voltype vol;

    lwlib_Zero(vol);
    vol.isStereo = isStereo;

    return vol;
}

voltype Vol_Mono(int volume)
{
    voltype vol;

    lwlib_Zero(vol);
    vol.u.monoVolume = volume;

    return vol;
}

void Vol_SetMonoVolume(voltype *vol, int volume)
{
    assert(!vol->isStereo);
    vol->u.monoVolume = volume;
}

void Vol_SetStereoVolume(voltype *vol, int side, int volume)
{
    assert(vol->isStereo && side >= 0 && side <= 1);
    vol->u.stereoVolume[side] = volume;
}

bool Vol_IsZero(voltype *vol)
{
    if (vol->isStereo)
    {
        return vol->u.stereoVolume[0] == 0 && vol->u.stereoVolume[1] == 0;
    }
    return vol->u.monoVolume == 0;
}

void Vol_ApplyToChannel(voltype vol, int channel)
{
    if (vol.isStereo)
    {
        Mix_Volume(channel, SD_MaxVolume());
        Mix_SetPanning(channel, vol.u.stereoVolume[0], vol.u.stereoVolume[1]);
    }
    else
    {
        Mix_Volume(channel, vol.u.monoVolume);
        Mix_SetPanning(channel, 255, 255);
    }
}

voltype Vol_Adjust(voltype vol, fixedpt scale)
{
    int side;

    scale = fixedpt_mul(scale, 
        fixedpt_fromint(ASM_GetSndVolume()) / MIX_MAX_VOLUME);

    if (!vol.isStereo)
    {
        vol.u.monoVolume = fixedpt_toint(scale * vol.u.monoVolume);
        vol.u.monoVolume = lwlib_MIN(vol.u.monoVolume, MIX_MAX_VOLUME);
    }
    else
    {
        for (side = 0; side < 2; side++)
        {
            vol.u.stereoVolume[side] = 
                fixedpt_toint(scale * vol.u.stereoVolume[side]);
            vol.u.stereoVolume[side] = lwlib_MIN(vol.u.stereoVolume[side], 
                255);
        }
    }

    return vol;
}

