// WL_ACT2.C

#include <stdio.h>
#include <math.h>
#include "wl_def.h"
#include "lw.h"
#if defined _WIN32
#pragma hdrstop
#endif

/*
=============================================================================

                               LOCAL CONSTANTS

=============================================================================
*/

#define PROJECTILESIZE  0xc000l
#define EXTENDEDPROJSIZE  0x10000l
#define PLAYERHEADHEIGHT 0.65

#define BJRUNSPEED      2048
#define BJJUMPSPEED     680


/*
=============================================================================

                              GLOBAL VARIABLES

=============================================================================
*/



/*
=============================================================================

                              LOCAL VARIABLES

=============================================================================
*/


dirtype dirtable[9] = {northwest,north,northeast,west,nodir,east,
    southwest,south,southeast};

void    A_StartDeathCam (objtype *ob);


void A_BecomeBlocking (objtype *ob)
{
    ob->flags |= FL_BLOCK;
}

void A_BecomeShootable (objtype *ob)
{
    ob->flags |= FL_SHOOTABLE;
}

void A_BecomeVulnerable (objtype *ob)
{
    ob->flags &= ~FL_NODAMAGE;
}

void A_BecomeUnkillable (objtype *ob)
{
    ob->flags |= FL_UNKILLABLE;
}

void A_BecomeNoPain (objtype *ob)
{
    ob->flags |= FL_SUPPRESSPAIN;
}

void A_BecomeUnderground (objtype *ob)
{
    ob->flags |= FL_UNDERGROUND;
}

void A_RemoveBlocking (objtype *ob)
{
    ob->flags &= ~FL_BLOCK;
}

void A_RemoveShootable (objtype *ob)
{
    ob->flags &= ~FL_SHOOTABLE;
}

void A_RemoveUnkillable (objtype *ob)
{
    ob->flags &= ~FL_UNKILLABLE;
}

void A_RemoveNoPain (objtype *ob)
{
    ob->flags &= ~FL_SUPPRESSPAIN;
}

void A_RemoveUnderground (objtype *ob)
{
    ob->flags &= ~FL_UNDERGROUND;
}

void Refire (objtype *ob, short maxrepeat, statetype *state, void (*Action)(objtype *));
void SplashDamage (objtype *ob, short maxdamage, damage_t damagetype, short blastradius, boolean hurtenemies);
boolean CheckGround (objtype *ob, boolean digdown);
boolean CheckRangeAttack (objtype *ob);
boolean CheckMeleeAttack (objtype *ob, int32_t move);
boolean CheckMeleeRange (objtype *ob, int32_t move);

/*
=============================================================================

                              PROJECTILES

=============================================================================
*/

void A_Smoke (objtype *ob)
{
    SpawnBulletPuff(OB_GetX(ob),OB_GetY(ob),
        ED_STATE(ob->proj->smokeState),ob->sproffset,ob->heightoff);
}

void A_RocketBlast (objtype *ob)
{
    Tremor_ObjStart(trem_explosion, ob);
    SplashDamage(ob,50,dam_rocket,5,false);
}

void A_BJRocketBlast (objtype *ob)
{
    Tremor_ObjStart(trem_explosion, ob);
    SplashDamage(ob,200,dam_rocket,5,true);
}

void A_FlameBurn (objtype *ob)
{
    objtype *check;
    const fixed minDist = FP(2);
    const int damage = US_RndT_Range(3, 10);

    for (ITEROBJ(check, 0))
    {
        if (check == ob || check->areanumber != ob->areanumber)
        {
            continue;
        }

        if (check != player && 
            (check->flags & FL_SHOOTABLE) == 0)
        {
            continue;
        }

        if (!OB_CheckCloserThan(check, ob, minDist))
        {
            continue;
        }

        if (check == player)
        {
            TakeDamage(damage, ob);
        }
        else
        {
            DamageActor(check, ob, damage, dam_fire);
        }
    }
}

/*
=================
=
= ReboundProjectile
=
=================
*/

bool ReboundProjectile(objtype *ob, lwlib::Plane3f plane, 
    lwlib::Point2i *tilePos)
{
    int side1, side2;
    physobjtype *physobj;
    const ED_ProjPhys_t *projPhys;
    lwlib::Point3f posPlane;
    lwlib::Point3f tilePosf;
    lwlib::Ray3f ray;
    lwlib::Box3f tileBox;

    physobj = &ob->u.phys;
    projPhys = ED_ProjPhys(ed, ob->proj->physName);

    const double eps = 0.0;
    const double secs = (double)(tics) / SEC2TICS(1);

    const lwlib::Point3f p1 = physobj->pos;
    side1 = vec_side(plane, p1, eps);

    const lwlib::Vector3f disp = vec_scale(physobj->vel, secs);
    const lwlib::Point3f p2 = vec_add(physobj->pos, disp);
    side2 = vec_side(plane, p2, eps);

    if (tilePos != NULL)
    {
        ray = ray3f(physobj->pos, disp);
        lwlib_Z(ray.p[0]) += 0.5;
        tilePosf = lwlib::vec3f(lwlib_X(*tilePos), lwlib_Y(*tilePos), 0.0);
        tileBox = lwlib::box_cell<double, 3>(tilePosf);
    }

#ifdef DEBUG_PROJECTILE_PHYSICS
    {
        lwlib_String_t s1, s2;
        fprintf(stderr, "%s %s %d %d %d %d\n", 
            vec_to_string(s1, "%f", p1), vec_to_string(s2, "%f", p2), 
            side1, side2, tilePos != NULL, 
            tilePos != NULL ? ray3f_hit_box(ray, tileBox) : -1);
    }
#endif
    if (side1 > 0 && side2 <= 0 && (tilePos == NULL ||
        ray3f_hit_box(ray, tileBox)))
    {
        physobj->vel = lwlib::vec_reflect(physobj->vel, 
            lwlib_plane_normal(plane));
        physobj->vel = lwlib::vec_scale(physobj->vel, 
            -projPhys->bounceDamping);

        posPlane = lwlib::vec_proj(physobj->pos, plane);

        const double offplanedist = 0.01;
        physobj->pos = lwlib::vec_add(posPlane, lwlib::vec_scale(
            plane.p[1], offplanedist));

        if (projPhys->bounceLimit != 0)
        {
            physobj->numBounces++;
            
            if (physobj->numBounces >= projPhys->bounceLimit)
            {
                return false;
            }

            if (projPhys->bounceSound[0] != '\0')
            {
                PlaySoundLocActor(ED_SND(projPhys->bounceSound), ob);
            }
        }
    }

    return true;
}


/*
===================
=
= ProjectileTryMove
=
= returns true if move ok
===================
*/

#define PROJSIZE        0x2000

boolean ProjectileTryMove (objtype *ob)
{
    int i,xl,yl,xh,yh,x,y;
    objtype *check;
    const ED_ProjPhys_t *projPhys;
    fixed projSize = PROJSIZE;
    lwlib::Point2i tilePos;
    lwlib::Point3f origin, normal;
    physobjtype *physobj;
    lwlib::Plane3f plane;

    projPhys = ED_ProjPhys(ed, ob->proj->physName);

    if (projPhys != NULL)
    {
        physobj = &ob->u.phys;
        projSize = EXTENDEDPROJSIZE;
    }
    xl = (OB_GetX(ob)-projSize) >> TILESHIFT;
    yl = (OB_GetY(ob)-projSize) >> TILESHIFT;

    xh = (OB_GetX(ob)+projSize) >> TILESHIFT;
    yh = (OB_GetY(ob)+projSize) >> TILESHIFT;

    //
    // check for solid walls
    //
    for (y=yl;y<=yh;y++)
    {
        for (x=xl;x<=xh;x++)
        {
            tilePos = lwlib::vec2(x, y);

            check = actorat[x][y];
            if (check && !ISPOINTER(check) && tilemap[x][y])
            {
                if (tilemap[x][y] == LANDSCAPETILE)
                {
                    OB_SetState(ob, NULL);
                    // WSJ: projectile flew off into the distance...
                    return true;
                }

                if (projPhys != NULL)
                {
                    for (i = 0; i < 2; i++)
                    {
                        origin = normal = lwlib::vec3f_zero();

                        if (lwlib_C(physobj->vel, i) >= 0.0)
                        {
                            lwlib_C(origin, i) = lwlib_C(tilePos, i);
                            lwlib_C(normal, i) = -1.0;
                        }
                        else
                        {
                            lwlib_C(origin, i) = lwlib_C(tilePos, i) + 1;
                            lwlib_C(normal, i) = 1.0;
                        }

                        plane = lwlib::plane3f(origin, normal);
                        if (!ReboundProjectile(ob, plane, &tilePos))
                        {
                            if (ob->proj->destroysWalls && 
                                ED_ActorIsBreakableWall(ed, MAPSPOT(x, y, 1)))
                            {
                                DestroyWall(x, y);
                            }
                            return false;
                        }
                    }
                }
                else
                {
                    return false;
                }
            }
            else if (ob->proj->hitsScenery &&
                check && !ISPOINTER(check) && !tilemap[x][y])
            {
                return false;
            }
        }
    }

    return true;
}


void ProjectileHit (objtype *ob, boolean dodamage, objtype *hitactor)
{
    int damage;
    int hitSnd;
    const ED_Proj_t *proj;
    ED_State_t *explState;
    boolean ignorearmor = false;

    proj = ob->proj;
    damage_t damagetype = dam_bullet;

    if (proj->damageType != dam_none)
    {
        if (proj->rndDamageDivisor != 0)
        {
            damage = proj->baseDamage + (US_RndT() * 
                proj->rndDamageMultiplier / proj->rndDamageDivisor);
        }
        else
        {
            damage = proj->baseDamage;
        }
        damagetype = proj->damageType;

        explState = ED_STATE(proj->explState);
        if (explState != NULL)
        {
            OB_SetState(ob, explState);
        }
        else
        {
            OB_SetState(ob, NULL); // mark for removal
        }

        if (proj->playHitSound)
        {
            hitSnd = (proj->hitSound[0] != 0 ? 
                ED_SND(proj->hitSound) : ED_SND("MISSILEHITSND"));
            PlaySoundLocActor(hitSnd,ob);
        }
    }
    else
    {
        damage = 0;
        OB_SetState(ob, NULL);      // mark for removal
    }

    if (damage && dodamage)
    {
        if (hitactor == player)
            TakeSpecialDamage (damage,ob,damagetype,ignorearmor,true);
        else
            DamageActor (hitactor,ob,damage,damagetype);
    }
}


/*
=================
=
= CheckProjectileHitsPlayer
=
=================
*/

bool CheckProjectileHitsPlayer(objtype *ob)
{
    int32_t deltax,deltay;
    const ED_ProjPhys_t *projPhys;
    physobjtype *physobj;

    projPhys = ED_ProjPhys(ed, ob->proj->physName);
    physobj = &ob->u.phys;

    deltax = LABS(OB_GetX(ob) - OB_GetX(player));
    deltay = LABS(OB_GetY(ob) - OB_GetY(player));

    const double h = PLAYERHEADHEIGHT * 2.0 - 1.0;

    return deltax < PROJECTILESIZE && deltay < PROJECTILESIZE && 
        ob->proj->hitsPlayer && (projPhys == NULL || 
        lwlib_Z(physobj->pos) <= h);
}

/*
=================
=
= T_Projectile
=
=================
*/

void T_Projectile (objtype *ob)
{
    int i;
    int32_t deltax,deltay;
    int32_t speed;
    physobjtype *physobj;
    lwlib::Plane3f plane;
    const ED_ProjPhys_t *projPhys;
    lwlib::Point3f disp, nextPos;

    const double secs = (double)(tics) / SEC2TICS(1);

    projPhys = ED_ProjPhys(ed, ob->proj->physName);

    if (projPhys != NULL)
    {
        physobj = &ob->u.phys;

        physobj->timer += secs;

        if (projPhys->explodeTime > 0.0 && 
            physobj->timer > projPhys->explodeTime)
        {
            ProjectileHit(ob, false, NULL);
            return;
        }

        lwlib_Z(physobj->vel) -= projPhys->gravity * secs;

        plane = lwlib::plane3f(lwlib::vec3f(0.0, 0.0, -1.0),
            lwlib::vec3f(0.0, 0.0, 1.0));
        if (!ReboundProjectile(ob, plane, NULL))
        {
            ProjectileHit(ob, false, NULL);
            return;
        }

        plane = lwlib::plane3f(lwlib::vec3f(0.0, 0.0, 1.0),
            lwlib::vec3f(0.0, 0.0, -1.0));
        if (!ReboundProjectile(ob, plane, NULL))
        {
            ProjectileHit(ob, false, NULL);
            return;
        }
    }
    else
    {
        speed = (int32_t)ob->speed*tics;

        deltax = FixedMul(speed,finecostable[ob->fineangle]);
        deltay = -FixedMul(speed,finesintable[ob->fineangle]);

        if (deltax>0x10000l)
            deltax = 0x10000l;
        if (deltay>0x10000l)
            deltay = 0x10000l;

        OB_SetX(ob, OB_GetX(ob) + deltax);
        OB_SetY(ob, OB_GetY(ob) + deltay);
    }

    if (!ProjectileTryMove (ob))
    {
        ProjectileHit(ob, false, NULL);
        return;
    }

    if (CheckProjectileHitsPlayer(ob))
    {       // hit the player
        ProjectileHit (ob,true,player);
        return;
    }
    if (ob->proj->hitsEnemies)
    {
        objtype *check;

        for (ITEROBJ(check, 0))
        {
            if (check == player || !(check->flags & FL_SHOOTABLE))
                continue;

            deltax = LABS(OB_GetX(ob) - OB_GetX(check));
            deltay = LABS(OB_GetY(ob) - OB_GetY(check));

            const fixed enemyProjSize = (PROJECTILESIZE / 2);
            if (deltax < enemyProjSize && deltay < enemyProjSize && 
                (ob->projEnemyIndex == 0 || ob->projEnemyIndex != 
                check->enemyIndex))
            {
                ProjectileHit (ob,true,check);
                return;
            }
        }
    }

    if (projPhys != NULL)
    {
        disp = vec_scale(physobj->vel, secs);
        nextPos = vec_add(physobj->pos, disp);

        // stop projectiles before they escape map limits
        for (i = 0; i < 2; i++)
        {
            if (lwlib_C(nextPos, i) <= 0.0 || lwlib_C(nextPos, i) >= 64.0)
            {
                OB_SetState(ob, NULL);
                return;
            }
        }

        physobj->pos = nextPos;

        OB_SetX(ob, fixedpt_rconst(lwlib_X(physobj->pos)));
        OB_SetY(ob, fixedpt_rconst(lwlib_Y(physobj->pos)));
        ob->heightoff = fixedpt_rconst(lwlib_Z(physobj->pos));
    }

    ob->tilex = (short)(OB_GetX(ob) >> TILESHIFT);
    ob->tiley = (short)(OB_GetY(ob) >> TILESHIFT);
}



/*
=================
=
= SplashDamage
=
= WSJ: based on Matthew's explosion code
=
=================
*/

void SplashDamage (objtype *ob, short maxdamage, damage_t damagetype, short blastradius, boolean hurtenemies)
{
    objtype *check;
    int damage;
    short obmaxdamage;

    const double dist = vec_length(vec_sub(OB_GetPos(player), OB_GetPos(ob)));
    if (dist > 0 && dist < blastradius && CheckLine(ob))
    {
        damage = maxdamage / (dist * dist);
        if (damage > maxdamage)
        {
            damage = maxdamage;
        }
        TakeSpecialDamage(damage,ob,damagetype,false,true);
    }

    if (!hurtenemies)
        return;

    for (ITEROBJ(check, 1))
    {
        if (check->flags & FL_SHOOTABLE)
        {
            const double dist = vec_length(vec_sub(OB_GetPos(check),
                OB_GetPos(ob)));
            if (dist > 0 && dist < blastradius && CheckObjLine(ob,check))
            {
                obmaxdamage = maxdamage;
                if (damagetype == dam_rocket &&
                    (check->flags & FL_WEAKTOROCKETS) != 0)
                {
                    obmaxdamage *= 2;
                }

                damage = obmaxdamage / (dist * dist);
                if (damage > obmaxdamage)
                {
                    damage = obmaxdamage;
                }
                DamageActor(check,ob,damage,damagetype);
            }
        }
    }
}


//
// bullet effects
//




/*
=================
=
= SpawnBulletPuff
=
=================
*/

void SpawnBulletPuff (fixed puffx, fixed puffy, statetype *state, 
    short spriteoffset, fixed puffz)
{
    if (state == NULL)
    {
        return;
    }

    GetNewActor ();
    OB_SetState(newobj, state);
    newobj->sproffset = spriteoffset;
    newobj->ticcount = 1;
    OB_SetX(newobj, puffx);
    OB_SetY(newobj, puffy);
    newobj->tilex = (short)(puffx >> TILESHIFT);
    newobj->tiley = (short)(puffy >> TILESHIFT);
    newobj->active = ac_yes;
    newobj->flags = FL_NEVERMARK;
    newobj->heightoff = puffz;
}

void SpawnRicochet (fixed puffx, fixed puffy, short spriteoffset, byte tile,
    fixed puffz)
{
    ED_Ricochet_t *ricochet;

    ricochet = ED_WallRicochet(ed, tile);
    if (ricochet != NULL)
    {
        PlaySoundLocGlobal(ED_SND(ricochet->soundDef), puffx, puffy,
            SD_MaxVolume(), false, false);
        SpawnBulletPuff(puffx, puffy, ED_STATE(ricochet->state),
            spriteoffset, puffz);
    }
}

void SpawnBloodPuff (objtype *ob, fixed puffx, fixed puffy, fixed puffz)
{
    ED_State_t *state;
    const ED_EnemyBloodPuff_t *bloodPuff;

    bloodPuff = &ob->enemy->bloodPuff;

    if (bloodPuff->soundDef[0] != '\0')
    {
        PlaySoundLocGlobal(ED_SND(bloodPuff->soundDef), puffx, puffy,
            SD_MaxVolume(), false, false);
    }

    state = ED_STATE(bloodPuff->state);
    if (state != NULL)
    {
        SpawnBulletPuff(puffx,puffy,state,
            ED_SPR(bloodPuff->spriteOffset.newStart) -
            ED_SPR(bloodPuff->spriteOffset.oldStart),puffz);
    }
}


/*
=============================================================================

ACTOR SOUND STUFF

=============================================================================
*/

/*
=================
=
= PlayActorVoice
=
=================
*/

int PlayActorVoice (int sound, objtype *ob)
{
    bool foreground;
    bool noDistAtt;
    ED_Boss_t *boss;

    boss = ob->enemy->boss;
    foreground = false;//(boss != NULL);
    noDistAtt = (boss != NULL);

    if (sound == -1)
    {
        return -1;
    }

    if (!areabyplayer[ob->areanumber] && !(ob->flags&FL_VISABLE))
    {
        return -1;
    }
    
    return PlaySoundLocActorVol (sound,ob,SD_MaxVolume(),foreground,
        noDistAtt);
}

void PlayActorWeapon (int sound, objtype *ob)
{
    PlaySoundLocActor(sound,ob);
    MakeNoise(ob);
}

/*
==================
=
= A_AlertSound
=
==================
*/

void A_AlertSound (objtype *ob)
{
    if (ED_SND(ob->enemy->alertSound) != -1)
    {
        PlayActorVoice(ED_SND(ob->enemy->alertSound), ob);
    }
}


/*
==================
=
= A_DeathScream
=
==================
*/

void A_DeathScream (objtype *ob)
{
    const ED_EnemyDeath_t *death;
    death = &ob->enemy->death;
    PlayActorVoice(ED_SND(death->screamSound), ob);
}


/*
==================
=
= A_FlameScream
=
==================
*/

void A_FlameScream (objtype *ob)
{
    ED_SoundIndex_t snd;
    const ED_EnemyDeath_t *death;

    death = &ob->enemy->death;
    
    snd = ED_SND(death->burnScreamSound); 
    if (PlayActorVoice(snd, ob) == -1)
    {
        snd = ED_SND(death->screamSound); 
        PlayActorVoice(snd, ob);
    }
}


////////////////////////////////////////////////////////
//
// A_Slurpie
//
////////////////////////////////////////////////////////
void A_Slurpie (objtype *ob)
{
    PlayActorVoice(ED_SND("SLURPIESND"),ob);
}

/*
=============================================================================

GUARD

=============================================================================
*/

static int EnemyShootDamage(objtype *ob, int base, int extra,
    fixed distLog2Scale)
{
    const int distLog2 = lwlib_Log2(1 + OBJTILEDIST(ob, player));
    const int distLog2Scaled = fixedpt_toint(distLog2Scale * distLog2);
    return (base + US_RndT() % extra)/(1 + distLog2Scaled);
}

//
// guards
//

void T_PistolShoot (objtype *ob)
{
    const int damage = EnemyShootDamage(ob, 4, 12, FP(75) / 100);
    FireBullet(ob,damage,dam_bullet,2,1,0);
    PlayActorWeapon(ED_SND("ATKPISTOLSND"),ob);
}

void T_RifleShoot (objtype *ob)
{
    const int damage = EnemyShootDamage(ob, 10, 10, FP(50) / 100);
    FireBullet(ob,damage,dam_rifle,1,1,0);
    PlayActorWeapon(ED_SND("ATKKAR98SND"),ob);
}

void T_MG42Shoot (objtype *ob)
{
    const int damage = EnemyShootDamage(ob, 2, 10, FP(75) / 100);
    FireBullet(ob,damage,dam_bullet,3,1,0);
}

void T_MG42ShootS (objtype *ob)
{
    T_MG42Shoot(ob);
    PlayActorWeapon(ED_SND("ATKMG42SND"),ob);
}

void T_MG42GrdLoop (objtype *ob)
{
    Refire(ob,10,ED_STATE("s_grdmg42shoot2"),T_MG42ShootS);
}

//
// dogs
//

void T_DogBite (objtype *ob)
{
    PlayActorVoice(ED_SND("DOGATTACKSND"),ob);
    if (CheckMeleeRange(ob,TILEGLOBAL))
    {
        if (US_RndT()<180)
        {
            TakeSpecialDamage (US_RndT()>>4,ob,dam_melee,true,true);
        }
    }
}

//
// officers
//

void T_OfcRefire (objtype *ob)
{
    Refire(ob,4,ED_STATE("s_ofcshoot2"),NULL);
}

void T_MG42OfcLoop (objtype *ob)
{
    Refire(ob,10,ED_STATE("s_ofcmg42shoot2"),T_MG42ShootS);
}

//
// Scientist
//

void T_SciRefire (objtype *ob)
{
    Refire(ob,4,ED_STATE("s_scishoot2"),NULL);
}

void T_SciMelee (objtype *ob)
{
    if (CheckMeleeRange(ob,TILEGLOBAL))
    {
        if (US_RndT()<220)
        {
            TakeSpecialDamage (US_RndT()>>4,ob,dam_mutagen,true,true);
        }
    }
}

void T_SciThrow (objtype *ob)
{
    T_FireProj(ob);
}

//
// SS
//

void T_MP40Shoot (objtype *ob)
{
    const int damage = EnemyShootDamage(ob, 3, 10, FP(30) / 100);
    FireBullet(ob,damage,dam_bullet,2,1,0);
    PlayActorWeapon(ED_SND("ATKMP40SND"),ob);
}

void T_STG44Shoot (objtype *ob)
{
    const int damage = EnemyShootDamage(ob, 10, 20, FP(75) / 100);
    FireBullet(ob,damage,dam_bullet,1,1,0);
    PlayActorWeapon(ED_SND("ATKSTG44SND"),ob);
}

void T_SSRefire (objtype *ob)
{
    Refire(ob,3,ED_STATE("s_ssshoot3"),NULL);
}

void T_SSOfcRefire (objtype *ob)
{
    Refire(ob,3,ED_STATE("s_ssofcshoot3"),NULL);
}

void T_MG42SSLoop (objtype *ob)
{
    Refire(ob,10,ED_STATE("s_ssmg42shoot2"),T_MG42ShootS);
}

//
// mutant
//

void T_MG42MutLoop (objtype *ob)
{
    Refire(ob,10,ED_STATE("s_mutmg42shoot2"),T_MG42ShootS);
}

//
// Skeleton
//

void T_ZWake (objtype *ob)
{
    ob->flags|=FL_SHOOTABLE;
    if(!(ob->flags&FL_ATTACKMODE))
        FirstSighting(ob);
}

void T_StartDig (objtype *ob)
{
    if(US_RndT()&1)
        CheckGround(ob,true);
}

void T_StartDigInChase (objtype *ob)
{
    if(US_RndT() < 128)
    {
        CheckGround(ob,true);
    }
}

void T_StartRise (objtype *ob)
{
    if(US_RndT()&1)
        CheckGround(ob,false);
}

void T_BoneShrapnel (objtype *ob)
{
    SplashDamage(ob,200,dam_rocket,5,true);
    ProjFire(ob,0,ED_Proj(ed, "bone"),ED_STATE("s_bone1"),0x2000l,0);
    ProjFire(ob,ANGLEQUAD,ED_Proj(ed, "bone"),ED_STATE("s_bone2"),0x2000l,0);
    ProjFire(ob,ANGLEQUAD*2,ED_Proj(ed, "bone"),ED_STATE("s_bone3"),0x2000l,0);
    ProjFire(ob,ANGLEQUAD*3,ED_Proj(ed, "bone"),ED_STATE("s_bone4"),0x2000l,0);
}

void T_FireProj (objtype *ob)
{
    ProjFire(ob,0,ED_Proj(ed, ob->enemy->projName),NULL,ob->enemy->projSpeed,0);
}

void T_FireShotgun (objtype *ob)
{
    FireBullet (ob,(2 + US_RndT()%10),dam_chaingun,5,4,0);
    PlayActorWeapon(ED_SND("ATKSHOTGUNSND"),ob);
}

void T_SkelMeleeLoop (objtype *ob)
{
    if(CheckMeleeRange(ob,TILEGLOBAL))
        NewState(ob,ED_STATE("s_skelmelee3"));
}

void T_MG42SkelLoop (objtype *ob)
{
    Refire(ob,10,ED_STATE("s_skelmg42shoot2"),T_MG42ShootS);
}

void T_SkelAttack (objtype *ob)
{
    T_ZWake(ob);
    if(CheckMeleeRange(ob,TILEGLOBAL))
        NewState(ob,ED_STATE("s_skelmelee1"));
}

void T_SkelBuried (objtype *ob)
{
    if(SightPlayer(ob,ED_STATE("s_skeldigout1")))
        A_BecomeBlocking(ob);
}

void T_SkelFireBomb (objtype *ob)
{
    SplashDamage(ob,400,dam_fire,3,true);
}


//
// Zombie SS
//

void T_ZSSRefire (objtype *ob)
{
    Refire(ob,3,ED_STATE("s_zssshoot3"),NULL);
}

void T_MG42ZSSLoop (objtype *ob)
{
    Refire(ob,10,ED_STATE("s_zssmg42shoot2"),T_MG42ShootS);
}

void T_ZSSAttack (objtype *ob)
{
    T_ZWake(ob);
    if(CheckLine(ob))
    {
        ob->temp1=0;
        NewState(ob,ED_STATE("s_zssshoot1"));
    }
}

void T_ZSSBuried (objtype *ob)
{
    if(SightPlayer(ob,ED_STATE("s_zssdigout1")))
        A_BecomeBlocking(ob);
}

/*
=============================================================================

                            BOSSES

=============================================================================
*/

//
// hans
//

void T_BossShoot (objtype *ob)
{
    const int damage = EnemyShootDamage(ob, 1, 5, FP(30) / 100);
    FireBullet(ob,damage,dam_chaingun,2,2,0);
}

void T_BossShootS (objtype *ob)
{
    T_BossShoot(ob);
    PlayActorWeapon(ED_SND("ATKCHAINGUNSND"),ob);
}

void T_BossRefire (objtype *ob)
{
    Refire(ob,3,ED_STATE("s_bossshoot2"),T_BossShootS);
}

void T_BossShootFireProj(objtype *ob)
{
    T_FireProj(ob);
    T_BossShoot(ob);
}

void A_BossBurnStart(objtype *ob)
{
    A_FlameScream(ob);
    ob->temp1 = 0;
}

void A_BossBurnRepeat(objtype *ob)
{
    ED_State_t *burnState;

    if (ob->temp1 >= 4)
    {
        ob->temp1 = 0;
        return; // finish burning death
    }

    ob->temp1++;
    burnState = ED_STATE(ob->enemy->death.burnState);
    NewState(ob, burnState);
}

/*
=============================================================================

                           SPAWN FUNCTIONS

=============================================================================
*/

/*
===============
=
= SpawnObjFlags
=
===============
*/
static void SpawnObjFlags(void)
{
    int i;
    const ED_EnemySpawnFlags_t *spawnFlags;

    spawnFlags = &newobj->enemy->spawnFlags;

    if (spawnFlags->override)
    {
        for (i = 0; (size_t)i < lwlib_CountArray(spawnFlags->flags); i++)
        {
            newobj->flags |= spawnFlags->flags[i];
        }
    }
    else
    {
        newobj->flags |= FL_SHOOTABLE|FL_BLOCK;
    }
}


/*
===============
=
= SpawnStand
=
===============
*/

void SpawnStand (int tilex, int tiley, int dir, bool deafguard)
{
    int hitpoints;
    ED_SkillProfile_t *skillProfile;

    hitpoints = newobj->enemy->health;
    if (newobj->enemy->killpoints != 0)
    {
        hitpoints = SKILLAMOUNT(hitpoints);
    }
    if (gamestate.difficulty < ED_MAX_SKILLPROFILES)
    {
        skillProfile = newobj->enemy->skillProfiles[gamestate.difficulty];
        if (skillProfile != NULL)
        {
            hitpoints = skillProfile->hitpoints;
        }
    }

    newobj->hitpoints = newobj->starthitpoints = hitpoints;

    newobj->killpoints = SKILLAMOUNT(newobj->enemy->killpoints);
    if (newobj->killpoints && !loadedgame)
        gamestate.killtotal++;
    if ((dirtype)dir == nodir)
        newobj->dir = nodir;
    else
        newobj->dir = (dirtype)(dir * 2);
    SpawnObjFlags();
    if (deafguard)
        newobj->flags |= FL_AMBUSH;

}


/*
===============
=
= SpawnPatrol
=
===============
*/

void SpawnPatrol (int tilex, int tiley, int dir)
{
    if ((dirtype)dir == nodir)
        newobj->dir = nodir;
    else
        newobj->dir = (dirtype)(dir * 2);
    newobj->hitpoints = newobj->starthitpoints =
        SKILLAMOUNT(newobj->enemy->health);
    newobj->killpoints = SKILLAMOUNT(newobj->enemy->killpoints);
    if (newobj->killpoints && !loadedgame)
        gamestate.killtotal++;
    newobj->distance = TILEGLOBAL;
    SpawnObjFlags();
    newobj->active = ac_yes;

    actorat[newobj->tilex][newobj->tiley] = NULL;           // don't use original spot

    switch (dir)
    {
        case 0:
            newobj->tilex++;
            break;
        case 1:
            newobj->tiley--;
            break;
        case 2:
            newobj->tilex--;
            break;
        case 3:
            newobj->tiley++;
            break;
    }

    actorat[newobj->tilex][newobj->tiley] = newobj;
}

//#define DEBUG_SPAWNENEMY

/*
===============
=
= SpawnActor
=
= LinuxWolf: Spawns static and enemy actors.
=
===============
*/
void SpawnActor (int tilex, int tiley, int tile)
{
    int type;
    int surpriseatk;
    ED_Enemy_t *enemy;
    const ED_Spawn_t *spawn;
    const ED_Ambient_t *ambient;
    bool deafguard;
    ED_Boss_t *boss;

    type = ED_ActorStatInfo(ed, tile);
    spawn = ED_ActorSpawn(ed, tile);
    ambient = ED_ActorAmbient(ed, tile);

    if (type != -1)
    {
        SpawnStatic(tilex, tiley, type);
    }
    else if (spawn != NULL)
    {
        enemy = ED_Enemy(ed, spawn->enemyName);
        if (enemy != NULL)
        {
            if (!enemy->surpriseAttack.noAmbushReset &&
                enemy->surpriseAttack.spawnEnemyName[0] != '\0')
            {
                surpriseatk = (int)MAPSPOT(tilex, tiley, 0) - 
                    AREATILE;
                surpriseatk = CLIP(surpriseatk, 0,
                    (int)surpriseatk_max - 1);

                // fix ambiance cut off on surprise mutant tile
                MAPSPOT(tilex, tiley, 0) = 0;
                PlaceValidArea(tilex, tiley);
            }
            else
            {
                surpriseatk = surpriseatk_default;
            }

            deafguard = (MAPSPOT(tilex, tiley, 0) == AMBUSHTILE);
            SpawnNewObj (tilex, tiley, ED_STATE(spawn->startState));

            newobj->speed = enemy->speed;
            OB_SetEnemy(newobj, enemy);
			newobj->proj = ED_ProjByIndex(ed, newobj->projIndex);
            newobj->sproffset = 
                ED_SPR(enemy->spriteOffset.newStart) -
                ED_SPR(enemy->spriteOffset.oldStart);
            newobj->surpriseatk = (surpriseatktype)surpriseatk;

            if (spawn->patrol)
            {
                SpawnPatrol(tilex, tiley, spawn->dir);
            }
            else
            {
                SpawnStand(tilex, tiley, spawn->dir, deafguard);
            }

            boss = newobj->enemy->boss;
            if (boss != NULL && boss->ownsMusic)
            {
                gamestate.bossmusicowner = newobj - objlist;
            }

            if (newobj->enemy->guarded)
            {
                gamestate.guardedObjNum = newobj - objlist;
            }

            #ifdef DEBUG_SPAWNENEMY
            fprintf(stderr, "spawned enemy %s,%d\n", spawn->enemyName, newobj->enemyIndex);
            #endif
        }
    }
    else if (ambient != NULL)
    {
        if (VALIDAREA(tilex, tiley))
        {
            ambiance[AreaAt(tilex, tiley)] |= (1 << (ambient - 
                ED_Ambients(ed)));
        }
    }
}


/*
=================
=
= A_Respawn
=
=================
*/

void A_Respawn (objtype *ob)
{
    ob->flags &= ~FL_NONMARK;
    ob->flags |= FL_SHOOTABLE|FL_BLOCK;
    actorat[ob->tilex][ob->tiley] = ob;
    FirstSighting (ob);
}


/*
============================================================================

STAND

============================================================================
*/


/*
===============
=
= T_Stand
=
===============
*/

void T_Stand (objtype *ob)
{
    SightPlayer (ob,NULL);
}


/*
===============
=
= T_MG42Wait
=
= WSJ: enemy just stands there and waits until the player
= is in his sights, then he fires
=
===============
*/

boolean CanUseMG42 (objtype *ob)
{
    if (!(ob->flags&FL_ATTACKMODE) || !(ob->flags&FL_SHOOTABLE) || ob->areanumber!=player->areanumber)
        return false;
    return ob->enemy->mg42.canUse;
}

void T_MG42Wait (objtype *ob)
{
    if (gamestate.victoryflag)
        return;

    if (!(ob->flags & FL_ATTACKMODE))
    {
        SightPlayer (ob,NULL);
        return;
    }

    if (CheckLinePlayerBody(ob))  // got a shot at player?
    {
        if ( (unsigned) US_RndT() < (tics<<3) )
        {
            ob->temp1 = 0;
            NewState (ob,ED_STATE(ob->enemy->mg42.shootState));
        }
    }
}


/*
============================================================================

CHASE

============================================================================
*/

/*
=================
=
= T_Chase
=
=================
*/

void T_Chase (objtype *ob)
{
    int32_t move;
    boolean dodge;
    int dist;
    watermutobjtype *watermut;

    watermut = &ob->u.watermut;

    if (gamestate.victoryflag)
        return;

    OB_ChooseGoalObj(ob);

    if (ob->enemy->runAwayDist > 0)
    {
        dist = OBJTILEDIST(ob, player);
        ob->runAway = dist < ob->enemy->runAwayDist;
    }

    dodge = false;
    if (CheckLine(ob)) // got a shot at player?
    {
        ob->hidden = false;
        if (CheckRangeAttack(ob))
        {
            return;
        }
        dodge = (ob->reaction != react_findmg42) &&
            !OB_GuardedObjInSight(ob);
    }
    else
    {
        ob->hidden = true;
        if (ob->enemy->waterMut != NULL && watermut->waterbound)
        {
            CheckGround(ob,false);
        }
    }

    if (ob->dir == nodir)
    {
        if (dodge)
            SelectDodgeDir (ob);
        else
            SelectChaseDir (ob);
        if (ob->dir == nodir)
            return;                                                 // object is blocked in
    }

    move = ob->speed*tics;

    while (move)
    {
        //
        // check for melee range
        //
        if (CheckMeleeAttack(ob,move))
            return;

        if (ob->waitdoornum)
        {
            //
            // waiting for a door to open
            //
            OpenDoor (ob->waitdoornum - 1);
            if (doorobjlist[ob->waitdoornum - 1].action != dr_open)
                return;
            // go ahead, the door is now open
            ob->distance = TILEGLOBAL;
            ob->waitdoornum = 0;
        }

        if (move < ob->distance)
        {
            MoveObj (ob,move);
            break;
        }

        //
        // reached goal tile, so select another one
        //

        //
        // fix position to account for round off during moving
        //
        OB_SetX(ob, ((int32_t)ob->tilex<<TILESHIFT)+TILEGLOBAL/2);
        OB_SetY(ob, ((int32_t)ob->tiley<<TILESHIFT)+TILEGLOBAL/2);

        move -= ob->distance;

        if (dodge)
            SelectDodgeDir (ob);
        else
            SelectChaseDir (ob);

        if (ob->dir == nodir)
            return;                                                 // object is blocked in
    }

    CheckGoal(ob);
}


/*
=================
=
= T_Ghosts
=
=================
*/

void T_Ghosts (objtype *ob)
{
    int32_t move;

    if (ob->dir == nodir)
    {
        SelectNextDir(ob, (byte)player->tilex, (byte)player->tiley);
        if (ob->dir == nodir)
            return;                                                 // object is blocked in
    }

    move = ob->speed*tics;

    while (move)
    {
        if (move < ob->distance)
        {
            MoveObj (ob,move);
            break;
        }

        //
        // reached goal tile, so select another one
        //

        //
        // fix position to account for round off during moving
        //
        OB_SetX(ob, ((int32_t)ob->tilex<<TILESHIFT)+TILEGLOBAL/2);
        OB_SetY(ob, ((int32_t)ob->tiley<<TILESHIFT)+TILEGLOBAL/2);

        move -= ob->distance;

        SelectNextDir(ob, (byte)player->tilex, (byte)player->tiley);

        if (ob->dir == nodir)
            return;                                                 // object is blocked in
    }
}

/*
=================
=
= T_DogChase
=
=================
*/

void T_DogChase (objtype *ob)
{
    int32_t    move;

    if (ob->dir == nodir)
    {
        SelectDodgeDir (ob);
        if (ob->dir == nodir)
            return;                                                 // object is blocked in
    }

    move = ob->speed*tics;

    while (move)
    {
        //
        // check for melee range
        //
        if (CheckMeleeAttack(ob,move))
            return;

        if (move < ob->distance)
        {
            MoveObj (ob,move);
            break;
        }

        //
        // reached goal tile, so select another one
        //

        //
        // fix position to account for round off during moving
        //
        OB_SetX(ob, ((int32_t)ob->tilex<<TILESHIFT)+TILEGLOBAL/2);
        OB_SetY(ob, ((int32_t)ob->tiley<<TILESHIFT)+TILEGLOBAL/2);

        move -= ob->distance;

        SelectDodgeDir (ob);

        if (ob->dir == nodir)
            return;                                                 // object is blocked in
    }
}

/*
=================
=
= ValidGround
=
= WSJ: determines if the enemy (ob) can burrow into
= the floor texture at tilex,tiley
=
=================
*/

boolean ValidGround (objtype *ob, int tilex, int tiley)
{
    int floor;

    if (TILEDIST(tilex, tiley, player->tilex, player->tiley) < 2)
        return false;

    floor = MAPSPOT(tilex, tiley, 2) & 0xff;
    if (ob->enemy->waterMut == NULL && ED_CanBurrowFloor(ed, floor))
    {
        return true;
    }
    else if (ob->enemy->waterMut != NULL && ED_FloorIsWater(ed, floor))
    {
        return true;
    }

    return false;
}

/*
=================
=
= CheckGround
=
= WSJ: if true, start digging in/out of underground
=
=================
*/

boolean CheckGround (objtype *ob, boolean digdown)
{
    int i;
    int tilex, tiley;
    boolean doDig;

    if (ob->reaction == react_findmg42)
        return false;

    doDig = false;
    tilex = GLOBALTILE(OB_GetX(ob));
    tiley = GLOBALTILE(OB_GetY(ob));

    if (digdown)
    {
        if (ValidGround(ob, tilex, tiley))
        {
            for (i = 0; i < 4; i++)
            {
                tilex = GLOBALTILE(OB_GetX(ob) + 0x8000 * 
                    ((i & 1) * 2 - 1));
                tiley = GLOBALTILE(OB_GetY(ob) + 0x8000 * 
                    ((i / 2) * 2 - 1));
                if (!ValidGround(ob, tilex, tiley))
                {
                    break;
                }
            }
            if (i == 4)
            {
                doDig = true;
            }
        }
    }
    else
    {
        doDig = ValidGround(ob, tilex, tiley);
    }

    if (doDig)
    {
        if (digdown)
        {
            ob->flags &= ~FL_SHOOTABLE;
            NewState(ob,ED_STATE(ob->enemy->burrow.digInState));
        }
        else
        {
            A_BecomeBlocking(ob);
            NewState(ob,ED_STATE(ob->enemy->burrow.digOutState));
        }
    }

    return doDig;
}

/*
=================
=
= T_Underground
=
= WSJ: Undead stealth chase
=
=================
*/

void T_Underground (objtype *ob)
{
    int32_t move;
    boolean dodge;

    ob->hidden = 2;
    dodge = false;

    if (US_RndT()&2)
        dodge = true;

    if (ob->dir == nodir)
    {
        if (dodge)
            SelectDodgeDir (ob);
        else
            SelectChaseDir (ob);
        if (ob->dir == nodir)
            return;                                                 // object is blocked in
    }

    move = ob->speed*tics;

    while (move)
    {
        if (move < ob->distance)
        {
            MoveObj (ob,move);
            break;
        }

        //
        // reached goal tile, so select another one
        //

        //
        // fix position to account for round off during moving
        //
        OB_SetX(ob, ((int32_t)ob->tilex<<TILESHIFT)+TILEGLOBAL/2);
        OB_SetY(ob, ((int32_t)ob->tiley<<TILESHIFT)+TILEGLOBAL/2);

        move -= ob->distance;

        if (OBJTILEDIST(ob,player) <2)
            SelectRunDir (ob);
        else if (dodge)
            SelectDodgeDir (ob);
        else
            SelectChaseDir (ob);

        if (ob->dir == nodir)
            return;                                                 // object is blocked in
    }
}


/*
============================================================================

                                    PATH

============================================================================
*/


/*
===============
=
= SelectPathDir
=
===============
*/

void SelectPathDir (objtype *ob)
{
    unsigned spot;

    spot = MAPSPOT(ob->tilex,ob->tiley,1)-ICONARROWS;

    if (spot<NUMICONARROWS)
    {
        // new direction
        ob->dir = (dirtype)(spot);
    }

    ob->distance = TILEGLOBAL;

    if (!TryWalk (ob))
        ob->dir = nodir;
}


/*
===============
=
= T_Path
=
===============
*/

void T_Path (objtype *ob)
{
    int32_t    move;

    if (SightPlayer (ob,NULL))
        return;

    if (ob->dir == nodir)
    {
        SelectPathDir (ob);
        if (ob->dir == nodir)
            return;                                 // all movement is blocked
    }


    move = ob->speed*tics;

    while (move)
    {
        if (ob->distance < 0)
        {
            //
            // waiting for a door to open
            //
            OpenDoor (-ob->distance-1);
            if (doorobjlist[-ob->distance-1].action != dr_open)
                return;
            ob->distance = TILEGLOBAL;      // go ahead, the door is now open
            DEMOIF_SDL
            {
                TryWalk(ob);
            }
        }

        if (move < ob->distance)
        {
            MoveObj (ob,move);
            break;
        }

        if (ob->tilex>MAPSIZE || ob->tiley>MAPSIZE)
        {
            sprintf (str, "T_Path hit a wall at %u,%u, dir %u",
                ob->tilex,ob->tiley,ob->dir);
            Quit (str);
        }

        OB_SetX(ob, ((int32_t)ob->tilex<<TILESHIFT)+TILEGLOBAL/2);
        OB_SetY(ob, ((int32_t)ob->tiley<<TILESHIFT)+TILEGLOBAL/2);
        move -= ob->distance;

        SelectPathDir (ob);

        if (ob->dir == nodir)
            return;                                 // all movement is blocked
    }
}


/*
=============================================================================

                                    FIGHT

=============================================================================
*/


/*
=================
=
= Refire
=
= WSJ: for looped attacks (ex: machine gun fire)
=
=================
*/

void Refire (objtype *ob, short maxrepeat, statetype *state, void (*Action)(objtype *))
{
    if (++ob->temp1 >= maxrepeat || !CheckLine(ob))
        return;

    NewState(ob, state);

    if (Action != NULL)
        Action (ob);
}


/*
============
=
= T_Aim
=
============
*/

void T_Aim (objtype *ob)
{
    int32_t deltax,deltay;
    float   angle;
    int     iangle;

    deltax = OB_GetX(player) - OB_GetX(ob);
    deltay = OB_GetY(ob) - OB_GetY(player);
    angle = (float) atan2((float) deltay, (float) deltax);
    if (angle<0)
        angle = (float) (M_PI*2+angle);
    iangle = (int) (angle/(M_PI*2)*FINEANGLES);
    ob->fineangle = iangle;
    ob->angle = ob->fineangle * ANGLES / FINEANGLES;
}

void T_Aim2 (objtype *ob)
{
    if (CheckLine(ob))
        T_Aim(ob);
}


/*
===============
=
= CheckRangeAttack
=
= WSJ: checks if the enemy has a shooting animation.
= If so, then return true and put the enemy into his shooting state.
=
===============
*/

boolean CheckRangeAttack (objtype *ob)
{
    int32_t target;
    int     chance;
    int     dist = OBJTILEDIST(ob,player);
    const ED_EnemyChanceConfig_t *enemyChanceConfig;
    watermutobjtype *watermut;

    watermut = &ob->u.watermut;

    if (dist)
    {
        switch (ob->enemy->rangeAttackVariant)
        {
        case ED_EnemyRangeAttackVariant(Normal):
            chance = (tics<<5)/dist;
            break;
        case ED_EnemyRangeAttackVariant(Long):
            chance = (tics<<3); // WSJ: enemy has a long range weapon
            break;
        }
    }
    else
        chance = 300;

    if (dist == 1)
    {
        target = abs(OB_GetX(ob) - OB_GetX(player));
        if (target < 0x14000l)
        {
            target = abs(OB_GetY(ob) - OB_GetY(player));
            if (target < 0x14000l)
                chance = 300;
        }
    }

    if (chance<300 && (ob->reaction == react_findmg42 || 
        OB_CanDefendGuardedObj(ob)))
    {
        chance /= 4;
    }

    enemyChanceConfig = &ob->enemy->shootChanceConfig;
    if (enemyChanceConfig->timeout > 0)
    {
        ob->shootTimer += TICS2MILLI(tics);
        if (ob->shootTimer < enemyChanceConfig->timeout)
        {
            chance = 0;
        }
        else
        {
            ob->shootTimer = 0;
            chance = OB_ComputeChance(ob, 
                enemyChanceConfig->ranges);
        }
    }

    if ( US_RndT()<chance && !invisiblemode)
    {
        if (ob->enemy->altShootChance > 0 &&
            US_RndT() < ob->enemy->altShootChance * 255 / 100 &&
            ED_STATE(ob->enemy->altShootState))
        {
            NewState (ob,ED_STATE(ob->enemy->altShootState));
            return true;
        }

        //
        // go into attack frame
        //
        ob->temp1 = 0;
        int deadObjNum = -1;
        if (ob->enemy->summoner != NULL && 
            US_RndT() < ob->enemy->summoner->summonChance * 255 / 100 &&
            OB_CheckSummonBodyInSight(ob, deadObjNum))
        {
            PlayActorVoice(ED_SND(ob->enemy->summoner->summonSound), ob);
            NewState (ob,ED_STATE(ob->enemy->summoner->summonState));
            return true;
        }
        else if (ob->enemy->waterMut != NULL && watermut->waterbound)
        {
            NewState (ob,ED_STATE(ob->enemy->waterMut->shootState));
            return true;
        }
        else if (ob->enemy->defenseChance > 0 &&
            US_RndT() < ob->enemy->defenseChance * 255 / 100 &&
            ED_STATE(ob->enemy->defenseState))
        {
            ob->flags |= FL_NODAMAGE;
            NewState (ob,ED_STATE(ob->enemy->defenseState));
            return true;
        }
        else if (ED_STATE(ob->enemy->shootState))
        {
            NewState (ob,ED_STATE(ob->enemy->shootState));
            return true;
        }
    }

    return false;
}


/*
===============
=
= CheckMeleeAttack
=
= WSJ: checks if the enemy can use a melee attack.
= If so, then return true and put the enemy into his melee state.
=
===============
*/

boolean CheckMeleeRange (objtype *ob, int32_t move)
{
    int32_t    dx,dy;

    if (invisiblemode)
        return false;

    dx = OB_GetX(player) - OB_GetX(ob);
    if (dx<0)
        dx = -dx;
    dx -= move;
    if (dx <= MINACTORDIST)
    {
        dy = OB_GetY(player) - OB_GetY(ob);
        if (dy<0)
            dy = -dy;
        dy -= move;
        if (dy <= MINACTORDIST)
            return true;
    }
    return false;
}

boolean CheckMeleeAttack (objtype *ob, int32_t move)
{
    ED_State_t *meleeState;

    if (CheckMeleeRange(ob,move))
    {
        //
        // go into attack frame
        //
        meleeState = ED_STATE(ob->enemy->meleeState);
        if (meleeState != NULL)
        {
            NewState (ob,meleeState);
            return true;
        }
    }

    return false;
}


/*
===============
=
= T_Melee
=
===============
*/

void T_Melee (objtype *ob)
{
    bool hit;
    int damagePoints;
    ED_Melee_t *melee;

    hit = false;

    if (CheckMeleeRange(ob,TILEGLOBAL))
    {
        if ((ob->enemy->meleeDamageChance != 0 && 
            US_RndT() <= ob->enemy->meleeDamageChance * 255 / 100) || 
            US_RndT() < 180)
        {
            damagePoints = ob->enemy->meleeDamageBase;
            if (ob->enemy->meleeDamageDivisor != 0)
            {
                damagePoints += US_RndT() / ob->enemy->meleeDamageDivisor;
            }
            else
            {
                damagePoints += US_RndT()>>4;
            }
            TakeSpecialDamage (damagePoints,ob,dam_melee,false,true);

            hit = true;
        }
    }

    melee = ob->enemy->melee;
    if (melee)
    {
        PlayActorVoice(ED_SND(hit ? melee->hitSound : melee->missSound), ob);
    }
}


/*
===============
=
= ProjFire
=
= WSJ: projectile attack
=
===============
*/

void ProjFire (objtype *ob, short angleoffset, const ED_Proj_t *proj, statetype *state, int32_t speed, short spriteoffset)
{
    int iangle;
    const ED_ProjPhys_t *projPhys;
    physobjtype *physobj;

    if (proj->shootSound[0] != '\0')
    {
        SD_PlaySound (ED_SND(proj->shootSound));
    }

    projPhys = ED_ProjPhys(ed, proj->physName);

    if (state == NULL)
    {
        state = ED_STATE(proj->startState);
    }

    GetNewActor ();
    OB_SetState(newobj, state);
    newobj->ticcount = 1;

    newobj->tilex = ob->tilex;
    newobj->tiley = ob->tiley;
    OB_SetX(newobj, OB_GetX(ob));
    OB_SetY(newobj, OB_GetY(ob));
    newobj->dir = nodir;

    iangle = ob->fineangle;
    iangle += (int)angleoffset * FINEANGLES / ANGLES;
    if (projPhys != NULL)
    {
        iangle += ED_RandomInt_Eval(&projPhys->angleOffset) *
            FINEANGLES / ANGLES;
    }
    if (iangle<0)
        iangle+=FINEANGLES;
    else if (iangle>=FINEANGLES)
        iangle-=FINEANGLES;

    newobj->fineangle = iangle;
    newobj->angle = newobj->fineangle * ANGLES / FINEANGLES;
    newobj->projIndex = ED_ProjIndex(ed, proj->name);
    newobj->proj = proj;
    newobj->speed = speed;
    newobj->flags = FL_NEVERMARK;
    newobj->active = ac_yes;
    newobj->sproffset = spriteoffset +	
        ED_SPR(proj->spriteOffset.newStart) -
        ED_SPR(proj->spriteOffset.oldStart);
    newobj->projEnemyIndex = ob->enemyIndex;

    if (projPhys != NULL)
    {
        lwlib::Vector3f u;
        lwlib::Mat3f rotMat;

        physobj = &newobj->u.phys;
        physobj->pos = lwlib::vec3f(fixedpt_toconst(OB_GetX(newobj)),
            fixedpt_toconst(OB_GetY(newobj)), 0.0);

        u = lwlib::vec3f(1.0, 0.0, 0.0);

        const double pitch = -ED_RandomConst_Eval(&projPhys->pitch);
        rotMat = lwlib::mat3f_rot(pitch, 1);
        u = lwlib::mat_xform(rotMat, u);

        rotMat = lwlib::mat3f_rot(-(double)newobj->fineangle * ANGLES / FINEANGLES, 2);
        u = lwlib::mat_xform(rotMat, u);
        physobj->vel = lwlib::vec_scale(u, 
            ED_RandomConst_Eval(&projPhys->speed));
    }
}

//
// cagemut
//

void T_CageMutWait(objtype *ob)
{
    int dist;
    bool doAwake;
    int chance;
    int rndval;
    cagemutobjtype *cagemutobj = &ob->u.cagemut;
    ED_State_t *chaseState;
    ED_EnemyConfigCheck_t *check;

    check = &(ED_EnemyConfig(ed)->wakeCheck);
    
    cagemutobj->waitTimer += TICS2MILLI(tics);
    if (cagemutobj->waitTimer < check->timeout)
    {
        return;
    }
    cagemutobj->waitTimer = 0;

    doAwake = false;
    dist = OBJTILEDIST(ob, player);

    if ((ob->surpriseatk == surpriseatk_delay ||
        ob->surpriseatk == surpriseatk_prox_delay) &&
        cagemutobj->sighted)
    {
        cagemutobj->delayTimer += check->timeout;
        if (cagemutobj->delayTimer > check->delayTime)
        {
            doAwake = true;
        }
    }

    if (ob->killActorAlert)
    {
        doAwake = true;
    }

    if (!CheckLine(ob) && !doAwake)
    {
        if (ob->surpriseatk == surpriseatk_out_of_sight && 
            cagemutobj->sighted)
        {
            doAwake = true;
        }
        else
        {
            return;
        }
    }

    switch (ob->surpriseatk)
    {
        case surpriseatk_default:
            chance = OB_ComputeChance(ob, check->chanceRanges);

            rndval = US_RndT();
            if (rndval < chance)
            {
                doAwake = true;
            }
            break;
        case surpriseatk_hit:
            break;
        case surpriseatk_delay:
            cagemutobj->sighted = true;
            break;
        case surpriseatk_prox:
            if (dist <= 4)
            {
                doAwake = true;
            }
            break;
        case surpriseatk_prox_delay:
            if (dist <= 4)
            {
                cagemutobj->sighted = true;
            }
            break;
        case surpriseatk_out_of_sight:
            cagemutobj->sighted = true;
            break;

        default:
            break;
    }

    if (doAwake)
    {
        A_AlertSound (ob);

        chaseState = ED_STATE(ob->enemy->chaseState);
        if (chaseState)
        {
            NewState(ob, chaseState);
        }

        ob->flags |= FL_ATTACKMODE;
    }
}

void T_CageMutAwake(objtype *ob)
{
    bool doEscape;
    int chance;
    int rndval;
    ED_State_t *escapeState;
    cagemutobjtype *cagemutobj = &ob->u.cagemut;
    ED_EnemyConfigCheck_t *check;

    check = &(ED_EnemyConfig(ed)->escapeCheck);
    
    if (gamestate.victoryflag)
        return;

    if (CheckLine(ob))
    {
        if (CheckRangeAttack(ob))
            return;
    }

    doEscape = false;
    if (!CheckLine(ob))
    {
        doEscape = true;
    }
    else
    {
        cagemutobj->escapeTimer += TICS2MILLI(tics);
        if (cagemutobj->escapeTimer < check->timeout)
        {
            return;
        }
        cagemutobj->escapeTimer = 0;

        chance = OB_ComputeChance(ob, check->chanceRanges);

        rndval = US_RndT();
        if (rndval < chance)
        {
            doEscape = true;
        }
    }

    if (doEscape)
    {
        ob->flags |= FL_UNKILLABLE;
        ob->flags |= FL_SUPPRESSPAIN;
        escapeState = ED_STATE(ob->enemy->escapeState);
        if (escapeState)
        {
            NewState(ob, escapeState);
        }
    }
}

static void A_CageMutSpawnMutant(objtype *ob)
{
    fixed u, v;

    u = fixedpt_fromint(1) / 4;
    v = 0;
    if (ob->enemy == ED_Enemy(ed, "en_laddermut"))
    {
        v = fixedpt_fromint(1) / 4;
    }

    OB_SpawnSurpriseEnemy(ob, u, v, false);
}

void A_CageMutEscapeSound(objtype *ob)
{
    PlayActorVoice(ED_SND("CAGEMUTESCAPESND"), ob);
}

void A_CageMutFinishEscape(objtype *ob)
{
    A_RemoveShootable(ob);
    A_CageMutSpawnMutant(ob);
}

//
// ropemut
//

void T_RopeMutWait(objtype *ob)
{
    bool doAwake;
    int chance;
    int rndval;
    int dist;
    ropemutobjtype *ropemutobj = &ob->u.ropemut;
    ED_EnemyConfigCheck_t *check;
    ED_State_t *chaseState;

    check = &(ED_EnemyConfig(ed)->wakeCheck);
    
    ropemutobj->waitTimer += TICS2MILLI(tics);
    if (ropemutobj->waitTimer < check->timeout)
    {
        return;
    }
    ropemutobj->waitTimer = 0;

    doAwake = false;
    dist = OBJTILEDIST(ob, player);

    if ((ob->surpriseatk == surpriseatk_delay ||
        ob->surpriseatk == surpriseatk_prox_delay) &&
        ropemutobj->sighted)
    {
        ropemutobj->delayTimer += check->timeout;
        if (ropemutobj->delayTimer > check->delayTime)
        {
            doAwake = true;
        }
    }

    if (ob->killActorAlert)
    {
        doAwake = true;
    }

    if (!CheckLine(ob) && !doAwake)
    {
        if (ob->surpriseatk == surpriseatk_out_of_sight && 
            ropemutobj->sighted)
        {
            doAwake = true;
        }
        else
        {
            return;
        }
    }

    if (actorat[ob->tilex][ob->tiley] != NULL)
    {
        return;
    }

    switch (ob->surpriseatk)
    {
        case surpriseatk_default:
            chance = OB_ComputeChance(ob, check->chanceRanges);

            rndval = US_RndT();
            if (rndval < chance)
            {
                doAwake = true;
            }
            break;
        case surpriseatk_hit:
            break;
        case surpriseatk_delay:
            ropemutobj->sighted = true;
            break;
        case surpriseatk_prox:
            if (dist <= 4)
            {
                doAwake = true;
            }
            break;
        case surpriseatk_prox_delay:
            if (dist <= 4)
            {
                ropemutobj->sighted = true;
            }
            break;
        case surpriseatk_out_of_sight:
            ropemutobj->sighted = true;
            break;

        default:
            break;
    }

    if (doAwake)
    {
        ob->flags |= FL_UNKILLABLE;
        A_AlertSound (ob);
        A_BecomeShootable(ob);
        A_BecomeBlocking(ob);
        chaseState = ED_STATE(ob->enemy->chaseState);
        if (chaseState)
        {
            NewState(ob, chaseState);
        }

        ob->flags |= FL_ATTACKMODE;
    }
}

void T_RopeMutFinishSlide(objtype *ob)
{
    A_RemoveBlocking(ob);
    A_RemoveShootable(ob);
    A_CageMutSpawnMutant(ob);
}

//
// laddermut
//

void T_LadderMutWait(objtype *ob)
{
    T_CageMutWait(ob);

    if (ob->state == ED_STATE(ob->enemy->chaseState))
    {
        A_BecomeShootable(ob);
        ob->flags |= FL_SUPPRESSPAIN;
        ob->flags |= FL_ATTACKMODE;
    }
}

void T_LadderMutFinishClimb(objtype *ob)
{
    ob->flags &= ~FL_SUPPRESSPAIN;
}

//
// knightskel
//

void T_KnightSkelWait(objtype *ob)
{
    int dist;
    bool doAwake;
    int chance;
    int rndval;
    knightskelobjtype *knightskelobj = &ob->u.knightskel;
    ED_State_t *chaseState;
    ED_EnemyConfigCheck_t *check;

    check = &(ED_EnemyConfig(ed)->wakeCheck);
    
    knightskelobj->waitTimer += TICS2MILLI(tics);
    if (knightskelobj->waitTimer < check->timeout)
    {
        return;
    }
    knightskelobj->waitTimer = 0;

    doAwake = false;
    dist = OBJTILEDIST(ob, player);

    if ((ob->surpriseatk == surpriseatk_delay ||
        ob->surpriseatk == surpriseatk_prox_delay) &&
        knightskelobj->sighted)
    {
        knightskelobj->delayTimer += TICS2MILLI(tics);
        if (knightskelobj->delayTimer > 300)
        {
            doAwake = true;
        }
    }

    if (!CheckLine(ob) && !doAwake)
    {
        if (ob->surpriseatk == surpriseatk_out_of_sight && 
            knightskelobj->sighted)
        {
            doAwake = true;
        }
        else
        {
            return;
        }
    }

    switch (ob->surpriseatk)
    {
        case surpriseatk_default:
            chance = OB_ComputeChance(ob, check->chanceRanges);

            rndval = US_RndT();
            if (rndval < chance)
            {
                doAwake = true;
            }
            break;
        case surpriseatk_hit:
            break;
        case surpriseatk_delay:
            knightskelobj->sighted = true;
            break;
        case surpriseatk_prox:
            if (dist <= 4)
            {
                doAwake = true;
            }
            break;
        case surpriseatk_prox_delay:
            if (dist <= 4)
            {
                knightskelobj->sighted = true;
            }
            break;
        case surpriseatk_out_of_sight:
            knightskelobj->sighted = true;
            break;
        default:
            break;
    }

    if (doAwake)
    {
        chaseState = ED_STATE(ob->enemy->chaseState);
        if (chaseState)
        {
            NewState(ob, chaseState);
        }
    }
}

void A_KnightSkelFinishDrop(objtype *ob)
{
    A_RemoveShootable(ob);
    A_RemoveBlocking(ob);
    A_CageMutSpawnMutant(ob);
}

//
// hangedmut
//

void A_HangedMutFinishDrop(objtype *ob)
{
    A_CageMutFinishEscape(ob);
    A_RemoveBlocking(ob);
}

//
// brownhorse
//

void A_BrownHorseRandomDelay(objtype *ob)
{
    ob->ticcount = US_RndT_Range(200, 800);
}

void A_BrownHorseResetAnimalTimer(objtype *ob)
{
    ob->enemy->animalSoundTimer = gamestate.TimeCount;
    ob->enemy->animalSoundPeriod = US_RndT_Range(1000, 2000);
}

void A_BrownHorsePlaySound(objtype *ob)
{
    ED_Enemy_t *enemy;
    const ED_EnemyDeath_t *death;

    enemy = ob->enemy;
    death = &enemy->death;

    if (ED_SND(death->screamSound) != -1)
    {
        PlayActorVoice(ED_SND(death->screamSound), ob);
    }
    else
    {
        if (enemy->animalSoundTimer == 0 || 
            gamestate.TimeCount < enemy->animalSoundTimer)
        {
            A_BrownHorseResetAnimalTimer(ob);
        }
        if (gamestate.TimeCount >= enemy->animalSoundTimer + 
            enemy->animalSoundPeriod)
        {
            A_BrownHorseResetAnimalTimer(ob);
            PlayActorVoice(ED_SND(enemy->alertSound), ob);
        }
    }
}

//
// barrel
//

void A_BarrelSplashDamage(objtype *ob)
{
    objtype *o;
    fixed radius;
    fixed radius_sq;			// damage radius^2
    unsigned damage = 0;		// amount

    Tremor_ObjStart(trem_explosion, ob);

    // Explosive barrel splash damage
    damage = 70 + US_RndT() * 30 / 256;
    radius = FP(5);

    if ( !damage )
    {
        return;
    }

    radius_sq = fixedpt_mul(radius, radius);

    // for each actor
    for (ITEROBJ(o, 0))
    {
        fixed dx, dy;
        fixed dist_sq;
        fixed dist;
        fixed dmg;
        int scaled_dmg;

        if (o == ob)
        {
            continue;
        }

        if ( o != player && 
            (o->flags & FL_SHOOTABLE) == 0 )
        {
            continue;
        }

        if (!CheckObjLine(ob, o))
        {
            continue;
        }

        dx = o->x - ob->x;
        dy = o->y - ob->y;
        dist_sq = FP_SQ(dx) + FP_SQ(dy);
        if ( dist_sq > radius_sq )
        {
            continue;
        }

        // We want a chain reaction. So we silently decrease radius
        // to 1 for other barrels.
        if ( o->enemy == ob->enemy && dist_sq > FP_ONE )
        {
            dist_sq = FP_ONE;
        }

        // scale damage proportionally to distance
        // FIXME: linear distribution is probably bad
        dist = fixedpt_sqrt( dist_sq );
        dmg = fixedpt_div( (radius - dist)*damage, radius );
        scaled_dmg = MAX( 0, FP_TRUNC( dmg ) );

        if ( !scaled_dmg )
        {
            continue;
        }

        // damage now
        if ( o == player )
        {
            TakeDamage( scaled_dmg, ob );
        }
        else
        {
            DamageActor( o, ob, scaled_dmg, dam_rocket );
        }
    }
}

void A_BarrelDebrisFall(objtype *ob)
{
    A_RemoveBlocking(ob);
    actorat[ob->tilex][ob->tiley] = NULL;
    ob->flags |= FL_NONMARK | FL_NEVERMARK;
}

void T_BarrelDebrisHurt(objtype *ob)
{
    objtype *damageObj;

    damageObj = OB_GetObjOn(ob);

    if (damageObj != NULL && damageObj->enemy->boss == NULL)
    {
        NewState(ob, ED_STATE(ob->enemy->shootState));
        OB_Damage(damageObj, ob, ob->enemy->speed, dam_fire, false);
    }
}

//
// greentoxbarrel
//

void A_GreenToxBarrelHurt(objtype *ob)
{
    objtype *damageObj;

    damageObj = OB_GetObjOn(ob);

    if (damageObj != NULL && 
        (damageObj->flags & FL_SHOOTABLE) != 0)
    {
        OB_Damage(damageObj, ob, ob->enemy->speed, dam_mutagen, true);
    }
}

//
// damagemut
//

void T_DamageMut(objtype *ob)
{
    while (ob->hitpoints < ob->starthitpoints)
    {
        ob->hitpoints += 60;
        NewState(ob, ED_STATE(ob->state->next));
        if (ob->state->tictime != 0)
        {
            break;
        }
    }
}

//
// tablemutdamage
//

void T_TableMutDamage(objtype *ob)
{
    while (ob->hitpoints < ob->starthitpoints)
    {
        ob->hitpoints += 60;
        NewState(ob, ED_STATE(ob->state->next));

        if (ob->hitpoints < ob->starthitpoints && 
            ob->state->tictime == 10)
        {
            NewState(ob, ED_STATE(ob->state->next));
        }
        else
        {
            if (ob->state->tictime == 1)
            {
                break;
            }
        }
    }
}

//
// villager
//

// villager with arms crossed with random slouch
//
// SPR_VILLAGER1_1
// randomly triggers SPR_VILLAGER1_2 -> SPR_VILLAGER1_3 -> SPR_VILLAGER1_1

// villager with arms crossed with random shrug
//
// SPR_VILLAGER1_1
// randomly triggers SPR_VILLAGER1_14 -> SPR_VILLAGER1_1

// unsure villager with arms down
// can be questioned
//
// SPR_VILLAGER1_15
// triggers SPR_VILLAGER1_11 -> ... -> SPR_VILLAGER1_4 -> SPR_VILLAGER1_15 when
// used

// answering villager with arms down
//
// SPR_VILLAGER1_15
// triggers SPR_VILLAGER1_10 -> SPR_VILLAGER1_11 -> SPR_VILLAGER1_15 when used

// refusing villager with arms down
//
// SPR_VILLAGER1_15
// triggers SPR_VILLAGER1_12 -> SPR_VILLAGER1_13 -> SPR_VILLAGER1_15 when used

// answering villager in grey shirt
//
// SPR_VILLAGER2_1

lwlib_EnumStr(
	villagertype,
    "villager_armscrossed_slouch",
    "villager_armscrossed_shrug",
    "villager_armsdown_unsure",
    "villager_armsdown_answering",
    "villager_armsdown_refusing",
    "villager_grey",
    "villager_grey2",
);

static void VillagerArmsDownUnsure_Start(objtype *ob)
{
    NewState(ob, ED_STATE(ob->enemy->meleeState));
}

static void VillagerArmsDownUnsure_Think(objtype *ob)
{
    if (OB_CheckUsed(ob))
    {
        NewState(ob, ED_STATE(ob->enemy->chaseState));
    }
}

static void VillagerArmsDownAnswering_Start(objtype *ob)
{
    NewState(ob, ED_STATE(ob->enemy->meleeState));
}

static void VillagerArmsDownAnswering_Think(objtype *ob)
{
    if (OB_CheckUsed(ob))
    {
        NewState(ob, ED_STATE(ob->enemy->death.gibState));
    }
}

static void VillagerArmsDownRefuse_Start(objtype *ob)
{
    NewState(ob, ED_STATE(ob->enemy->meleeState));
}

static void VillagerArmsDownRefuse_Think(objtype *ob)
{
    if (OB_CheckUsed(ob))
    {
        NewState(ob, ED_STATE(ob->enemy->death.headShotState));
    }
}

static void VillagerArmsCrossedSlouch_Start(objtype *ob)
{
    NewState(ob, ED_STATE(ob->enemy->escapeState));
}

static void VillagerArmsCrossedSlouch_Think(objtype *ob)
{
    if (OB_CheckUsed(ob))
    {
        A_VillagerFinishResponse(ob);
    }
}

static void VillagerArmsCrossedShrug_Start(objtype *ob)
{
    NewState(ob, ED_STATE(ob->enemy->death.burnState));
}

static void VillagerArmsCrossedShrug_Think(objtype *ob)
{
    if (OB_CheckUsed(ob))
    {
        A_VillagerFinishResponse(ob);
    }
}

static void VillagerGrey_Start(objtype *ob)
{
    NewState(ob, ED_STATE(ob->enemy->death.dieState));
}

static void VillagerGrey_Think(objtype *ob)
{
    if (OB_CheckUsed(ob))
    {
        A_VillagerFinishResponse(ob);
    }
}

typedef struct
{
    void (*start)(objtype *ob);
    void (*think)(objtype *ob);
} villagerfunctable_t;

static villagerfunctable_t villagerfunctables[villager_max] =
{
    {
        VillagerArmsCrossedSlouch_Start, // start
        VillagerArmsCrossedSlouch_Think, // think
    }, // villager_armscrossed_slouch
    {
        VillagerArmsCrossedShrug_Start, // start
        VillagerArmsCrossedShrug_Think, // think
    }, // villager_armscrossed_shrug
    {
        VillagerArmsDownUnsure_Start, // start
        VillagerArmsDownUnsure_Think, // think
    }, // villager_armsdown_unsure
    {
        VillagerArmsDownAnswering_Start, // start
        VillagerArmsDownAnswering_Think, // think
    }, // villager_armsdown_answering
    {
        VillagerArmsDownRefuse_Start, // start
        VillagerArmsDownRefuse_Think, // think
    }, // villager_armsdown_refusing
    {
        VillagerGrey_Start, // start
        VillagerGrey_Think, // think
    }, // villager_grey
    {
        VillagerGrey_Start, // start
        VillagerGrey_Think, // think
    }, // villager_grey2
};

void A_VillagerStart(objtype *ob)
{
    int area;
    villagertype type;
    villagerobjtype *villager;
    villagerfunctable_t *fntable;

    type = villager_armscrossed_slouch;

    area = MAPSPOT(ob->tilex, ob->tiley, 0) - AREATILE;
    if (area >= 0 && area < (int)villager_max)
    {
        type = (villagertype)area;
    }

    villager = &ob->u.villager;
    villager->type = type;

    fntable = &villagerfunctables[villager->type];
    if (fntable->start)
    {
        fntable->start(ob);
    }
}

void T_Villager(objtype *ob)
{
    villagerobjtype *villager;
    villagerfunctable_t *fntable;

    villager = &ob->u.villager;

    fntable = &villagerfunctables[villager->type];
    if (fntable->think)
    {
        fntable->think(ob);
    }
}

void A_VillagerFinishResponse(objtype *ob)
{
    villagerobjtype *villager;
    villagerfunctable_t *fntable;

    villager = &ob->u.villager;
    LogDiscScreens(villagertype_Str[villager->type]);

    fntable = &villagerfunctables[villager->type];
    if (fntable->start)
    {
        fntable->start(ob);
    }
}

void A_VillagerRandomDelay(objtype *ob)
{
    ob->ticcount = US_RndT_Range(200, 800);
}

//
// mech mutant 1
//

void A_MechMut1SpawnSuit(objtype *ob)
{
    OB_SpawnSurpriseEnemy(ob, 0, 0, true);
}

//
// mech mutant 2
//

void A_MechMut2Evacuate(objtype *ob)
{
    A_RemoveBlocking(ob);
    OB_SpawnSurpriseEnemy(ob, 0, FP(1) / 4, true);
    OB_SpawnSurpriseEnemy(ob, 0, -FP(1) / 4, true);
}

// mech mutant 4

void A_MechMut4Evacuate(objtype *ob)
{
    A_RemoveBlocking(ob);
    A_SpawnSurpriseEnemy(ob);
}

void T_MechMut4Chase(objtype *ob)
{
    T_Chase(ob);

    if (ob->hitpoints < MECHA_HP_ONEARM(ob->starthitpoints))
    {
        SpawnBulletPuff(OB_GetX(ob), OB_GetY(ob),
            ED_STATE("s_mechboom1"), 0, 0);
        newobj->ticcount = 6;
        T_Aim(newobj);
        OB_AngleMove(newobj, FP(1) / 4, FP(1) / 4);

        NewState(ob, ED_STATE("s_mechmut4onearmchase1") + 
            (ob->state - ED_STATE("s_mechmut4chase1")));
    }
}

void T_MechMut4OneArmChase(objtype *ob)
{
    T_Chase(ob);

    if (ob->state == ED_STATE("s_mechmut4shoot1"))
    {
        NewState(ob, ED_STATE("s_mechmut4onearmshoot1"));
    }
}

void A_MechMut4NoArms(objtype *ob)
{
    SpawnBulletPuff(OB_GetX(ob), OB_GetY(ob),
        ED_STATE("s_mechboom1"), 0, 0);
    newobj->ticcount = 6;
    T_Aim(newobj);
    OB_AngleMove(newobj, FP(1) / 4, -FP(1) / 4);

    T_Aim(ob);
    ProjFire (ob,0,ED_Proj(ed, "mechmut4noarms"),NULL,0x2000l,0);
}


void A_MechMut4Explode(objtype *ob)
{
    PlayActorVoice(ED_SND("EXPLOSIONSND"), ob);
}

void A_MechaSound(objtype *ob)
{
    PlayActorVoice(ED_SND("FSTEP_MECH_SND"), ob);
    Tremor_ObjStart(trem_mechastep, ob);
}

//
// mecha suit
//

static void T_MechaSuitCheckEnter(objtype *ob)
{
    mecha_t *mecha;
    controldir_t cdir;

    mecha = &gamestate.mecha;

    if (gamestate.weapon == wp_mecha || 
        player->state != ED_STATE("s_player") ||
        !OB_CheckUsed(ob))
    {
        return;
    }

    buttonheld[bt_use] = true;

    if (ob->enemy->lightWeight)
    {
        cdir = Player_CardinalDir();
    }
    else
    {
        cdir = di_north;
        switch (ob->dir)
        {
        case east:
            cdir = di_east;
            break;
        case west:
            cdir = di_west;
            break;
        case north:
            cdir = di_north;
            break;
        case south:
            cdir = di_south;
            break;
        default:
            break;
        }
    }

    if (EnterMountedWeapon(wp_mecha, ob->tilex, ob->tiley, cdir))
    {
        lwlib_Zero(*mecha);
        mecha->starthp = ob->starthitpoints;
        mecha->playerarmor = gamestate.armor;
        if (!ob->enemy->lightWeight)
        {
            gamestate.armor = ob->hitpoints;
        }
        mecha->lightWeight = ob->enemy->lightWeight;
        OB_SetState(ob, NULL);
        WeaponBobble_Reset();
        DrawPlayScreen();
    }
}

void T_MechaSuit(objtype *ob)
{
    if (!ob->enemy->lightWeight)
    {
        if (ob->hitpoints < MECHA_HP_ONEARM(ob->starthitpoints))
        {
            SpawnBulletPuff(OB_GetX(ob), OB_GetY(ob),
                ED_STATE("s_mechboom1"), 0, 0);
            newobj->ticcount = 6;
            T_Aim(newobj);
            OB_AngleMove(newobj, FP(1) / 4, FP(1) / 4);

            NewState(ob, ED_STATE("s_mechasuitonearm"));
        }
    }

    T_MechaSuitCheckEnter(ob);
}

void T_MechaSuitOneArm(objtype *ob)
{
    if (ob->hitpoints < MECHA_HP_NOARMS(ob->starthitpoints))
    {
        SpawnBulletPuff(OB_GetX(ob), OB_GetY(ob),
            ED_STATE("s_mechboom1"), 0, 0);
        newobj->ticcount = 6;
        T_Aim(newobj);
        OB_AngleMove(newobj, FP(1) / 4, -FP(1) / 4);

        NewState(ob, ED_STATE("s_mechasuitnoarms"));
    }

    T_MechaSuitCheckEnter(ob);
}

void T_MechaSuitNoArms(objtype *ob)
{
    T_MechaSuitCheckEnter(ob);
}

void A_PlayMechBoomSound(objtype *ob)
{
    PlaySoundLocActor(ED_SND("MECHAARMBOOMSND"),ob);
}

void T_MechaSuitEnemyEnter(objtype *ob)
{
    int dir;
    int goalObjIndex;
    ED_Spawn_t *mechMutSpawn;
    objtype *goalObj, *mechaSuitObj;
    const char *mechMutSpawnName;

    goalObjIndex = ob->objIndexArr[oit_goaltarget];
    goalObj = &objlist[goalObjIndex];

    mechaSuitObj = goalObj;

    if (mechaSuitObj->enemy->lightWeight)
    {
        mechMutSpawnName = "en_mechmut1";
        dir = 8;
    }
    else
    {
        mechMutSpawnName = "en_mechmut4";
        dir = mechaSuitObj->dir / 2;
    }
    mechMutSpawn = ED_SpawnByEnemyName(ed, mechMutSpawnName, dir);
    if (mechMutSpawn == NULL)
    {
        return;
    }

    PlayActorVoice(ED_SND(weaponinfo[wp_mecha].mountedUseSound), ob);

    SpawnActor(mechaSuitObj->tilex, mechaSuitObj->tiley, 
        mechMutSpawn->actor);
    OB_SetState(mechaSuitObj, NULL);
    OB_SetState(ob, NULL);
    gamestate.killtotal--;

    FirstSighting(newobj);

    if (!mechaSuitObj->enemy->lightWeight)
    {
        if (!mechaSuitObj->enemy->lightWeight &&
            mechaSuitObj->hitpoints < 
                MECHA_HP_ONEARM(mechaSuitObj->starthitpoints))
        {
            OB_SetState(newobj, ED_STATE("s_mechmut4onearmchase1"));
            newobj->hitpoints = MECHA_HP_ONEARM(newobj->starthitpoints) - 1;
        }
    }
}

//
// blackcat
//

void A_BlackCatPlaySound(objtype *ob)
{
    int dist;

    dist = OB_TileDist(player, ob);
    if (dist <= 1 || (dist == 2 && US_RndT() < 96))
    {
        A_AlertSound(ob);
        NewState(ob, ED_STATE(ob->enemy->shootState));
        ob->ticcount = 100;
    }
    else
    {
        A_BrownHorsePlaySound(ob);
    }
}

//
// water mutant
//

void A_WaterMutEnter(objtype *ob)
{
    watermutobjtype *watermut;

    A_BecomeNoPain(ob);
    A_BecomeUnderground(ob);

    watermut = &ob->u.watermut;
    watermut->waterbound = true;
}

void A_WaterMutExit(objtype *ob)
{
    watermutobjtype *watermut;

    A_RemoveNoPain(ob);
    A_RemoveUnderground(ob);

    watermut = &ob->u.watermut;
    watermut->waterbound = false;
}

//
// sewer boss
//

void A_SewerBossShoot(objtype *ob)
{
    PlaySoundLocActorVol(ED_SND("SEWERBOSSSHOOTSND"), ob, SD_MaxVolume(), 
        false, false);
    T_FireProj(ob);
}

//
// generic
//

void A_Dead(objtype *ob)
{
    if (ob->enemy->death.doVictory)
    {
        VictoryTile();
    }
}

//#define OB_DEBUG_XY

void OB_SetX(objtype *ob, fixed x)
{
    ob->x = x;
    #ifdef OB_DEBUG_XY
    lwlib_String_t fpStr;
    fixedpt_str(ob->x, fpStr);
    fprintf(stderr, "ob[%d]->x = %s\n", ob->enemyIndex, fpStr);
    #endif
}

void OB_SetY(objtype *ob, fixed y)
{
    ob->y = y;
    #ifdef OB_DEBUG_XY
    lwlib_String_t fpStr;
    fixedpt_str(ob->y, fpStr);
    fprintf(stderr, "ob[%d]->y = %s\n", ob->enemyIndex, fpStr);
    #endif
}

int OB_ComputeChance(objtype *ob, ED_ChanceRangeSeq_t chanceRanges)
{
    int i;
    int chance;
    int dist;
    ED_ChanceRange_t *range;

    chance = 256;
    dist = OBJTILEDIST(ob, player);

    for (i = 0; (size_t)i < vec_size(chanceRanges); i++)
    {
        range = &vec_at(chanceRanges, i);
        if (range->startDist != range->endDist &&
            dist >= range->startDist && 
            (range->endDist == -1 || dist < range->endDist))
        {
            if (range->endDist == -1)
            {
                chance = range->startChance;
            }
            else
            {
                chance = range->startChance + (range->endChance - 
                    range->startChance) * (dist - range->startDist) /
                    (range->endDist - range->startDist) * 256 / 100;
            }
            break;
        }
    }

    return chance;
}

void OB_SetEnemy(objtype *ob, ED_Enemy_t *enemy)
{
    assert(enemy != NULL);
    ob->enemyIndex = ED_EnemyIndex(ed, enemy->name);
    ob->enemy = enemy;
}

void OB_StateEntered(objtype *ob)
{
    ED_PlayVoice_t *playVoice;

    playVoice = ob->state->playVoice;
    if (playVoice != NULL)
    {
        PlayActorVoice(ED_SND(playVoice->sound), ob);
    }
}

void OB_SetState(objtype *ob, statetype *state)
{
    statetype *oldState;

    if (state != NULL)
    {
        oldState = ob->state;

        ob->stateIndex = ED_StateIndex(ed, state->name);
        ob->state = state;

        if (state != oldState)
        {
            OB_StateEntered(ob);
        }
    }
    else
    {
        ob->stateIndex = 0;
        ob->state = NULL;
    }
}

int OB_TileDist(objtype *ob1, objtype *ob2)
{
    return TILEDIST(ob1->tilex, ob1->tiley, ob2->tilex, ob2->tiley);
}

boolean OB_CheckCoordLine(objtype *ob1, objtype *ob2)
{
    return CheckCoordLine(OB_GetX(ob1), OB_GetY(ob1), 
        OB_GetX(ob2), OB_GetY(ob2), false);
}

bool OB_InShootDelta(objtype *ob)
{
    return OB_InShootDistDelta(ob, 1);
}

bool OB_InShootDistDelta(objtype *ob, int divisor)
{
    int viewDelta;

    viewDelta = abs(ob->viewx - centerx);
    if (viewDelta >= shootdelta / divisor)
    {
        return false;
    }

    return ray3f_hit_box(Player_GetViewRay(), 
        OB_GetBoundingBox(ob, 1.0 / divisor, 1.0));
}

bool OB_CheckUsed(objtype *ob)
{
    int dist;
    dist = OB_TileDist(player, ob);
    return dist <= 1 && OB_InShootDelta(ob) &&
        (buttonstate[bt_use] && !buttonheld[bt_use]);
}

void OB_AngleMove(objtype *ob, fixed u, fixed v)
{
    fixed x, y;
    fixed moveDist;

    x = OB_GetX(ob);
    y = OB_GetY(ob);

    if (u != 0)
    {
        moveDist = u;
        x += FixedMul(moveDist, finecostable[ob->fineangle]);
        y -= FixedMul(moveDist, finesintable[ob->fineangle]);
    }

    if (v != 0)
    {
        moveDist = v;
        x += FixedMul(moveDist, -finesintable[ob->fineangle]);
        y -= FixedMul(moveDist, finecostable[ob->fineangle]);
    }

    OB_SetX(ob, x);
    OB_SetY(ob, y);
}

void OB_SpawnSurpriseEnemy(objtype *ob, fixed u, fixed v,
    bool resetHitPoints)
{
    ED_Enemy_t *enemy;
    const ED_EnemySurpriseAttack_t *surpriseAttack;

    surpriseAttack = &ob->enemy->surpriseAttack;

    enemy = ED_Enemy(ed, surpriseAttack->spawnEnemyName);
    SpawnNewObj (ob->tilex, ob->tiley, ED_STATE(enemy->chaseState));

    newobj->speed = enemy->chaseSpeed;
    OB_SetEnemy(newobj, enemy);

    if (resetHitPoints)
    {
        newobj->hitpoints = newobj->starthitpoints = enemy->health;
        newobj->killpoints = enemy->killpoints;
        if (newobj->killpoints)
        {
            gamestate.killtotal++;
        }
    }
    else
    {
        newobj->hitpoints = newobj->starthitpoints = ob->hitpoints;
        newobj->killpoints = ob->killpoints;
    }
    newobj->dir = nodir;
    newobj->flags |= FL_SHOOTABLE|FL_BLOCK;
    newobj->x = ob->x;
    newobj->y = ob->y;

    T_Aim(newobj);
    if (OB_TileDist(player, newobj) < 2)
    {
        u = 0;
    }
    OB_AngleMove(newobj, u, v);
}

void OB_KillActorAlert(objtype *ob)
{
    using namespace lwlib;

    if (US_RndT() >= 64)
    {
        return;
    }

    const int maxActors = 8;
    objtype *foundActors[maxActors];
    int numActors = 0;

    const int maxRadius = 16;
    for (int r = 1; r < maxRadius && numActors < maxActors; r++)
    {
        for (int i = 0; i < (r * 2) * 4 && numActors < maxActors; i++)
        {
            const int j = i % (r * 2);
            Point2i tilePos = vec_add(vec_rot_quadrant(
                vec_add(vec2i(r, r), vec2i(0, -j)), i / (r * 2)
                ), vec2i(ob->tilex, ob->tiley));
            if (tilePosInMap(tilePos))
            {
                int x, y;
                x = tilePos.v[0];
                y = tilePos.v[1];

                objtype *check = actorat[x][y];
                if (check != NULL && ISPOINTER(check) && 
                    check->areanumber == ob->areanumber &&
                    (check->flags & FL_ATTACKMODE) == 0)
                {
                    foundActors[numActors++] = check;
                }
            }
        }
    }

    if (numActors > 0)
    {
        const int numToAlert = US_RndT() % (numActors + 1);
        //const int numToAlert = numActors;
        const int startActor = US_RndT() % numActors;
        for (int i = 0; i < numToAlert; i++)
        {
            objtype *check = foundActors[(startActor + i) % numActors];
            check->killActorAlert = 1;
        }
    }
}

void A_SpawnSurpriseEnemy(objtype *ob)
{
    OB_SpawnSurpriseEnemy(ob, 0, 0, false);
}

void OB_Damage(objtype *ob, objtype *attacker, int damage, 
    damage_t damagetype, boolean ignorearmor)
{
    if (ob == player)
    {
        TakeSpecialDamage(damage, attacker, damagetype, ignorearmor, true);
    }
    else
    {
        DamageActor(ob, attacker, damage, damagetype);
    }
}

objtype *OB_GetObjOn(objtype *ob)
{
    int dist;
    objtype *objOn;

    objOn = NULL;

    if (actorat[ob->tilex][ob->tiley] != NULL)
    {
        objOn = actorat[ob->tilex][ob->tiley];
    }

    dist = OB_TileDist(player, ob);
    if (dist <= 0)
    {
        objOn = player;
    }

    return objOn;
}

int OB_AngleTo(objtype *from, objtype *to)
{
    float fangle;
    int iangle;
    fixed dx, dy;

    dx = OB_GetX(to) - OB_GetX(from);
    dy = OB_GetY(from) - OB_GetY(to);

    fangle = (float)atan2((float)dy, (float)dx);
    if (fangle<0)
    {
        fangle = (float)(M_PI * 2 + fangle);
    }

    iangle = (int)(fangle / (M_PI * 2) * ANGLES);

    return iangle;
}

void OB_AngleMoveTo(objtype *from, objtype *to, fixed dist)
{
    int viewangle;

    viewangle = OB_AngleTo(from, to);

    from->x += FixedMul(dist, costable[viewangle]);
    from->y -= FixedMul(dist, sintable[viewangle]);
}

bool OB_CheckCloserThan(objtype *ob1, objtype *ob2, fixed minDist)
{
    return LABS(OB_GetX(ob1) - OB_GetX(ob2)) < minDist && 
        LABS(OB_GetY(ob1) - OB_GetY(ob2)) < minDist;
}

lwlib::Point2f OB_PointFromView(objtype *ob, double u, double v)
{
    double angle;
    lwlib::Point2f udir, vdir;

    angle = lwlib_DegToRad((double)ob->fineangle * ANGLES / FINEANGLES);

    udir = lwlib::vec2f(cos(angle), -sin(angle));
    vdir = vec_neg(vec_rot90(udir));

    return vec_add(vec_scale(udir, u), vec_scale(vdir, v));
}

bool OB_TileCentred(objtype *ob, fixed dist)
{
    return fixedpt_fracpart(OB_GetX(ob) - 0x8000) <= dist &&
        fixedpt_fracpart(OB_GetY(ob) - 0x8000) <= dist;
}

lwlib::Box3f OB_GetBoundingBox(objtype *ob, double size, double height)
{
    lwlib::Point3f center;
    lwlib::Vector3f span;

    span = lwlib::vec3f(size / 2, size / 2, height / 2);

    center = lwlib::vec3f(fixedpt_toconst(OB_GetX(ob)),
        fixedpt_toconst(OB_GetY(ob)), 0.5);

    return lwlib::box3f(vec_sub(center, span), vec_add(center, span));
}

bool OB_NaziFriends(objtype *ob, objtype *check)
{
    if (ob == check)
    {
        return true;
    }

    if (ob == player || check == player)
    {
        return false;
    }

    return ob->enemy != NULL && check->enemy != NULL &&
        ob->enemy->naziFriends && check->enemy->naziFriends;
}

bool OB_GuardedObjInSight(objtype *ob)
{
    if (ob->enemy->boss != NULL && gamestate.guardedObjNum != 0)
    {
        objtype *guardedObj = objlist + gamestate.guardedObjNum;
        if (ob != guardedObj && ob->areanumber == guardedObj->areanumber &&
            (CheckLine(ob) || OB_CheckCoordLine(ob, guardedObj)))
        {
            return true;
        }
    }

    return false;
}

bool OB_CanDefendGuardedObj(objtype *ob)
{
    objtype *guardedObj = objlist + gamestate.guardedObjNum;
    return OB_GuardedObjInSight(ob) &&
        OBJTILEDIST(ob, guardedObj) >= 3;
}

bool OB_GuardedObjBlocksMove(objtype *ob, int tilex, int tiley)
{
    return false;
#ifdef NOTYET
    if (gamestate.guardedObjNum == 0)
    {
        return false;
    }

    objtype *guardedObj = objlist + gamestate.guardedObjNum;
    if (ob == guardedObj)
    {
        return false;
    }

    if (!OB_GuardedObjInSight(ob))
    {
        return false;
    }

    const lwlib::Point2f start = OB_GetPos(player);
    const lwlib::Vector2f normal = vec_rot90(
        vec_normalize(vec_sub(OB_GetPos(guardedObj), start)));
    const lwlib::Plane2f plane = lwlib::vec_pair(start, normal);

    const lwlib::Point2f destPos = vec_add(lwlib::vec2f(tilex, tiley),
        lwlib::vec2f(0.5, 0.5));
    return fabs(vec_planedot(destPos, plane)) < 1.0;
#endif
}

const lwlib::Point2f OB_GetPos(objtype *ob)
{
    return lwlib::vec2f(fixedpt_toconst(OB_GetX(ob)),
        fixedpt_toconst(OB_GetY(ob)));
}

const lwlib::Ray3f OB_GetViewRay(objtype *ob)
{
    lwlib::Ray3f ray;
    lwlib::Point3f origin;
    lwlib::Vector3f u;
    lwlib::Mat3f rotMat;

    origin = lwlib::vec3f(fixedpt_toconst(OB_GetX(ob)),
        fixedpt_toconst(OB_GetY(ob)), 0.5);
    rotMat = lwlib::mat3f_rot(
        -(double)ob->fineangle * ANGLES / FINEANGLES, 2);
    u = lwlib::vec3f(1.0, 0.0, 0.0);
    u = lwlib::mat_xform(rotMat, u);
    ray = ray3f(origin, u);

    return ray;
}
