// WL_AGENT.C

#include "wl_def.h"
#include "lw_vec.h"
#pragma hdrstop

/*
=============================================================================

                                LOCAL CONSTANTS

=============================================================================
*/

#define MAXMOUSETURN    10

#define MOVESCALE       150l
#define BACKMOVESCALE   100l
#define ANGLESCALE      20

#define FOOTSTEP_MAX_SPEED (FP(1) / 5)
#define FOOTSTEP_MAX_VOLUME (SD_MaxVolume() * 4 / 5)

#define FUELBARCOLOR    0x2f
#define FUELBARHLIN     0x2c
#define FUELBARVLIN     0x28

#define LOWFUELBARCOLOR 0x4c
#define LOWFUELBARHLIN  0x49
#define LOWFUELBARVLIN  0x4b

#define MECHAMOVE       20
#define MECHARUN        50
#define MECHASLOW       10
#define LWMECHAMOVE     30
#define LWMECHASLOW     20

#define MECHASIDERATIO  1 / 2
#define MECHAMOVESIDE   (MECHAMOVE * MECHASIDERATIO)
#define MECHARUNSIDE    (MECHARUN * MECHASIDERATIO)
#define MECHASLOWSIDE   (MECHASLOW * MECHASIDERATIO)
#define LWMECHAMOVESIDE (LWMECHAMOVE * MECHASIDERATIO)
#define LWMECHASLOWSIDE (LWMECHASLOW * MECHASIDERATIO)

/*
=============================================================================

                                GLOBAL VARIABLES

=============================================================================
*/

//
// player state info
//
int32_t         thrustspeed;

word            plux,pluy;          // player coordinates scaled to unsigned

short           anglefrac;

damage_t       KillDamageType;

/*
=============================================================================

                                                 LOCAL VARIABLES

=============================================================================
*/

static weaponspr_t cyringe_weaponsprs[] =
{
    { "SPR_INJECTION_FULL" },
    { "SPR_INJECTION_NEEDLE1" },
    { "SPR_INJECTION_NEEDLE2" },
    { "SPR_INJECTION_NEEDLE3" },
    { "SPR_INJECTION_BLOOD" },
    { "SPR_INJECTION_SLEEVE3" },
    { "SPR_INJECTION_SLEEVE4" },
    { "SPR_INJECTION_FINISHED" },
    { NULL },
};

//
// Weapon Info
//
weaponinfo_t weaponinfo[NUMWEAPONS] =
{
    // wp_cyringe
    {
        NULL, // readySprite
        cyringe_weaponsprs, // weaponSprs
        GUYCYRINGE1PIC, // guyPic
        bt_readyknife, // selectkey
        0, // clip
        ammo_antidote, // ammotype
        0, // noBobble
        1, // discardAtEnd
        1, // skipfallback
    },
    // wp_knife
    {
        "SPR_KNIFEREADY", // readySprite
        NULL, // weaponSprs
        GUYKNIFE1PIC, // guyPic
        bt_readyknife, // selectkey
        0, // clip
        ammo_none, // ammotype
        0, // noBobble
    },
    // wp_pistol
    {
        "SPR_PISTOLREADY", // readySprite
        NULL, // weaponSprs
        GUYLUGER1PIC, // guyPic
        bt_readypistol, // selectkey
        8, // clip
        ammo_bullet, // ammotype
        0, // noBobble
    },
    // wp_kar98k
    {
        "SPR_KAR98KREADY", // readySprite
        NULL, // weaponSprs
        GUYKAR98K1PIC, // guyPic
        bt_readyrifle, // selectkey
        5, // clip
        ammo_rifle, // ammotype
        1, // noBobble
    },
    // wp_sniper
    {
        "SPR_SNIPERREADY", // readySprite
        NULL, // weaponSprs
        GUYSNIPER1PIC, // guyPic
        bt_readyrifle, // selectkey
        3, // clip
        ammo_rifle, // ammotype
        1, // noBobble
    },
    // wp_mp40
    {
        "SPR_MP40READY", // readySprite
        NULL, // weaponSprs
        GUYMP401PIC, // guyPic
        bt_readymachinegun, // selectkey
        32, // clip
        ammo_bullet, // ammotype
        0, // noBobble
    },
    // wp_stg44
    {
        "SPR_STG44READY", // readySprite
        NULL, // weaponSprs
        GUYSTG441PIC, // guyPic
        bt_readymachinegun, // selectkey
        30, // clip
        ammo_stg44, // ammotype
        0, // noBobble
    },
    // wp_shotgun
    {
        "SPR_SHOTGUNREADY", // readySprite
        NULL, // weaponSprs
        GUYSHOTGUN1PIC, // guyPic
        bt_readyshotgun, // selectkey
        2, // clip
        ammo_shotgun, // ammotype
        0, // noBobble
    },
    // wp_flamethrower
    {
        "SPR_FLAMETHROWERREADY", // readySprite
        NULL, // weaponSprs
        GUYFLAMETHROWER1PIC, // guyPic
        bt_readyflamethrower, // selectkey
        0, // clip
        ammo_flame, // ammotype
        0, // noBobble
    },
    // wp_chaingun
    {
        "SPR_CHAINGUNREADY", // readySprite
        NULL, // weaponSprs
        GUYCHAINGUN1PIC, // guyPic
        bt_readychaingun, // selectkey
        0, // clip
        ammo_chaingun, // ammotype
        0, // noBobble
    },
    // wp_panzerschreck
    {
        "SPR_PANZERSCHRECKREADY", // readySprite
        NULL, // weaponSprs
        GUYPANZERSCHRECK1PIC, // guyPic
        bt_readylauncher, // selectkey
        0, // clip
        ammo_rocket, // ammotype
        0, // noBobble
        0, // discardAtEnd
        1, // skipfallback
    },
    // wp_wirecutters
    {
        "SPR_WIRECUTTERATK1", // readySprite
        NULL, // weaponSprs
        GUYWIRECUTTERS1PIC, // guyPic
        bt_readyknife, // selectkey
        0, // clip
        ammo_none, // ammotype
        0, // noBobble
        1, // discardAtEnd
        1, // skipfallback
        true, // villageFriendly
    },
    // wp_mg42_mounted
    {
        "SPR_MOUNTEDMG42READY", // readySprite
        NULL, // weaponSprs
        GUYLUGER1PIC, // guyPic
        bt_nobutton, // selectkey
        0, // clip
        ammo_none, // ammotype
        1, // noBobble
        0, // discardAtEnd
        1, // skipfallback
        false, // villageFriendly
        "MG42", // name
        "USEMG42SND", // mountedUseSound
    },
    // wp_mecha
    {
        "SPR_MECHAGUNS_READY", // readySprite
        NULL, // weaponSprs
        GUYMECHA1PIC, // guyPic
        bt_nobutton, // selectkey
        0, // clip
        ammo_fuel, // ammotype
        0, // noBobble
        1, // discardAtEnd
        1, // skipfallback
        false, // villageFriendly
        "Mecha", // name
        "USEMECHASND", // mountedUseSound
        "EXITMECHASND", // mountedExitSound
        "T_MechaExit", // mountedExitThink
    },
};

ammoinf_t ammoinfo[NUMAMMOTYPES] =
{
    {
        "Bullets", // name
        { 100, 150, 250 }, // maxammo
        35, // giveBackpackAmmo
        false, // discardAtEnd
    }, // ammo_bullet
    {
        "Rifle Rounds", // name
        { 50, 80, 100 }, // maxammo
        25, // giveBackpackAmmo
        false, // discardAtEnd
    }, // ammo_rifle
    {
        "STG44 Clip", // name
        { 100, 200, 400 }, // maxammo
        50, // giveBackpackAmmo
        false, // discardAtEnd
    }, // ammo_stg44
    {
        "Flamer Gas", // name
        { 150, 300, 600 }, // maxammo
        100, // giveBackpackAmmo
        false, // discardAtEnd
    }, // ammo_flame
    {
        "Chaingun Bullets", // name
        { 200, 300, 400 }, // maxammo
        125, // giveBackpackAmmo
        false, // discardAtEnd
    }, // ammo_chaingun
    {
        "Panzerschreck Rockets", // name
        { 15, 20, 25 }, // maxammo
        5, // giveBackpackAmmo
        false, // discardAtEnd
    }, // ammo_rocket
    {
        "Shotgun Shells", // name
        { 50, 70, 100 }, // maxammo
        25, // giveBackpackAmmo
        false, // discardAtEnd
    }, // ammo_shotgun
    {
        "Fuel",
        { 400, 500, 800 }, // maxammo
        100, // giveBackpackAmmo
        true, // discardAtEnd
    }, // ammo_fuel
    {
        "Antidote",
        { 5, 10, 20 }, // maxammo
        3, // giveBackpackAmmo
        false, // discardAtEnd
    }, // ammo_antidote
};


/*
WSJ: Attack Reference

   -4 - end of reloading sequence
   -3 - end of firing sequence if no ammo left (bolt sound)
   -2 - end of firing sequence if no ammo left
   -1 - end of firing sequence
    1 - gun attack
    2 - knife attack
    3 - mp40 refire
    4 - chaingun refire
    5 - gun attack (no sound)
    6 - chaingun refire (no sound)
    7 - alternate gun attack
    8 - reload full clip
    9 - reload 1 round
   10 - go back 2 frames if clip isn't full yet

   21-26 - recoil shots (decreasing accuracy)
*/

struct atkinf
{
    int8_t    tics,attack,frame;
} attackinfo[NUMWEAPONS][2][15] =
{
// wp_cyringe
  { { {16,0,1},{16,0,2},{16,0,3},{16,0,4},{16,0,5},{16,1,6},{16,-1,7} }, },
// wp_knife
  { { {6,0,1},{6,2,2},{6,0,3},{6,-1,4} }, },
// wp_pistol
  { { {6,0,1},{6,1,2},{6,0,3},{6,-1,4} },
    { {16,0,5},{10,0,6},{10,0,7},{12,0,8},{24,0,9},{12,0,10},{12,8,11},{12,0,12},{12,0,13},{20,-4,14} }, },
// wp_kar98k
  { { {6,0,0},{6,1,1},{10,0,2},{10,-3,3},{16,0,4},{6,0,5},{16,0,6},{16,0,7},{10,0,8},{6,-1,9} },
    { {16,0,10},{12,0,11},{24,0,12},{12,0,13},{12,0,14},{10,8,15},{10,0,16},{12,0,17},{12,0,18},{16,-4,19} }, },
// wp_sniper
  { { {6,0,0},{6,1,1},{6,-1,0},{6,0,0},{6,1,0},{6,-1,0} },
    { {16,0,7},{10,0,8},{36,0,9},{16,8,10},{12,0,11},{12,0,12},{12,0,13},{16,-4,14} }, },
// wp_mp40
  { { {5,0,1},{5,1,2},{5,3,3},{5,-1,1},{10,21,4},{10,22,5},{10,23,6},{5,24,7},{5,-1,7} },
    { {16,0,8},{12,0,9},{24,0,10},{12,0,11},{12,0,12},{12,8,13},{16,0,14},{24,-4,15} }, },
// wp_stg44
  { { {5,0,1},{5,1,2},{5,3,3},{5,-1,1},{10,21,4},{10,22,5},{10,23,6},{5,24,7},{5,-1,7} },
    { {16,0,8},{12,0,9},{24,0,10},{12,0,11},{12,0,12},{12,8,13},{16,0,14},{24,-4,15} }, },
// wp_shotgun
  { { {6,0,0},{6,1,1},{6,0,0},{6,-1,0},{6,0,0},{6,1,2},{6,0,0},{6,-1,0},{6,0,0},{6,7,3},{6,0,0},{6,-1,0}, },
    { {16,0,4},{24,0,5},{12,9,6},{24,0,7},{12,9,8},{12,0,9},{12,0,10},{16,-4,11},{16,0,4},{24,0,5},{12,9,8},{12,0,9},{12,0,10},{16,-4,11}, }, },
// wp_flamethrower
  { { {4,0,0},{4,1,1},{4,5,2},{4,5,1},{4,5,2},{4,5,1},{4,5,2},{4,5,1},{4,6,2},{6,-1,0} }, },
// wp_chaingun
  { { {4,0,1},{4,1,3},{4,5,4},{4,5,3},{4,5,4},{4,1,3},{4,5,4},{4,5,3},{4,6,4},{6,-1,1} }, },
// wp_panzerschreck
  { { {6,0,1},{6,1,2},{6,0,3},{64,0,4},{6,-1,5} }, },
// wp_wirecutters
  { { {6,1,1},{6,-1,1} }, },
// wp_mg42_mounted
  { { {5,0,1},{5,1,2},{5,5,3},{5,1,2},{5,6,3},{6,-1,4} }, },
// wp_mecha
  { { {5,0,1},{5,1,2},{5,5,3},{5,1,2},{5,6,3},{6,-1,3} }, },
};

//===========================================================================

//----------

void Attack (void);
void Use (void);
void Search (objtype *ob);
void SelectWeapon (void);
void SelectItem (void);

//----------

void T_Player (objtype *ob);

void ClipMove (objtype *ob, int32_t xmove, int32_t ymove);

void Cmd_Reload (void);

/*
=============================================================================

                                CONTROL STUFF

=============================================================================
*/

/*
======================
=
= CheckWeaponChange
=
= Keys 1-8 change weapons
=
======================
*/

// WSJ: check the type of ammo used by weapon
ammo_t CheckAmmoType (int weapon)
{
    if (weapon > -1 && weapon < NUMWEAPONS)
        return weaponinfo[weapon].ammotype;

    return ammo_none;
}

boolean WeaponAvailable (int weapon)
{
    if (weapon >= wp_mg42_mounted && gamestate.weapon == weapon)
    {
        if (CheckAmmoType(weapon) == ammo_none)
        {
            return true;
        }
        if (HASAMMO(weapon))
        {
            return true;
        }
    }

    if (!HASWEAPON(weapon))
        return false;

    return HASAMMO(weapon);
}

boolean WeaponVillageFriendly (int weapon)
{
    return (GetFeatureFlags() & FF_VILLAGE) == 0 || 
        weaponinfo[weapon].villageFriendly;
}

boolean ReadyToFire (int weapon)
{
    assert(weapon == gamestate.weapon);

    if (Mecha_Equipped() && Mecha_HasNoArms())
    {
        return false;
    }

    if (!WeaponVillageFriendly(weapon))
        return false;

    if (!WeaponAvailable(weapon))
        return false;

    if (EMPTYCLIP(weapon))
        return false;

    return true;
}

weapontype BestAvailableWeapon (boolean skipdangerweapon)
{
    boolean skipweapon;
    int bestweapon = 0;

    for (int i = 0; i <= gamestate.bestweapon; i++)
    {
        if(skipdangerweapon)
            skipweapon = weaponinfo[i].skipfallback;
        else
            skipweapon = false;

        if(WeaponAvailable(i) && !skipweapon)
            bestweapon = i;
    }

    return (weapontype)bestweapon;
}

weapontype CheckWeaponChoice (short wpkey)
{
    weapontype firstchoice = (weapontype)-1;
    weapontype lastchosen = gamestate.lastchosenweapon[wpkey];

    if (wpkey != WEAPONKEY(gamestate.weapon) && lastchosen > 0 &&
        WeaponAvailable(lastchosen) && WeaponVillageFriendly(lastchosen))
    {
        return lastchosen;
    }

    for (int i = 0; i < NUMWEAPONS; i++)
    {
        if (!WeaponAvailable(i) || !WeaponVillageFriendly(i) ||
            WEAPONKEY(i) != wpkey)
        {
            continue;
        }

        if (firstchoice == -1)
        {
            firstchoice = (weapontype)i;
        }

        if (wpkey == WEAPONKEY(gamestate.weapon) && i > gamestate.weapon)
        {
            return (weapontype)i;
        }
    }

    return firstchoice;
}

void ChangeWeapon (int weapon)
{
    if (gamestate.weapon == weapon)
        return;

    gamestate.weapon = (weapontype)weapon;
    gamestate.snipermode = 0;
    DrawWeapon ();
    DrawAmmo ();
    if (EMPTYCLIP(weapon))
        Cmd_Reload ();
}

void CheckWeaponChange (void)
{
    int i, c;
    int newWeapon = -1;
    weapontype bestweapon = BestAvailableWeapon(false);

    if (gamestate.weapon >= wp_mg42_mounted)
    {
        return;
    }

#ifdef _arch_dreamcast
    int joyx, joyy;
    IN_GetJoyFineDelta (&joyx, &joyy);
    if(joyx < -64)
        buttonstate[bt_prevweapon] = true;
    else if(joyx > 64)
        buttonstate[bt_nextweapon] = true;
#endif

    if(buttonstate[bt_nextweapon] && !buttonheld[bt_nextweapon])
    {
        i = gamestate.weapon;
        c = 0;

        while (c < bestweapon + 1)
        {
            i = (i + 1) % (bestweapon + 1);
            c++;

            if(WeaponAvailable(i) && WeaponVillageFriendly(i))
            {
                newWeapon = i;
                break;
            }
        }
    }
    else if(buttonstate[bt_prevweapon] && !buttonheld[bt_prevweapon])
    {
        i = gamestate.weapon;
        c = 0;

        while (c < bestweapon + 1)
        {
            i = (i + (bestweapon + 1) - 1) % (bestweapon + 1);
            c++;

            if(WeaponAvailable(i) && WeaponVillageFriendly(i))
            {
                newWeapon = i;
                break;
            }
        }
    }
    else
    {
        for(int i = bt_readyfirst; i <= bt_readylast; i++)
        {
            if (buttonstate[i] && !buttonheld[i])
            {
                newWeapon = CheckWeaponChoice(i);
                break;
            }
        }
    }

    if(newWeapon != -1)
    {
        if (gamestate.weapon != newWeapon)
        {
            OB_SetState(player, ED_STATE("s_player"));
            gamestate.attackframe = gamestate.weaponframe = 0;
        }

        ChangeWeapon(newWeapon);
        gamestate.chosenweapon = (weapontype) newWeapon;
        gamestate.lastchosenweapon[WEAPONKEY(newWeapon)] = (weapontype) newWeapon;
    }
}


/*
=======================
=
= ControlMovement
=
= Takes controlx,controly, and buttonstate[bt_strafe]
=
= Changes the player's angle and position
=
= There is an angle hack because when going 70 fps, the roundoff becomes
= significant
=
=======================
*/

void ControlMovement (objtype *ob)
{
    int     angle;
    int     angleunits;
    mecha_t *mecha;
    int     controlxlim;

    mecha = &gamestate.mecha;

    Mecha_ControlRunningCharge();

#ifdef WEAPONBOBBLE
    gamestate.bobberx = OB_GetX(player);
    gamestate.bobbery = OB_GetY(player);
#endif

    if (gamestate.snipermode)    // WSJ: slower movement in sniper mode
    {
        controlx /= 4;
        controly /= 4;
    }
    if (Mecha_Equipped())
    {
        const int mechaMove = mecha->lightWeight ? RUNMOVE : BASEMOVE;
        controlxlim = mechaMove * tics;
        controlx = lwlib_CLIP(controlx, -controlxlim, controlxlim);

        if (abs(controly) < BASEMOVE)
        {
            controly = 0;
        }
    }

    thrustspeed = 0;

    if (Mecha_Equipped() && mecha->moveVel != 0)
    {
        angle = ob->fineangle;
        if (mecha->strafeMove)
        {
            angle += FINEANGLES/4;
            if (angle >= FINEANGLES)
                angle -= FINEANGLES;
        }
        else
        {
            if (Mecha_RunningChargeActive())
            {
                angle = mecha->runningCharge.angle;
            }
        }
        if (buttonstate[bt_strafe])
        {
            controlx = 0;
        }
        controly = 0;

        Thrust (angle, mecha->moveVel * tics);
    }

    //
    // forward/backwards move
    //
    if (controly < 0 || Mecha_RunningChargeActive())
    {
        angle = ob->fineangle;

        if (Mecha_Equipped())
        {
            if (gamestate.ammo[ammo_fuel] > 0)
            {
                if (Mecha_RunningChargeActive())
                {
                    mecha->moveVel = MECHARUN * MOVESCALE;
                }
                else if (mecha->lightWeight)
                {
                    mecha->moveVel = (Mecha_FuelLow() ? LWMECHASLOW : 
                        LWMECHAMOVE) * MOVESCALE;
                }
                else
                {
                    mecha->moveVel = (Mecha_FuelLow() ? MECHASLOW : 
                        MECHAMOVE) * MOVESCALE;
                }
                mecha->bobbleFactor = 1;
                mecha->strafeMove = false;
            }
            else
            {
                mecha->moveVel = 0;
            }
        }
        else
        {
            Thrust (angle, -controly * MOVESCALE); // move forwards
        }
    }
    else if (controly > 0)
    {
        angle = ob->fineangle + FINEANGLES/2;
        if (angle >= FINEANGLES)
            angle -= FINEANGLES;

        if (Mecha_Equipped())
        {
            if (gamestate.ammo[ammo_fuel] > 0)
            {
                if (mecha->lightWeight)
                {
                    mecha->moveVel = (Mecha_FuelLow() ? -LWMECHASLOW : 
                        -LWMECHAMOVE) * BACKMOVESCALE;
                }
                else
                {
                    mecha->moveVel = (Mecha_FuelLow() ? -MECHASLOW : 
                        -MECHAMOVE) * BACKMOVESCALE;
                }
                mecha->bobbleFactor = 2;
                mecha->strafeMove = false;
            }
            else
            {
                mecha->moveVel = 0;
            }
        }
        else
        {
            Thrust (angle, controly * BACKMOVESCALE); // move backwards
        }
    }

    if (mecha->moveVel == 0 &&
        buttonstate[bt_strafeleft] && Mecha_Equipped())
    {
        buttonstate[bt_strafe] = true;
        controlx = -1;
    }

    if (mecha->moveVel == 0 &&
        buttonstate[bt_straferight] && Mecha_Equipped())
    {
        buttonstate[bt_strafe] = true;
        controlx = 1;
    }

    //
    // side to side move
    //
    if (buttonstate[bt_strafe])
    {
        buttonstate[bt_strafe] = false;

        Mecha_NotifyRotating(false);

        //
        // strafing
        //
        //
        if (controlx > 0)
        {
            angle = ob->fineangle - FINEANGLES/4;
            if (angle < 0)
                angle += FINEANGLES;
            if (Mecha_Equipped())
            {
                if (gamestate.ammo[ammo_fuel] > 0)
                {
                    if (mecha->lightWeight)
                    {
                        mecha->moveVel = (Mecha_FuelLow() ?
                            -LWMECHASLOWSIDE : -LWMECHAMOVESIDE) * MOVESCALE;
                    }
                    else
                    {
                        mecha->moveVel = (Mecha_FuelLow() ?
                            -MECHASLOWSIDE : -MECHAMOVESIDE) * MOVESCALE;
                    }
                    mecha->bobbleFactor = 3;
                    mecha->strafeMove = true;
                }
                else
                {
                    mecha->moveVel = 0;
                }
            }
            else
            {
                Thrust (angle,controlx*MOVESCALE);      // move to left
            }
        }
        else if (controlx < 0)
        {
            angle = ob->fineangle + FINEANGLES/4;
            if (angle >= FINEANGLES)
                angle -= FINEANGLES;
            if (Mecha_Equipped())
            {
                if (gamestate.ammo[ammo_fuel] > 0)
                {
                    if (mecha->lightWeight)
                    {
                        mecha->moveVel = (Mecha_FuelLow() ?
                            LWMECHASLOWSIDE : LWMECHAMOVESIDE) * MOVESCALE;
                    }
                    else
                    {
                        mecha->moveVel = (Mecha_FuelLow() ?
                            MECHASLOWSIDE : MECHAMOVESIDE) * MOVESCALE;
                    }
                    mecha->bobbleFactor = 3;
                    mecha->strafeMove = true;
                }
                else
                {
                    mecha->moveVel = 0;
                }
            }
            else
            {
                Thrust (angle,-controlx*MOVESCALE);     // move to right
            }
        }
    }
    else
    {
        if (Mecha_Equipped())
        {
            if (gamestate.ammo[ammo_fuel] == 0)
            {
                //controlx = 0;
            }
            if (controlx != 0)
            {
                Mecha_NotifyRotating(true);
            }
            else
            {
                Mecha_NotifyRotating(false);
            }
        }
        else
        {
            Mecha_NotifyRotating(false);
        }

        //
        // not strafing
        //
        anglefrac += controlx * (FINEANGLES / ANGLES);
        angleunits = anglefrac/ANGLESCALE;
        anglefrac -= angleunits*ANGLESCALE;
        ob->fineangle -= angleunits;

        if (ob->fineangle >= FINEANGLES)
            ob->fineangle -= FINEANGLES;
        if (ob->fineangle < 0)
            ob->fineangle += FINEANGLES;

        ob->angle = ob->fineangle * ANGLES / FINEANGLES;
    }

    if(buttonstate[bt_strafeleft] && !Mecha_Equipped())
    {
        angle = ob->fineangle + FINEANGLES/4;
        if(angle >= FINEANGLES)
            angle -= FINEANGLES;
        if(buttonstate[bt_run])
            Thrust(angle, RUNMOVE * MOVESCALE * tics);
        else
            Thrust(angle, BASEMOVE * MOVESCALE * tics);
    }

    if(buttonstate[bt_straferight] && !Mecha_Equipped())
    {
        angle = ob->fineangle - FINEANGLES/4;
        if(angle < 0)
            angle += FINEANGLES;
        if(buttonstate[bt_run])
            Thrust(angle, RUNMOVE * MOVESCALE * tics );
        else
            Thrust(angle, BASEMOVE * MOVESCALE * tics);
    }

    WeaponBobble_Update();
    Tremor_Update();

    if (gamestate.victoryflag)              // watching the BJ actor
        return;
}

/*
=============================================================================

                            STATUS WINDOW STUFF

=============================================================================
*/


/*
==================
=
= StatusDrawPic
=
==================
*/

void StatusDrawPic (unsigned x, unsigned y, unsigned picnum)
{
    LatchDrawPicScaledCoord ((SCREEN_XOFF / 8) + 
        scaleFactor*x, screenHeight-scaleFactor*(STATUSLINES-y),
        picnum, 8);
}

/*
==================
=
= StatusDrawPicNoScaleX
=
==================
*/

void StatusDrawPicNoScaleX (unsigned x, unsigned y, unsigned picnum)
{
    LatchDrawPicScaledCoord (SCREEN_XOFF + scaleFactor*x, 
        screenHeight-scaleFactor*(STATUSLINES-y),
        picnum, 1);
}

void StatusDrawFace(unsigned picnum)
{
    if (gamestate.infected && picnum >= FACE1APIC && picnum <= 
        GOTGATLINGPIC)
    {
        picnum = picnum - FACE1APIC + INFECTFACE1APIC;
    }
    StatusDrawPic(17, 4, picnum);
}


/*
==================
=
= DrawFace
=
==================
*/

void DrawFace (void)
{
    if(viewsize == 21 && ingame) return;
    if (SD_SoundPlaying() == ED_SND("GETGATLINGSND"))
        StatusDrawFace(GOTGATLINGPIC);
    else if (gamestate.health)
    {
#ifdef SPEAR
        if (godmode)
            StatusDrawFace(GODMODEFACE1PIC+gamestate.faceframe);
        else
#endif
            StatusDrawFace(FACE1APIC+3*((100-gamestate.health)/16)+gamestate.faceframe);
    }
    else
    {
#ifndef SPEAR
        if (KillDamageType == dam_mutagen || KillDamageType == dam_infection)
            StatusDrawFace(MUTANTBJPIC);
        else
#endif
            StatusDrawFace(FACE8APIC);
    }
}

/*
===============
=
= UpdateFace
=
= Calls draw face if time to change
=
===============
*/

int facecount = 0;
int facetimes = 0;

void UpdateFace (void)
{
    // don't make demo depend on sound playback
    if(demoplayback || demorecord)
    {
        if(facetimes > 0)
        {
            facetimes--;
            return;
        }
    }
    else if(SD_SoundPlaying() == ED_SND("GETGATLINGSND"))
        return;

    facecount += tics;
    if (facecount > US_RndT())
    {
        gamestate.faceframe = (US_RndT()>>6);
        if (gamestate.faceframe==3)
            gamestate.faceframe = 1;

        facecount = 0;
        DrawFace ();
    }
    
    Player_UpdateInfection();
}



/*
===============
=
= LatchNumber
=
= right justifies and pads with blanks
=
===============
*/

static void LatchNumber (int x, int y, unsigned width, int32_t number)
{
    unsigned length,c;
    char    str[20];

    ltoa (number,str,10);

    length = (unsigned) strlen (str);

    while (length<width)
    {
        StatusDrawPic (x,y,N_BLANKPIC);
        x++;
        width--;
    }

    c = length <= width ? 0 : length-width;

    while (c<length)
    {
        if (number < 0)
            StatusDrawPic (x,y,N_BLANKPIC);
        else
            StatusDrawPic (x,y,str[c]-'0'+ N_0PIC);
        x++;
        c++;
    }
}


/*
===============
=
= DrawHealth
=
===============
*/

void DrawHealth (void)
{
    if(viewsize == 21 && ingame) return;
    LatchNumber (21,16,3,gamestate.health);
}


/*
===============
=
= TakeDamage
=
===============
*/

void TakeDamage (int points,objtype *attacker)
{
    TakeSpecialDamage(points,attacker,dam_bullet,false,true);
}

void TakeSpecialDamage (int points,objtype *attacker,damage_t damagetype,boolean ignorearmor,boolean useflash)
{
    int damagepoints;
    int oldarmor;
    mecha_t *mecha;

    if (damagetype == dam_mutagen || damagetype == dam_infection)
    {
        ignorearmor = true;
    }

    if (gamestate.victoryflag || godmode)
        return;

    points = SKILLDAMAGE(points);
    if (points <= 0)
        return;

    mecha = &gamestate.mecha;

    if (Mecha_Equipped() && damagetype != dam_infection && 
        !mecha->lightWeight)
    {
        if (damagetype == dam_mutagen)
        {
            return;
        }

        if (damagetype == dam_melee)
        {
            if (!(attacker != NULL && 
                attacker->enemy->meleeHurtsMecha))
            {
                return;
            }
        }

        oldarmor = gamestate.armor;

        if (gamestate.armor >= points)
        {
            gamestate.armor -= points;
            damagepoints = 0;
        }
        else
        {
            damagepoints = points - gamestate.armor;
            gamestate.armor = 0;
        }
        DrawWeapon ();

        if (attacker != NULL && gamestate.armor < oldarmor)
        {
            if (oldarmor >= MECHA_HP_ONEARM(mecha->starthp) &&
                gamestate.armor < MECHA_HP_ONEARM(mecha->starthp))
            {
                SpawnBulletPuff(OB_GetX(player), OB_GetY(player),
                    ED_STATE("s_mechboom1"), 0, 0);
                newobj->fineangle = Mecha_GetViewingFineAngle();
                newobj->angle = newobj->fineangle * ANGLES / FINEANGLES;
                OB_AngleMove(newobj, FP(1) * 5 / 8, FP(1) / 4);
            }
            else if (oldarmor >= MECHA_HP_NOARMS(mecha->starthp) &&
                gamestate.armor < MECHA_HP_NOARMS(mecha->starthp))
            {
                SpawnBulletPuff(OB_GetX(player), OB_GetY(player),
                    ED_STATE("s_mechboom1"), 0, 0);
                newobj->fineangle = Mecha_GetViewingFineAngle();
                newobj->angle = newobj->fineangle * ANGLES / FINEANGLES;
                OB_AngleMove(newobj, FP(1) * 5 / 8, -FP(1) / 4);
            }
            else
            {
                SpawnBulletPuff(OB_GetX(player), OB_GetY(player), 
                    ED_STATE("s_firespark1"), 0, 0);
                OB_AngleMoveTo(newobj, attacker, FP(1));
                SD_PlaySound(ED_SND("RICOCHETMETALSND"));
            }
        }
    }
    else
    {
        if (Mecha_Equipped() && mecha->lightWeight && !ignorearmor)
        {
            const int armor = 100;
            points = (points/((100+armor)/50))+1;

            if (attacker != NULL)
            {
                SpawnBulletPuff(OB_GetX(player), OB_GetY(player), 
                    ED_STATE("s_firespark1"), 0, 0);
                OB_AngleMoveTo(newobj, attacker, FP(1));
                SD_PlaySound(ED_SND("RICOCHETMETALSND"));
            }
        }

        if (gamestate.armor && !ignorearmor)
        {
            damagepoints = (points/((100+gamestate.armor)/50))+1;
            gamestate.armor -= (points/2)+1;
            if (gamestate.armor < 0)
                gamestate.armor = 0;
            DrawWeapon ();
        }
        else
            damagepoints = points;
    }

    if (damagepoints <= 0)
        return;

    if (damagepoints > 0 && damagetype == dam_mutagen && 
        Player_CanInfect() && US_RndT() < 128)
    {
        if (HASWEAPON(wp_cyringe) && HASAMMO(wp_cyringe))
        {
            GetMessage("You are infected! Use the syringe!");
        }
        else
        {
            GetMessage("You are infected! Find the antidote!");
        }
        SD_PlaySound(ED_SND("INFECTEDSND"));
        gamestate.infected = 1;
    }

    gamestate.health -= damagepoints;

    if (gamestate.health<=0)
    {
        KillDamageType = damagetype;
        gamestate.health = 0;
        playstate = ex_died;
        killerobj = attacker;
    }

    if (godmode != 2 && useflash == true)
        StartDamageFlash (damagepoints);

    DrawHealth ();
    DrawFace ();

    //
    // MAKE BJ'S EYES BUG IF MAJOR DAMAGE!
    //
#ifdef SPEAR
    if (points > 30 && gamestate.health!=0 && !godmode && viewsize != 21)
    {
        StatusDrawFace(BJOUCHPIC);
        facecount = 0;
    }
#endif
}

/*
===============
=
= HealSelf
=
===============
*/

void HealSelf (int points)
{
    if (gamestate.health >= 100)
        return;
    gamestate.health += points;
    if (gamestate.health>100)
        gamestate.health = 100;

    DrawHealth ();
    DrawFace ();
}

/*
===============
=
= GiveArmor
=
===============
*/

void GiveArmor (int points)
{
    short &armor = PlayerArmor();

    if (armor >= 100)
        return;
    armor += points;
    if (armor > 100)
        armor = 100;

    DrawWeapon ();
}

/*
===============
=
= PlayerArmor
=
===============
*/

short &PlayerArmor (void)
{
    mecha_t *mecha;

    mecha = &gamestate.mecha;

    if (Mecha_Equipped() && !mecha->lightWeight)
    {
        return gamestate.mecha.playerarmor;
    }

    return gamestate.armor;
}

//===========================================================================


/*
===============
=
= DrawLevel
=
===============
*/

void DrawLevel (void)
{
    if(viewsize == 21 && ingame) return;
#ifdef SPEAR
    if (gamestate.mapon == 20)
        LatchNumber (3,16,2,18);
    else
#endif
        LatchNumber (3,16,2,gamestate.mapon+1);
}

//===========================================================================


/*
===============
=
= DrawLives
=
===============
*/

void DrawLives (void)
{
    int i, x, y;
    int picw, pich;
    pictabletype picattr;
    if(viewsize == 21 && ingame) return;
#ifdef USE_LIVES
    LatchNumber (15,16,1,gamestate.lives);
#endif

    const int sf = (int)scaleFactor;

    const int areaw = sf * (3 * 8 - 1);
    const int areah = sf * 13;
    const int areax = (screenWidth - sf * 320) / 2 +
        sf * (14 * 8 - 2);
    const int areay = screenHeight - sf * 
        (STATUSLINES - 16);
    VWB_BarScaledCoord (areax, areay, areaw, areah, 0);

    x = sf * 2;
    y = sf * 2;

    for (i = 0; i < gamestate.medkits; i++)
    {
        LatchDrawPicScaledCoord(areax + x, areay + y, MEDKITPIC, 1);
        LatchGetPicAttr(MEDKITPIC, &picattr);
        picw = picattr.width * sf;
        pich = picattr.height * sf;

        x += picw;
        if (x + picw > areaw)
        {
            x = sf * 2;
            y += pich + (sf * 2);
            if (y + pich + (sf * 2) > areah)
            {
                break;
            }
        }
    }
}


/*
===============
=
= GiveExtraMan
=
===============
*/

void GiveExtraMan (void)
{
#ifdef USE_LIVES
    if (gamestate.lives<9)
        gamestate.lives++;
    DrawLives ();
    SD_PlaySound (ED_SND("BONUS1UPSND"));
#endif
}

//===========================================================================

/*
===============
=
= DrawScore
=
===============
*/

void DrawScore (void)
{
    if(viewsize == 21 && ingame) return;
    LatchNumber (6,16,7,gamestate.score);
}

/*
===============
=
= GivePoints
=
===============
*/

void GivePoints (int32_t points)
{
    gamestate.score += points;
#ifdef USE_LIVES
    while (gamestate.score >= gamestate.nextextra)
    {
        gamestate.nextextra += EXTRAPOINTS;
        if (gamestate.lives < 9)
            GetMessage("You earned an extra life!");
        GiveExtraMan ();
    }
#endif
    DrawScore ();
}

//===========================================================================

/*
==================
=
= DrawWeapon
=
= WSJ: now draws armor
=
==================
*/

void DrawWeapon (void)
{
    if(viewsize == 21 && ingame)
    {
        return;
    }

    //StatusDrawPic (32,8,KNIFEPIC+gamestate.weapon);
    LatchNumber (33,16,3,gamestate.armor);

    const int sf = (int)scaleFactor;
    const int areax = (screenWidth - sf * 320) / 2 +
        sf * (14 * 8 - 2 + 4);
    const int areay = screenHeight - sf * 
        (STATUSLINES - 29);
    LatchDrawPicScaledCoord(areax, areay, 
        HASWEAPON(wp_wirecutters) ? WIRECUTTERSPIC : NOWIRECUTTERSPIC, 
        1);
}


/*
==================
=
= DrawKeys
=
==================
*/

void DrawKeys (void)
{
    if(viewsize == 21 && ingame) return;

    StatusDrawPic (37,4,KEY0PIC);

    if (gamestate.keys & 1)
        StatusDrawPic (37,4,KEY1PIC);

    if (gamestate.keys & 2)
        StatusDrawPic (37,12,KEY2PIC);

    if (gamestate.keys & 4)
        StatusDrawPic (37,20,KEY3PIC);

    if (gamestate.keys & 8)
        StatusDrawPic (37,28,KEY4PIC);
}

/*
==================
=
= GiveWeapon
=
==================
*/

void GiveWeapon (int weapon)
{
    if (weapon >= wp_mg42_mounted)
        return;    // WSJ: mounted weapons can only be used, not picked up

    if (MAXEDWEAPON(weapon))
    {
        return;
    }

    if (!HASWEAPON(weapon))
    {   // WSJ: first time the player has picked up the weapon
        gamestate.gotweapon[weapon] = true;
        if (gamestate.weapon<wp_mg42_mounted)
        {
            gamestate.lastchosenweapon[WEAPONKEY(weapon)] =
                (weapontype)weapon;
            gamestate.chosenweapon = (weapontype)weapon;
        }

        if (CLIPSIZE(weapon) > 0)
            gamestate.clip[weapon] = CLIPSIZE(weapon);

        if (!gamestate.attackframe && !gamestate.snipermode && 
            gamestate.weapon<wp_mg42_mounted)
        {
            ChangeWeapon(gamestate.chosenweapon);
        }

        if (gamestate.bestweapon<weapon)
            gamestate.bestweapon = (weapontype)weapon;
    }
    else if (CLIPSIZE(weapon) > 0)
        GiveAmmo (CLIPSIZE(weapon),CheckAmmoType(weapon));
}


/*
==================
=
= DiscardWeapon
=
==================
*/

void DiscardWeapon (weapontype weapon)
{
    int i;
    mecha_t *mecha;

    mecha = &gamestate.mecha;

    if (weapon == wp_mecha && Mecha_Equipped())
    {
        if (!mecha->lightWeight)
        {
            gamestate.armor = mecha->playerarmor;
        }
        lwlib_Zero(*mecha);

        weapon = BestAvailableWeapon(true);

        ChangeWeapon(weapon);
        gamestate.chosenweapon = weapon;
        gamestate.lastchosenweapon[WEAPONKEY(weapon)] = weapon;
        return;
    }

    if (weapon >= wp_mg42_mounted || !HASWEAPON(weapon))
        return;

    assert(weapon <= gamestate.bestweapon);
    if (weapon == gamestate.bestweapon && gamestate.bestweapon > 0)
    {
        gamestate.bestweapon = (weapontype)(weapon - 1);
    }

    for (i = 0; i < gamestate.bestweapon; i++)
    {
        if (HASWEAPON(i) && WEAPONKEY(i) == WEAPONKEY(weapon))
        {
            gamestate.lastchosenweapon[WEAPONKEY(i)] = (weapontype)i;
        }
    }

    gamestate.gotweapon[weapon] = false;
    gamestate.clip[weapon] = 0;

    if (gamestate.weapon == weapon)
    {
        weapon = BestAvailableWeapon(true);

        ChangeWeapon(weapon);
        gamestate.chosenweapon = weapon;
        gamestate.lastchosenweapon[WEAPONKEY(weapon)] = weapon;
    }
}


/*
==================
=
= DiedReset
=
==================
*/

void DiedReset (void)
{
    int i;
    gamestate.health = 100;
    gamestate.armor = 0;
    gamestate.medkits = 0;
    for (i=0; i<NUMWEAPONS; i++) { gamestate.gotweapon[i] = 0; gamestate.clip[i] = 0; }
    for (i=0; i<NUMBUTTONS; i++) { gamestate.lastchosenweapon[i] = (weapontype)0; }
    for (i=0; i<NUMAMMOTYPES; i++)
    {
        gamestate.ammo[i] = 0;
        gamestate.maxammoindex = 0;
    }
    gamestate.gotweapon[wp_knife] =
    gamestate.gotweapon[wp_pistol] = true;
    gamestate.clip[wp_pistol] = STARTAMMO;
    gamestate.weapon = gamestate.bestweapon
        = gamestate.chosenweapon = wp_pistol;
    gamestate.keys = 0;
    pwallstate = pwallpos = 0;
    gamestate.attackframe = gamestate.attackcount =
        gamestate.weaponframe = 0;
}

/*
==================
=
= DiscardWeaponsAtEnd
=
==================
*/

void DiscardWeaponsAtEnd (void)
{
    gamestate.snipermode = 0;

    for (int i = 0; i < NUMWEAPONS; i++)
    {
        if (weaponinfo[i].discardAtEnd)
        {
            DiscardWeapon((weapontype)i);
        }
    }
}

/*
==================
=
= DiscardAmmoAtEnd
=
==================
*/

void DiscardAmmoAtEnd (void)
{
    for (int i = 0; i < NUMAMMOTYPES; i++)
    {
        if (ammoinfo[i].discardAtEnd)
        {
            gamestate.ammo[i] = 0;
        }
    }
}

//===========================================================================

/*
===============
=
= DrawAmmo
=
===============
*/

void DrawAmmo (void)
{
    int x, y;
    int w, h;

    if(viewsize == 21 && ingame)
    {
        return;
    }

    if (CheckAmmoType(gamestate.weapon) == ammo_none)
    {
        LatchNumber (26,16,6,-1);
        return;
    }

    x = 26 * 8;
    y = 200 - STATUSLINES + 16;
    h = STATUSLINES - 24;
    VWB_Bar(x, y, 48, h, 0);

    if (Mecha_Equipped())
    {
        if (Mecha_FuelLow())
        {
            w = 48 * gamestate.ammo[ammo_fuel] / MECHA_LOW_FUEL;
            if (w > 0)
            {
                VWB_Bar(x, y, w, h, LOWFUELBARCOLOR);
                DrawOutline(x, y, w - 1, h - 1, LOWFUELBARHLIN, 
                    LOWFUELBARVLIN);
            }
        }
        else
        {
            w = 48 * gamestate.ammo[ammo_fuel] / CURMAXAMMO(ammo_fuel);
            if (w > 0)
            {
                VWB_Bar(x, y, w, h, FUELBARCOLOR);
                DrawOutline(x, y, w - 1, h - 1, FUELBARHLIN, FUELBARVLIN);
            }
        }
    }
    else
    {
        if (CLIPSIZE(gamestate.weapon) > 0)
            LatchNumber (26,16,2,CURCLIP);
        else
            LatchNumber (26,16,3,-1);
        LatchNumber (29,16,3,CURAMMO);
    }
}

/*
===============
=
= GiveAmmo
=
===============
*/

void GiveAmmo (int ammo, ammo_t ammotype)
{
    gamestate.ammo[ammotype] += ammo;
    if (HASMAXAMMO(ammotype))
        gamestate.ammo[ammotype] = CURMAXAMMO(ammotype);
    if (gamestate.weapon!=gamestate.chosenweapon &&
        WeaponAvailable(gamestate.chosenweapon) && 
        !gamestate.attackframe && gamestate.weapon<wp_mg42_mounted)
    {
        ChangeWeapon(gamestate.chosenweapon);
    }

    DrawAmmo ();
}

/*
===============
=
= ConsumeAmmo
=
===============
*/

void ConsumeAmmo (int ammo)
{
    if (ammo > CURAMMO)
    {
        ammo = CURAMMO;
    }

    if (ammo > 0)
    {
        CURAMMO -= ammo;
        DrawAmmo();

        if (Mecha_Equipped())
        {
            if (CURAMMO == 0)
            {
                GetMessage ("Your Mecha is out of fuel!");
                SD_PlaySound (ED_SND("MECHAOUTOFFUELSND"));

                Mecha_StopRunningCharge();
            }
            else if (CURAMMO + ammo >= MECHA_LOW_FUEL && 
                CURAMMO < MECHA_LOW_FUEL)
            {
                GetMessage ("Your Mecha is running out of fuel!");
                SD_PlaySound (ED_SND("MECHAALARMSND"));
            }
        }
    }
}

//===========================================================================

/*
==================
=
= GiveKey
=
==================
*/

void GiveKey (int key)
{
    gamestate.keys |= (1<<key);
    DrawKeys ();
}

/*
==================
=
= GiveLives
=
==================
*/

void GiveLives (int lives)
{
#ifdef USE_LIVES
    gamestate.lives += lives;
    if (gamestate.lives > MAXLIVES)
    {
        gamestate.lives = MAXLIVES;
    }
    DrawLives ();
#endif
}



/*
=============================================================================

                                MOVEMENT

=============================================================================
*/


/*
===================
=
= CheckExitType
=
= WSJ: checks whether to go to the next level or the secret level
=
===================
*/

exit_t CheckExitType (void)
{
    exit_t exittype;

    switch (MAPSPOT(player->tilex,player->tiley,0))
    {
        case 0:
        case ALTELEVATORTILE:
            exittype = ex_secretlevel;
            break;
        default:
            exittype = ex_completed;
    }

    return exittype;
}

//===========================================================================

/*
===================
=
= CanGetBonus
=
===================
*/
bool CanGetBonus(statobj_t *check)
{
    int i;
    ED_Bonus_t *bonus;
    ED_AmmoBonus_t *ammoBonus;

    if (check->bonusIndex == -1)
    {
        return false;
    }

    bonus = &ED_Bonuses(ed)[check->bonusIndex];

    if (bonus->usePickup)
    {
        if (buttonstate[bt_use] && !buttonheld[bt_use])
        {
            // continue
        }
        else
        {
            return false;
        }
    }

    if (bonus->healthPoints > 0 && gamestate.health < 100)
    {
        return true;
    }

    for (i = 0; (size_t)i < lwlib_CountArray(bonus->ammoBonuses); i++)
    {
        ammoBonus = &bonus->ammoBonuses[i];
        if (ammoBonus->ammoType != ammo_none && ammoBonus->ammo > 0 &&
            !HASMAXAMMO(ammoBonus->ammoType))
        {
            return true;
        }
    }

    if (bonus->isKey)
    {
        return true;
    }

    if (bonus->weapon != wp_none && !MAXEDWEAPON(bonus->weapon))
    {
        return true;
    }

    if (bonus->armorPoints > 0)
    {
        short &armor = PlayerArmor();
        if (bonus->armorJumpTo)
        {
            if (armor < bonus->armorPoints)
            {
                return true;
            }
        }
        else
        {
            if (armor < 100)
            {
                return true;
            }
        }
    }

    if (bonus->regenHealth && !gamestate.gotregenhealth)
    {
        return true;
    }

    if (bonus->regenArmor && !gamestate.gotregenarmor)
    {
        return true;
    }

#ifdef USE_LIVES
    if (bonus->lives && gamestate.lives < MAXLIVES)
    {
        return true;
    }
#endif

    if (bonus->damagePoints > 0)
    {
        return true;
    }

    if (bonus->scorePoints > 0)
    {
        return true;
    }

    if (bonus->increaseMedkitLimit && gamestate.medkitlim < MAX_MEDKITS)
    {
        return true;
    }

    if (bonus->medkits > 0 && gamestate.medkits < gamestate.medkitlim)
    {
        return true;
    }

    if (bonus->increaseAmmoLimit && gamestate.maxammoindex < 
        NUMAMMOLIMITS - 1)
    {
        return true;
    }

    if (bonus->giveBackpackAmmo)
    {
        for (i = 0; i < NUMAMMOTYPES; i++)
        {
            if (!HasWeaponForAmmoType((ammo_t)i))
            {
                continue;
            }
            if (gamestate.ammo[i] < CURMAXAMMO(i))
            {
                return true;
            }
        }
    }

    return false;
}

/*
===================
=
= GetBonus
=
===================
*/
void GetBonus (statobj_t *check)
{
    int i;
    ED_Bonus_t *bonus;
    ED_AmmoBonus_t *ammoBonus;

    if (playstate == ex_died)
        return;    // WSJ: don't pick up anything while dying!

    boolean usedsprite = false;

    if (check->bonusIndex >= 0)
    {
        if (!CanGetBonus(check))
        {
            return;
        }

        bonus = &ED_Bonuses(ed)[check->bonusIndex];
        if (bonus->healthPoints > 0)
        {
            HealSelf(bonus->healthPoints);
        }

        if (bonus->isKey)
        {
            GiveKey(bonus->key);
        }

        if (bonus->message[0] != '\0')
        {
            GetMessage(bonus->message);
        }

        if (bonus->sound[0] != '\0')
        {
            SD_PlaySound(ED_SND(bonus->sound));
        }

        if (bonus->scorePoints > 0)
        {
            GivePoints(bonus->scorePoints);
        }

        if (bonus->isTreasure)
        {
            gamestate.treasurecount++;
        }

        for (i = 0; (size_t)i < lwlib_CountArray(bonus->ammoBonuses); i++)
        {
            ammoBonus = &bonus->ammoBonuses[i];
            if (ammoBonus->ammoType != ammo_none && ammoBonus->ammo > 0)
            {
                GiveAmmo(ammoBonus->ammo, ammoBonus->ammoType);
            }
        }

        if (bonus->weapon != wp_none)
        {
            GiveWeapon(bonus->weapon);
        }

        if (bonus->armorPoints > 0)
        {
            if (bonus->armorJumpTo)
            {
                short &armor = PlayerArmor();
                armor = bonus->armorPoints;
            }
            else
            {
                GiveArmor(bonus->armorPoints);
            }
        }

        if (bonus->regenHealth)
        {
            gamestate.gotregenhealth = true;
        }

        if (bonus->regenArmor)
        {
            gamestate.gotregenarmor = true;
        }

        if (bonus->lives)
        {
            GiveLives(bonus->lives);
        }

        if (bonus->usedSprite)
        {
            usedsprite = true;
        }

        if (bonus->damagePoints > 0)
        {
            TakeDamage(bonus->damagePoints, NULL);
        }

        if (bonus->usePickup && !usedsprite)
        {
            actorat[check->tilex][check->tiley] = NULL;
        }

        if (bonus->increaseMedkitLimit && gamestate.medkitlim < MAX_MEDKITS)
        {
            gamestate.medkitlim += 1;
        }

        if (bonus->medkits > 0 && gamestate.medkits < gamestate.medkitlim)
        {
            gamestate.medkits += bonus->medkits;
            if (gamestate.medkits > gamestate.medkitlim)
            {
                gamestate.medkits = gamestate.medkitlim;
            }
            DrawLives();
        }

        if (bonus->logChoice[0] != '\0')
        {
            LogDiscScreens(bonus->logChoice);
        }

        if (bonus->increaseAmmoLimit && 
            gamestate.maxammoindex < NUMAMMOLIMITS - 1)
        {
            gamestate.maxammoindex++;
        }

        if (bonus->giveBackpackAmmo)
        {
            for (i = 0; i < NUMAMMOTYPES; i++)
            {
                if (!HasWeaponForAmmoType((ammo_t)i))
                {
                    continue;
                }
                GiveAmmo(ammoinfo[i].giveBackpackAmmo, (ammo_t)i);
            }
        }
    }

    if (check->itemnumber == bo_mg42_mounted)
    {
        return;
    }

    StartBonusFlash ();
    check->flags &= ~FL_BONUS;
    if (usedsprite)
    {
        // WSJ: item doesn't disappear, but goes to "used state"
        if (ED_SPR(bonus->usedSpriteDef) != -1)
        {
            check->shapenum = ED_SPR(bonus->usedSpriteDef);
        }
        else
        {
            check->shapenum += 1;
        }
    }
    else
    {
        RemoveStaticObj(check);
    }
}

/*
===================
=
= TryMove
=
= returns true if move ok
= debug: use pointers to optimize
===================
*/

bool TryMove (objtype *ob, TryMoveInfo_t *tryMoveInfo)
{
    int         xl,yl,xh,yh,x,y;
    objtype    *check;
    int32_t     deltax,deltay;

    lwlib_Zero(*tryMoveInfo);

    xl = (OB_GetX(ob)-PLAYERSIZE) >>TILESHIFT;
    yl = (OB_GetY(ob)-PLAYERSIZE) >>TILESHIFT;

    xh = (OB_GetX(ob)+PLAYERSIZE) >>TILESHIFT;
    yh = (OB_GetY(ob)+PLAYERSIZE) >>TILESHIFT;

#define PUSHWALLMINDIST PLAYERSIZE

    //
    // check for solid walls
    //
    for (y=yl;y<=yh;y++)
    {
        for (x=xl;x<=xh;x++)
        {
            if 
                (
                    MAPSPOT(x,y,1) == MG42TILE || 
                    (
                        MAPSPOT(x,y,1) == ENDLEVELTILE &&
                        !(
                            gamestate.weaponframe == 0 &&
                            gamestate.attackframe == 0
                        )
                    )
                )
            {
                tryMoveInfo->hitSolid = true;
                return false;
            }

            check = actorat[x][y];
            if (check && !ISPOINTER(check))
            {
                if(tilemap[x][y]==64 && x==pwallx && y==pwally)
                    // back of moving pushwall?
                {
                    switch(pwalldir)
                    {
                        case di_north:
                            if(OB_GetY(ob)-PUSHWALLMINDIST<=
                                (pwally<<TILESHIFT)+((63-pwallpos)<<10))
                            {
                                tryMoveInfo->hitSolid = true;
                                return false;
                            }
                            break;
                        case di_west:
                            if(OB_GetX(ob)-PUSHWALLMINDIST<=
                                (pwallx<<TILESHIFT)+((63-pwallpos)<<10))
                            {
                                tryMoveInfo->hitSolid = true;
                                return false;
                            }
                            break;
                        case di_east:
                            if(OB_GetX(ob)+PUSHWALLMINDIST>=
                                (pwallx<<TILESHIFT)+(pwallpos<<10))
                            {
                                tryMoveInfo->hitSolid = true;
                                return false;
                            }
                            break;
                        case di_south:
                            if(OB_GetY(ob)+PUSHWALLMINDIST>=
                                (pwally<<TILESHIFT)+(pwallpos<<10))
                            {
                                tryMoveInfo->hitSolid = true;
                                return false;
                            }
                            break;
                    }
                }
                else
                {
                    tryMoveInfo->hitSolid = true;
                    return false;
                }
            }
        }
    }

    //
    // check for actors
    //
    if (yl>0)
        yl--;
    if (yh<MAPSIZE-1)
        yh++;
    if (xl>0)
        xl--;
    if (xh<MAPSIZE-1)
        xh++;

    for (y=yl;y<=yh;y++)
    {
        for (x=xl;x<=xh;x++)
        {
            check = actorat[x][y];
            if (ISPOINTER(check) && check != player && (check->flags & FL_BLOCK) )
            {
                deltax = OB_GetX(ob) - OB_GetX(check);
                if (deltax < -MINACTORDIST || deltax > MINACTORDIST)
                    continue;
                deltay = OB_GetY(ob) - OB_GetY(check);
                if (deltay < -MINACTORDIST || deltay > MINACTORDIST)
                    continue;

                tryMoveInfo->hitObj = check;
                return false;
            }
        }
    }

    return true;
}


/*
===================
=
= ClipMove
=
===================
*/

void ClipMove (objtype *ob, int32_t xmove, int32_t ymove)
{
    int32_t    basex,basey;
    TryMoveInfo_t tryMoveInfo;

    basex = OB_GetX(ob);
    basey = OB_GetY(ob);

    OB_SetX(ob, basex+xmove);
    OB_SetY(ob, basey+ymove);
    if (TryMove (ob, &tryMoveInfo))
        return;

    if (tryMoveInfo.hitSolid)
    {
        Mecha_NotifyHitSolid();
    }
    if (tryMoveInfo.hitObj != NULL)
    {
        Mecha_NotifyHitObj(tryMoveInfo.hitObj);
    }

#ifndef REMDEBUG
    if (noclip && OB_GetX(ob) > 2*TILEGLOBAL && OB_GetY(ob) > 2*TILEGLOBAL
        && OB_GetX(ob) < (((int32_t)(mapwidth-1))<<TILESHIFT)
        && OB_GetY(ob) < (((int32_t)(mapheight-1))<<TILESHIFT) )
        return;         // walk through walls
#endif

    OB_SetX(ob, basex+xmove);
    OB_SetY(ob, basey);
    if (TryMove (ob, &tryMoveInfo))
        return;

    OB_SetX(ob, basex);
    OB_SetY(ob, basey+ymove);
    if (TryMove (ob, &tryMoveInfo))
        return;

    OB_SetX(ob, basex);
    OB_SetY(ob, basey);
}

//==========================================================================

/*
===================
=
= VictoryTile
=
===================
*/

void VictoryTile (void)
{
    playstate = ex_victorious;
}

/*
===================
=
= WeaponBobble_Update
=
===================
*/
void WeaponBobble_Update(void)
{
#ifdef WEAPONBOBBLE
    #define WEAPONBOBBER_UNIT (FP(1) / 40)
    #define WEAPONBOBBER_COUNTERPERIOD 8
    #define WEAPONBOBBER_MECHA_SWING 0//(ANGLES * 1 / 32)
    #define WEAPONBOBBER_MECHA_SWAY 0//(FP(1) / 4)

    int volume;
    int floorTile;
    fixed dist, speed;
    bool playSound; 
    const char *footStepSound;
    mecha_t *mecha;
    fixed u, v, t;
    objtype tmpob;

    mecha = &gamestate.mecha;

    playSound = false;

    if (mecha->moveVel != 0)
    {
        dist = fixedpt_abs(mecha->moveVel) * tics *
            mecha->bobbleFactor;
    }
    else
    {
        dist = Math_Distance2(OB_GetX(player), OB_GetY(player), 
            gamestate.bobberx, gamestate.bobbery);
    }

    gamestate.bobberdist += dist;
    while (gamestate.bobberdist >= WEAPONBOBBER_UNIT)
    {
        if (gamestate.bobdir != 1)
        {
            gamestate.bobcnt++;
        }
        else
        {
            gamestate.bobcnt--;
        }

        if (gamestate.bobdir2 != 1)
        {
            gamestate.bobcnt2++;
        }
        else
        {
            gamestate.bobcnt2--;
        }

        if (gamestate.bobcnt % WEAPONBOBBER_COUNTERPERIOD == 0)
        {
            if (gamestate.bobdir != 1)
                gamestate.bobber++;
            else
                gamestate.bobber--;
            if (gamestate.bobber >= 5 || gamestate.bobber <= 0)
            {
                gamestate.bobdir ^= 1;
            }
            if (gamestate.bobdir2 != 1)
                gamestate.bobber2++;
            else
                gamestate.bobber2--;
            if (gamestate.bobber2 >= 10 || gamestate.bobber2 <= 0)
            {
                gamestate.bobdir2 ^= 1;
                playSound = true;

                if (Mecha_Equipped())
                {
                    gamestate.bobberdist = WEAPONBOBBER_UNIT;
                    mecha->moveVel = 0;
                }
            }
        }

        gamestate.bobberdist -= WEAPONBOBBER_UNIT;
    }

    if (playSound)
    {
        floorTile = MAPSPOT(player->tilex, player->tiley, 2) & 0xff;
        #ifdef DEBUG_FOOTSTEP
        fprintf(stderr, "Stepped on floor tile %d\n", floorTile);
        #endif

        if (Mecha_Equipped())
        {
            footStepSound = ED_FootStepByFloorTile(ed, 256);

            // fuel drain while moving
            ConsumeAmmo(1);
        }
        else
        {
            footStepSound = ED_FootStepByFloorTile(ed, floorTile);
            if (footStepSound == NULL)
            {
                footStepSound = ED_FootStepByFloorTile(ed, 0);
            }
        }

        if (footStepSound != NULL)
        {
            speed = tics > 0 ? dist / tics : 0;
            speed = lwlib_MIN(speed, FOOTSTEP_MAX_SPEED);

            volume = fixedpt_toint(fixedpt_div(speed, FOOTSTEP_MAX_SPEED) *
                FOOTSTEP_MAX_VOLUME);
            volume = lwlib_MIN(volume, FOOTSTEP_MAX_VOLUME);

            SD_PlaySoundVolume(ED_SND(footStepSound), volume);
        }
    }

    if (Mecha_Equipped())
    {
        t = FP(gamestate.bobcnt) / 40;
        const int bobbleFactorOrOne = mecha->bobbleFactor ? 
            mecha->bobbleFactor : 1;
        gamestate.camz = (FP(1) / 8) + (t / (6 * bobbleFactorOrOne));

        t = FP(gamestate.bobcnt2) / 80;
        mecha->motionTurnAngle = (WEAPONBOBBER_MECHA_SWING / 2) - 
            fixedpt_toint(WEAPONBOBBER_MECHA_SWING * t);

        // fuel drain while standing
        if (dist == 0)
        {
            mecha->standTics += tics;
            if (mecha->standTics >= 60)
            {
                mecha->standTics = 0;
                if (!mecha->lightWeight)
                {
                    ConsumeAmmo(1);
                }
            }
        }

        u = 0;//gamestate.camz * 2;
        v = -(WEAPONBOBBER_MECHA_SWAY / 2) + 
            fixedpt_mul(WEAPONBOBBER_MECHA_SWAY, t);
        lwlib_Zero(tmpob);
        tmpob.angle = player->angle;
        tmpob.fineangle = player->fineangle;
        OB_AngleMove(&tmpob, u, v);
        gamestate.camx = OB_GetX(&tmpob);
        gamestate.camy = OB_GetY(&tmpob);
    }
    else
    {
        gamestate.camz = 0;
        gamestate.camx = 0;
        gamestate.camy = 0;
    }

#endif
}

/*
===================
=
= WeaponBobble_Reset
=
===================
*/
void WeaponBobble_Reset(void)
{
    gamestate.bobberdist = gamestate.bobdir = gamestate.bobdir2 =
        gamestate.bobber = gamestate.bobber2 = gamestate.bobcnt =
        gamestate.bobcnt2 = 0;
}

//========================================================================

#define TREMOR_REFRESHTICS 4
#define TREMOR_MAXMAGNITUDE 0.2
#define TREMOR_WEAPONBOBBLEFACTOR 200
#define TREMOR_WEAPONBOBBLEMAXX 20
#define TREMOR_WEAPONBOBBLEMAXY 20

static tremorprofile_t tremorProfiles[trem_max] =
{
    {
        16, // tileRadius
        0.1, // magnitude
        0.3, // mechaFactor
        MILLI2TICS(300), // duration
        true, // localized
        false, // directional
        {
            0.0, 
            0.1 / MILLI2TICS(300),
            0.0,
        }, // falloff
        {
            0.0, 
            1.0,
            0.0,
        }, // camDistFallOff
        false, // playerFocus
    }, // trem_mechastep
    {
        20, // tileRadius
        0.2, // magnitude
        0.6, // mechaFactor
        MILLI2TICS(400), // duration
        true, // localized
        true, // directional
        {
            0.0, 
            0.2 / MILLI2TICS(400),
            0.0,
        }, // falloff
        {
            0.0, 
            1.0,
            0.0,
        }, // camDistFallOff
        false, // playerFocus
    }, // trem_explosion
    {
        0, // tileRadius
        0.1, // magnitude
        0.0, // mechaFactor
        MILLI2TICS(250), // duration
        true, // localized
        false, // directional
        {
            0.0, 
            0.1 / MILLI2TICS(250),
            0.0,
        }, // falloff
        {
            0.0, 
            0.0,
            0.0,
        }, // camDistFallOff
        true, // playerFocus
    }, // trem_shotgun
    {
        0, // tileRadius
        0.2, // magnitude
        0.0, // mechaFactor
        MILLI2TICS(400), // duration
        true, // localized
        false, // directional
        {
            0.2 / lwlib_Sq(MILLI2TICS(400)), 
            0.0,
            0.0,
        }, // falloff
        {
            0.0, 
            0.0,
            0.0,
        }, // camDistFallOff
        true, // playerFocus
    }, // trem_shotgunalt
    {
        0, // tileRadius
        0.08, // magnitude
        0.0, // mechaFactor
        MILLI2TICS(200), // duration
        true, // localized
        false, // directional
        {
            0.0, 
            0.08 / MILLI2TICS(200),
            0.0,
        }, // falloff
        {
            0.0, 
            0.0,
            0.0,
        }, // camDistFallOff
        true, // playerFocus
    }, // trem_stg44
};

/*
===================
=
= Tremor_Update
=
===================
*/

void Tremor_Update(void)
{
    int i;
    double distSq, dist;
    double magnitude;
    double *falloff, *camDistFallOff;
    double dirLenSq;
    tremor_t *tremor;
    tremorsource_t *tremorSource;
    tremorprofile_t *tremorProfile;
    lwlib::Point2f dir, uVec, vVec;

    tremor = &gamestate.tremor;

    tremor->tics += tics;
    if (tremor->tics < TREMOR_REFRESHTICS)
    {
        return;
    }

    dir = lwlib::vec2f_zero();

    for (i = 0; i < TREMOR_MAX_SOURCES; i++)
    {
        tremorSource = &tremor->sources[i];
        if (!tremorSource->active)
        {
            continue;
        }
        tremorProfile = &tremorProfiles[tremorSource->profileType];

        tremorSource->totalTics += tremor->tics;
        if (tremorProfile->duration != 0 && tremorSource->totalTics >= 
            tremorProfile->duration)
        {
            tremorSource->active = false;
            continue;
        }

        falloff = tremorProfile->falloff;
        magnitude = tremorProfile->magnitude - 
            (falloff[0] * lwlib_Sq(tremorSource->totalTics) + falloff[1] * 
            tremorSource->totalTics + falloff[2]);
        if (magnitude <= 0)
        {
            tremorSource->active = false;
            continue;
        }

        if (!tremorProfile->playerFocus)
        {
            if (abs(player->tilex - tremorSource->tilex) > 
                tremorProfile->tileRadius ||
                abs(player->tiley - tremorSource->tiley) > 
                tremorProfile->tileRadius)
            {
                continue;
            }

            camDistFallOff = tremorProfile->camDistFallOff;

            distSq = lwlib_Sq(fixedpt_toconst(player->x - tremorSource->x)) +
                lwlib_Sq(fixedpt_toconst(player->y - tremorSource->y));

            dist = 0;
            if (camDistFallOff[1] != 0)
            {
                dist = sqrt(distSq);
            }

            if (Mecha_Equipped())
            {
                magnitude *= tremorProfile->mechaFactor;
            }

            magnitude = magnitude / 
                (camDistFallOff[0] * distSq + camDistFallOff[1] * dist + 
                camDistFallOff[2]);
        }

        dir = vec_add(dir, vec_scale(tremorSource->dir, magnitude));
    }

    dirLenSq = vec_length_sq(dir);
    if (dirLenSq > 0)
    {
        if (dirLenSq >= lwlib_Sq(TREMOR_MAXMAGNITUDE))
        {
            dir = vec_scale(vec_normalize(dir), TREMOR_MAXMAGNITUDE);
        }

        tremor->camxOff = fixedpt_rconst(lwlib_X(dir));
        tremor->camyOff = fixedpt_rconst(lwlib_Y(dir));

        uVec = OB_PointFromView(player, 1.0, 0.0);
        vVec = OB_PointFromView(player, 0.0, 1.0);

        tremor->weaponx = -vec_dot(dir, vVec) * TREMOR_WEAPONBOBBLEFACTOR;
        tremor->weapony = -vec_dot(dir, uVec) * TREMOR_WEAPONBOBBLEFACTOR;
        if (tremor->weapony < 0)
        {
            tremor->weapony = 0;
        }

        tremor->weaponx = lwlib_CLIP(tremor->weaponx,
            -TREMOR_WEAPONBOBBLEMAXX, TREMOR_WEAPONBOBBLEMAXX);
        tremor->weapony = lwlib_CLIP(tremor->weapony,
            -TREMOR_WEAPONBOBBLEMAXY, TREMOR_WEAPONBOBBLEMAXY);
    }
    else
    {
        tremor->camxOff = 0;
        tremor->camyOff = 0;
        tremor->weaponx = 0;
        tremor->weapony = 0;
    }
    tremor->tics = 0;
}

/*
===================
=
= Tremor_Start
=
===================
*/

void Tremor_Start(tremorprofiletype_t profileType, fixed x, fixed y,
    int angle)
{
    int i;
    double radAngle;
    tremor_t *tremor;
    tremorsource_t *tremorSource;

    tremor = &gamestate.tremor;

    for (i = 0; i < TREMOR_MAX_SOURCES; i++)
    {
        tremorSource = &tremor->sources[i];
        if (!tremorSource->active)
        {
            break;
        }
    }
    if (i == TREMOR_MAX_SOURCES)
    {
        return;
    }

    radAngle = lwlib_DegToRad(angle);

    lwlib_Zero(*tremorSource);
    tremorSource->active = true;
    tremorSource->x = x;
    tremorSource->y = y;
    tremorSource->tilex = fixedpt_toint(x);
    tremorSource->tiley = fixedpt_toint(y);
    tremorSource->totalTics = 0;
    tremorSource->profileType = profileType;
    tremorSource->dir = lwlib::vec2f(cos(radAngle), -sin(radAngle));
}

/*
===================
=
= Tremor_ObjStart
=
===================
*/

void Tremor_ObjStart(tremorprofiletype_t profileType, objtype *ob)
{
    int angle;
    tremorprofile_t *tremorProfile;

    tremorProfile = &tremorProfiles[profileType];

    if (tremorProfile->directional)
    {
        angle = OB_AngleTo(ob, player);
    }
    else
    {
        angle = US_RndT_Range(0, ANGLES - 1);
    }

    Tremor_Start(profileType, OB_GetX(ob), OB_GetY(ob), angle);
}

/*
===================
=
= Tremor_WeaponStart
=
===================
*/

void Tremor_WeaponStart(tremorprofiletype_t profileType)
{
    Tremor_Start(profileType, OB_GetX(player), OB_GetY(player), 
        player->angle + 180);
}

//========================================================================

/*
===================
=
= Thrust
=
===================
*/

// For player movement in demos exactly as in the original Wolf3D v1.4 source code
static fixed FixedByFracOrig(fixed a, fixed b)
{
    int sign = 0;
    if(b == 65536) b = 65535;
    else if(b == -65536) b = 65535, sign = 1;
    else if(b < 0) b = (-b), sign = 1;

    if(a < 0)
    {
        a = -a;
        sign = !sign;
    }
    fixed res = (fixed)(((int64_t) a * b) >> 16);
    if(sign)
        res = -res;
    return res;
}

void Thrust (int angle, int32_t speed)
{
    int32_t xmove,ymove;
    unsigned offset;


    thrustspeed += speed;
    //
    // moving bounds speed
    //
    if (speed >= MINDIST*2)
        speed = MINDIST*2-1;

    xmove = DEMOCHOOSE_ORIG_SDL(
                FixedByFracOrig(speed, finecostable[angle]),
                FixedMul(speed,finecostable[angle]));
    ymove = DEMOCHOOSE_ORIG_SDL(
                -FixedByFracOrig(speed, finesintable[angle]),
                -FixedMul(speed,finesintable[angle]));

    ClipMove(player,xmove,ymove);

    player->tilex = (short)(OB_GetX(player) >> TILESHIFT);                // scale to tile values
    player->tiley = (short)(OB_GetY(player) >> TILESHIFT);

    offset = (player->tiley<<mapshift)+player->tilex;
    if ( (*(mapsegs[0] + offset)) >= AREATILE)
        player->areanumber = *(mapsegs[0] + offset) -AREATILE;
    if (!areabyplayer[player->areanumber] && player->areanumber<NUMAREAS)
        ConnectAreas ();    // WSJ: added so player can cross over to other area codes without using doors

    switch (*(mapsegs[1] + offset))
    {
        case EXITTILE:
            VictoryTile ();
            break;
        case ENDLEVELTILE:
            playstate = CheckExitType();
            break;
        case ALERTENEMYTILE:
            lastalertedarea = alertedarea = player->areanumber;
            break;
    }
}


/*
=============================================================================

                                ACTIONS

=============================================================================
*/


/*
===============
=
= EnterMountedWeapon
=
= WSJ: player spawns at the position of the mounted
= weapon and takes control of it
=
===============
*/

bool EnterMountedWeapon (weapontype weapon, int newx, int newy, 
    controldir_t dir)
{
    lwlib_String_t message;

    if (!HASAMMO(weapon))
    {
        GetMessage(lwlib_VaString(message, 
            "Not enough %s to control %s!",
            ammoinfo[CheckAmmoType(weapon)].name,
            weaponinfo[weapon].name));
    }
    else
    {
        GetMessage(lwlib_VaString(message, "You now control the %s!",
            weaponinfo[weapon].name));
    }

    gamestate.chosenweapon = gamestate.weapon;

    buttonstate[bt_attack] = false;
    buttonheld[bt_attack] = false;
    gamestate.attackframe = 0;
    gamestate.weaponframe = 0;
    SpawnPlayer(newx, newy, dir);
    gamestate.weapon = weapon;
    DrawWeapon ();
    DrawAmmo ();
    SD_PlaySound (ED_SND(weaponinfo[weapon].mountedUseSound));

    return true;
}

/*
===============
=
= ExitMountedWeapon
=
= WSJ: player leaves the mounted weapon and spawns at a
= position that's one tile away from it
=
===============
*/

#define CHECKEXITSPOT(x,y)                            \
{                                                     \
    blocked = false;                                  \
    check = actorat[x][y];                            \
    if (check && !ISPOINTER(check))                   \
        blocked = true;                               \
    if (ISPOINTER(check) && (check->flags & FL_BLOCK) != 0) \
        blocked = true;                               \
    if (!foundspot && !blocked)                       \
    {                                                 \
        newx = x;                                     \
        newy = y;                                     \
        foundspot = true;                             \
    }                                                 \
}

void ExitMountedWeapon (controldir_t dir)
{
    int        x,y,newx,newy,facedir;
    boolean    foundspot,blocked;
    objtype    *check;
    lwlib_String_t message;

    x = player->tilex;
    y = player->tiley;
    foundspot = false;

    switch (dir)
    {
    case di_east:
        facedir = EAST;
        CHECKEXITSPOT(x-1,y);
        CHECKEXITSPOT(x,y-1);
        CHECKEXITSPOT(x,y+1);
        CHECKEXITSPOT(x+1,y);
        break;
    case di_north:
        facedir = NORTH;
        CHECKEXITSPOT(x,y+1);
        CHECKEXITSPOT(x+1,y);
        CHECKEXITSPOT(x-1,y);
        CHECKEXITSPOT(x,y-1);
        break;
    case di_west:
        facedir = WEST;
        CHECKEXITSPOT(x+1,y);
        CHECKEXITSPOT(x,y-1);
        CHECKEXITSPOT(x,y+1);
        CHECKEXITSPOT(x-1,y);
        break;
    case di_south:
        facedir = SOUTH;
        CHECKEXITSPOT(x,y-1);
        CHECKEXITSPOT(x+1,y);
        CHECKEXITSPOT(x-1,y);
        CHECKEXITSPOT(x,y+1);
        break;
    }

    if (!foundspot)
    {
        GetMessage ("Not enough room to leave!");
        return;
    }

    GetMessage (lwlib_VaString(message, "You stopped using the %s",
        weaponinfo[gamestate.weapon].name));
    SD_PlaySound (ED_SND(weaponinfo[gamestate.weapon].mountedExitSound));

    buttonstate[bt_attack] = false;
    buttonheld[bt_attack] = false;
    gamestate.attackframe = 0;
    gamestate.weaponframe = 0;
    SpawnPlayer(newx,newy,facedir);
    gamestate.weapon = gamestate.chosenweapon;
    DrawPlayScreen();
}

//===========================================================================


/*
===============
=
= Cmd_Fire
=
===============
*/

void NewWeaponFrame (short frame, boolean reload)
{
    gamestate.weaponframe = 0;

    if (reload == 1)
        OB_SetState(player, ED_STATE("s_reload"));
    else
        OB_SetState(player, ED_STATE("s_attack"));

    gamestate.attackframe = frame;
    gamestate.attackcount =
        attackinfo[gamestate.weapon][reload][gamestate.attackframe].tics;
    gamestate.weaponframe =
        attackinfo[gamestate.weapon][reload][gamestate.attackframe].frame;
}

void Cmd_Fire (void)
{
    if (!WeaponAvailable(gamestate.weapon) || 
        !WeaponVillageFriendly(gamestate.weapon))
    {
        return;
    }

    if (Mecha_Equipped() && Mecha_HasNoArms())
    {
        Mecha_StartRunningCharge();
        return;
    }

    if (gamestate.weapon >= wp_mg42_mounted &&
        CURAMMOTYPE != ammo_none && CURAMMO == 0)
    {
        return;
    }

    buttonheld[bt_attack] = true;

    if (gamestate.weapon==wp_shotgun && CURCLIP==1)
        NewWeaponFrame (4,0);
    else if (gamestate.weapon==wp_sniper && gamestate.snipermode)
        NewWeaponFrame (3,0);
    else
        NewWeaponFrame (0,0);
}

//===========================================================================

/*
===============
=
= Cmd_AltFire
=
===============
*/

void Cmd_AltFire (void)
{
    if (!WeaponAvailable(gamestate.weapon) || 
        !WeaponVillageFriendly(gamestate.weapon))
    {
        return;
    }

    if (gamestate.attackframe)
        return;

    buttonheld[bt_altattack] = true;

    switch (gamestate.weapon)
    {
        case wp_sniper:
            gamestate.snipermode ^= 1;
            break;
        case wp_shotgun:
            if (CURCLIP == CLIPSIZE(gamestate.weapon))
            {
                NewWeaponFrame (8,0);
            }
            else if (CURCLIP == 1)
            {
                NewWeaponFrame (4,0);
            }
            break;
        case wp_mecha:
            Mecha_StartRunningCharge();
            break;
        default:
            break;
    }
}

//===========================================================================

/*
===============
=
= Cmd_Reload
=
===============
*/

void Cmd_Reload (void)
{   
    if (!CLIPSIZE(gamestate.weapon) || CURAMMOTYPE == ammo_none)
        return;    // weapon never reloads

    if (!CURAMMO || CURCLIP == CLIPSIZE(gamestate.weapon))
        return;    // can't reload - either clip is full, or there's no ammo left in the reserve

    buttonheld[bt_reload] = true;

    gamestate.snipermode = 0;

    if (gamestate.weapon==wp_shotgun && CURCLIP==1)
        NewWeaponFrame (8,1);
    else
        NewWeaponFrame (0,1);
}

/*
===============
=
= ReloadWeapon
=
===============
*/

void ReloadWeapon (int ammo)
{
    if (!CURAMMO)
        return;

    if (ammo > CURAMMO)
        ammo = CURAMMO;

    CURCLIP += ammo;
    CURAMMO -= ammo;

    DrawAmmo ();
}

//===========================================================================

/*
===============
=
= Cmd_Use
=
===============
*/

void Cmd_Use (void)
{
    int     checkx,checky,doornum;
    controldir_t dir;
#ifdef USE_ELEVATORS
    boolean elevatorok;
#endif
    Player_Facing_t facing;
    void (*think) (objtype *);

    //
    // find which cardinal direction the player is facing
    //
    Player_GetFacing(&facing);
    dir = facing.dir;
    checkx = facing.checkx;
    checky = facing.checky;
#ifdef USE_ELEVATORS
    elevatorok = facing.elevatorok;
#endif

    doornum = tilemap[checkx][checky];
    if (*(mapsegs[1]+(checky<<mapshift)+checkx) == PUSHABLETILE)
    {
        //
        // pushable wall
        //

        PushWall (checkx,checky,dir);
        return;
    }
#ifdef USE_ELEVATORS
    if (!buttonheld[bt_use] && doornum == ELEVATORTILE && elevatorok)
    {
        //
        // use elevator
        //
        buttonheld[bt_use] = true;

        tilemap[checkx][checky]++;              // flip switch
        playstate = CheckExitType();
        SD_PlaySound (ED_SND("LEVELDONESND"));
        SD_WaitSoundDone();
    }
    else
#endif
    if (!buttonheld[bt_use] && doornum & 0x80)
    {
        buttonheld[bt_use] = true;
        OperateDoor (doornum & ~0x80);
    }
    else if (!buttonheld[bt_use] && gamestate.weapon >= wp_mg42_mounted
        && !gamestate.attackframe && !buttonstate[bt_attack])
    {
        //
        // WSJ: stop using MG42
        //
        buttonheld[bt_use] = true;

        think = ED_THINK(weaponinfo[gamestate.weapon].mountedExitThink);
        if (think != NULL)
        {
            think(player);
        }
        else
        {
            ExitMountedWeapon (dir);
        }
    }
    else if (!buttonheld[bt_use] && !gamestate.attackframe
        && !buttonstate[bt_attack])
    {
        // cannot use static object in sniper mode
        if (gamestate.snipermode)
        {
            return;
        }

        //
        // WSJ: use static object?
        //
        statobj_t *checkstatic;
        for (checkstatic = &statobjlist[0] ; checkstatic !=laststatobj ; checkstatic++)
        {
            if (checkstatic->tilex == checkx && 
                checkstatic->tiley == checky &&
                (checkstatic->flags & FL_ATTACKMODE) == 0)
            {
                switch (checkstatic->itemnumber)
                {
                 case bo_mg42_mounted:
                    buttonheld[bt_use] = true;
                    EnterMountedWeapon (wp_mg42_mounted,checkx,checky,dir);
                    return;
                }
            }
        }
    }
}

/*
=============================================================================

                                PLAYER CONTROL

=============================================================================
*/



/*
===============
=
= SpawnPlayer
=
===============
*/

void SpawnPlayer (int tilex, int tiley, int dir)
{
    player->enemy = ED_EnemyByIndex(ed, player->enemyIndex);
    player->active = ac_yes;
    player->tilex = tilex;
    player->tiley = tiley;
    player->areanumber = (byte) *(mapsegs[0]+(player->tiley<<mapshift)+player->tilex);
    OB_SetX(player, ((int32_t)tilex<<TILESHIFT)+TILEGLOBAL/2);
    OB_SetY(player, ((int32_t)tiley<<TILESHIFT)+TILEGLOBAL/2);
    OB_SetState(player, ED_STATE("s_player"));
    player->angle = (1-dir)*90;
    if (player->angle<0)
        player->angle += ANGLES;
    player->fineangle = player->angle * FINEANGLES / ANGLES;
    player->flags = FL_NEVERMARK | FL_SHOOTABLE;
    Thrust (0,0);                           // set some variables

    InitAreas ();
}


//===========================================================================

/*
===============
=
= T_KnifeAttack
=
= Update player hands, and try to do damage when the proper frame is reached
=
===============
*/

void    KnifeAttack (objtype *ob)
{
    objtype *check,*closest;
    int32_t  dist;

    // actually fire
    dist = 0x7fffffff;
    closest = NULL;
    for (ITEROBJ(check, 1))
    {
        if ( (check->flags & FL_SHOOTABLE) && (check->flags & FL_VISABLE)
            && OB_InShootDistDelta(check, 1))
        {
            if (check->transx < dist)
            {
                dist = check->transx;
                closest = check;
            }
        }
    }

    if (!closest || dist > 0x18000l)
    {
        // missed
        SD_PlaySound (ED_SND("ATKKNIFESND"));
        return;
    }

    // hit something
    if (DamageActor (closest,player,US_RndT() >> 3,dam_melee))
        SpawnBloodPuff (closest, OB_GetX(closest), OB_GetY(closest), 0);
    SD_PlaySound (ED_SND("HITENEMYSND"));
}


/*
===============
=
= GunAttack
=
===============
*/

void    GunAttack (objtype *ob, int8_t attacknum)
{
    int weaponsound = -1;
    int8_t     recoil;
    bool noiseless = false;

    if (attacknum > 20)
        recoil = attacknum-20;
    else
        recoil = 0;

    byte ammopershot = 1;

    switch (gamestate.weapon)
    {
        case wp_pistol:
            weaponsound = ED_SND("ATKPISTOLSND");
            FireBullet (ob,(10 + US_RndT()%25),dam_bullet,2,1,0);
            break;
        case wp_kar98k:
            weaponsound = ED_SND("ATKKAR98SND");
            FireBullet (ob,(75 + US_RndT()%10),dam_rifle,0,1,0);
            break;
        case wp_sniper:
            weaponsound = ED_SND("ATKSNIPERSND");
            if (gamestate.snipermode)
                FireBullet (ob,(100 + US_RndT()%10),dam_rifle,0,1,0);
            else
                FireBullet (ob,(50 + US_RndT()%10),dam_bullet,1+recoil,1,0,
                    1 + (recoil * 5));
            break;
        case wp_mp40:
            weaponsound = ED_SND("ATKMP40SND");
            FireBullet (ob,(8 + US_RndT()%22),dam_bullet,2+recoil,1,0,
                recoil * 5);
            break;
        case wp_stg44:
            weaponsound = ED_SND("ATKSTG44SND");
            FireBullet (ob,(20 + US_RndT()%22),dam_bullet,1+recoil,1,0,
                recoil * 5);
            Tremor_WeaponStart(trem_stg44);
            break;
        case wp_shotgun:
            weaponsound = ED_SND("ATKSHOTGUNSND");
            if (attacknum == 7)
            {
                FireBullet (ob,(5 + US_RndT()%22),dam_chaingun,5,16,0);
                ammopershot = 2;
                Tremor_WeaponStart(trem_shotgunalt);
            }
            else
            {
                FireBullet (ob,(5 + US_RndT()%22),dam_bullet,4,8,0);
                Tremor_WeaponStart(trem_shotgun);
            }
            break;
        case wp_flamethrower:
            weaponsound = ED_SND("FLAMETHROWERSND");
            ProjFire (ob,0,ED_Proj(ed, "bjfire"),NULL,0x3000l,0);
            break;
        case wp_chaingun:
            weaponsound = ED_SND("ATKCHAINGUNSND");
            FireBullet (ob,(10 + US_RndT()%30),dam_chaingun,4,1,0);
            break;
        case wp_panzerschreck:
            weaponsound = ED_SND("MISSILEFIRESND");
            ProjFire (ob,0,ED_Proj(ed, "bjrocket"),NULL,0x5000l,0);
            break;
        case wp_wirecutters:
            Player_UseWireCutters();
            noiseless = true;
            break;
        case wp_cyringe:
            Player_UseCyringe();
            noiseless = true;
            break;
        case wp_mg42_mounted:
            weaponsound = ED_SND("ATKMG42SND");
            FireBullet (ob,(12 + US_RndT()%28),dam_bullet,2,1,0);
            break;
        case wp_mecha:
            weaponsound = ED_SND("ATKMECHASND");
            FireBullet (ob,(10 + US_RndT()%25),dam_chaingun,2,1,0);
            ammopershot = 1;
            break;
        default:
            break;
    }

    if (!noiseless)
    {
        MakeNoise (player);
    }

    if (attacknum != 5 && attacknum != 6)
    {
        SD_PlaySound (weaponsound);
    }

    if (!ammocheat && CURAMMOTYPE != ammo_none)
    {
        if (CLIPSIZE(gamestate.weapon) > 0)
            CURCLIP -= ammopershot;
        else
            ConsumeAmmo(ammopershot);
    }
    DrawAmmo ();
}


/*
===============
=
= Player_UseWireCutters
=
= Cut wires in front of player
=
===============
*/

void Player_UseWireCutters (void)
{
    bool doCut;
    int statObjNum;
    statobj_t *check;
    Player_Facing_t facing;

    doCut = false;
    Player_GetFacing(&facing);

    statObjNum = statat[facing.checkx][facing.checky];
    if (statObjNum != -1)
    {
        check = &statobjlist[statObjNum];
        if ((check->flags & FL_FENCING) != 0 && check->statInfoNum >= 2)
        {
            doCut = true;
        }
    }

    if (doCut)
    {
        RemoveStaticObj(check);
        SpawnStatic(check->tilex, check->tiley, check->statInfoNum - 2);
        SD_PlaySound (ED_SND("CUTFENCINGSND"));
    }
    else
    {
        SD_PlaySound (ED_SND("USEWIRECUTTERSSND"));
        KnifeAttack(player);
    }
}


/*
===============
=
= Player_UseCyringe
=
= Inject antidote using the cyringe
=
===============
*/

void Player_UseCyringe (void)
{
    if (gamestate.infected)
    {
        GetMessage("You are cured of the infection!");
        SD_PlaySound(ED_SND("INJECTCURESND"));
        gamestate.infected = 0;
    }
    else
    {
        SD_PlaySound(ED_SND("INJECTSND"));
    }
    gamestate.antidotetics = SEC2TICS(60);
}


/*
===============
=
= Player_CanInfect
=
= Player can be infected by mutagen
=
===============
*/

bool Player_CanInfect (void)
{
    return !gamestate.infected && gamestate.antidotetics == 0;
}


/*
===============
=
= Player_UpdateInfection
=
= Infection does damage against player
=
===============
*/

void Player_UpdateInfection (void)
{
    if (gamestate.infected)
    {
        const int damagetics = 60;
        gamestate.infectedtics += tics;
        if (gamestate.infectedtics > damagetics)
        {
            gamestate.infectedtics -= damagetics;
            TakeSpecialDamage(1,NULL,dam_infection,true,true);
        }
    }

    if (gamestate.antidotetics > 0)
    {
        gamestate.antidotetics -= tics;
        if (gamestate.antidotetics < 0)
        {
            gamestate.antidotetics = 0;
        }
    }
}


/*
===============
=
= Player_ClearInfection
=
= Clear infection and antidote effect
=
===============
*/

void Player_ClearInfection (void)
{
    gamestate.infected = gamestate.antidotetics = 0;
}


//
// WSJ: assisted aiming, based on the original attack code
//
short PlayerBulletAngle (objtype *ob)
{
    if (aimassist)
    {
        objtype *check,*closest;
        int32_t  dist;

        // WSJ: find potential targets
        dist = 0x7fffffff;
        closest = NULL;
        for (ITEROBJ(check, 1))
        {
            if ( (check->flags & FL_SHOOTABLE) && (check->flags & FL_VISABLE)
                && OB_InShootDistDelta(check, 2))
            {
                if (check->transx < dist && CheckLine(check))
                {
                    dist = check->transx;
                    closest = check;
                }
            }
        }

        if (closest != NULL)
        {
            // WSJ: calculate the angle required for the bullet to hit closest
            int32_t deltax,deltay;
            float   angle;
            short   iangle;

            deltax = OB_GetX(closest) - OB_GetX(ob);
            deltay = OB_GetY(ob) - OB_GetY(closest);
            angle = (float) atan2((float) deltay, (float) deltax);
            if (angle<0)
                angle = (float) (M_PI*2+angle);
            iangle = (short) (angle/(M_PI*2)*FINEANGLES);
            return iangle;
        }

        assert(ob == player);
    }

    return Mecha_GetViewingFineAngle();    // WSJ: missed
}


//
// AlumiuN's new weaponry code - trace bullets
//
#define ENEMYCOLSIZE 0x5000l
#define PLAYERCOLSIZE 0x3000l    // WSJ: make the player harder to hit
#define STEPDIST 0x400l
#define MINTRACERDIST 0x36000l

void FireBullet (objtype *ob, short gundamage, damage_t damagetype, short accuracy, short numshots, short spriteoffset, short pitchangle)
{
    int32_t         bulletx,bullety,bulletz,colsize;
    short           iangle,bangle,pangle,damage;
    boolean         hit;
    objtype         *check;

    damage = gundamage;
    if (ob == player && !gamestate.snipermode)
    {
        iangle = PlayerBulletAngle (ob);
    }
    else
    {
        if (ob == player)
        {
            iangle = Mecha_GetViewingFineAngle();
        }
        else
        {
            iangle = ob->angle * FINEANGLES / ANGLES;
        }
    }

    for (int i = 0; i<numshots; i++)
    {
        int shotdamage = damage;

        bulletx = OB_GetX(ob);
        bullety = OB_GetY(ob);
        bulletz = 0;
        bangle = iangle;
        pangle = pitchangle * FINEANGLES / ANGLES;

        if (accuracy)
        {
            int angleoff = 0;

            const int maxangle = accuracy * FINEANGLES / ANGLES;

            angleoff = US_RndT() % maxangle;
            if (US_RndT() > 127)
            {
                bangle -= angleoff;
                if (bangle < 0)
                {
                    bangle += FINEANGLES;
                }
            }
            else
            {
                bangle += angleoff;
                if (bangle >= FINEANGLES)
                {
                    bangle -= FINEANGLES;
                }
            }

            // lose up to 50% damage from accuracy
            shotdamage = damage - ((angleoff * damage / maxangle) / 2);

            const int pangleoff = US_RndT() % maxangle;
            if (US_RndT() > 127)
            {
                pangle -= pangleoff;
                if (pangle < 0)
                {
                    pangle += FINEANGLES;
                }
            }
            else
            {
                pangle += pangleoff;
                if (pangle >= ANGLES)
                {
                    pangle -= FINEANGLES;
                }
            }
        }

        hit = false;    // Assume the worst! Oh, and reset the variable. :)

        while(1)
        {
            //
            // check for solid walls
            //
            check = actorat[bulletx >> TILESHIFT][bullety >> TILESHIFT];
            if (check && !ISPOINTER(check) && (uintptr_t)check != 64)
            {
                if ((uintptr_t)check < 128)      // Hit a wall
                {
                    hit = true;
                    if (damagetype != dam_none && (uintptr_t)check != 
                        LANDSCAPETILE)
                    {
                        SpawnRicochet(bulletx,bullety,spriteoffset,
                            (uintptr_t)check,bulletz);
                    }
                }
                else if (doorposition[(uintptr_t)check - 128] < 0xdfff/2)   // Cheap hack - improve later?
                {
                    hit = true;
                    if (damagetype != dam_none)
                    {
                        SpawnRicochet(bulletx,bullety,spriteoffset,0,bulletz);
                    }
                }
            }
            else if (fixedpt_abs(bulletz) >= FP_HALF)
            {
                hit = true;
                SpawnRicochet(bulletx,bullety,spriteoffset,
                    (uintptr_t)check,bulletz);
            }

            if (hit)
                break;

            for (ITEROBJ(check, 0))
            {
                if (check != ob && (check->flags & FL_SHOOTABLE) && 
                    !OB_NaziFriends(ob, check))
                {
                    if (check == player)
                        colsize = PLAYERCOLSIZE;
                    else
                        colsize = ENEMYCOLSIZE;
                    if (LABS(bulletx - OB_GetX(check)) < colsize && LABS(bullety - OB_GetY(check)) < colsize)
                    {
                        hit = true;
                        if (damagetype != dam_none)
                        {
                            if (check == player)
                            {
                                TakeSpecialDamage (shotdamage, ob, damagetype, false, true);
                            }
                            else
                            {
                                if (DamageActor (check, ob, shotdamage, damagetype))
                                    SpawnBloodPuff (check, bulletx, bullety, bulletz);
                                else
                                    SpawnRicochet(bulletx,bullety,spriteoffset,0,bulletz);
                            }
                        }
                    }
                }
            }

            if (hit)
                break;

            bulletx += FixedMul(STEPDIST,finecostable[bangle]);
            bullety -= FixedMul(STEPDIST,finesintable[bangle]);
            bulletz += FixedMul(STEPDIST,finesintable[pangle]);
        }
    }
}

//===========================================================================

/*
===============
=
= VictorySpin
=
===============
*/

void VictorySpin (void)
{
    int32_t    desty;

    if (player->angle > 270)
    {
        player->angle -= (short)(tics * 3);
        if (player->angle < 270)
            player->angle = 270;
    }
    else if (player->angle < 270)
    {
        player->angle += (short)(tics * 3);
        if (player->angle > 270)
            player->angle = 270;
    }

    desty = (((int32_t)player->tiley-5)<<TILESHIFT)-0x3000;

    if (OB_GetY(player) > desty)
    {
        OB_SetY(player, OB_GetY(player) - tics*4096);
        if (OB_GetY(player) < desty)
            OB_SetY(player, desty);
    }
}


//===========================================================================

/*
===============
=
= T_Attack
=
===============
*/

short   shotcount;

void    WeaponAnimation (objtype *ob, boolean reload)
{
    int shapenum, newshapenum;
    struct  atkinf  *cur;
    lwlib_String_t shapeSoundName;
    const char *spriteName;

    if (reload && buttonstate[bt_use])
    {
        Cmd_Use();
    }

    UpdateFace ();
    if (reload)
    {
        CheckWeaponChange();
    }

    if (gamestate.victoryflag)              // watching the BJ actor
    {
        VictorySpin ();
        return;
    }

    if ( buttonstate[bt_use] && !buttonheld[bt_use] )
        buttonstate[bt_use] = false;

    if ( buttonstate[bt_attack] && !buttonheld[bt_attack])
        buttonstate[bt_attack] = false;

    ControlMovement (ob);
    if (gamestate.victoryflag)              // watching the BJ actor
        return;

    plux = (word) (OB_GetX(player) >> UNSIGNEDSHIFT);                     // scale to fit in unsigned
    pluy = (word) (OB_GetY(player) >> UNSIGNEDSHIFT);
    player->tilex = (short)(OB_GetX(player) >> TILESHIFT);                // scale to tile values
    player->tiley = (short)(OB_GetY(player) >> TILESHIFT);

    //
    // change frame and fire
    //
    gamestate.attackcount -= (short) tics;
    while (gamestate.attackcount <= 0)
    {
        cur = &attackinfo[gamestate.weapon][reload][gamestate.attackframe];
        switch (cur->attack)
        {
            case -4:
                OB_SetState(ob, ED_STATE("s_player"));
                gamestate.attackframe = gamestate.weaponframe = 0;
                return;
            case -3:
                if (ReadyToFire(gamestate.weapon)) SD_PlaySound (ED_SND("KAR98BOLTSND"));
            case -2:
                if (ReadyToFire(gamestate.weapon)) break;
            case -1:
                OB_SetState(ob, ED_STATE("s_player"));
                if (gamestate.weapon >= wp_mg42_mounted && 
                    CURAMMOTYPE != ammo_none)
                {
                    // no weapon change
                }
                else if (gamestate.weapon == wp_cyringe ||
                    !WeaponAvailable(gamestate.weapon))
                {
                    ChangeWeapon(BestAvailableWeapon(true));
                    if (gamestate.chosenweapon < gamestate.weapon)
                        gamestate.chosenweapon = gamestate.weapon;
                }
                else
                {
                    if (gamestate.weapon != gamestate.chosenweapon && 
                        gamestate.weapon < wp_mg42_mounted)
                    {
                        if (WeaponAvailable(gamestate.chosenweapon))
                            ChangeWeapon(gamestate.chosenweapon);
                    }
                }
                shotcount = 0;
                gamestate.attackframe = gamestate.weaponframe = 0;
                if (EMPTYCLIP(gamestate.weapon) && CURAMMO)
                    Cmd_Reload ();
                return;

            case 4:
            case 6:
                if (!ReadyToFire(gamestate.weapon))
                    break;
                if (buttonstate[bt_attack])
                    gamestate.attackframe = gamestate.weaponframe = 0;
            case 1:
            case 5:
            case 7:

            // recoil shots:
            case 21:
            case 22:
            case 23:
            case 24:
            case 25:
            case 26:
                if (!ReadyToFire(gamestate.weapon))
                {       // can only happen with chain gun
                    gamestate.attackframe++;
                    break;
                }
                GunAttack (ob,cur->attack);
                break;

            case 2:
                KnifeAttack (ob);
                break;

            case 3:
                if (shotcount > 10)
                {
                    shotcount = 0;
                    gamestate.attackframe += 2;
                }
                else if (ReadyToFire(gamestate.weapon) && buttonstate[bt_attack])
                {
                    shotcount++;
                    gamestate.attackframe = gamestate.weaponframe = 0;
                }
                break;

            case 8:    // reload a full clip
                ReloadWeapon (CLIPSIZE(gamestate.weapon)-CURCLIP);
                break;

            case 9:    // reload 1 round
                ReloadWeapon (1);
                break;

            case 10:   // go back 2 frames if clip isn't full yet
                if (CURCLIP < CLIPSIZE(gamestate.weapon) && CURAMMO)
                    gamestate.attackframe -= 2;
                break;
        }

        gamestate.attackcount += cur->tics;
        gamestate.attackframe++;

        shapenum = GetPlayerWeaponShapeNum();
        gamestate.weaponframe =
            attackinfo[gamestate.weapon][reload]
            [gamestate.attackframe].frame;

        newshapenum = GetPlayerWeaponShapeNum();
        if (newshapenum != shapenum)
        {
            spriteName = ED_SpriteName(ed, newshapenum);
            if (strncmp(spriteName, "SPR_", 4) == 0)
            {
                lwlib_Snprintf(shapeSoundName, "SPR%sSND", 
                    spriteName + 4);
                SD_PlaySound(ED_SND(shapeSoundName));
            }
        }
    }
}

void T_Attack (objtype *ob)    { WeaponAnimation(ob,0); }
void T_Reload (objtype *ob)    { WeaponAnimation(ob,1); }


//===========================================================================

/*
===============
=
= T_Player
=
===============
*/

void    T_Player (objtype *ob)
{
    mecha_t *mecha;

    mecha = &gamestate.mecha;

    if (gamestate.victoryflag)              // watching the BJ actor
    {
        VictorySpin ();
        return;
    }

    UpdateFace ();
    CheckWeaponChange ();

    if ( buttonstate[bt_use])
        Cmd_Use ();

    if ( buttonstate[bt_reload] && !buttonheld[bt_reload])
        Cmd_Reload ();
    else if ( buttonstate[bt_altattack] && !buttonheld[bt_altattack])
        Cmd_AltFire ();
    else if ( buttonstate[bt_attack] && !buttonheld[bt_attack])
        Cmd_Fire ();

    ControlMovement (ob);
    if (gamestate.victoryflag)              // watching the BJ actor
        return;

    if (Mecha_Equipped() && !mecha->lightWeight && gamestate.armor <= 0)
    {
        T_MechaEject(player);
    }

    plux = (word) (OB_GetX(player) >> UNSIGNEDSHIFT);                     // scale to fit in unsigned
    pluy = (word) (OB_GetY(player) >> UNSIGNEDSHIFT);
    player->tilex = (short)(OB_GetX(player) >> TILESHIFT);                // scale to tile values
    player->tiley = (short)(OB_GetY(player) >> TILESHIFT);
}

//===========================================================================

/*
===============
=
= HasWeaponForAmmoType
=
===============
*/

bool    HasWeaponForAmmoType(ammo_t ammotype)
{
    int i;
    weaponinfo_t *wpinf;

    for (i = 0; i <= gamestate.bestweapon; i++)
    {
        wpinf = &weaponinfo[i];
        if (HASWEAPON(i) && wpinf->ammotype == ammotype)
        {
            return true;
        }
    }

    return false;
}
