#ifndef WL_DEF_H
#define WL_DEF_H

// Defines which version shall be built and configures supported extra features
#include "version.h"

#include <assert.h>
#include <fcntl.h>
#include <math.h>
#ifdef USE_RAND
#include <ctime>
#endif
#include <ctype.h>
#include <stdio.h>
#include <stdlib.h>
#if defined(_arch_dreamcast)
#	include <kos.h>
#elif !defined(_WIN32)
#	include <stdint.h>
#	include <string.h>
#	include <stdarg.h>
#endif
#include <SDL.h>

#if !defined O_BINARY
#	define O_BINARY 0
#endif

#pragma pack(1)

#include "foreign.h"

#ifndef SPEAR
    #include "audiowl6.h"
    #ifdef UPLOAD
        #include "gfxv_apo.h"
    #else
        #ifdef GOODTIMES
            #include "gfxv_wl6.h"
        #else
            #include "gfxv_apo.h"
        #endif
    #endif
#else
    #include "audiosod.h"
    #include "gfxv_sod.h"
    #include "f_spear.h"
#endif

#ifdef _arch_dreamcast
typedef uint8 uint8_t;
typedef uint16 uint16_t;
typedef uint32 uint32_t;
typedef int8 int8_t;
typedef int16 int16_t;
typedef int32 int32_t;
typedef int64 int64_t;
typedef ptr_t uintptr_t;
#endif

typedef uint8_t byte;
typedef uint16_t word;
typedef int32_t fixed;
typedef uint32_t longword;
typedef int8_t boolean;
typedef void * memptr;

typedef struct
{
    int x,y;
} Point;
typedef struct
{
    Point ul,lr;
} Rect;

void Quit(const char *errorStr, ...);

#include "id_pm.h"
#pragma pack(0)
#ifdef VIEASM
    #include "asmcomp.h"
#else
    #include "id_sd.h"
#endif
#pragma pack(1)
#include "id_in.h"
#include "id_vl.h"
#include "id_vh.h"
#include "id_us.h"
#include "id_ca.h"

#include "wl_menu.h"
#pragma pack(0)
#include "wl_ed.h"
#include "lw_vec.h"
#pragma pack(1)

#ifdef SPEAR
#include "f_spear.h"
#endif

#define MAPSPOT(x,y,plane) mapsegs[plane][((y) << mapshift) + (x)]

#define SIGN(x)         ((x)>0?1:-1)
#define ABS(x)          ((int)(x)>0?(x):-(x))
#define LABS(x)         ((int32_t)(x)>0?(x):-(x))

#define abs(x) ABS(x)

#ifndef MIN
#define MIN(x,y) ((x)<(y)?(x):(y))
#endif

#ifndef MAX
#define MAX(x,y) ((x)>(y)?(x):(y))
#endif

#ifndef CLIP
#define CLIP(x,a,b) MIN(MAX(x,a),b)
#endif

#define SWAP_VALS(x,y,t) \
do { \
    t = x; \
    x = y; \
    y = t; \
} while(0)

#define SKILLAMOUNT(x)  ((x) + (((x) / 5) * gamestate.difficulty *((x) > 4)))
//#define SKILLDAMAGE(x)  ((x) + ((x) * gamestate.difficulty / 10))

#define SKILLDAMAGE(x)  \
    ( \
        gamestate.difficulty == gd_easy && (x) >= 4 ? \
        (x) >> 2 : (x) \
    )

#define TILEDIST(tx1,ty1,tx2,ty2) ((int)abs(tx1-tx2)>abs(ty1-ty2) ? abs(tx1-tx2) : abs(ty1-ty2))
#define OBJTILEDIST(ob1,ob2) TILEDIST(ob1->tilex,ob1->tiley,ob2->tilex,ob2->tiley)
#define TILECENTER(x) (((int32_t)(x)<<TILESHIFT)+TILEGLOBAL/2)
#define GLOBALTILE(x) ((word)((x)>>TILESHIFT))
#define FACINGTILEX(ob) (ob->tilex+(ob->dir==east)-(ob->dir==west))
#define FACINGTILEY(ob) (ob->tiley+(ob->dir==south)-(ob->dir==north))
#define TICS2MILLI(tics) ((tics) * 100 / 7)
#define MILLI2TICS(ms) ((ms) * 7 / 100)
#define SEC2TICS(sec) MILLI2TICS((sec) * 1000)

/*
=============================================================================

                            GLOBAL CONSTANTS

=============================================================================
*/

#define MAXTICS 10
#define DEMOTICS        4

#define MAXACTORS       2500        // max number of nazis, etc / map
#define MAXSTATS        1000        // max number of lamps, bonus, etc
#define MAXDOORS        64          // max number of sliding doors
#define MAXWALLTILES    256         // max number of wall tiles

//
// tile ordering
//

// 1. wall textures
// 2. floor/ceiling textures
// 3. wall patches
// 4. sky textures
// 5. door textures

#define DOORPAGE_START_DEFAULT 185
#define DOORPAGE_START         (ffDataBottomRight != 0 ? \
    ffDataBottomRight : DOORPAGE_START_DEFAULT)
#define DOORPAGE_ABSOLUTE(x) (DOORPAGE_START + ((x) * 2))

//
// tile constants
//

#define PLAYERSPAWNTILE 19
#define MG42TILE        72
#define ICONARROWS      90
#define NUMICONARROWS   8
#define PUSHABLETILE    98
#define EXITTILE        99          // at end of castle
#define ENDLEVELTILE    100
#define ENEMYBLOCK      101
#define ALERTENEMYTILE  102
#define KEY1ENEMYBLOCK  103
#define KEY2ENEMYBLOCK  104
#define KEY3ENEMYBLOCK  105
#define KEY4ENEMYBLOCK  106
#define DEADGUARDTILE   124

#define AREATILE        107         // first of NUMAREAS floor tiles
#define NUMAREAS        39
#define LASTAREA        (AREATILE+NUMAREAS-1)
#define LANDSCAPETILE   1
#define NORMALWALLTILE  2
#define ELEVATORTILE    21
#define BLOCKTILE       64
#define AMBUSHTILE      106
#define ALTELEVATORTILE 107
#define DOORTILE_FIRST  90
#define DOORTILE_LAST   105

#define NUMBERCHARS     9


//----------------

#define EXTRAPOINTS     40000

#define PLAYERSPEED     3000
#define RUNSPEED        6000

#define SCREENSEG       0xa000

#define SCREENBWIDE     80

#define HEIGHTRATIO     0.50            // also defined in id_mm.c

#define BORDERCOLOR     3
#define FLASHCOLOR      5
#define FLASHTICS       4

#ifndef SPEAR
    #define LRpack      ED_LRPack(ed)   // # of levels to store in endgame
#else
    #define LRpack      20
#endif
#define MAX_LRPACK      30

#define PLAYERSIZE      MINDIST         // player radius
#define MINACTORDIST    0xd800l         // minimum dist from player center
                                        // to any actor center
#define TOTENDELAYTICS  500

#define NUMLATCHPICS    256

#undef M_PI
#define PI              3.141592657
#define M_PI PI

#define GLOBAL1         (1l<<16)
#define TILEGLOBAL      GLOBAL1
#define PIXGLOBAL       (GLOBAL1/64)
#define TILESHIFT       16l
#define UNSIGNEDSHIFT   8

#define ANGLES          360             // must be divisable by 4
#define ANGLEQUAD       (ANGLES/4)
#define FINEANGLES      3600
#define ANG90           (FINEANGLES/4)
#define ANG180          (ANG90*2)
#define ANG270          (ANG90*3)
#define ANG360          (ANG90*4)
#define VANG90          (ANGLES/4)
#define VANG180         (VANG90*2)
#define VANG270         (VANG90*3)
#define VANG360         (VANG90*4)

#define MINDIST         (0x5800l)

#define mapshift        6
#define MAPSIZE         (1<<mapshift)
#define maparea         MAPSIZE*MAPSIZE
#define FLATKERNELSECTORS 8

#define mapheight       MAPSIZE
#define mapwidth        MAPSIZE

#ifdef USE_HIRES

#define TEXTURESHIFT    7
#define TEXTURESIZE     (1<<TEXTURESHIFT)
#define TEXTUREFROMFIXEDSHIFT 2
#define TEXTUREMASK     (TEXTURESIZE*(TEXTURESIZE-1))

#define SPRITESCALEFACTOR 1

#else

#define TEXTURESHIFT    6
#define TEXTURESIZE     (1<<TEXTURESHIFT)
#define TEXTUREFROMFIXEDSHIFT 4
#define TEXTUREMASK     (TEXTURESIZE*(TEXTURESIZE-1))

#define SPRITESCALEFACTOR 2

#endif

#define NORTH   0
#define EAST    1
#define SOUTH   2
#define WEST    3


#define STATUSLINES     40

#define SCREENSIZE      (SCREENBWIDE*208)
#define PAGE1START      0
#define PAGE2START      (SCREENSIZE)
#define PAGE3START      (SCREENSIZE*2u)
#define FREESTART       (SCREENSIZE*3u)


#define PIXRADIUS       512

#define STARTAMMO       8



/*
=============================================================================

                               GLOBAL TYPES

=============================================================================
*/

typedef enum {
    di_north,
    di_east,
    di_south,
    di_west
} controldir_t;

static inline controldir_t cdir_reverse(controldir_t cdir)
{
    return (controldir_t)(((int)cdir + 2) % 4);
}

typedef enum {
    dr_normal1,
    dr_lock1,
    dr_lock2,
    dr_lock3,
    dr_lock4,
    dr_normal2,
    dr_normal3,
    dr_normal4,
    dr_lock_fw,
    dr_lock_bw,
    dr_max,
} door_t;

typedef enum {
  dr_orient_horz,
  dr_orient_vert,
  dr_orient_max
} door_orient_t;

typedef enum {
  dr_face_front,
  dr_face_side,
  dr_face_max,
} door_face_t;

typedef enum {
    ac_badobject = -1,
    ac_no,
    ac_yes,
    ac_allways
} activetype;

typedef enum {
    east,
    northeast,
    north,
    northwest,
    west,
    southwest,
    south,
    southeast,
    nodir
} dirtype;


typedef ED_State_t statetype;

//---------------------
//
// trivial actor structure
//
//---------------------

typedef struct statstruct
{
    byte      tilex,tiley;
    short     shapenum;           // if shapenum == -1 the obj has been removed
    byte      *visspot;
    uint32_t  flags;
    byte      itemnumber;
    byte      animframes;
    ED_BonusIndex_t bonusIndex;
    int       statInfoNum;
} statobj_t;


//---------------------
//
// door actor structure
//
//---------------------

typedef enum
{
    dr_open,dr_closed,dr_opening,dr_closing
} doortype;

typedef struct doorstruct
{
    byte     tilex,tiley;
    boolean  vertical;
    byte     lock;
    doortype action;
    short    ticcount;
} doorobj_t;


//--------------------
//
// thinking actor structure
//
//--------------------

typedef enum
{
    react_none,
    react_sight,
    react_hear,
    react_findbody,

// WSJ: special reactions:
    react_dead,
    react_findmg42,
    react_usingmg42
} reactiontype;

typedef enum
{
    death_none,
    death_normal,
    death_burned,
    death_gibbed,
    death_headshot,
} deathtype;

typedef enum
{
    surpriseatk_default,
    surpriseatk_hit,
    surpriseatk_delay,
    surpriseatk_prox,
    surpriseatk_prox_delay,
    surpriseatk_out_of_sight,
    surpriseatk_max,
} surpriseatktype;

lwlib_EnumStrExtern(surpriseatktype);

typedef struct cagemutobjstruct
{
    int32_t waitTimer;
    int32_t escapeTimer;
    int32_t delayTimer;
    bool sighted;
} cagemutobjtype;

typedef struct ropemutobjstruct
{
    int32_t waitTimer;
    int32_t delayTimer;
    bool sighted;
} ropemutobjtype;

typedef struct knightskelobjstruct
{
    int32_t waitTimer;
    int32_t delayTimer;
    bool sighted;
} knightskelobjtype;

typedef enum
{
    villager_armscrossed_slouch,
    villager_armscrossed_shrug,
    villager_armsdown_unsure,
    villager_armsdown_answering,
    villager_armsdown_refusing,
    villager_grey,
    villager_grey2,
    villager_max,
} villagertype;

lwlib_EnumStrExtern(villagertype);

typedef struct villagerobjstruct
{
    villagertype type;
} villagerobjtype;

typedef struct physobjstruct
{
    lwlib::Point3f pos;
    lwlib::Point3f vel;
    int numBounces;
    double timer;
} physobjtype;

typedef enum objindextype_e
{
    oit_goaltarget,
    oit_goalseeker,
    oit_max,
} objindextype_t;

typedef struct watermutobjstruct
{
    bool waterbound;
} watermutobjtype;

typedef struct objstruct
{
    byte        alloced;
    activetype  active;
    short       ticcount;
    ED_EnemyIndex_t     enemyIndex;
    ED_Enemy_t          *enemy;
    ED_ProjIndex_t      projIndex;
    const ED_Proj_t     *proj;
    ED_StateIndex_t     stateIndex;
    statetype           *state;

    uint32_t    flags;              // FL_SHOOTABLE, etc

    int32_t     distance;           // if negative, wait for that door to open
    int         waitdoornum;
    dirtype     dir;

    fixed       x,y;
    word        tilex,tiley;
    byte        areanumber;

    short       viewx;
    word        viewheight;
    fixed       transx,transy;      // in global coord

    int         angle, fineangle;
    short       hitpoints,killpoints,starthitpoints;
    int32_t     speed;

    short       sproffset;
    reactiontype   reaction;

    short       temp1,temp2,hidden;
    union
    {
        cagemutobjtype    cagemut;
        ropemutobjtype    ropemut;
        knightskelobjtype knightskel;
        villagerobjtype   villager;
        physobjtype       phys;
        watermutobjtype   watermut;
    } u;
    int32_t     shootTimer;
    int         projEnemyIndex;
    surpriseatktype surpriseatk;
    bool        runAway;
    fixed       heightoff;
    deathtype   deathtyp;
    // only used to reference other objects except player
    int         objIndexArr[oit_max];
    int         goalitemnum;
    byte        killActorAlert;
} objtype;

enum
{
    bt_nobutton=-1,
    bt_attack=0,
    bt_altattack,
    bt_strafe,
    bt_run,
    bt_use,
    bt_reload,
    bt_usemedkit,
    bt_readyknife,
    bt_readyfirst=bt_readyknife,
    bt_readypistol,
    bt_readyrifle,
    bt_readymachinegun,
    bt_readyshotgun,
    bt_readyflamethrower,
    bt_readychaingun,
    bt_readylauncher,
    bt_readylast=bt_readylauncher,
    bt_nextweapon,
    bt_prevweapon,
    bt_esc,
    bt_pause,
    bt_strafeleft,
    bt_straferight,
    bt_moveforward,
    bt_movebackward,
    bt_turnleft,
    bt_turnright,
    NUMBUTTONS
};

typedef ED_WeaponType_t weapontype;
typedef ED_AmmoType_t ammo_t;
typedef ED_DamageType_t damage_t;

enum
{
    gd_baby,
    gd_easy,
    gd_medium,
    gd_hard
};

#define MECHA_HP_ONEARM(starthp) ((starthp) * 5 / 12)
#define MECHA_HP_NOARMS(starthp) ((starthp) * 1 / 4)

#define MECHA_HP_GLASS(starthp, level, maxlevel) \
    ( \
        MECHA_HP_ONEARM(starthp) + \
        (((starthp) - MECHA_HP_ONEARM(starthp)) * (level) / (maxlevel)) \
    )

#define MECHA_LOW_FUEL 48

typedef struct MechaRunningCharge_s
{
    bool active;
    int32_t ticsPassed;
    int angle;
} MechaRunningCharge_t;

typedef struct mecha_s
{
    int hp;
    int starthp;
    short playerarmor;
    int32_t standTics;
    bool rotating;
    int32_t rotateTics;
    int32_t stopTics;
    MechaRunningCharge_t runningCharge;
    short motionTurnAngle;
    fixed camx, camy;
    int32_t moveVel;
    int bobbleFactor;
    bool strafeMove;
    bool lightWeight;
} mecha_t;

#define TREMOR_MAX_SOURCES 8

typedef enum tremorprofiletype_e
{
    trem_mechastep,
    trem_explosion,
    trem_shotgun,
    trem_shotgunalt,
    trem_stg44,
    trem_max,
} tremorprofiletype_t;

typedef struct tremorprofile_s
{
    int tileRadius;
    double magnitude;
    double mechaFactor;
    int32_t duration;
    bool localized;
    bool directional;
    double falloff[3];
    double camDistFallOff[3];
    bool playerFocus;
} tremorprofile_t;

typedef struct tremorsource_s
{
    bool active;
    fixed x;
    fixed y;
    int tilex;
    int tiley;
    int32_t totalTics;
    tremorprofiletype_t profileType;
    lwlib::Point2f dir;
} tremorsource_t;

typedef struct tremor_s
{
    int32_t tics;
    fixed camxOff;
    fixed camyOff;
    int weaponx;
    int weapony;
    tremorsource_t sources[TREMOR_MAX_SOURCES];
} tremor_t;

//---------------
//
// gamestate structure
//
//---------------

#define GAMEMSGLENGTH 128
#define INITIAL_MEDKIT_LIM 5
#define MAX_MEDKITS 10

typedef struct
{
    short       difficulty;
    short       mapon,altmapon;
    int32_t     oldscore,score,nextextra;
    short       lives;
    short       health,armor;
    short       ammo[NUMAMMOTYPES];
    short       clip[NUMWEAPONS];
    boolean     snipermode;
    short       keys;
    weapontype  bestweapon,weapon,chosenweapon,lastchosenweapon[NUMBUTTONS];
    boolean     gotweapon[NUMWEAPONS],gotregenarmor,gotregenhealth;

#ifdef WEAPONBOBBLE
    boolean            bobdir, bobdir2;
    short              bobber, bobber2;
    int                bobcnt, bobcnt2;
    fixed              bobberdist;
    fixed              bobberx, bobbery;
#endif

    short       faceframe;
    short       attackframe,attackcount,weaponframe;

    short       episode,secretcount,treasurecount,killcount,
                secrettotal,treasuretotal,killtotal;
    int32_t     TimeCount;
    int32_t     killx,killy;
    boolean     victoryflag;            // set during victory animations
    char        message[GAMEMSGLENGTH];
    mecha_t     mecha;
    fixed       camx, camy, camz;
    int         ambiance[NUMAREAS];
    int         medkits;
    int         medkitlim;
    int         maxammoindex;
    boolean     infected;
    int         infectedtics;
    int         antidotetics;
    tremor_t    tremor;
    int         activebosses;
    uint32_t    bosskeydrop;
    int         bossmusicowner;
    int         guardedObjNum;
    int         bossmusictracknum;
} gametype;


typedef enum
{
    ex_stillplaying,
    ex_completed,
    ex_died,
    ex_warped,
    ex_resetgame,
    ex_loadedgame,
    ex_victorious,
    ex_abort,
    ex_demodone,
    ex_secretlevel
} exit_t;


extern word *mapsegs[MAPPLANES];
extern int mapon;

/*
=============================================================================

                             WL_MAIN DEFINITIONS

=============================================================================
*/

#ifdef VIEASM
extern bool allowwindow;
extern byte soundvol, musicvol;
extern bool reversestereo;
#endif

extern  boolean  loadedgame;
extern  fixed    focallength;
extern  int      viewscreenx, viewscreeny;
extern  int      viewwidth;
extern  int      viewheight;
extern  short    centerx;
extern  int32_t  heightnumerator;
extern  fixed    scale;

extern  int      dirangle[9];

extern  int      mouseadjustment;
extern  int      shootdelta;
extern  unsigned screenofs;

extern  boolean  startgame;
extern  char     str[80];
extern  char     configname[13];

//
// Command line parameter variables
//
extern  boolean  param_debugmode;
extern  boolean  param_nowait;
extern  int      param_difficulty;
extern  int      param_tedlevel;
extern  int      param_joystickindex;
extern  int      param_joystickhat;
extern  int      param_samplerate;
extern  int      param_audiobuffer;
extern  int      param_mission;
extern  boolean  param_goodtimes;
extern  boolean  param_ignorenumchunks;
extern  boolean  param_editmode;
extern  boolean  param_gl;

extern  ED_t     ed;

#define SAVE_FIELD(x) \
do { \
    fwrite(&(x), sizeof(x), 1, file); \
    checksum = DoChecksum((byte *)&(x),sizeof(x),checksum); \
} while (0)

#define LOAD_FIELD(x) \
do { \
    fread(&(x), sizeof(x), 1, file); \
    checksum = DoChecksum((byte *)&(x),sizeof(x),checksum); \
} while (0)

void            NewGame (int difficulty,int episode);
void            CalcProjection (int32_t focal);
void            NewViewSize (int width);
boolean         SetViewSize (unsigned width, unsigned height);
boolean         LoadTheGame(FILE *file,int x,int y);
boolean         SaveTheGame(FILE *file,int x,int y);
void            ShowViewSize (int width);
void            ShutdownId (void);


/*
=============================================================================

                         WL_GAME DEFINITIONS

=============================================================================
*/

#define SEARCHDATASZ    0x10000

extern  gametype        gamestate;
extern  byte            bordercol;
extern  byte            searchdata[SEARCHDATASZ];
extern  byte*           sdqstart;
extern  SDL_Surface     *latchpics[NUMLATCHPICS];
extern  char            demoname[13];

#ifdef DEBUGKEYS
extern  boolean         infinitelives;    // WSJ: used with tedlevel
#endif

extern int leftchannel, rightchannel, channeldist;

void    SetupGameLevel (void);
void    PlaceValidArea (int tilex, int tiley);
void    GameLoop (void);
void    DrawPlayBorder (void);
void    DrawStatusBorder (byte color);
void    DrawPlayScreen (void);
void    DrawPlayBorderSides (void);
void    ShowActStatus();
void    SetSoundLoc(fixed gx,fixed gy,int maxTileDist);

void    PlayDemo (int demonumber);
void    RecordDemo (void);


#ifdef SPEAR
extern  int32_t            spearx,speary;
extern  unsigned        spearangle;
extern  boolean         spearflag;
#endif


#define ClearMemory SD_StopDigitized


// JAB
#define PlaySoundLocTile(s,tx,ty) \
    PlaySoundLocGlobal(s,(((int32_t)(tx) << TILESHIFT) + \
        (1L << (TILESHIFT - 1))),(((int32_t)ty << TILESHIFT) + \
        (1L << (TILESHIFT - 1))),SD_MaxVolume(),false,false)
#define PlaySoundLocActor(s,ob) \
    PlaySoundLocGlobal(s,OB_GetX(ob),OB_GetY(ob),SD_MaxVolume(),false, \
        false)
#define PlaySoundLocActorVol(s,ob,volume,foreground,noDistAtt) \
    PlaySoundLocGlobal(s,OB_GetX(ob),OB_GetY(ob),volume,foreground,\
        noDistAtt)
int  PlaySoundLocGlobal(word s,fixed gx,fixed gy,int volume,
	bool foreground,bool noDistAtt);
void UpdateSoundLoc(void);

void DiscardWeaponsAtEnd (void);
void DiscardAmmoAtEnd (void);

/*
=============================================================================

                            WL_PLAY DEFINITIONS

=============================================================================
*/

#define BASEMOVE                35
#define RUNMOVE                 70

#define JOYSCALE                2

typedef enum tileflags_e
{
	TILEFLAGS_DESTROYED = 0x01,
} tileflags_t;

// equals (floor + (ceil << 8))
typedef uint32_t flat_t; 

extern  byte            tilemap[MAPSIZE][MAPSIZE];      // wall values only
extern  byte            wallpatch[MAPSIZE][MAPSIZE];    // WSJ: wall patch values
extern  byte            spotvis[MAPSIZE][MAPSIZE];
extern  flat_t          flatmap[MAPSIZE][MAPSIZE];
extern  objtype         *actorat[MAPSIZE][MAPSIZE];
extern  int             statat[MAPSIZE][MAPSIZE];
extern  int             ambiance[NUMAREAS];    // WSJ: set during SetupGameLevel
extern  byte            tileflags[MAPSIZE][MAPSIZE];    // wall flags

extern  objtype         *player;

extern  unsigned        tics;
extern  unsigned        lightningtics;    // WSJ: amount of time everything is lit up during a lightning flash
extern  int             viewsize;

extern  int             lastgamemusicoffset;

//
// current user input
//
extern  int         controlx,controly;              // range from -100 to 100
extern  boolean     buttonstate[NUMBUTTONS];
extern  objtype     objlist[MAXACTORS];
extern  int         objlistend;
extern  int         objcount;
extern  boolean     buttonheld[NUMBUTTONS];
extern  exit_t      playstate;
extern  boolean     madenoise;
extern  byte        alertedarea,lastalertedarea;
extern  statobj_t   statobjlist[MAXSTATS];
extern  statobj_t   *laststatobj;
extern  objtype     *newobj,*killerobj;
extern  doorobj_t   doorobjlist[MAXDOORS];
extern  doorobj_t   *lastdoorobj;
extern  int         godmode,invisiblemode;

extern  boolean     demorecord,demoplayback;
extern  int8_t      *demoptr, *lastdemoptr;
extern  memptr      demobuffer;

//
// control info
//
extern  boolean     mouseenabled,joystickenabled,alwaysrun,aimassist;
extern  int         dirscan[4];
extern  int         buttonscan[NUMBUTTONS][2];
extern  int         buttonmouse[5];
extern  int         buttonjoy[32];

#define ITEROBJ(ob, c) IterObjReset(&(ob), c); IterObjMore(ob); \
    IterObjNext(&(ob))

void    InitActorList (void);
void    GetNewActor (void);
void    PlayLoop (void);

void    CenterWindow(word w,word h);

void    InitRedShifts (void);
void    FinishPaletteShifts (void);

void    RemoveObj (objtype *gone);

void    IterObjReset(objtype **ob, int c);
bool    IterObjMore(objtype *ob);
void    IterObjNext(objtype **ob);

void    PollControls (void);
int     StopMusic(void);
void    StartMusic(void);
void    ContinueMusic(int offs);
void    UpdateBossMusic(objtype *ob, bool newActiveBoss);
void    StartDamageFlash (int damage);
void    StartBonusFlash (void);

void    SetPlayerAmbiance(ED_Ambient_t *ambient, bool on);

extern  objtype     *objfreelist;     // *obj,*player,*lastobj,

extern  boolean     noclip,ammocheat;
extern  int         singlestep, extravbls;

void    MakeNoise (objtype *ob);

// mecha

boolean Mecha_Equipped(void);

boolean Mecha_FuelLow(void);

void T_MechaExit(objtype *ob);

void T_MechaEject(objtype *ob);

bool Mecha_HasNoArms(void);

void Mecha_SetAmbiance(void);

void Mecha_NotifyRotating(bool rotating);

void Mecha_StartRunningCharge(void);

void Mecha_StopRunningCharge(void);

void Mecha_ControlRunningCharge(void);

void Mecha_NotifyHitSolid(void);

void Mecha_NotifyHitObj(objtype *ob);

short Mecha_GetViewingAngle(void);

short Mecha_GetViewingFineAngle(void);

bool Mecha_RunningChargeActive(void);

int Mecha_GetStatusBarPic(void);

fixed Mecha_GetX(void);

fixed Mecha_GetY(void);

// player

typedef struct Player_Facing_s
{
    int checkx, checky;
    controldir_t dir;
    boolean elevatorok;
} Player_Facing_t;

controldir_t Player_CardinalDir(void);

void Player_GetFacing(Player_Facing_t *facing);

fixed Player_CamGetX(void);

fixed Player_CamGetY(void);

int Player_CamGetTileX(void);

int Player_CamGetTileY(void);

const lwlib::Ray3f Player_GetViewRay(void);

/*
=============================================================================

                                WL_INTER

=============================================================================
*/

void IntroScreen (void);
void PG13(void);
void DrawHighScores(void);
void CheckHighScore (int32_t score,word other);
void Victory (void);
void LevelCompleted (void);
void ClearSplitVWB (void);

void PreloadGraphics(void);


/*
=============================================================================

                                WL_DEBUG

=============================================================================
*/

int DebugKeys (void);

/*
=============================================================================

                            WL_DRAW DEFINITIONS

=============================================================================
*/

//
// math tables
//
extern  short *pixelangle;
extern  int32_t finetangent[FINEANGLES/4];
extern  fixed sintable[];
extern  fixed *costable;
extern  fixed finesintable[];
extern  fixed *finecostable;
extern  int *wallheight;
extern  word horizwall[],vertwall[];
extern  int32_t    lasttimecount;
extern  int32_t    frameon;
extern  int        messagetime;

extern  unsigned screenloc[3];

extern  boolean fizzlein, fpscounter;

extern  fixed   viewx,viewy;                    // the focal point
extern  fixed   viewsin,viewcos;

void    ThreeDRefresh (void);
void    DrawScaleds (void);
void    CalcTics (void);
int     GetPlayerWeaponShapeNum(void);

typedef struct
{
    word leftpix,rightpix;
    word dataofs[64];
// table data after dataofs[rightpix-leftpix+1]
} t_compshape;

void    GetMessage (const char *lastmessage);
void    DrawMessage (void);
boolean CheckVisTile (short tilex, short tiley);
fixed   Math_Distance2(fixed x1, fixed y1, fixed x2, fixed y2);

int WallMidY(int ywcount, int bot);

int InvWallMidY(int y, int bot);

flat_t GetGameLevelFlat(int x, int y);

void CA_CopyPlanes(word dst[MAPPLANES][maparea], int dx, int dy,
    word src[MAPPLANES][maparea], int sx, int sy, int w, int h);

/*
=============================================================================

                       WL_SNOWMUTANTS DEFINITIONS

=============================================================================
*/

namespace SnowMutants
{
    extern void convertSprites(void);
}

/*
=============================================================================

                             WL_STATE DEFINITIONS

=============================================================================
*/
#define TURNTICS        10
#define SPDPATROL       512
#define SPDDOG          1500


void    InitHitRect (objtype *ob, unsigned radius);
void    SpawnNewObj (unsigned tilex, unsigned tiley, statetype *state);
void    NewState (objtype *ob, statetype *state);

namespace SearchMemChecker
{
    extern boolean selectNextDir(objtype *ob, int dtilex, int dtiley);
}

boolean TryWalk (objtype *ob);
boolean SelectNextDir(objtype *ob, int dtilex, int dtiley);
void    SelectChaseDir (objtype *ob);
void    SelectDodgeDir (objtype *ob);
void    SelectRunDir (objtype *ob);
void    MoveObj (objtype *ob, int32_t move);
void    MoveObjInDir (objtype *ob, int32_t move, dirtype dir);
boolean SightPlayer (objtype *ob, statetype *state);
void    FirstSighting (objtype *ob);

void    KillActor (objtype *ob, objtype *killer, damage_t damagetype);
boolean DamageActor (objtype *ob, objtype *attacker, unsigned damage, damage_t damagetype);

boolean CheckCoordLine (fixed fx1, fixed fy1, fixed fx2, fixed fy2, 
    boolean binarydoor);
#define CheckObjLine(ob,check)    CheckCoordLine(OB_GetX(ob),OB_GetY(ob),OB_GetX(check),OB_GetY(check),false)
boolean CheckTileLine (unsigned tx1, unsigned ty1, unsigned tx2, unsigned ty2,
    boolean binarydoor);
boolean CheckLinePlayerBody (objtype *ob);
boolean CheckLine (objtype *ob);
boolean CheckSight (objtype *ob);
void    CheckGoal (objtype *ob);
void    ResetGoal (objtype *ob);
bool    CheckPlaysMusic (objtype *ob);
boolean CheckObjSight (objtype *ob, objtype *check);

/*
=============================================================================

                             WL_AGENT DEFINITIONS

=============================================================================
*/

#define NUMAMMOLIMITS 3

typedef struct TryMoveInfo_s
{
    bool hitSolid;
    objtype *hitObj;
} TryMoveInfo_t;

typedef struct weaponspr
{
    const char *sprName;
} weaponspr_t;

typedef struct weaponinfo
{
    const char *readySprite;
    weaponspr_t *weaponSprs;
    int        guyPic;
    short      selectkey;
    short      clip;
    ammo_t     ammotype;
    boolean    noBobble;
    boolean    discardAtEnd;
    boolean    skipfallback;
    boolean    villageFriendly;
    const char *name;
    const char *mountedUseSound;
    const char *mountedExitSound;
    ED_ThinkDef_t mountedExitThink;
} weaponinfo_t;

typedef struct ammoinf
{
    const char *name;
    short maxammo[NUMAMMOLIMITS];
    int giveBackpackAmmo;
    bool discardAtEnd;
} ammoinf_t;

extern weaponinfo_t weaponinfo[NUMWEAPONS];
extern ammoinf_t ammoinfo[NUMAMMOTYPES];

extern  short    anglefrac;
extern  int      facecount, facetimes;
extern  word     plux,pluy;         // player coordinates scaled to unsigned
extern  int32_t  thrustspeed;

#define HASWEAPON(wp) (gamestate.gotweapon[wp])
#define WEAPONKEY(wp) (weaponinfo[wp].selectkey)
#define CLIPSIZE(wp) (weaponinfo[wp].clip)
#define CURAMMOTYPE (CheckAmmoType(gamestate.weapon))
#define CURAMMO (gamestate.ammo[CURAMMOTYPE])
#define CURCLIP (gamestate.clip[gamestate.weapon])
#define HASAMMO(wp) (CheckAmmoType(wp)==ammo_none||gamestate.ammo[CheckAmmoType(wp)]>0||gamestate.clip[wp]>0)
#define CURHASAMMO HASAMMO(gamestate.weapon)
#define CURMAXAMMO(ammotype) ammoinfo[ammotype].maxammo[ \
    gamestate.maxammoindex]
#define HASMAXAMMO(ammotype) (gamestate.ammo[ammotype] >= \
    CURMAXAMMO(ammotype))
#define MAXEDWEAPON(wp) (HASWEAPON(wp)&&HASMAXAMMO(CheckAmmoType(wp)))
#define EMPTYCLIP(wp) (CLIPSIZE(wp)>0 && !gamestate.clip[wp])
#define MAXLIVES 9

bool    HasWeaponForAmmoType(ammo_t ammotype);

void    WeaponBobble_Update(void);
void    WeaponBobble_Reset(void);
void    Tremor_Update(void);
void    Tremor_Start(tremorprofiletype_t profileType, fixed x, fixed y,
    int angle);
void    Tremor_ObjStart(tremorprofiletype_t profileType, objtype *ob);
void    Tremor_WeaponStart(tremorprofiletype_t profileType);
void    Thrust (int angle, int32_t speed);
void    SpawnPlayer (int tilex, int tiley, int dir);
void    TakeDamage (int points,objtype *attacker);
void    TakeSpecialDamage (int points,objtype *attacker,damage_t damagetype,boolean ignorearmor,boolean useflash);
void    GivePoints (int32_t points);
void    GetBonus (statobj_t *check);
void    GiveWeapon (int weapon);
void    GiveAmmo (int ammo, ammo_t ammotype);
void    ConsumeAmmo (int ammo);
void    GiveKey (int key);
void    GiveLives (int lives);
void    GiveArmor (int points);
short   &PlayerArmor (void);
void    FireBullet (objtype *ob, short gundamage, damage_t damagetype, short accuracy, short numshots, short spriteoffset, short pitchangle = 0);

//
// player state info
//

void    StatusDrawFace(unsigned picnum);
void    DrawFace (void);
void    DrawHealth (void);
void    HealSelf (int points);
void    DrawLevel (void);
void    DrawLives (void);
void    GiveExtraMan (void);
void    DrawScore (void);
void    DrawWeapon (void);
void    DrawKeys (void);
void    DrawAmmo (void);

void    T_Player (objtype *ob);
void    T_Attack (objtype *ob);
void    T_Reload (objtype *ob);

bool    EnterMountedWeapon (weapontype weapon, int newx, int newy, 
    controldir_t dir);
void    ExitMountedWeapon (controldir_t dir);

ammo_t  CheckAmmoType (int weapon);
bool    TryMove (objtype *ob, TryMoveInfo_t *tryMoveInfo);

void    Player_UseWireCutters (void);
void    Player_UseCyringe (void);
bool    Player_CanInfect (void);
void    Player_ClearInfection (void);
void    Player_UpdateInfection (void);

void    VictoryTile (void);

void    DiedReset (void);

/*
=============================================================================

                             WL_ACT1 DEFINITIONS

=============================================================================
*/

extern  doorobj_t doorobjlist[MAXDOORS];
extern  doorobj_t *lastdoorobj;
extern  short     doornum;

extern  word      doorposition[MAXDOORS];

extern  byte      areaconnect[NUMAREAS][NUMAREAS];

extern  boolean   areabyplayer[NUMAREAS];

extern word     pwallstate;
extern word     pwallpos;        // amount a pushable wall has been moved (0-63)
extern word     pwallx,pwally;
extern byte     pwalldir,pwalltile,pwallpatch;


void InitDoorList (void);
void InitStaticList (void);
void SpawnStatic (int tilex, int tiley, int type);
void SpawnSpecialStatic (int tilex, int tiley, int type, short picnum, 
    uint32_t specialFlags, byte animframes, const ED_BonusDef_t bonusDef,
    int statInfoNum);
#define SpawnSimpleStatic(x,y,picnum) SpawnSpecialStatic(x,y,none,picnum,0,0,"",-1)
void RemoveStaticObj(statobj_t *check);
void SpawnDoor (int tilex, int tiley, boolean vertical, int lock);
void MoveDoors (void);
void MovePWalls (void);
void OpenDoor (int door);
void PlaceItemType (const ED_BonusDef_t bonusDef, int tilex, int tiley);
void PushWall (int checkx, int checky, int dir);
void OperateDoor (int door);
void InitAreas (void);
void ConnectAreas (void);
boolean AreaByArea (int area1, int area2);
#define AreaByActor(ob,area)    AreaByArea(ob->areanumber,area)
#define AreaAt(x,y)    (GetMapSpot(x,y,0)-AREATILE)

word GetMapSpot(int x, int y, int plane);
void DestroyWall(int x, int y);

/*
=============================================================================

                             WL_ACT2 DEFINITIONS

=============================================================================
*/

void SpawnStand (int tilex, int tiley, int dir, bool deafguard);
void SpawnPatrol (int tilex, int tiley, int dir);
void SpawnActor (int tilex, int tiley, int tile);

void SpawnDeadGuard (int tilex, int tiley);
void SpawnBoss (int tilex, int tiley);

void A_AlertSound (objtype *ob);
void A_DeathScream (objtype *ob);
void A_FlameScream (objtype *ob);
void SpawnBJVictory (void);
void SpawnBulletPuff (fixed puffx, fixed puffy, statetype *state, 
    short spriteoffset, fixed heightoff);
void SpawnBloodPuff (objtype *ob, fixed puffx, fixed puffy, fixed puffz);
void SpawnRicochet (fixed puffx, fixed puffy, short spriteoffset, byte tile, fixed puffz);
void ProjFire (objtype *ob, short angleoffset, const ED_Proj_t *proj, statetype *state, int32_t speed, short spriteoffset);

int PlayActorVoice (int sound, objtype *ob);
boolean ValidGround (objtype *ob, int tilex, int tiley);
boolean CanUseMG42 (objtype *ob);

void A_BJRocketBlast (objtype *ob);
void A_RocketBlast (objtype *ob);
void A_Smoke (objtype *ob);
void A_FlameBurn (objtype *ob);

void A_BecomeBlocking (objtype *ob);
void A_RemoveBlocking (objtype *ob);
void A_BecomeShootable (objtype *ob);
void A_RemoveShootable (objtype *ob);
void A_BecomeVulnerable (objtype *ob);
void A_BecomeUnkillable (objtype *ob);
void A_RemoveUnkillable (objtype *ob);
void A_BecomeNoPain (objtype *ob);
void A_RemoveNoPain (objtype *ob);
void A_BecomeUnderground (objtype *ob);
void A_RemoveUnderground (objtype *ob);

void A_Slurpie (objtype *ob);
void A_HitlerMorph (objtype *ob);
void A_MechaSound (objtype *ob);

void    T_Schabb (objtype *ob);
void    T_SchabbThrow (objtype *ob);
void    T_Fake (objtype *ob);
void    T_Ghosts (objtype *ob);

void    T_Path (objtype *ob);
void    T_Aim (objtype *ob);
void    T_Aim2 (objtype *ob);
void    T_Shoot (objtype *ob);
void    T_Melee (objtype *ob);
void    T_Chase (objtype *ob);
void    T_DogChase (objtype *ob);
void    T_Projectile (objtype *ob);
void    T_Stand (objtype *ob);
void    T_MG42Wait (objtype *ob);
void    T_Underground (objtype *ob);

void A_DeathScream (objtype *ob);
void T_BoneShrapnel (objtype *ob);
void T_FireProj (objtype *ob);
void T_FireShotgun (objtype *ob);

void T_BossShoot (objtype *ob);
void T_BossShootS (objtype *ob);
void T_BossRefire (objtype *ob);
void T_BossShootFireProj(objtype *ob);
void A_BossBurnStart(objtype *ob);
void A_BossBurnRepeat(objtype *ob);

void T_DogBite (objtype *ob);
void T_MG42GrdLoop (objtype *ob);
void T_MG42MutLoop (objtype *ob);
void T_MG42OfcLoop (objtype *ob);
void T_MG42ShootS (objtype *ob);
void T_MG42Shoot (objtype *ob);
void T_MG42SkelLoop (objtype *ob);
void T_MG42SSLoop (objtype *ob);
void T_MG42ZSSLoop (objtype *ob);
void T_ZSSAttack (objtype *ob);
void T_ZSSBuried (objtype *ob);
void T_MP40Shoot (objtype *ob);
void T_STG44Shoot (objtype *ob);
void T_SSRefire (objtype *ob);
void T_SSOfcRefire (objtype *ob);
void T_OfcRefire (objtype *ob);
void T_PistolShoot (objtype *ob);
void T_RifleShoot (objtype *ob);
void T_SciRefire (objtype *ob);
void T_SciMelee (objtype *ob);
void T_SciThrow (objtype *ob);
void T_SkelAttack (objtype *ob);
void T_SkelBuried (objtype *ob);
void T_SkelFireBomb (objtype *ob);
void T_ZSSRefire (objtype *ob);
void T_SkelMeleeLoop (objtype *ob);
void T_StartDig (objtype *ob);
void T_StartDigInChase (objtype *ob);
void T_StartRise (objtype *ob);
void T_ZWake (objtype *ob);

//
// cagemut
//

void T_CageMutWait(objtype *ob);

void T_CageMutAwake(objtype *ob);

void A_CageMutEscapeSound(objtype *ob);

void A_CageMutFinishEscape(objtype *ob);

//
// ropemut
//

void T_RopeMutWait(objtype *ob);

void T_RopeMutFinishSlide(objtype *ob);

//
// laddermut
//

void T_LadderMutWait(objtype *ob);

void T_LadderMutFinishClimb(objtype *ob);

//
// knightskel
//

void T_KnightSkelWait(objtype *ob);

void A_KnightSkelFinishDrop(objtype *ob);

//
// hangedmut
//

void A_HangedMutFinishDrop(objtype *ob);

//
// brownhorse
//

void A_BrownHorseRandomDelay(objtype *ob);

void A_BrownHorsePlaySound(objtype *ob);

//
// barrel
//

void A_BarrelSplashDamage(objtype *ob);

void A_BarrelDebrisFall(objtype *ob);

void T_BarrelDebrisHurt(objtype *ob);

//
// greentoxbarrel
//

void A_GreenToxBarrelHurt(objtype *ob);

//
// damagemut
//

void T_DamageMut(objtype *ob);

//
// tablemutdamage
//

void T_TableMutDamage(objtype *ob);

//
// villager
//

void A_VillagerStart(objtype *ob);

void T_Villager(objtype *ob);

void A_VillagerFinishResponse(objtype *ob);

void A_VillagerRandomDelay(objtype *ob);

//
// mech mutant 1
//

void A_MechMut1SpawnSuit(objtype *ob);

//
// mech mutant 2
//

void A_MechMut2Evacuate(objtype *ob);

//
// mech mutant 4
//

void A_MechMut4Evacuate(objtype *ob);

void T_MechMut4Chase(objtype *ob);

void T_MechMut4OneArmChase(objtype *ob);

void A_MechMut4Explode(objtype *ob);

void A_MechMut4NoArms(objtype *ob);

//
// mecha suit
//

void T_MechaSuit(objtype *ob);

void T_MechaSuitOneArm(objtype *ob);

void T_MechaSuitNoArms(objtype *ob);

void A_PlayMechBoomSound(objtype *ob);

void T_MechaSuitEnemyEnter(objtype *ob);

//
// blackcat
//

void A_BlackCatPlaySound(objtype *ob);

//
// water mutant
//

void A_WaterMutEnter(objtype *ob);

void A_WaterMutExit(objtype *ob);

//
// sewer boss
//

void A_SewerBossShoot(objtype *ob);

//
// generic
//

void A_Dead(objtype *ob);

void OB_SetX(objtype *ob, fixed x);

void OB_SetY(objtype *ob, fixed y);

int OB_ComputeChance(objtype *ob, ED_ChanceRangeSeq_t chanceRanges);

void OB_SetEnemy(objtype *ob, ED_Enemy_t *enemy);

void OB_SetState(objtype *ob, statetype *state);

int OB_TileDist(objtype *ob1, objtype *ob2);

bool OB_InShootDelta(objtype *ob);

bool OB_InShootDistDelta(objtype *ob, int divisor);

bool OB_CheckUsed(objtype *ob);

void OB_AngleMove(objtype *ob, fixed u, fixed v);

void OB_SpawnSurpriseEnemy(objtype *ob, fixed u, fixed v,
    bool resetHitPoints);

void A_SpawnSurpriseEnemy(objtype *ob);

void OB_Damage(objtype *ob, objtype *attacker, int damage, 
    damage_t damagetype, boolean ignorearmor);

objtype *OB_GetObjOn(objtype *ob);

int OB_AngleTo(objtype *from, objtype *to);

void OB_AngleMoveTo(objtype *from, objtype *to, fixed dist);

bool OB_CheckCloserThan(objtype *ob1, objtype *ob2, fixed minDist);

lwlib::Point2f OB_PointFromView(objtype *ob, double u, double v);

bool OB_TileCentred(objtype *ob, fixed dist);

boolean OB_CheckCoordLine(objtype *ob1, objtype *ob2);

lwlib::Box3f OB_GetBoundingBox(objtype *ob, double size, double height);

bool OB_NaziFriends(objtype *ob, objtype *check);

bool OB_GuardedObjInSight(objtype *ob);

bool OB_CanDefendGuardedObj(objtype *ob);

bool OB_GuardedObjBlocksMove(objtype *ob, int tilex, int tiley);

const lwlib::Point2f OB_GetPos(objtype *ob);

const lwlib::Ray3f OB_GetViewRay(objtype *ob);

void OB_KillActorAlert(objtype *ob);

static inline fixed OB_GetX(objtype *ob)
{
    return ob->x;
}

static inline fixed OB_GetY(objtype *ob)
{
    return ob->y;
}

/*
=============================================================================

                             WL_GOAL DEFINITIONS

=============================================================================
*/

void OB_ChooseGoalObj(objtype *ob);

bool OB_CanUseGoalObj(objtype *ob, int x, int y);

/*
=============================================================================

                          WL_RESURRECT DEFINITIONS

=============================================================================
*/

bool OB_CanResurrect(objtype *ob);

bool OB_CanTargetObjResurrect(objtype *ob, objtype *check);

void A_ReachedBody(objtype *ob);

void A_ResurrectBody(objtype *ob);

bool OB_CheckSummonBodyInSight(objtype *ob, int &deadObjNum);

void A_SummonGhostFromDead(objtype *ob);

void A_ResetSpawnedGhostFlag(objtype *ob);

/*
=============================================================================

                             WL_TEXT DEFINITIONS

=============================================================================
*/

extern  char    helpfilename[],endfilename[];

extern  void    HelpScreens(void);
extern  void    EndText(void);
extern  void    IntermissionScreens(void);

extern  void    LogDiscScreensEx(const char* choice, bool perLevel);
extern  void    LogDiscScreens(const char* choice);
extern  void    LogDiscScreensNoLev(const char* choice);

/*
=============================================================================

                            DREAMCAST DEFINITIONS

=============================================================================
*/

#ifdef _arch_dreamcast

// defined in dc_main.cpp
void DC_Main(void);
void DC_CheckParameters(void);
int DC_MousePresent(void);

// defined in dc_vmu.cpp
void StatusDrawLCD(int lcd_index);
int DC_SaveToVMU(char *src, int tp);
int DC_LoadFromVMU(char *dst);

#else

// ignore calls to this function
#define StatusDrawLCD(x)

#endif


/*
=============================================================================

                               GP2X DEFINITIONS

=============================================================================
*/

#ifdef GP2X

void GP2X_Init();
void GP2X_Shutdown();
void GP2X_ButtonDown(int button);
void GP2X_ButtonUp(int button);

#endif


/*
=============================================================================

                             MISC DEFINITIONS

=============================================================================
*/

#define FP(x) ((x) << TILESHIFT)
#define FP_TRUNC(x) ((x) >> 16)
#define FP_ONE 0x10000l
#define FP_HALF 0x8000l
#define FP_SQ(x) fixedpt_mul(x,x)

static inline fixed FixedMul(fixed a, fixed b)
{
	return (fixed)(((int64_t)a * b + 0x8000) >> 16);
}

#define TILEINMAP(x, y) ((x) >= 0 && (x) < MAPSIZE && \
	(y) >= 0 && (y) < MAPSIZE)

static inline bool tilePosInMap(lwlib::Point2i tilePos)
{
    return tilePos.v[0] >= 0 && tilePos.v[0] < MAPSIZE &&
        tilePos.v[1] >= 0 && tilePos.v[1] < MAPSIZE;
}

#define MAPSPOT_ISSOLID(x, y) \
	((x) < 0 || (x) >= MAPSIZE || (y) < 0 || (y) >= MAPSIZE || \
	(MAPSPOT(x, y, 0) > 0 && MAPSPOT(x, y, 0) < AMBUSHTILE))

#ifdef PLAYDEMOLIKEORIGINAL
    #define DEMOCHOOSE_ORIG_SDL(orig, sdl) ((demorecord || demoplayback) ? (orig) : (sdl))
    #define DEMOCOND_ORIG                  (demorecord || demoplayback)
    #define DEMOIF_SDL                     if(DEMOCOND_SDL)
#else
    #define DEMOCHOOSE_ORIG_SDL(orig, sdl) (sdl)
    #define DEMOCOND_ORIG                  false
    #define DEMOIF_SDL
#endif
#define DEMOCOND_SDL                   (!DEMOCOND_ORIG)

#define GetTicks() ((SDL_GetTicks()*7)/100)

#define ISPOINTER(x) ((((uintptr_t)(x)) & ~0xffff) != 0)
#define VALIDAREATILE(t) (t>=AREATILE && t<=LASTAREA)
#define VALIDAREA(x,y)   VALIDAREATILE(GetMapSpot(x,y,0))

#define CHECKMALLOCRESULT(x) if(!(x)) Quit("Out of memory at %s:%i", __FILE__, __LINE__)

#ifdef _WIN32
    #define strcasecmp stricmp
    #define strncasecmp strnicmp
#else
    static inline char* itoa(int value, char* string, int radix)
    {
	    sprintf(string, "%d", value);
	    return string;
    }

    static inline char* ltoa(long value, char* string, int radix)
    {
	    sprintf(string, "%ld", value);
	    return string;
    }
#endif

#define lengthof(x) (sizeof(x) / sizeof(*(x)))
#define endof(x)    ((x) + lengthof(x))

static inline word READWORD(byte *&ptr)
{
    word val = ptr[0] | ptr[1] << 8;
    ptr += 2;
    return val;
}

static inline longword READLONGWORD(byte *&ptr)
{
    longword val = ptr[0] | ptr[1] << 8 | ptr[2] << 16 | ptr[3] << 24;
    ptr += 4;
    return val;
}


/*
=============================================================================

                           FEATURE DEFINITIONS

=============================================================================
*/

#ifdef USE_FEATUREFLAGS
    // The currently available feature flags
    #define FF_STARSKY      0x0001
    #define FF_PARALLAXSKY  0x0002
    #define FF_CLOUDSKY     0x0004
    #define FF_RAIN         0x0010
    #ifndef USE_SNOW
        #define FF_SHADESKY     0x0020
    #else
        #define FF_SNOW         0x0020
    #endif
    #define FF_VILLAGE      0x0040
    #define FF_THUNDER      0x0080
    #define FF_WEAPONRESET  0x0100
    #ifdef USE_SNOW
        #define FF_SHADESKY     0x0200
    #endif

    // The ffData... variables contain the 16-bit values of the according corners of the current level.
    // The corners are overwritten with adjacent tiles after initialization in SetupGameLevel
    // to avoid interpretation as e.g. doors.
    extern int ffDataTopLeft, ffDataTopRight, ffDataBottomLeft, ffDataBottomRight,
        ffDataMusic, ffDataBossMusic, ffDataSecretLevel, ffDataThunderVol;

    /*************************************************************
     * Current usage of ffData... variables:
     * ffDataTopLeft:     lower 8-bit: shade color; higher 8-bit: shade fog
     * ffDataTopRight:    FeatureFlags
     * ffDataBottomLeft:  lower 8-bit: floor color; higher 8-bit: ceiling color, CloudSkyDefID, ParallaxStartTexture
     * ffDataBottomRight: DoorStartTexture
     * ffDataMusic: Music Track Number
     * ffDataBossMusic: Boss Music Track Number
     *************************************************************/

    // The feature flags are stored as a wall in the upper right corner of each level
    static inline word GetFeatureFlags()
    {
        return ffDataTopRight;
    }

    #define TILEINRAIN(x,y) (MAPSPOT(x,y,2)<256 && !tilemap[x][y] && GetFeatureFlags()&FF_RAIN)
    #define TILEINSNOW(x,y) (MAPSPOT(x,y,2)<256 && !tilemap[x][y] && GetFeatureFlags()&FF_SNOW)
#else
    #define TILEINRAIN(x,y) (MAPSPOT(x,y,2)<256 && !tilemap[x][y])
#endif

#define BRIGHTTILE(x,y) (TILEINRAIN(x,y) && lightningtics)
#define TILEINWATER(x,y) (ED_FloorIsWater(ed, (MAPSPOT(x,y,2) & 0xff)))

#ifdef USE_FLOORCEILINGTEX
    void DrawFloorAndCeiling(byte *vbuf, unsigned vbufPitch, int min_wallheight);
#endif

#ifdef USE_PARALLAX
    void DrawParallax(byte *vbuf, unsigned vbufPitch);
#endif

#ifdef USE_DIR3DSPR
    void Scale3DShape(byte *vbuf, unsigned vbufPitch, int shapenum, uint32_t flags, int tilex, int tiley);
#endif

void GenerateShadeTable(byte destRed, byte destGreen, 
	byte destBlue, SDL_Color *palette, int fog);

#endif
