// WL_DRAW.C

#include "wl_def.h"
#pragma hdrstop

#include "wl_cloudsky.h"
#include "wl_atmos.h"
#include "wl_shade.h"
#include "fixedptc.h"

/*
=============================================================================

                               LOCAL CONSTANTS

=============================================================================
*/

#define ACTORSIZE       0x4000
#define FOCALLENGTH    0x5700L    // WSJ: added for zooming

/*
=============================================================================

                              GLOBAL VARIABLES

=============================================================================
*/

byte *vbuf = NULL;
unsigned vbufPitch = 0;

int32_t    lasttimecount;
int32_t    frameon;
int        messagetime=0;
boolean fpscounter;

int fps_frames=0, fps_time=0, fps=0;

int *wallheight;
int min_wallheight;

//
// math tables
//
short *pixelangle;
int32_t finetangent[FINEANGLES/4];
fixed sintable[ANGLES+ANGLES/4];
fixed *costable = sintable+(ANGLES/4);
fixed finesintable[FINEANGLES+FINEANGLES/4];
fixed *finecostable = finesintable+(FINEANGLES/4);

//
// refresh variables
//
fixed   viewx,viewy;                    // the focal point
short   viewangle;
fixed   viewsin,viewcos;

void    TransformActor (objtype *ob);
void    BuildTables (void);
void    ClearScreen (void);
int     CalcRotate (objtype *ob);
void    DrawScaleds (void);
void    CalcTics (void);
void    ThreeDRefresh (void);



//
// wall optimization variables
//
int     lastside;               // true for vertical
int32_t    lastintercept;
int     lasttilehit,lastpatchhit;
int     lasttexture;
int     lastdoortile;

//
// ray tracing variables
//
short    focaltx,focalty,viewtx,viewty;
longword xpartialup,xpartialdown,ypartialup,ypartialdown;

int     midangle,angle;

word    tilehit,patchhit;
int     pixx;

short   xtile,ytile,shadex,shadey;
short   xtilestep,ytilestep;
int32_t    xintercept,yintercept;
word    xstep,ystep;
word    xspot,yspot;
int     texdelta;

word horizwall[MAXWALLTILES],vertwall[MAXWALLTILES];

/*
============================================================================

                           3 - D  DEFINITIONS

============================================================================
*/

/*
========================
=
= TransformActor
=
= Takes paramaters:
=   gx,gy               : globalx/globaly of point
=
= globals:
=   viewx,viewy         : point of view
=   viewcos,viewsin     : sin/cos of viewangle
=   scale               : conversion from global value to screen value
=
= sets:
=   screenx,transx,transy,screenheight: projected edge location and size
=
========================
*/


//
// transform actor
//
void TransformActor (objtype *ob)
{
    fixed gx,gy,gxt,gyt,nx,ny;

//
// translate point to view centered coordinates
//
    gx = OB_GetX(ob)-viewx;
    gy = OB_GetY(ob)-viewy;

//
// calculate newx
//
    gxt = FixedMul(gx,viewcos);
    gyt = FixedMul(gy,viewsin);
    nx = gxt-gyt-ACTORSIZE;         // fudge the shape forward a bit, because
                                    // the midpoint could put parts of the shape
                                    // into an adjacent wall

//
// calculate newy
//
    gxt = FixedMul(gx,viewsin);
    gyt = FixedMul(gy,viewcos);
    ny = gyt+gxt;

//
// calculate perspective ratio
//
    ob->transx = nx;
    ob->transy = ny;

    if (nx<MINDIST)                 // too close, don't overflow the divide
    {
        ob->viewheight = 0;
        return;
    }

    ob->viewx = (word)(centerx + ny*scale/nx);

//
// calculate height (heightnumerator/(nx>>8))
//
    ob->viewheight = (word)(heightnumerator/(nx>>8));
}

//==========================================================================

/*
========================
=
= TransformTile
=
= Takes paramaters:
=   tx,ty               : tile the object is centered in
=
= globals:
=   viewx,viewy         : point of view
=   viewcos,viewsin     : sin/cos of viewangle
=   scale               : conversion from global value to screen value
=
= sets:
=   screenx,transx,transy,screenheight: projected edge location and size
=
= Returns true if the tile is withing getting distance
=
========================
*/

boolean TransformTile (int tx, int ty, short *dispx, short *dispheight)
{
    fixed gx,gy,gxt,gyt,nx,ny,nxbonus;

//
// translate point to view centered coordinates
//
    gx = ((int32_t)tx<<TILESHIFT)+0x8000-viewx;
    gy = ((int32_t)ty<<TILESHIFT)+0x8000-viewy;

//
// calculate newx
//
    gxt = FixedMul(gx,viewcos);
    gyt = FixedMul(gy,viewsin);
    nx = gxt-gyt-0x2000;            // 0x2000 is size of object

//
// calculate newy
//
    gxt = FixedMul(gx,viewsin);
    gyt = FixedMul(gy,viewcos);
    ny = gyt+gxt;


//
// calculate height / perspective ratio
//
    if (nx<MINDIST)                 // too close, don't overflow the divide
        *dispheight = 0;
    else
    {
        *dispx = (short)(centerx + ny*scale/nx);
        *dispheight = (short)(heightnumerator/(nx>>8));
    }

//
// see if it should be grabbed
//
    nxbonus = TILEGLOBAL;
    if (buttonstate[bt_use] && !buttonheld[bt_use])
    {
        nxbonus = nxbonus * 3 / 2;
    }

    if (nx<nxbonus && ny>-TILEGLOBAL/2 && ny<TILEGLOBAL/2)
        return true;
    else
        return false;
}

//==========================================================================

/*
====================
=
= CalcHeight
=
= Calculates the height of xintercept,yintercept from viewx,viewy
=
====================
*/

int CalcHeight()
{
    fixed z = FixedMul(xintercept - viewx, viewcos)
        - FixedMul(yintercept - viewy, viewsin);
    if(z < MINDIST) z = MINDIST;
    int height = heightnumerator / (z >> 8);
    if(height < min_wallheight) min_wallheight = height;
    return height;
}

//==========================================================================

/*
===================
=
= ScalePost
=
===================
*/

byte *postsource,*patchsource;
bool patchglow;
int postx;
int postwidth;

void ScalePost()
{
    int ywcount, yoffs, yw, yd, yendoffs;
    byte col;
    int midy;

#ifdef USE_SHADING
    byte *curshades, *patchshades;

    if(BRIGHTTILE(shadex,shadey))
        curshades = patchshades = shadetable[ShadeGlow::off][0];
    else
    {
        curshades = patchshades = 
            shadetable[ShadeGlow::off][GetShade(wallheight[postx])];
    
        if (patchglow)
        {
            patchshades = 
                shadetable[ShadeGlow::on][GetShade(wallheight[postx])];
        }
    }
#endif

    ywcount = yd = wallheight[postx] >> 3;
    if(yd <= 0) yd = 100;

    midy = (viewheight / 2) - WallMidY(ywcount, -1);

    yoffs = midy * vbufPitch;
    if(yoffs < 0) yoffs = 0;
    yoffs += postx;

    yendoffs = midy + (ywcount * 2) - 1;
    yw=TEXTURESIZE-1;

    while(yendoffs >= viewheight)
    {
        ywcount -= TEXTURESIZE/2;
        while(ywcount <= 0)
        {
            ywcount += yd;
            yw--;
        }
        yendoffs--;
    }
    if(yw < 0) return;

#ifdef USE_SHADING
#ifdef USE_WALLPATCHES
    if (patchsource[yw] != 0xff)
        col = patchshades[patchsource[yw]];
    else
#endif
    if (postsource[yw] == 0xff)
        col = 0xff;
    else
        col = curshades[postsource[yw]];
#else
#ifdef USE_WALLPATCHES
    if (patchsource[yw] != 0xff)
        col = patchsource[yw];
    else
#endif
        col = postsource[yw];
#endif
    yendoffs = yendoffs * vbufPitch + postx;
    while(yoffs <= yendoffs)
    {
        if (col != 0xff)    // WSJ: don't draw transparent color
            vbuf[yendoffs] = col;
        ywcount -= TEXTURESIZE/2;
        if(ywcount <= 0)
        {
            do
            {
                ywcount += yd;
                yw--;
            }
            while(ywcount <= 0);
            if(yw < 0) break;
#ifdef USE_SHADING
#ifdef USE_WALLPATCHES
            if (patchsource[yw] != 0xff)
                col = patchshades[patchsource[yw]];
            else
#endif
            if (postsource[yw] == 0xff)
                col = 0xff;
            else
                col = curshades[postsource[yw]];
#else
#ifdef USE_WALLPATCHES
            if (patchsource[yw] != 0xff)
                col = patchsource[yw];
#endif
            else
                col = postsource[yw];
#endif
        }
        yendoffs -= vbufPitch;
    }
}

void GlobalScalePost(byte *vidbuf, unsigned pitch)
{
    vbuf = vidbuf;
    vbufPitch = pitch;
    ScalePost();
}

byte *DefaultPatchSource(void)
{
    int size;
    static byte *patchsource = NULL;

    if (patchsource == NULL)
    {
        size = TEXTURESIZE * TEXTURESIZE;
        patchsource = (byte *)malloc(size);
        CHECKMALLOCRESULT(patchsource);
        memset(patchsource, 0xff, size);
    }

    return patchsource;
}

/*
====================
=
= HitVertWall
=
= tilehit bit 7 is 0, because it's not a door tile
= if bit 6 is 1 and the adjacent tile is a door tile, use door side pic
=
====================
*/

void HitVertWall (void)
{
    int patchpic;
    int wallpic;
    int texture;
    int doornum;
    int doortile;
    int doorpage;
    int animatedtilehit;
    int animatedpatchhit;
    doorobj_t doorobj;

    texture = ((yintercept+texdelta)>>TEXTUREFROMFIXEDSHIFT)&TEXTUREMASK;
    if (xtilestep == -1)
    {
        texture = TEXTUREMASK-texture;
        xintercept += TILEGLOBAL;
    }
    else
    {
        if (ED_TileIsTruckTile(ed, tilehit))
            texture = TEXTUREMASK-texture;
    }

    ytile = (short)(yintercept>>TILESHIFT);
    doortile = tilemap[xtile-xtilestep][ytile];

    if (lastside==1 && lastintercept==xtile && lasttilehit==tilehit && 
        lastpatchhit==patchhit && lastdoortile == doortile)
    {
        if((pixx&3) && texture == lasttexture)
        {
            ScalePost();
            postx = pixx;
            wallheight[pixx] = wallheight[pixx-1];
            return;
        }
        ScalePost();
        wallheight[pixx] = CalcHeight();
        postsource+=texture-lasttexture;
        patchsource+=texture-lasttexture;
        postwidth=1;
        postx=pixx;
        lasttexture=texture;
        return;
    }

    if(lastside!=-1) ScalePost();

    lastside=1;
    lastintercept=xtile;
    lasttilehit=tilehit;
    lasttexture=texture;
    lastpatchhit=patchhit;
    lastdoortile=doortile;
    wallheight[pixx] = CalcHeight();
    postx = pixx;
    postwidth = 1;

    animatedtilehit = ED_AnimateWallTile(ed, tilehit);
    animatedpatchhit = ED_AnimateWallTile(ed, patchhit);

    if (doortile & 0x80)
    {
        doornum = doortile & 0x7f;
        doorobj = doorobjlist[doornum];

        doorpage = (doorobj.lock * 2) + (int)dr_face_side;
        wallpic = DOORPAGE_ABSOLUTE(doorpage) + dr_orient_vert;

        animatedpatchhit = 0;
    }
    else
        wallpic = vertwall[animatedtilehit];

    if (animatedpatchhit != 0)
    {
        patchpic = vertwall[animatedpatchhit];
        patchsource = PM_GetTexture(patchpic) + texture;
        patchglow = ED_PatchPicIsGlow(patchpic);
    }
    else
    {
        patchsource = DefaultPatchSource() + texture;
        patchglow = false;
    }

    postsource = PM_GetTexture(wallpic) + texture;
}


/*
====================
=
= HitHorizWall
=
= tilehit bit 7 is 0, because it's not a door tile
= if bit 6 is 1 and the adjacent tile is a door tile, use door side pic
=
====================
*/

void HitHorizWall (void)
{
    int patchpic;
    int wallpic;
    int texture;
    int doornum;
    int doortile;
    doorobj_t doorobj;
    int doorpage;
    int animatedtilehit;
    int animatedpatchhit;

    texture = ((xintercept+texdelta)>>TEXTUREFROMFIXEDSHIFT)&TEXTUREMASK;
    if (ytilestep == -1)
        yintercept += TILEGLOBAL;
    else
    {
        if (!ED_TileIsTruckTile(ed, tilehit))
            texture = TEXTUREMASK-texture;
    }

    xtile = (short)(xintercept>>TILESHIFT);
    doortile = tilemap[xtile][ytile-ytilestep];

    if(lastside==0 && lastintercept==ytile && lasttilehit==tilehit && 
        lastpatchhit==patchhit && lastdoortile==doortile)
    {
        if((pixx&3) && texture == lasttexture)
        {
            ScalePost();
            postx=pixx;
            wallheight[pixx] = wallheight[pixx-1];
            return;
        }
        ScalePost();
        wallheight[pixx] = CalcHeight();
        postsource+=texture-lasttexture;
        patchsource+=texture-lasttexture;
        postwidth=1;
        postx=pixx;
        lasttexture=texture;
        return;
    }

    if(lastside!=-1) ScalePost();

    lastside=0;
    lastintercept=ytile;
    lasttilehit=tilehit;
    lasttexture=texture;
    lastpatchhit=patchhit;
    lastdoortile=doortile;
    wallheight[pixx] = CalcHeight();
    postx = pixx;
    postwidth = 1;

    animatedtilehit = ED_AnimateWallTile(ed, tilehit);
    animatedpatchhit = ED_AnimateWallTile(ed, patchhit);

    if (doortile & 0x80)
    {
        doornum = doortile & 0x7f;
        doorobj = doorobjlist[doornum];

        doorpage = (doorobj.lock * 2) + dr_face_side;
        wallpic = DOORPAGE_ABSOLUTE(doorpage) + dr_orient_horz;

        animatedpatchhit = 0;
    }
    else
        wallpic = horizwall[animatedtilehit];

    if (animatedpatchhit != 0)
    {
        patchpic = horizwall[animatedpatchhit];
        patchsource = PM_GetTexture(patchpic) + texture;
        patchglow = ED_PatchPicIsGlow(patchpic);
    }
    else
    {
        patchsource = DefaultPatchSource() + texture;
        patchglow = false;
    }

    postsource = PM_GetTexture(wallpic) + texture;
}

//==========================================================================

/*
====================
=
= HitHorizDoor
=
====================
*/

void HitHorizDoor (void)
{
    int patchpic;
    int doorpage;
    int doornum;
    int texture;
    doorobj_t doorobj;

    doornum = tilehit&0x7f;
    texture = ((xintercept-doorposition[doornum])>>TEXTUREFROMFIXEDSHIFT)&TEXTUREMASK;

    if(lasttilehit==tilehit && lastpatchhit==patchhit)
    {
        if((pixx&3) && texture == lasttexture)
        {
            ScalePost();
            postx=pixx;
            wallheight[pixx] = wallheight[pixx-1];
            return;
        }
        ScalePost();
        wallheight[pixx] = CalcHeight();
        postsource+=texture-lasttexture;
        patchsource+=texture-lasttexture;
        postwidth=1;
        postx=pixx;
        lasttexture=texture;
        return;
    }

    if(lastside!=-1) ScalePost();

    lastside=2;
    lasttilehit=tilehit;
    lasttexture=texture;
    wallheight[pixx] = CalcHeight();
    postx = pixx;
    postwidth = 1;

    doorobj = doorobjlist[doornum];
    if (doorobj.lock < dr_max)
    {
        doorpage = (doorobj.lock * 2) + dr_face_front;
        doorpage = DOORPAGE_ABSOLUTE(doorpage) + dr_orient_horz;
    }

    postsource = PM_GetTexture(doorpage) + texture;

    if (patchhit != 0)
    {
        patchpic = horizwall[patchhit];
        patchsource = PM_GetTexture(patchpic) + texture;
        patchglow = ED_PatchPicIsGlow(patchpic);
    }
    else
    {
        patchsource = DefaultPatchSource() + texture;
        patchglow = false;
    }
}

//==========================================================================

/*
====================
=
= HitVertDoor
=
====================
*/

void HitVertDoor (void)
{
    int patchpic;
    int doorpage;
    int doornum;
    int texture;
    doorobj_t doorobj;

    doornum = tilehit&0x7f;
    texture = ((yintercept-doorposition[doornum])>>TEXTUREFROMFIXEDSHIFT)&TEXTUREMASK;

    if(lasttilehit==tilehit)
    {
        if((pixx&3) && texture == lasttexture)
        {
            ScalePost();
            postx=pixx;
            wallheight[pixx] = wallheight[pixx-1];
            return;
        }
        ScalePost();
        wallheight[pixx] = CalcHeight();
        postsource+=texture-lasttexture;
        patchsource+=texture-lasttexture;
        postwidth=1;
        postx=pixx;
        lasttexture=texture;
        return;
    }

    if(lastside!=-1) ScalePost();

    lastside=2;
    lasttilehit=tilehit;
    lasttexture=texture;
    wallheight[pixx] = CalcHeight();
    postx = pixx;
    postwidth = 1;

    doorobj = doorobjlist[doornum];
    if (doorobj.lock < dr_max)
    {
        doorpage = (doorobj.lock * 2) + dr_face_front;
        doorpage = DOORPAGE_ABSOLUTE(doorpage) + dr_orient_vert;
    }

    postsource = PM_GetTexture(doorpage) + texture;

    if (patchhit != 0)
    {
        patchpic = vertwall[patchhit];
        patchsource = PM_GetTexture(patchpic) + texture;
        patchglow = ED_PatchPicIsGlow(patchpic);
    }
    else
    {
        patchsource = DefaultPatchSource() + texture;
        patchglow = false;
    }
}

//==========================================================================

#define HitHorizBorder HitHorizWall
#define HitVertBorder HitVertWall

//==========================================================================

byte vgaCeiling[]=
{
#ifndef SPEAR
 0x1d,0x1d,0x1d,0x1d,0x1d,0x1d,0x1d,0x1d,0x1d,0xbf,
 0x4e,0x4e,0x4e,0x1d,0x8d,0x4e,0x1d,0x2d,0x1d,0x8d,
 0x1d,0x1d,0x1d,0x1d,0x1d,0x2d,0xdd,0x1d,0x1d,0x98,

 0x1d,0x9d,0x2d,0xdd,0xdd,0x9d,0x2d,0x4d,0x1d,0xdd,
 0x7d,0x1d,0x2d,0x2d,0xdd,0xd7,0x1d,0x1d,0x1d,0x2d,
 0x1d,0x1d,0x1d,0x1d,0xdd,0xdd,0x7d,0xdd,0xdd,0xdd
#else
 0x6f,0x4f,0x1d,0xde,0xdf,0x2e,0x7f,0x9e,0xae,0x7f,
 0x1d,0xde,0xdf,0xde,0xdf,0xde,0xe1,0xdc,0x2e,0x1d,0xdc
#endif
};

/*
=====================
=
= VGAClearScreen
=
=====================
*/

void VGAClearScreen (void)
{
#ifdef USE_FEATUREFLAGS
    byte floor=ffDataBottomLeft & 0x00ff;
    byte ceiling=ffDataBottomLeft >> 8;
#else
    byte floor=0x19;
    byte ceiling=vgaCeiling[gamestate.episode*10+mapon];
#endif

    int y;
    byte *ptr = vbuf;
#ifdef USE_SHADING
    for(y = 0; y < viewheight / 2; y++, ptr += vbufPitch)
        memset(ptr, shadetable[ShadeGlow::off][GetShade((viewheight / 2 - y) << 3)][ceiling], viewwidth);
    for(; y < viewheight; y++, ptr += vbufPitch)
        memset(ptr, shadetable[ShadeGlow::off][GetShade((y - viewheight / 2) << 3)][floor], viewwidth);
#else
    for(y = 0; y < viewheight / 2; y++, ptr += vbufPitch)
        memset(ptr, ceiling, viewwidth);
    for(; y < viewheight; y++, ptr += vbufPitch)
        memset(ptr, 0x19, viewwidth);
#endif
}

//==========================================================================

/*
=====================
=
= CalcRotate
=
=====================
*/

int CalcRotate (objtype *ob)
{
    int angle, viewangle;

    // this isn't exactly correct, as it should vary by a trig value,
    // but it is close enough with only eight rotations

    viewangle = Mecha_GetViewingAngle() + (centerx - ob->viewx)/8;

    if (ob->state->rotate == 1 && ob->dir != nodir)
    {
        ob->angle = dirangle[ob->dir];
        ob->fineangle = ob->angle * ANGLES / FINEANGLES;
    }

    angle = (viewangle-180) - ob->angle;

    angle+=ANGLES/16;
    while (angle>=ANGLES)
        angle-=ANGLES;
    while (angle<0)
        angle+=ANGLES;

    return angle/(ANGLES/8);
}

void ScaleShape (int xcenter, int shapenum, unsigned height, 
    uint32_t flags, int tilex, int tiley, fixed heightoff)
{
    t_compshape *shape;
    unsigned scale,pixheight;
    unsigned starty,endy;
    word *cmdptr;
    byte *cline;
    byte *line;
    byte *vmem;
    int actx,i,upperedge;
    short newstart;
    int scrstarty,screndy,lpix,rpix,pixcnt,ycnt;
    unsigned j;
    byte col;

#ifdef USE_SHADING
    byte *curshades;
    if(flags & FL_FULLBRIGHT || BRIGHTTILE(tilex,tiley))
        curshades = shadetable[ShadeGlow::off][0];
    else if(flags & FL_GLOWBRIGHT)
        curshades = shadetable[ShadeGlow::on][GetShade(height)];
    else
        curshades = shadetable[ShadeGlow::off][GetShade(height)];
#endif

    shape = (t_compshape *) PM_GetSprite(shapenum);

    scale=height>>3;                 // low three bits are fractional
    if(!scale) return;   // too close or far away

    pixheight=scale*SPRITESCALEFACTOR;
    actx=xcenter-scale;
    upperedge = (viewheight / 2) - WallMidY(scale, -1);
    upperedge -= FP_TRUNC(FixedMul(heightoff, FP(scale)));

    cmdptr=(word *) shape->dataofs;

    for(i=shape->leftpix,pixcnt=i*pixheight,rpix=(pixcnt>>6)+actx;i<=shape->rightpix;i++,cmdptr++)
    {
        lpix=rpix;
        if(lpix>=viewwidth) break;
        pixcnt+=pixheight;
        rpix=(pixcnt>>6)+actx;
        if(lpix!=rpix && rpix>0)
        {
            if(lpix<0) lpix=0;
            if(rpix>viewwidth) rpix=viewwidth,i=shape->rightpix+1;
            cline=(byte *)shape + *cmdptr;
            while(lpix<rpix)
            {
                if(wallheight[lpix]<=(int)height)
                {
                    line=cline;
                    while((endy = READWORD(line)) != 0)
                    {
                        endy >>= 1;
                        newstart = READWORD(line);
                        starty = READWORD(line) >> 1;
                        j=starty;
                        ycnt=j*pixheight;
                        screndy=(ycnt>>6)+upperedge;
                        if(screndy<0) vmem=vbuf+lpix;
                        else vmem=vbuf+screndy*vbufPitch+lpix;
                        for(;j<endy;j++)
                        {
                            scrstarty=screndy;
                            ycnt+=pixheight;
                            screndy=(ycnt>>6)+upperedge;
                            if(scrstarty!=screndy && screndy>0)
                            {
#ifdef USE_SHADING
                                col=curshades[((byte *)shape)[newstart+j]];
#else
                                col=((byte *)shape)[newstart+j];
#endif
                                if(scrstarty<0) scrstarty=0;
                                if(screndy>viewheight) screndy=viewheight,j=endy;

                                while(scrstarty<screndy)
                                {
                                    *vmem=col;
                                    vmem+=vbufPitch;
                                    scrstarty++;
                                }
                            }
                        }
                    }
                }
                lpix++;
            }
        }
    }
}

void SimpleScaleShape (int xcenter, int shapenum, unsigned height)
{
    t_compshape   *shape;
    unsigned scale,pixheight;
    unsigned starty,endy;
    word *cmdptr;
    byte *cline;
    byte *line;
    int actx,i,upperedge;
    short newstart;
    int scrstarty,screndy,lpix,rpix,pixcnt,ycnt;
    unsigned j;
    byte col;
    byte *vmem;

#ifdef USE_SHADING
    byte *curshades;
    curshades = shadetable[ShadeGlow::off][0];
#endif

    if (shapenum == -1)
    {
        return;
    }

    shape = (t_compshape *) PM_GetSprite(shapenum);

    scale=height>>1;
    pixheight=scale*SPRITESCALEFACTOR;
    actx=xcenter-scale;
    upperedge=viewheight/2-scale;

    cmdptr=shape->dataofs;

    for(i=shape->leftpix,pixcnt=i*pixheight,rpix=(pixcnt>>6)+actx;i<=shape->rightpix;i++,cmdptr++)
    {
        lpix=rpix;
        if(lpix>=viewwidth) break;
        pixcnt+=pixheight;
        rpix=(pixcnt>>6)+actx;
        if(lpix!=rpix && rpix>0)
        {
            if(lpix<0) lpix=0;
            if(rpix>viewwidth) rpix=viewwidth,i=shape->rightpix+1;
            cline = (byte *)shape + *cmdptr;
            while(lpix<rpix)
            {
                line=cline;
                while((endy = READWORD(line)) != 0)
                {
                    endy >>= 1;
                    newstart = READWORD(line);
                    starty = READWORD(line) >> 1;
                    j=starty;
                    ycnt=j*pixheight;
                    screndy=(ycnt>>6)+upperedge;
                    if(screndy<0) vmem=vbuf+lpix;
                    else vmem=vbuf+screndy*vbufPitch+lpix;
                    for(;j<endy;j++)
                    {
                        scrstarty=screndy;
                        ycnt+=pixheight;
                        screndy=(ycnt>>6)+upperedge;
                        if(scrstarty!=screndy && screndy>0)
                        {
#ifdef USE_SHADING
                            col=curshades[((byte *)shape)[newstart+j]];
#else
                            col=((byte *)shape)[newstart+j];
#endif
                            if(scrstarty<0) scrstarty=0;
                            if(screndy>viewheight) screndy=viewheight,j=endy;

                            while(scrstarty<screndy)
                            {
                                *vmem=col;
                                vmem+=vbufPitch;
                                scrstarty++;
                            }
                        }
                    }
                }
                lpix++;
            }
        }
    }
}

/*
=====================
=
= CheckVisTile
=
= WSJ: determine if the player can see
= into the tile
=
=====================
*/

boolean CheckVisTile (short tilex, short tiley)
{
    byte     *tilespot,*visspot;
    unsigned spotloc;

    if (tilex<0 || tiley<0 || tilex>(MAPSIZE-1) || tiley>(MAPSIZE-1))
        return false;

    spotloc = (tilex<<mapshift)+tiley;   // optimize: keep in struct?
    visspot = &spotvis[0][0]+spotloc;
    tilespot = &tilemap[0][0]+spotloc;

    //
    // could be in any of the nine surrounding tiles
    //
    if (*visspot
        || ( *(visspot-1) && !*(tilespot-1) )
        || ( *(visspot+1) && !*(tilespot+1) )
        || ( *(visspot-65) && !*(tilespot-65) )
        || ( *(visspot-64) && !*(tilespot-64) )
        || ( *(visspot-63) && !*(tilespot-63) )
        || ( *(visspot+65) && !*(tilespot+65) )
        || ( *(visspot+64) && !*(tilespot+64) )
        || ( *(visspot+63) && !*(tilespot+63) ) )
        return true;

    return false;
}

/*
=====================
=
= DrawScaleds
=
= Draws all objects that are visable
=
=====================
*/

#define MAXVISABLE 2000

typedef struct
{
    short      viewx,
               viewheight,
               tilex,
               tiley,
               shapenum;
    uint32_t   flags;          // this must be changed to uint32_t, when you
                               // you need more than 16-flags for drawing
    fixed      heightoff;
} visobj_t;

visobj_t vislist[MAXVISABLE];
visobj_t *visptr,*visstep,*farthest;

void DrawScaleds (void)
{
    int      i,least,numvisable,height;

    statobj_t *statptr;
    objtype   *obj;

    visptr = &vislist[0];

//
// place static objects
//
    for (statptr = &statobjlist[0] ; statptr !=laststatobj ; statptr++)
    {
        if ((visptr->shapenum = statptr->shapenum) == -1)
            continue;                                               // object has been deleted

        if (statptr->flags & FL_ATTACKMODE)
            continue;                                               // WSJ: object is being used by enemy

        if (!*statptr->visspot)
            continue;                                               // not visable

        if (TransformTile (statptr->tilex,statptr->tiley,
            &visptr->viewx,&visptr->viewheight) && statptr->flags & FL_BONUS)
        {
            GetBonus (statptr);
            if(statptr->shapenum == -1)
                continue;                                           // object has been taken
        }

        if (!visptr->viewheight)
            continue;                                               // to close to the object

        if (statptr->animframes)
            visptr->shapenum+=(gamestate.TimeCount/6)%(statptr->animframes);

        if (visptr < &vislist[MAXVISABLE-1])    // don't let it overflow
        {
            visptr->flags = statptr->flags;
            visptr->heightoff = 0;
            visptr->tilex = (short) statptr->tilex;
            visptr->tiley = (short) statptr->tiley;
            visptr++;
        }
    }

//
// place active objects
//
    for (ITEROBJ(obj, 1))
    {
        if (obj->state == NULL)
        {
            continue;
        }

        if ((visptr->shapenum = ED_SPR(obj->state->shapenum))==0)
            continue;                                               // no shape

        if (visptr->shapenum == -1)
        {
            continue;
        }

        if (CheckVisTile(obj->tilex,obj->tiley))
        {
            obj->active = ac_yes;
            TransformActor (obj);
            if (!obj->viewheight)
                continue;                                               // too close or far away

            visptr->viewx = obj->viewx;
            visptr->viewheight = obj->viewheight;
            if (visptr->shapenum == -1)
            {
                visptr->shapenum = obj->temp1;  // special shape
            }
            else
            {
                visptr->shapenum += obj->sproffset;    // WSJ: added
                visptr->shapenum = ED_ReplaceSprite(obj->enemy,
                    visptr->shapenum);
            }

            if (obj->state->rotate)
                visptr->shapenum += CalcRotate (obj);

            if (obj->state->fullbright == true)
                obj->flags |= FL_FULLBRIGHT;    // actor is "lit up"
            else if (obj->flags & FL_FULLBRIGHT)
                obj->flags &= ~FL_FULLBRIGHT;   // back to normal lighting

            if (visptr < &vislist[MAXVISABLE-1])    // don't let it overflow
            {
                visptr->flags = obj->flags;
                visptr->heightoff = obj->heightoff;
                visptr->tilex = (short) obj->tilex;
                visptr->tiley = (short) obj->tiley;
                visptr++;
            }
            obj->flags |= FL_VISABLE;
        }
        else
            obj->flags &= ~FL_VISABLE;
    }

//
// draw from back to front
//
    numvisable = (int) (visptr-&vislist[0]);

    if (!numvisable)
        return;                                                                 // no visable objects

    for (i = 0; i<numvisable; i++)
    {
        least = 32000;
        for (visstep=&vislist[0] ; visstep<visptr ; visstep++)
        {
            height = visstep->viewheight;
            if (height < least)
            {
                least = height;
                farthest = visstep;
            }
        }
        //
        // draw farthest
        //
#ifdef USE_DIR3DSPR
        if(farthest->flags & FL_DIR_MASK)
            Scale3DShape(vbuf, vbufPitch, farthest->shapenum, farthest->flags, farthest->tilex, farthest->tiley);
        else
#endif
            ScaleShape(farthest->viewx, farthest->shapenum, 
                farthest->viewheight, farthest->flags, farthest->tilex, 
                farthest->tiley, farthest->heightoff);

        farthest->viewheight = 32000;
    }
}

//==========================================================================

/*
==============
=
= GetPlayerWeaponShapeNum
=
= Get sprite num of player weapon
=
==============
*/

int GetPlayerWeaponShapeNum(void)
{
    int i;
    int shapenum;
    const char *sprName;
    mecha_t *mecha;
    weaponinfo_t *wpinf;
    
    mecha = &gamestate.mecha;

    wpinf = NULL;
    if (gamestate.weapon != -1)
    {
        wpinf = &weaponinfo[gamestate.weapon];
    }

    if (gamestate.snipermode)
    {
        assert(wpinf != NULL);
        shapenum = (int)ED_SPR(wpinf->readySprite);
    }
    else if (gamestate.weapon != -1)
    {
        if (gamestate.weapon == wp_mecha)
        {
            if (mecha->lightWeight)
            {
                sprName = "SPR_MECHAGUNS_READY";
            }
            else if (gamestate.armor < MECHA_HP_NOARMS(mecha->starthp))
            {
                sprName = "SPR_MECHAGUNS_D5NOARMSREADY";
            }
            else if (gamestate.armor < MECHA_HP_ONEARM(mecha->starthp))
            {
                sprName = "SPR_MECHAGUNS_D4ONEARMREADY";
            }
            else if (gamestate.armor < MECHA_HP_GLASS(mecha->starthp, 1, 4))
            {
                sprName = "SPR_MECHAGUNS_D3READY";
            }
            else if (gamestate.armor < MECHA_HP_GLASS(mecha->starthp, 2, 4))
            {
                sprName = "SPR_MECHAGUNS_D2READY";
            }
            else if (gamestate.armor < MECHA_HP_GLASS(mecha->starthp, 3, 4))
            {
                sprName = "SPR_MECHAGUNS_D1READY";
            }
            else
            {
                sprName = "SPR_MECHAGUNS_READY";
            }
            shapenum = (int)ED_SPR(sprName);
        }
        else
        {
            shapenum = 0;
            if (wpinf->readySprite != NULL)
            {
                shapenum = (int)ED_SPR(wpinf->readySprite);
            }
        }
        shapenum = shapenum + gamestate.weaponframe;

        if (wpinf->weaponSprs != NULL)
        {
            for (i = 0; wpinf->weaponSprs[i].sprName != NULL; i++)
            {
                if (i == gamestate.weaponframe)
                {
                    shapenum = (int)ED_SPR(wpinf->weaponSprs[i].sprName);
                    break;
                }
            }
        }

        if((GetFeatureFlags() & FF_VILLAGE) != 0 && 
            !wpinf->villageFriendly)
        {
            shapenum = -1; // no weapon shown
        }
    }
    else
    {
        shapenum = 0;
    }

    assert(shapenum >= -1 && shapenum < ED_NumSprites(ed));
    return shapenum;
}

/*
==============
=
= DrawPlayerWeapon
=
= Draw the player's hands
=
==============
*/

void DrawPlayerWeapon (void)
{
    int shapenum;
#ifdef WEAPONBOBBLE
    int xcenter;
    unsigned height;
#endif 
    tremor_t *tremor;

    tremor = &gamestate.tremor;

#ifndef SPEAR
    if (gamestate.victoryflag)
    {
        return;
    }
#endif

    if (gamestate.snipermode)
    {
        shapenum = GetPlayerWeaponShapeNum();
        SimpleScaleShape(viewwidth/2,(shapenum-1),viewheight+1);

        shapenum = (int)ED_SPR("SPR_DEATHCAM");
        if (viewsize == 21)
        {
            SimpleScaleShape((viewwidth/4)-(viewwidth/16),shapenum,viewheight+1);
            SimpleScaleShape(viewwidth+(viewwidth/8),shapenum,viewheight+1);
        }
        else
        {
            SimpleScaleShape(viewwidth/4,shapenum,viewheight+1);
            SimpleScaleShape(viewwidth,shapenum,viewheight+1);
        }
    }
    else if (gamestate.weapon != -1)
    {
        shapenum = GetPlayerWeaponShapeNum();

#ifdef WEAPONBOBBLE
        const bool noBobble = weaponinfo[gamestate.weapon].noBobble;
        xcenter = (viewwidth / 2) + ((noBobble ? 0 : gamestate.bobber2 - 5) + 
            tremor->weaponx) * scaleFactor;
        height = viewheight + 1 + (gamestate.bobber + 
            tremor->weapony) * scaleFactor;
        SimpleScaleShape(xcenter, shapenum, height);
#else 
        SimpleScaleShape(viewwidth/2,shapenum,viewheight+1);
#endif
    }

    if (demorecord || demoplayback)
    {
        shapenum = (int)ED_SPR("SPR_DEMO");
        SimpleScaleShape(viewwidth/2,shapenum,viewheight+1);
    }
}


//==========================================================================


/*
=====================
=
= CalcTics
=
=====================
*/

void CalcTics (void)
{
//
// calculate tics since last refresh for adaptive timing
//
    if (lasttimecount > (int32_t) GetTimeCount())
        lasttimecount = GetTimeCount();    // if the game was paused a LONG time

    uint32_t curtime = SDL_GetTicks();
    tics = (curtime * 7) / 100 - lasttimecount;
    if(!tics)
    {
        // wait until end of current tic
        SDL_Delay(((lasttimecount + 1) * 100) / 7 - curtime);
        tics = 1;
    }

    lasttimecount += tics;

    if (tics>MAXTICS)
        tics = MAXTICS;
}


//==========================================================================

void AsmRefresh()
{
    int32_t xstep,ystep;
    longword xpartial,ypartial;
    boolean playerInPushwallBackTile = tilemap[focaltx][focalty] == 64;

    for(pixx=0;pixx<viewwidth;pixx++)
    {
        short angl=midangle+pixelangle[pixx];
        if(angl<0) angl+=FINEANGLES;
        if(angl>=3600) angl-=FINEANGLES;
        if(angl<900)
        {
            xtilestep=1;
            ytilestep=-1;
            xstep=finetangent[900-1-angl];
            ystep=-finetangent[angl];
            xpartial=xpartialup;
            ypartial=ypartialdown;
        }
        else if(angl<1800)
        {
            xtilestep=-1;
            ytilestep=-1;
            xstep=-finetangent[angl-900];
            ystep=-finetangent[1800-1-angl];
            xpartial=xpartialdown;
            ypartial=ypartialdown;
        }
        else if(angl<2700)
        {
            xtilestep=-1;
            ytilestep=1;
            xstep=-finetangent[2700-1-angl];
            ystep=finetangent[angl-1800];
            xpartial=xpartialdown;
            ypartial=ypartialup;
        }
        else if(angl<3600)
        {
            xtilestep=1;
            ytilestep=1;
            xstep=finetangent[angl-2700];
            ystep=finetangent[3600-1-angl];
            xpartial=xpartialup;
            ypartial=ypartialup;
        }
        yintercept=FixedMul(ystep,xpartial)+viewy;
        xtile=focaltx+xtilestep;
        xspot=(word)((xtile<<mapshift)+((uint32_t)yintercept>>16));
        xintercept=FixedMul(xstep,ypartial)+viewx;
        ytile=focalty+ytilestep;
        yspot=(word)((((uint32_t)xintercept>>16)<<mapshift)+ytile);
        texdelta=0;

        // Special treatment when player is in back tile of pushwall
        if(playerInPushwallBackTile)
        {
            if(    (pwalldir == di_east && xtilestep ==  1)
                || (pwalldir == di_west && xtilestep == -1))
            {
                int32_t yintbuf = yintercept - ((ystep * (64 - pwallpos)) >> 6);
                if((yintbuf >> 16) == focalty)   // ray hits pushwall back?
                {
                    if(pwalldir == di_east)
                        xintercept = (focaltx << TILESHIFT) + (pwallpos << 10);
                    else
                        xintercept = (focaltx << TILESHIFT) - TILEGLOBAL + ((64 - pwallpos) << 10);
                    yintercept = yintbuf;
                    ytile = (short) (yintercept >> TILESHIFT);
                    tilehit = pwalltile;
                    patchhit = pwallpatch;
                    HitVertWall();
                    continue;
                }
            }
            else if((pwalldir == di_south && ytilestep ==  1)
                ||  (pwalldir == di_north && ytilestep == -1))
            {
                int32_t xintbuf = xintercept - ((xstep * (64 - pwallpos)) >> 6);
                if((xintbuf >> 16) == focaltx)   // ray hits pushwall back?
                {
                    xintercept = xintbuf;
                    if(pwalldir == di_south)
                        yintercept = (focalty << TILESHIFT) + (pwallpos << 10);
                    else
                        yintercept = (focalty << TILESHIFT) - TILEGLOBAL + ((64 - pwallpos) << 10);
                    xtile = (short) (xintercept >> TILESHIFT);
                    tilehit = pwalltile;
                    HitHorizWall();
                    continue;
                }
            }
        }

        do
        {
            if(ytilestep==-1 && (yintercept>>16)<=ytile) goto horizentry;
            if(ytilestep==1 && (yintercept>>16)>=ytile) goto horizentry;
vertentry:
            if((uint32_t)yintercept>mapheight*65536-1 || (word)xtile>=mapwidth)
            {
                if(xtile<0) xintercept=0, xtile=0;
                else if(xtile>=mapwidth) xintercept=mapwidth<<TILESHIFT, xtile=mapwidth-1;
                else xtile=(short) (xintercept >> TILESHIFT);
                if(yintercept<0) yintercept=0, ytile=0;
                else if(yintercept>=(mapheight<<TILESHIFT)) yintercept=mapheight<<TILESHIFT, ytile=mapheight-1;
                yspot=0xffff;
                tilehit=0;
                patchhit=0;
                HitHorizBorder();
                break;
            }
            if(xspot>=maparea) break;
            tilehit=((byte *)tilemap)[xspot];
            patchhit=((byte *)wallpatch)[xspot];
            if(tilehit)
            {
                if(tilehit&0x80)
                {
                    int32_t yintbuf=yintercept+(ystep>>1);
                    if((yintbuf>>16)!=(yintercept>>16))
                        goto passvert;
                    if((word)yintbuf<doorposition[tilehit&0x7f])
                        goto passvert;
                    yintercept=yintbuf;
                    xintercept=(xtile<<TILESHIFT)|0x8000;
                    ytile = (short) (yintercept >> TILESHIFT);
                    HitVertDoor();
                }
                else
                {
                    if(tilehit==64)
                    {
                        if(pwalldir==di_west || pwalldir==di_east)
                        {
	                        int32_t yintbuf;
                            int pwallposnorm;
                            int pwallposinv;
                            if(pwalldir==di_west)
                            {
                                pwallposnorm = 64-pwallpos;
                                pwallposinv = pwallpos;
                            }
                            else
                            {
                                pwallposnorm = pwallpos;
                                pwallposinv = 64-pwallpos;
                            }
                            if((pwalldir == di_east && xtile==pwallx && 
                                ((uint32_t)yintercept>>16)==pwally) || 
                                (pwalldir == di_west && !(xtile==pwallx && 
                                ((uint32_t)yintercept>>16)==pwally)))
                            {
                                yintbuf=yintercept+((ystep*pwallposnorm)>>6);
                                if((yintbuf>>16)!=(yintercept>>16))
                                    goto passvert;

                                xintercept=(xtile<<TILESHIFT)+TILEGLOBAL-(pwallposinv<<10);
                                yintercept=yintbuf;
                                ytile = (short) (yintercept >> TILESHIFT);
                                tilehit=pwalltile;
                                patchhit=pwallpatch;
                                HitVertWall();
                            }
                            else
                            {
                                yintbuf=yintercept+((ystep*pwallposinv)>>6);
                                if((yintbuf>>16)!=(yintercept>>16))
                                    goto passvert;

                                xintercept=(xtile<<TILESHIFT)-(pwallposinv<<10);
                                yintercept=yintbuf;
                                ytile = (short) (yintercept >> TILESHIFT);
                                tilehit=pwalltile;
                                patchhit=pwallpatch;
                                HitVertWall();
                            }
                        }
                        else
                        {
                            int pwallposi = pwallpos;
                            if(pwalldir==di_north) pwallposi = 64-pwallpos;
                            if((pwalldir==di_south && 
                                (word)yintercept<(pwallposi<<10)) || 
                                (pwalldir==di_north && 
                                (word)yintercept>(pwallposi<<10)))
                            {
                                if(((uint32_t)yintercept>>16)==pwally && xtile==pwallx)
                                {
                                    if((pwalldir==di_south && 
                                        (int32_t)((word)yintercept)+
                                        ystep<(pwallposi<<10)) || 
                                        (pwalldir==di_north && 
                                        (int32_t)((word)yintercept)+
                                        ystep>(pwallposi<<10)))
                                        goto passvert;

                                    if(pwalldir==di_south)
                                        yintercept=(yintercept&0xffff0000)+(pwallposi<<10);
                                    else
                                        yintercept=(yintercept&0xffff0000)-TILEGLOBAL+(pwallposi<<10);
                                    xintercept=xintercept-((xstep*(64-pwallpos))>>6);
                                    xtile = (short) (xintercept >> TILESHIFT);
                                    tilehit=pwalltile;
                                    HitHorizWall();
                                }
                                else
                                {
                                    texdelta = -(pwallposi<<10);
                                    xintercept=xtile<<TILESHIFT;
                                    ytile = (short) (yintercept >> TILESHIFT);
                                    tilehit=pwalltile;
                                    patchhit=pwallpatch;
                                    HitVertWall();
                                }
                            }
                            else
                            {
                                if(((uint32_t)yintercept>>16)==pwally && xtile==pwallx)
                                {
                                    texdelta = -(pwallposi<<10);
                                    xintercept=xtile<<TILESHIFT;
                                    ytile = (short) (yintercept >> TILESHIFT);
                                    tilehit=pwalltile;
                                    patchhit=pwallpatch;
                                    HitVertWall();
                                }
                                else
                                {
                                    if((pwalldir==di_south && 
                                        (int32_t)((word)yintercept)+
                                        ystep>(pwallposi<<10)) || 
                                        (pwalldir==di_north && 
                                        (int32_t)((word)yintercept)+
                                        ystep<(pwallposi<<10)))
                                        goto passvert;

                                    if(pwalldir==di_south)
                                        yintercept=(yintercept&0xffff0000)-((64-pwallpos)<<10);
                                    else
                                        yintercept=(yintercept&0xffff0000)+((64-pwallpos)<<10);
                                    xintercept=xintercept-((xstep*pwallpos)>>6);
                                    xtile = (short) (xintercept >> TILESHIFT);
                                    tilehit=pwalltile;
                                    patchhit=pwallpatch;
                                    HitHorizWall();
                                }
                            }
                        }
                    }
                    else
                    {
                        xintercept=xtile<<TILESHIFT;
                        ytile = (short) (yintercept >> TILESHIFT);
                        HitVertWall();
                    }
                }
                break;
            }
passvert:
#ifdef USE_SHADING
            shadex=xtile;
#endif
            *((byte *)spotvis+xspot)=1;
            xtile+=xtilestep;
            yintercept+=ystep;
            xspot=(word)((xtile<<mapshift)+((uint32_t)yintercept>>16));
        }
        while(1);
        continue;

        do
        {
            if(xtilestep==-1 && (xintercept>>16)<=xtile) goto vertentry;
            if(xtilestep==1 && (xintercept>>16)>=xtile) goto vertentry;
horizentry:
            if((uint32_t)xintercept>mapwidth*65536-1 || (word)ytile>=mapheight)
            {
                if(ytile<0) yintercept=0, ytile=0;
                else if(ytile>=mapheight) yintercept=mapheight<<TILESHIFT, ytile=mapheight-1;
                else ytile=(short) (yintercept >> TILESHIFT);
                if(xintercept<0) xintercept=0, xtile=0;
                else if(xintercept>=(mapwidth<<TILESHIFT)) xintercept=mapwidth<<TILESHIFT, xtile=mapwidth-1;
                xspot=0xffff;
                tilehit=0;
                patchhit=0;
                HitVertBorder();
                break;
            }
            if(yspot>=maparea) break;
            tilehit=((byte *)tilemap)[yspot];
            patchhit=((byte *)wallpatch)[yspot];
            if(tilehit)
            {
                if(tilehit&0x80)
                {
                    int32_t xintbuf=xintercept+(xstep>>1);
                    if((xintbuf>>16)!=(xintercept>>16))
                        goto passhoriz;
                    if((word)xintbuf<doorposition[tilehit&0x7f])
                        goto passhoriz;
                    xintercept=xintbuf;
                    yintercept=(ytile<<TILESHIFT)+0x8000;
                    xtile = (short) (xintercept >> TILESHIFT);
                    HitHorizDoor();
                }
                else
                {
                    if(tilehit==64)
                    {
                        if(pwalldir==di_north || pwalldir==di_south)
                        {
                            int32_t xintbuf;
                            int pwallposnorm;
                            int pwallposinv;
                            if(pwalldir==di_north)
                            {
                                pwallposnorm = 64-pwallpos;
                                pwallposinv = pwallpos;
                            }
                            else
                            {
                                pwallposnorm = pwallpos;
                                pwallposinv = 64-pwallpos;
                            }
                            if((pwalldir == di_south && ytile==pwally && 
                                ((uint32_t)xintercept>>16)==pwallx) || 
                                (pwalldir == di_north && !(ytile==pwally && 
                                ((uint32_t)xintercept>>16)==pwallx)))
                            {
                                xintbuf=xintercept+((xstep*pwallposnorm)>>6);
                                if((xintbuf>>16)!=(xintercept>>16))
                                    goto passhoriz;

                                yintercept=(ytile<<TILESHIFT)+TILEGLOBAL-(pwallposinv<<10);
                                xintercept=xintbuf;
                                xtile = (short) (xintercept >> TILESHIFT);
                                tilehit=pwalltile;
                                patchhit=pwallpatch;
                                HitHorizWall();
                            }
                            else
                            {
                                xintbuf=xintercept+((xstep*pwallposinv)>>6);
                                if((xintbuf>>16)!=(xintercept>>16))
                                    goto passhoriz;

                                yintercept=(ytile<<TILESHIFT)-(pwallposinv<<10);
                                xintercept=xintbuf;
                                xtile = (short) (xintercept >> TILESHIFT);
                                tilehit=pwalltile;
                                patchhit=pwallpatch;
                                HitHorizWall();
                            }
                        }
                        else
                        {
                            int pwallposi = pwallpos;
                            if(pwalldir==di_west) pwallposi = 64-pwallpos;
                            if((pwalldir==di_east && 
                                (word)xintercept<(pwallposi<<10)) || 
                                (pwalldir==di_west && 
                                (word)xintercept>(pwallposi<<10)))
                            {
                                if(((uint32_t)xintercept>>16)==pwallx && ytile==pwally)
                                {
                                    if((pwalldir==di_east && 
                                        (int32_t)((word)xintercept)+
                                        xstep<(pwallposi<<10)) || 
                                        (pwalldir==di_west && 
                                        (int32_t)((word)xintercept)+
                                        xstep>(pwallposi<<10)))
                                        goto passhoriz;

                                    if(pwalldir==di_east)
                                        xintercept=(xintercept&0xffff0000)+(pwallposi<<10);
                                    else
                                        xintercept=(xintercept&0xffff0000)-TILEGLOBAL+(pwallposi<<10);
                                    yintercept=yintercept-((ystep*(64-pwallpos))>>6);
                                    ytile = (short) (yintercept >> TILESHIFT);
                                    tilehit=pwalltile;
                                    patchhit=pwallpatch;
                                    HitVertWall();
                                }
                                else
                                {
                                    texdelta = -(pwallposi<<10);
                                    yintercept=ytile<<TILESHIFT;
                                    xtile = (short) (xintercept >> TILESHIFT);
                                    tilehit=pwalltile;
                                    patchhit=pwallpatch;
                                    HitHorizWall();
                                }
                            }
                            else
                            {
                                if(((uint32_t)xintercept>>16)==pwallx && ytile==pwally)
                                {
                                    texdelta = -(pwallposi<<10);
                                    yintercept=ytile<<TILESHIFT;
                                    xtile = (short) (xintercept >> TILESHIFT);
                                    tilehit=pwalltile;
                                    patchhit=pwallpatch;
                                    HitHorizWall();
                                }
                                else
                                {
                                    if((pwalldir==di_east && 
                                        (int32_t)((word)xintercept)+
                                        xstep>(pwallposi<<10)) || 
                                        (pwalldir==di_west && 
                                        (int32_t)((word)xintercept)+
                                        xstep<(pwallposi<<10)))
                                        goto passhoriz;

                                    if(pwalldir==di_east)
                                        xintercept=(xintercept&0xffff0000)-((64-pwallpos)<<10);
                                    else
                                        xintercept=(xintercept&0xffff0000)+((64-pwallpos)<<10);
                                    yintercept=yintercept-((ystep*pwallpos)>>6);
                                    ytile = (short) (yintercept >> TILESHIFT);
                                    tilehit=pwalltile;
                                    patchhit=pwallpatch;
                                    HitVertWall();
                                }
                            }
                        }
                    }
                    else
                    {
                        yintercept=ytile<<TILESHIFT;
                        xtile = (short) (xintercept >> TILESHIFT);
                        HitHorizWall();
                    }
                }
                break;
            }
passhoriz:
#ifdef USE_SHADING
            shadey=ytile;
#endif
            *((byte *)spotvis+yspot)=1;
            ytile+=ytilestep;
            xintercept+=xstep;
            yspot=(word)((((uint32_t)xintercept>>16)<<mapshift)+ytile);
        }
        while(1);
    }
}

/*
====================
=
= WallRefresh
=
====================
*/

void WallRefresh (void)
{
    xpartialdown = viewx&(TILEGLOBAL-1);
    xpartialup = TILEGLOBAL-xpartialdown;
    ypartialdown = viewy&(TILEGLOBAL-1);
    ypartialup = TILEGLOBAL-ypartialdown;

    min_wallheight = viewheight;
    lastside = -1;                  // the first pixel is on a new wall
    AsmRefresh ();
    ScalePost ();                   // no more optimization on last post
}

void CalcViewVariables()
{
//
// WSJ: set zoom
//
    if (gamestate.snipermode)
        CalcProjection(FOCALLENGTH*8);
    else
        CalcProjection(FOCALLENGTH);

    viewangle = Mecha_GetViewingFineAngle();
    midangle = viewangle;
    viewsin = finesintable[viewangle];
    viewcos = finecostable[viewangle];
    //viewx = Player_CamGetX() - FixedMul(focallength,viewcos);
    //viewy = Player_CamGetY() + FixedMul(focallength,viewsin);
    if ( focallength < FOCALLENGTH )
    {
        viewx = Player_CamGetX() - FixedMul( focallength, viewcos );
        viewy = Player_CamGetY() + FixedMul( focallength, viewsin );
    }
    else
    {
        viewx = Player_CamGetX() - FixedMul( FOCALLENGTH, viewcos );
        viewy = Player_CamGetY() + FixedMul( FOCALLENGTH, viewsin );
    }

    focaltx = (short)(viewx>>TILESHIFT);
    focalty = (short)(viewy>>TILESHIFT);

    viewtx = (short)(Player_CamGetX() >> TILESHIFT);
    viewty = (short)(Player_CamGetY() >> TILESHIFT);
}

//==========================================================================

/*
========================
=
= ThreeDRefresh
=
========================
*/

void    ThreeDRefresh (void)
{
//
// clear out the traced array
//
    memset(spotvis,0,maparea);

    // Detect all sprites over player fix
    spotvis[Player_CamGetTileX()][Player_CamGetTileY()] = 1;

    vbuf = VL_LockSurface(screenBuffer);
    vbuf+=screenofs;
    vbufPitch = bufferPitch;

    CalcViewVariables();

//
// follow the walls from there to the right, drawing as we go
//
    VGAClearScreen ();
#if defined(USE_FEATUREFLAGS) && defined(USE_STARSKY)
    if(GetFeatureFlags() & FF_STARSKY)
        DrawStarSky(vbuf, vbufPitch);
#endif

#if defined(USE_FEATUREFLAGS) && defined(USE_PARALLAX)
    if(GetFeatureFlags() & FF_PARALLAXSKY)
        DrawParallax(vbuf, vbufPitch);
#endif
#if defined(USE_FEATUREFLAGS) && defined(USE_CLOUDSKY)
    if(GetFeatureFlags() & FF_CLOUDSKY)
        DrawClouds(vbuf, vbufPitch, min_wallheight);
#endif

    WallRefresh ();

#ifdef USE_FLOORCEILINGTEX
    DrawFloorAndCeiling(vbuf, vbufPitch, min_wallheight);
#endif

//
// draw all the scaled images
//
    DrawScaleds();                  // draw scaled stuff

#if defined(USE_FEATUREFLAGS) && defined(USE_RAIN)
    if(GetFeatureFlags() & FF_RAIN)
        DrawRain(vbuf, vbufPitch);
#endif
#if defined(USE_FEATUREFLAGS) && defined(USE_SNOW)
    if(GetFeatureFlags() & FF_SNOW)
        DrawSnow(vbuf, vbufPitch);
#endif

    if (invisiblemode != 1)
        DrawPlayerWeapon ();    // draw player's hands

    DrawPlayBorderSides (); // WSJ: draw in-game messages
    if (messagetime > 0)
    {
        DrawMessage ();
        messagetime-=tics;
    }

    if(Keyboard[sc_Tab] && viewsize == 21 && gamestate.weapon != -1)
        ShowActStatus();

    VL_UnlockSurface(screenBuffer);
    vbuf = NULL;

//
// show screen and time last cycle
//

    if (fizzlein)
    {
        FizzleFade(screenBuffer, screenSurf, 0, 0,
            screenWidth, screenHeight, 20, false);
        fizzlein = false;

        lasttimecount = GetTimeCount();          // don't make a big tic count
    }
    else
    {
#ifndef REMDEBUG
        if (fpscounter)
        {
            fontnumber = 0;
            SETFONTCOLOR(7,127);
            PrintX=4; PrintY=1;
            VWB_Bar(0,0,50,10,bordercol);
            US_PrintSigned(fps);
            US_Print(" fps");
        }
#endif
        SDL_BlitSurface(screenBuffer, NULL, screenSurf, NULL);
        SDL_UpdateRect(screenSurf, 0, 0, 0, 0);
    }

#ifndef REMDEBUG
    if (fpscounter)
    {
        fps_frames++;
        fps_time+=tics;

        if(fps_time>35)
        {
            fps_time-=35;
            fps=fps_frames<<1;
            fps_frames=0;
        }
    }
#endif
}

//===========================================================================

/*
========================
=
= GetMessage
=
========================
*/

void GetMessage (const char *lastmessage)
{
    messagetime = 200; // time for message to display

    strcpy(gamestate.message, lastmessage);
}

/*
========================
=
= DrawMessage
=
========================
*/

void DrawMessage (void)
{
    fontnumber = 0;
    SETFONTCOLOR(0xc,0x0);
    PrintY=4;
    US_GamePrint(gamestate.message);
}

//==========================================================================

fixed Math_Distance2(fixed x1, fixed y1, fixed x2, fixed y2)
{
    fixed dist;

    if (x1 != x2 && y1 != y2)
    {
        dist = fixedpt_sqrt(FP_SQ(x1 - x2) + FP_SQ(y1 - y2));
    }
    else if (x1 != x2)
    {
        dist = abs(x1 - x2);
    }
    else if (y1 != y2)
    {
        dist = abs(y1 - y2);
    }
    else
    {
        dist = 0;
    }

    return dist;
}

//===========================================================================

void CA_CopyPlanes(word dst[MAPPLANES][maparea], int dx, int dy,
    word src[MAPPLANES][maparea], int sx, int sy, int w, int h)
{
    int i;
    int x, y;

#define TRUNC_SX_W(sx, w) \
do { \
    if (sx + w <= 0) \
    { \
        return; \
    } \
    if (sx >= MAPSIZE) \
    { \
        return; \
    } \
    if (sx + w > MAPSIZE) \
    { \
        w = MAPSIZE - sx; \
    } \
    if (sx < 0) \
    { \
        w += sx; \
        sx = 0; \
    } \
} while (0)
    TRUNC_SX_W(sx, w);
    TRUNC_SX_W(sy, h);
    TRUNC_SX_W(dx, w);
    TRUNC_SX_W(dy, h);

    for (x = 0; x < w; x++)
    {
        for (y = 0; y < h; y++)
        {
            for (i = 0; i < MAPPLANES; i++)
            {
                dst[i][(dy + y) * MAPSIZE + (dx + x)] = 
                    src[i][(sy + y) * MAPSIZE + (sx + x)];
            }
        }
    }
#undef TRUNC_SX_W
}

//===========================================================================

int WallMidY(int ywcount, int bot)
{
    return fixedpt_toint((FP(1) + (bot * gamestate.camz * 2)) * ywcount);
}

int InvWallMidY(int y, int bot)
{
    return 
        fixedpt_toint(
            fixedpt_div(
                FP(y),
                FP(1) + (bot * gamestate.camz * 2)
                )
            ) + 1;
}

void CameraSlowVerticalPan(void)
{
    static fixed camz = 0;
    static int camdir = 1;
    static fixed camspeed = FP(1) / 2048;
    camz += tics * camspeed * camdir;
    if (camz >= 0x7000 - camspeed || camz <= -0x7000 + camspeed)
    {
        camdir = -camdir;
    }
    gamestate.camz = camz;
}
