//
//	WL Edit
//	WL_ED.cpp - Game Editor
//	By Linux Wolf - Team RayCAST
//

#include "lw.h"
#include "wl_def.h"
#include "wl_shade.h"

#define ED_FONTPATH "editor/textures/ui/font.tga"
#define ED_STATEFILE EDITSTATE_FILE

typedef enum ED_ToolId_e
{
	ED_ToolId_MapTool,
	ED_ToolId_TileViewer,
	ED_ToolId_PicViewer,
	ED_ToolId_Picker,
	ED_ToolId_ModelMaker,
	ED_ToolId_Last = ED_ToolId_ModelMaker,
} ED_ToolId_t;

lwlib_EnumStr(
	ED_ToolId,
	"mapTool",
	"tileViewer",
	"picViewer",
	"picker",
	"modelMaker",
);

typedef struct ED_StateHandler_s
{
	const char *name;
	ED_ThinkFunc_t thinkFunc;
} ED_StateHandler_t;

static ED_StateHandler_t ED_StateHandlers[] =
{
#define ED_LH(func) { #func, (ED_ThinkFunc_t)func }
	ED_LH(A_BJRocketBlast),
	ED_LH(A_DeathScream),
	ED_LH(A_RemoveBlocking),
	ED_LH(A_BecomeShootable),
	ED_LH(A_RemoveShootable),
	ED_LH(A_BecomeVulnerable),
	ED_LH(A_RocketBlast),
	ED_LH(A_Slurpie),
	ED_LH(A_Smoke),
	ED_LH(T_Aim),
	ED_LH(T_Aim2),
	ED_LH(T_Attack),
	ED_LH(T_BoneShrapnel),
	ED_LH(T_BossRefire),
	ED_LH(T_BossShoot),
	ED_LH(T_BossShootS),
	ED_LH(T_CageMutWait),
	ED_LH(T_CageMutAwake),
	ED_LH(A_CageMutFinishEscape),
	ED_LH(A_CageMutEscapeSound),
	ED_LH(T_LadderMutWait),
	ED_LH(T_LadderMutFinishClimb),
	ED_LH(T_Chase),
	ED_LH(T_DogBite),
	ED_LH(T_DogChase),
	ED_LH(T_Melee),
	ED_LH(T_MG42GrdLoop),
	ED_LH(T_MG42MutLoop),
	ED_LH(T_MG42OfcLoop),
	ED_LH(T_MG42Shoot),
	ED_LH(T_MG42ShootS),
	ED_LH(T_MG42SkelLoop),
	ED_LH(T_MG42SSLoop),
	ED_LH(T_MG42Wait),
	ED_LH(T_MG42ZSSLoop),
	ED_LH(T_MP40Shoot),
	ED_LH(T_STG44Shoot),
	ED_LH(T_OfcRefire),
	ED_LH(T_Path),
	ED_LH(T_PistolShoot),
	ED_LH(T_Player),
	ED_LH(T_Projectile),
	ED_LH(T_Reload),
	ED_LH(T_RifleShoot),
	ED_LH(T_RopeMutWait),
	ED_LH(T_RopeMutFinishSlide),
	ED_LH(T_SciRefire),
	ED_LH(T_SciMelee),
	ED_LH(T_SciThrow),
	ED_LH(T_SkelAttack),
	ED_LH(T_SkelBuried),
	ED_LH(T_SkelFireBomb),
	ED_LH(T_SkelMeleeLoop),
	ED_LH(T_SSRefire),
	ED_LH(T_SSOfcRefire),
	ED_LH(T_Stand),
	ED_LH(T_StartDig),
	ED_LH(T_StartDigInChase),
	ED_LH(T_StartRise),
	ED_LH(T_Underground),
	ED_LH(T_ZSSAttack),
	ED_LH(T_ZSSBuried),
	ED_LH(T_ZSSRefire),
	ED_LH(T_ZWake),
	ED_LH(T_KnightSkelWait),
	ED_LH(A_KnightSkelFinishDrop),
	ED_LH(T_BossShootFireProj),
	ED_LH(A_BossBurnStart),
	ED_LH(A_BossBurnRepeat),
	ED_LH(T_FireProj),
	ED_LH(T_FireShotgun),
	ED_LH(A_HangedMutFinishDrop),
	ED_LH(A_BrownHorseRandomDelay),
	ED_LH(A_BrownHorsePlaySound),
	ED_LH(A_BarrelSplashDamage),
	ED_LH(T_DamageMut),
	ED_LH(T_TableMutDamage),
	ED_LH(A_BecomeBlocking),
	ED_LH(A_SpawnSurpriseEnemy),
	ED_LH(A_VillagerStart),
	ED_LH(T_Villager),
	ED_LH(A_VillagerFinishResponse),
	ED_LH(A_VillagerRandomDelay),
	ED_LH(A_BarrelDebrisFall),
	ED_LH(T_BarrelDebrisHurt),
	ED_LH(A_GreenToxBarrelHurt),
	ED_LH(A_MechMut2Evacuate),
	ED_LH(A_MechMut4Evacuate),
	ED_LH(T_MechMut4Chase),
	ED_LH(T_MechMut4OneArmChase),
	ED_LH(T_MechaSuit),
	ED_LH(T_MechaSuitOneArm),
	ED_LH(T_MechaSuitNoArms),
	ED_LH(T_MechaExit),
	ED_LH(A_PlayMechBoomSound),
	ED_LH(A_BlackCatPlaySound),
	ED_LH(A_MechaSound),
	ED_LH(A_FlameScream),
	ED_LH(T_MechaSuitEnemyEnter),
	ED_LH(A_FlameBurn),
	ED_LH(A_MechMut4Explode),
	ED_LH(A_MechMut4NoArms),
	ED_LH(A_MechMut1SpawnSuit),
	ED_LH(A_ReachedBody),
	ED_LH(A_ResurrectBody),
	ED_LH(A_BecomeNoPain),
	ED_LH(A_RemoveNoPain),
	ED_LH(A_WaterMutEnter),
	ED_LH(A_WaterMutExit),
	ED_LH(A_BecomeUnderground),
	ED_LH(A_RemoveUnderground),
	ED_LH(A_Dead),
	ED_LH(A_SewerBossShoot),
	ED_LH(A_SummonGhostFromDead),
	ED_LH(A_ResetSpawnedGhostFlag),
#undef ED_LH
};

typedef struct ED_SDLApp_s
{
	bool visible;
	bool mouseFocus;
	bool keyboardFocus;
} ED_SDLApp_t;

typedef struct ED_Priv_s
{
	lwlib::Gl *gl;
	lwlib::Font *font;
	lwlib::Edit *edit;
	lwlib_Keys_t keys;
	lwlib::ColorRemap *colorRemap;
	SDL_Color *palette;
	lwlib::Edit_Persist_t persist;
	bool quit;
	ED_SpriteSeq_t sprites;
	lwlib_StrMap_t spriteMap;
	ED_StatInfoSeq_t statInfos;
	lwlib_IntMap_t statInfoMap;
	lwlib::MapTool *mapTool;
	lwlib::Picker *picker;
	lwlib::ToolPtrSeq_t toolPtrs;
	ED_ToolId_t activeTool;
	ED_StateHandler_t *stateHandlers;
	int numStateHandlers;
	lwlib_StrMap_t stateHandlerMap;
	ED_StateSeq_t states;
	lwlib_StrMap_t stateMap;
	ED_SpawnSeq_t spawns;
	lwlib_IntMap_t spawnMap;
	ED_EnemySeq_t enemies;
	lwlib_StrMap_t enemyMap;
	lwlib::Img_t signonImg;
	ED_SDLApp_t sdlApp;
	ED_EnemyConfig_t enemyConfig;
	ED_ProjSeq_t projs;
	lwlib_StrMap_t projMap;
	ED_ProjPhysSeq_t projPhysSeq;
	lwlib_StrMap_t projPhysMap;
	ED_Inject_t inject;
	lwlib::Mouse_t mouse;
	ED_SoundSeq_t sounds;
	lwlib_StrMap_t soundMap;
	ED_RandomSoundSeq_t randomSounds;
	lwlib_StrMap_t randomSoundMap;
	ED_BonusSeq_t bonuses;
	lwlib_StrMap_t bonusMap;
	lwlib::PicViewer *picViewer;
	ED_FootStepSeq_t footSteps;
	lwlib_IntMap_t footStepMap;
	ED_ChanceRangeSeq_t chanceRanges;
	ED_VSWAPInfo_t vswapInfo;
	ED_AmbientSeq_t ambients;
	lwlib_IntMap_t ambientMap;
	lwlib_StrMap_t ambientStrMap;
	ED_TileInfo_t tileInfo;
	ED_RicochetSeq_t ricochets;
	lwlib_StrMap_t ricochetMap;
	ED_WallAttrSeq_t wallAttrs;
	int lrPack;
	ED_AmbientSettings_t ambientSettings;
} ED_Priv_t;

extern byte signon[];

lwlib_EnumStr(
	ED_Stat,
	"none",
	"block",
	"bo_mg42_mounted",
	"bo_any",
);

lwlib_EnumStr(
	ED_DamageType,
    "dam_none",
    "dam_melee",
    "dam_bullet",
    "dam_mutagen",
    "dam_rifle",
    "dam_fire",
    "dam_chaingun",
    "dam_rocket",
    "dam_selfdestruct",
    "dam_infection",
);

lwlib_EnumStr(
	ED_ObjFlag,
	"FL_NONE",
	"FL_SHOOTABLE",
	"FL_BONUS",
	"FL_NEVERMARK",
	"FL_VISABLE",
	"FL_ATTACKMODE",
	"FL_FIRSTATTACK",
	"FL_AMBUSH",
	"FL_NONMARK",
	"FL_FULLBRIGHT",
//#ifdef USE_DIR3DSPR
	"FL_DIR_HORIZ_MID",
	"FL_DIR_HORIZ_FW",
	"FL_DIR_HORIZ_BW",
	"FL_DIR_VERT_MID",
	"FL_DIR_VERT_FW",
	"FL_DIR_VERT_BW",
	"FL_DIR_VERT_FLAG",
//#endif
	"FL_BLOCK",
	"FL_UNKILLABLE",
	"FL_SUPPRESSPAIN",
	"FL_FENCING",
	"FL_UNDERGROUND",
	"FL_WEAKTOROCKETS",
	"FL_NODAMAGE",
	"FL_SPAWNEDGHOST",
	"FL_GLOWBRIGHT",
);

lwlib_EnumVal(
	ED_ObjFlag,
	FL_NONE,
	FL_SHOOTABLE,
	FL_BONUS,
	FL_NEVERMARK,
	FL_VISABLE,
	FL_ATTACKMODE,
	FL_FIRSTATTACK,
	FL_AMBUSH,
	FL_NONMARK,
	FL_FULLBRIGHT,
//#ifdef USE_DIR3DSPR
	FL_DIR_HORIZ_MID,
	FL_DIR_HORIZ_FW,
	FL_DIR_HORIZ_BW,
	FL_DIR_VERT_MID,
	FL_DIR_VERT_FW,
	FL_DIR_VERT_BW,
	FL_DIR_VERT_FLAG,
//#endif
	FL_BLOCK,
	FL_UNKILLABLE,
	FL_SUPPRESSPAIN,
	FL_FENCING,
	FL_UNDERGROUND,
	FL_WEAKTOROCKETS,
);

lwlib_EnumStr(
	ED_AmmoType,
	"ammo_bullet",
	"ammo_rifle",
	"ammo_stg44",
	"ammo_flame",
	"ammo_chaingun",
	"ammo_rocket",
	"ammo_shotgun",
	"ammo_fuel",
	"ammo_antidote",
);

lwlib_EnumStr(
	ED_EnemyRangeAttackVariant,
	"Normal",
	"Long",
);

static const char *ED_WeaponType_Str[] =
{
	"wp_none",
	"wp_knife",
	"wp_pistol",
	"wp_kar98k",
	"wp_sniper",
	"wp_mp40",
	"wp_stg44",
	"wp_shotgun",
	"wp_flamethrower",
	"wp_chaingun",
	"wp_panzerschreck",
	"wp_wirecutters",
	"wp_cyringe",
	"wp_mg42_mounted",
	"wp_mecha",
};

static const int ED_WeaponType_Vals[] =
{
	wp_none,
	wp_knife,
	wp_pistol,
	wp_kar98k,
	wp_sniper,
	wp_mp40,
	wp_stg44,
	wp_shotgun,
	wp_flamethrower,
	wp_chaingun,
	wp_panzerschreck,
	wp_wirecutters,
	wp_cyringe,
	wp_mg42_mounted,
	wp_mecha,
};

static const char *ED_AmbientType_Str[] =
{
	"amb_area",
	"amb_rain",
	"amb_raincover",
	"amb_water",
	"amb_scenery",
};

static const int ED_AmbientType_Vals[] =
{
	amb_area,
	amb_rain,
	amb_raincover,
	amb_water,
	amb_scenery,
};

lwlib_EnumStr(
	ED_FeatureType,
	"feat_resurrection",
	"feat_watermut",
	"feat_boss",
	"feat_damagedebris",
	"feat_skillprofile",
	"feat_playvoice",
	"feat_melee",
	"feat_summoner",
	"feat_guarded",
	"feat_music",
);

static void ED_Sprite_AddToEdit(void *elem, lwlib::Edit *edit)
{
	ED_Sprite_t *sprite;

	sprite = (ED_Sprite_t *)elem;

	edit->addItem_string("name", sprite->name, sizeof(sprite->name));
}

static void ED_StatInfo_AddToEdit(void *elem, lwlib::Edit *edit)
{
	ED_StatInfo_t *statInfo;

	statInfo = (ED_StatInfo_t *)elem;

	edit->addItem_string("name", statInfo->name, sizeof(statInfo->name));
	edit->addItem_string("picnum", statInfo->picnum, 
		sizeof(statInfo->picnum));
	edit->addItem_enum("type", (int *)&statInfo->type, ED_Stat_Str, 
		NULL, lwlib_CountArray(ED_Stat_Str));
	edit->addItem_string("bonus", statInfo->bonus, 
		sizeof(statInfo->bonus));
	edit->addItem_enum("specialFlags", (int *)&statInfo->specialFlags,
		ED_ObjFlag_Str, ED_ObjFlag_Val, ED_ObjFlag_StrCount);
	edit->addItem_int("animframes", &statInfo->animframes);
	edit->addItem_int("actor", &statInfo->actor);
}

static void ED_Spawn_AddToEdit(void *elem, lwlib::Edit *edit)
{
	ED_Spawn_t *spawn;

	spawn = (ED_Spawn_t *)elem;

	edit->addItem_string("enemyName", spawn->enemyName, 
		sizeof(spawn->enemyName));
	edit->addItem_string("startState", spawn->startState,
		sizeof(spawn->startState));
	edit->addItem_int("dir", &spawn->dir);
	edit->addItem_bool("patrol", &spawn->patrol);
	edit->addItem_int("actor", &spawn->actor);
}

static void ED_Str_AddToEdit(void *elem, lwlib::Edit *edit)
{
	lwlib_Str_t *str;
	str = (lwlib_Str_t *)elem;
	edit->addItem_string("str", str->v, sizeof(str->v));
}

static void ED_RandomSound_AddToEdit(void *elem,
	lwlib::Edit *edit)
{
	ED_RandomSound_t *randomSound;

	randomSound = (ED_RandomSound_t *)elem;

	edit->addItem_string("name", randomSound->name, 
		sizeof(randomSound->name));
	lwlib_Edit_AddVec(edit, "soundDefs", randomSound->soundDefs, NULL, 
		NULL, ED_Str_AddToEdit);
}

static void ED_DropItem_AddToEdit(void *elem, lwlib::Edit *edit)
{
	ED_DropItem_t *dropItem;

	dropItem = (ED_DropItem_t *)elem;

	edit->addItem_string("bonusDef", dropItem->bonusDef,
		sizeof(dropItem->bonusDef));
	edit->addItem_int("chance", &dropItem->chance);
}

static void ED_Resurrection_AddToEdit(ED_Resurrection_t *resurrection, 
	lwlib::Edit *edit)
{
	edit->addItem_string("undeadName", resurrection->undeadName,
		sizeof(resurrection->undeadName));
	edit->addItem_string("resurrectState", 
		resurrection->resurrectState,
		sizeof(resurrection->resurrectState));
	edit->addItem_string("resurrectSound", 
		resurrection->resurrectSound,
		sizeof(resurrection->resurrectSound));
	edit->addItem_string("enterBodyState", 
		resurrection->enterBodyState,
		sizeof(resurrection->enterBodyState));
	edit->addItem_string("enterBodySound", 
		resurrection->enterBodySound,
		sizeof(resurrection->enterBodySound));
	edit->addItem_string("skinState", resurrection->skinState,
		sizeof(resurrection->skinState));
}

static void ED_Summoner_AddToEdit(ED_Summoner_t *summoner, 
	lwlib::Edit *edit)
{
	edit->addItem_int("summonChance", &summoner->summonChance);
	edit->addItem_string("summonState", summoner->summonState,
		sizeof(summoner->summonState));
	edit->addItem_string("summonSound", summoner->summonSound,
		sizeof(summoner->summonSound));
}

static void ED_Guarded_AddToEdit(ED_Guarded_t *guarded, 
	lwlib::Edit *edit)
{
}

static void ED_Music_AddToEdit(ED_Music_t *music, 
	lwlib::Edit *edit)
{
	edit->addItem_int("mapNum", &music->mapNum);
	edit->addItem_int("trackNum", &music->trackNum);
}

static void ED_WaterMut_AddToEdit(ED_WaterMut_t *waterMut, 
	lwlib::Edit *edit)
{
	edit->addItem_string("shootState", waterMut->shootState,
		sizeof(waterMut->shootState));
	edit->addItem_string("dieState", waterMut->dieState,
		sizeof(waterMut->dieState));
}

static void ED_MapNum_AddToEdit(void *elem, lwlib::Edit *edit)
{
	ED_MapNum_t *mapNum;
	mapNum = (ED_MapNum_t *)elem;
	edit->addItem_int("num", mapNum);
}

static void ED_Boss_AddToEdit(ED_Boss_t *boss, 
	lwlib::Edit *edit)
{
	lwlib_Edit_AddVec(edit, "musicMaps", boss->musicMaps, 
		NULL, NULL, ED_MapNum_AddToEdit);
	edit->addItem_int("lastMusicMap", &boss->lastMusicMap);
	edit->addItem_bool("ownsMusic", &boss->ownsMusic);
}

static void ED_DamageDebris_AddToEdit(ED_DamageDebris_t *damageDebris, 
	lwlib::Edit *edit)
{
	edit->addItem_string("projName", damageDebris->projName,
		sizeof(damageDebris->projName));
	edit->addItem_int("hitpoints", &damageDebris->hitpoints);
}

static void ED_SkillProfile_AddToEdit(ED_SkillProfile_t *skillProfile, 
	lwlib::Edit *edit)
{
	edit->addItem_int("hitpoints", &skillProfile->hitpoints);
}

static void ED_PlaySound_AddToEdit(ED_PlayVoice_t *playVoice, 
	lwlib::Edit *edit)
{
	edit->addItem_string("sound", playVoice->sound,
		sizeof(playVoice->sound));
}

static void ED_Melee_AddToEdit(ED_Melee_t *melee, 
	lwlib::Edit *edit)
{
	edit->addItem_string("hitSound", melee->hitSound,
		sizeof(melee->hitSound));
	edit->addItem_string("missSound", melee->missSound,
		sizeof(melee->missSound));
}

static void ED_Feature_AddToEdit(void *elem, lwlib::Edit *edit)
{
	ED_Feature_t *feature;

	feature = (ED_Feature_t *)elem;

	edit->addItem_enum("type", (int *)&feature->type, ED_FeatureType_Str, NULL, 
		lwlib_CountArray(ED_FeatureType_Str));
	switch (feature->type)
	{
	case feat_resurrection:
		lwlib_Edit_AddGroup(edit, "resurrection", 
			ED_Resurrection_AddToEdit(&feature->u.resurrection, edit));
		break;
	case feat_watermut:
		lwlib_Edit_AddGroup(edit, "waterMut", 
			ED_WaterMut_AddToEdit(&feature->u.waterMut, edit));
		break;
	case feat_boss:
		lwlib_Edit_AddGroup(edit, "boss", 
			ED_Boss_AddToEdit(&feature->u.boss, edit));
		break;
	case feat_damagedebris:
		lwlib_Edit_AddGroup(edit, "damageDebris", 
			ED_DamageDebris_AddToEdit(&feature->u.damageDebris, edit));
		break;
	case feat_skillprofile:
		lwlib_Edit_AddGroup(edit, "skillProfile", 
			ED_SkillProfile_AddToEdit(&feature->u.skillProfile, edit));
		break;
	case feat_playvoice:
		lwlib_Edit_AddGroup(edit, "playVoice", 
			ED_PlaySound_AddToEdit(&feature->u.playVoice, edit));
		break;
	case feat_melee:
		lwlib_Edit_AddGroup(edit, "melee", 
			ED_Melee_AddToEdit(&feature->u.melee, edit));
		break;
	case feat_summoner:
		lwlib_Edit_AddGroup(edit, "summoner", 
			ED_Summoner_AddToEdit(&feature->u.summoner, edit));
		break;
	case feat_guarded:
		lwlib_Edit_AddGroup(edit, "guarded", 
			ED_Guarded_AddToEdit(&feature->u.guarded, edit));
		break;
	case feat_music:
		lwlib_Edit_AddGroup(edit, "music", 
			ED_Music_AddToEdit(&feature->u.music, edit));
		break;
	}
}

static void ED_State_AddToEdit(void *elem, lwlib::Edit *edit)
{
	ED_State_t *state;

	state = (ED_State_t *)elem;

	edit->addItem_string("name", state->name, sizeof(state->name));
	edit->addItem_int("rotate", &state->rotate);
	edit->addItem_string("shapenum", state->shapenum, 
		sizeof(state->shapenum));
	edit->addItem_short("tictime", &state->tictime);
	edit->addItem_string("think", state->think, sizeof(state->think));
	edit->addItem_string("action", state->action, sizeof(state->action));
	edit->addItem_string("next", state->next, sizeof(state->next));
	edit->addItem_bool("fullbright", &state->fullbright);
	lwlib_Edit_AddVec(edit, "features", state->features,
		NULL, NULL, ED_Feature_AddToEdit);
}

static void ED_EnemyDeath_AddToEdit(ED_EnemyDeath_t *enemyDeath, lwlib::Edit *edit)
{
	edit->addItem_string("burnState",
		enemyDeath->burnState, sizeof(enemyDeath->burnState));
	edit->addItem_string("gibState",
		enemyDeath->gibState, sizeof(enemyDeath->gibState));
	edit->addItem_string("headShotState",
		enemyDeath->headShotState, sizeof(enemyDeath->headShotState));
	edit->addItem_string("dieState",
		enemyDeath->dieState, sizeof(enemyDeath->dieState));
	lwlib_Edit_AddVec(edit, "dropItems", enemyDeath->dropItems,
		NULL, NULL, ED_DropItem_AddToEdit);
	edit->addItem_int("burnHealth", &enemyDeath->burnHealth);
	edit->addItem_string("screamSound", enemyDeath->screamSound, 
		sizeof(enemyDeath->screamSound));
	edit->addItem_string("burnScreamSound", enemyDeath->burnScreamSound, 
		sizeof(enemyDeath->burnScreamSound));
	edit->addItem_bool("doVictory", &enemyDeath->doVictory);
}

static void ED_EnemyMg42_AddToEdit(ED_EnemyMg42_t *enemyMg42, lwlib::Edit *edit)
{
	edit->addItem_bool("canUse", &enemyMg42->canUse);
	edit->addItem_bool("canWalkThrough", &enemyMg42->canWalkThrough);
	edit->addItem_string("shootState",
		enemyMg42->shootState, sizeof(enemyMg42->shootState));
	edit->addItem_string("waitState",
		enemyMg42->waitState, sizeof(enemyMg42->waitState));
}

static void ED_EnemyReactionTime_AddToEdit(ED_EnemyReactionTime_t *enemyReactionTime, lwlib::Edit *edit)
{
	edit->addItem_int("baseTics", &enemyReactionTime->baseTics);
	edit->addItem_int("rndMultiplier", &enemyReactionTime->rndMultiplier);
	edit->addItem_int("rndDivisor", &enemyReactionTime->rndDivisor);
}

static void ED_SpriteRepl_AddToEdit(void *elem, lwlib::Edit *edit)
{
	ED_SpriteRepl_t *spriteRepl;

	spriteRepl = (ED_SpriteRepl_t *)elem;

	edit->addItem_string("oldSprite", spriteRepl->oldSprite,
		sizeof(spriteRepl->oldSprite));
	edit->addItem_string("lastOldSprite", spriteRepl->lastOldSprite,
		sizeof(spriteRepl->lastOldSprite));
	edit->addItem_string("newSprite", spriteRepl->newSprite,
		sizeof(spriteRepl->newSprite));
}

static void ED_SpriteOffset_AddToEdit(ED_SpriteOffset_t *spriteOffset, lwlib::Edit *edit)
{
	edit->addItem_string("newStart", spriteOffset->newStart, 
		sizeof(spriteOffset->newStart));
	edit->addItem_string("oldStart", spriteOffset->oldStart,
		sizeof(spriteOffset->oldStart));
}

static void ED_EnemyBloodPuff_AddToEdit(ED_EnemyBloodPuff_t *enemyBloodPuff, lwlib::Edit *edit)
{
	edit->addItem_string("state", enemyBloodPuff->state, 
		sizeof(enemyBloodPuff->state));
	edit->addItem_string("soundDef", enemyBloodPuff->soundDef, 
		sizeof(enemyBloodPuff->soundDef));
	lwlib_Edit_AddGroup(edit, "spriteOffset", 
		ED_SpriteOffset_AddToEdit(&enemyBloodPuff->spriteOffset, edit));
}

static void ED_EnemyBurrow_AddToEdit(ED_EnemyBurrow_t *enemyBurrow, lwlib::Edit *edit)
{
	edit->addItem_string("digInState", enemyBurrow->digInState, 
		sizeof(enemyBurrow->digInState));
	edit->addItem_string("digOutState", enemyBurrow->digOutState, 
		sizeof(enemyBurrow->digOutState));
}

static void ED_EnemySpawnFlags_AddToEdit(ED_EnemySpawnFlags_t *enemySpawnFlags, lwlib::Edit *edit)
{
	edit->addItem_bool("override", &enemySpawnFlags->override);
	edit->addItem_enum("flags[0]", (int *)&enemySpawnFlags->flags[0],
		ED_ObjFlag_Str, ED_ObjFlag_Val, ED_ObjFlag_StrCount);
	edit->addItem_enum("flags[1]", (int *)&enemySpawnFlags->flags[1],
		ED_ObjFlag_Str, ED_ObjFlag_Val, ED_ObjFlag_StrCount);
	edit->addItem_enum("flags[2]", (int *)&enemySpawnFlags->flags[2],
		ED_ObjFlag_Str, ED_ObjFlag_Val, ED_ObjFlag_StrCount);
}

static void ED_EnemySurpriseAttack_AddToEdit(ED_EnemySurpriseAttack_t *enemySurpriseAttack, lwlib::Edit *edit)
{
	edit->addItem_bool("noAmbushReset", &enemySurpriseAttack->noAmbushReset);
	edit->addItem_string("spawnEnemyName", 
		enemySurpriseAttack->spawnEnemyName, sizeof(enemySurpriseAttack->spawnEnemyName));
}

static void ED_EnemyChanceConfig_AddToEdit(
	ED_EnemyChanceConfig_t *chanceConfig, lwlib::Edit *edit)
{
	edit->addItem_int("timeout", &chanceConfig->timeout);
	edit->addItem_int("firstRange", &chanceConfig->firstRange);
	edit->addItem_int("numRanges", &chanceConfig->numRanges);
}

static void ED_Enemy_AddToEdit(void *elem, lwlib::Edit *edit)
{
	ED_Enemy_t *enemy;

	enemy = (ED_Enemy_t *)elem;

	edit->addItem_string("name", enemy->name, sizeof(enemy->name));
	lwlib_Edit_AddGroup(edit, "spriteOffset", 
		ED_SpriteOffset_AddToEdit(&enemy->spriteOffset, edit));
	lwlib_Edit_AddVec(edit, "spriteRepls", enemy->spriteRepls,
		NULL, NULL, ED_SpriteRepl_AddToEdit);
	edit->addItem_short("health", &enemy->health);
	edit->addItem_short("killpoints", &enemy->killpoints);
	edit->addItem_int("speed", &enemy->speed);
	edit->addItem_int("chaseSpeed", &enemy->chaseSpeed);
	edit->addItem_string("shootState", enemy->shootState,
		sizeof(enemy->shootState));
	edit->addItem_string("painStateOdd", enemy->painStateOdd,
		sizeof(enemy->painStateOdd));
	edit->addItem_string("painStateEven", enemy->painStateEven, 
		sizeof(enemy->painStateEven));
	edit->addItem_string("chaseState", enemy->chaseState, 
		sizeof(enemy->chaseState));
	edit->addItem_string("meleeState", enemy->meleeState, 
		sizeof(enemy->meleeState));
	edit->addItem_string("escapeState", enemy->escapeState, 
		sizeof(enemy->escapeState));
	lwlib_Edit_AddGroup(edit, "death", 
		ED_EnemyDeath_AddToEdit(&enemy->death, edit));
	lwlib_Edit_AddGroup(edit, "mg42", 
		ED_EnemyMg42_AddToEdit(&enemy->mg42, edit));
	edit->addItem_string("alertSound", enemy->alertSound, 
		sizeof(enemy->alertSound));
	edit->addItem_enum("rangeAttackVariant", 
		(int *)&enemy->rangeAttackVariant,
		ED_EnemyRangeAttackVariant_Str, NULL, 
		lwlib_CountArray(ED_EnemyRangeAttackVariant_Str));
	lwlib_Edit_AddGroup(edit, "reactionTime", 
		ED_EnemyReactionTime_AddToEdit(&enemy->reactionTime, edit));
	lwlib_Edit_AddGroup(edit, "bloodPuff", 
		ED_EnemyBloodPuff_AddToEdit(&enemy->bloodPuff, edit));
	lwlib_Edit_AddGroup(edit, "burrow", 
		ED_EnemyBurrow_AddToEdit(&enemy->burrow, edit));
	edit->addItem_bool("cannotOpenDoors", &enemy->cannotOpenDoors);
	lwlib_Edit_AddGroup(edit, "spawnFlags", 
		ED_EnemySpawnFlags_AddToEdit(&enemy->spawnFlags, edit));
	lwlib_Edit_AddGroup(edit, "surpriseAttack", 
		ED_EnemySurpriseAttack_AddToEdit(&enemy->surpriseAttack, edit));
	lwlib_Edit_AddGroup(edit, "shootChanceConfig", 
		ED_EnemyChanceConfig_AddToEdit(&enemy->shootChanceConfig, edit));
	edit->addItem_int("runAwayDist", &enemy->runAwayDist);
	edit->addItem_int("altShootChance", &enemy->altShootChance);
	edit->addItem_string("altShootState", enemy->altShootState,
		sizeof(enemy->altShootState));
	edit->addItem_string("projName", enemy->projName,
		sizeof(enemy->projName));
	edit->addItem_int("projSpeed", &enemy->projSpeed);
	edit->addItem_int("meleeDamageBase", &enemy->meleeDamageBase);
	edit->addItem_int("meleeDamageDivisor", &enemy->meleeDamageDivisor);
	edit->addItem_int("meleeDamageChance", &enemy->meleeDamageChance);
	edit->addItem_int("defenseChance", &enemy->defenseChance);
	edit->addItem_string("defenseState", enemy->defenseState,
		sizeof(enemy->defenseState));
	lwlib_Edit_AddVec(edit, "goalEnemyNames", enemy->goalEnemyNames, 
		NULL, NULL, ED_Str_AddToEdit);
	edit->addItem_string("reachedGoalThink", enemy->reachedGoalThink,
		sizeof(enemy->reachedGoalThink));
	edit->addItem_bool("lightWeight", &enemy->lightWeight);
	edit->addItem_bool("naziFriends", &enemy->naziFriends);
	edit->addItem_bool("meleeHurtsMecha", &enemy->meleeHurtsMecha);
	lwlib_Edit_AddVec(edit, "features", enemy->features,
		NULL, NULL, ED_Feature_AddToEdit);
}

static void ED_ChanceRange_AddToEdit(void *elem, lwlib::Edit *edit)
{
	ED_ChanceRange_t *chanceRange;

	chanceRange = (ED_ChanceRange_t *)elem;
	edit->addItem_int("startDist", &chanceRange->startDist);
	edit->addItem_int("endDist", &chanceRange->endDist);
	edit->addItem_int("startChance", &chanceRange->startChance);
	edit->addItem_int("endChance", &chanceRange->endChance);
}

static void ED_EnemyConfigCheck_AddToEdit(ED_EnemyConfigCheck_t *check, lwlib::Edit *edit)
{
	edit->addItem_int("timeout", &check->timeout);
	edit->addItem_int("delayTime", &check->delayTime);
	lwlib_Edit_AddVec(edit, "chanceRanges", check->chanceRanges,
		NULL, NULL, ED_ChanceRange_AddToEdit);
}

static void ED_EnemyConfig_AddToEdit(ED_EnemyConfig_t *enemyConfig, lwlib::Edit *edit)
{
	lwlib_Edit_AddGroup(edit, "wakeCheck", 
		ED_EnemyConfigCheck_AddToEdit(&enemyConfig->wakeCheck, edit));
	lwlib_Edit_AddGroup(edit, "escapeCheck", 
		ED_EnemyConfigCheck_AddToEdit(&enemyConfig->escapeCheck, edit));
}

static void ED_RandomConst_AddToEdit(ED_RandomConst_t *randomConst, 
	lwlib::Edit *edit)
{
	edit->addItem_double("min", &randomConst->min);
	edit->addItem_double("max", &randomConst->max);
}

static void ED_RandomInt_AddToEdit(ED_RandomInt_t *randomInt, 
	lwlib::Edit *edit)
{
	edit->addItem_int("min", &randomInt->min);
	edit->addItem_int("max", &randomInt->max);
}

static void ED_ProjPhys_AddToEdit(void *elem, lwlib::Edit *edit)
{
	ED_ProjPhys_t *projPhys;

	projPhys = (ED_ProjPhys_t *)elem;

	edit->addItem_string("name", projPhys->name, 
		sizeof(projPhys->name));
	lwlib_Edit_AddGroup(edit, "speed", 
		ED_RandomConst_AddToEdit(&projPhys->speed, edit));
	lwlib_Edit_AddGroup(edit, "pitch", 
		ED_RandomConst_AddToEdit(&projPhys->pitch, edit));
	lwlib_Edit_AddGroup(edit, "angleOffset", 
		ED_RandomInt_AddToEdit(&projPhys->angleOffset, edit));
	edit->addItem_double("gravity", &projPhys->gravity);
	edit->addItem_double("bounceDamping", &projPhys->bounceDamping);
	edit->addItem_int("bounceLimit", &projPhys->bounceLimit);
	edit->addItem_string("bounceSound", projPhys->bounceSound, 
		sizeof(projPhys->bounceSound));
	edit->addItem_double("explodeTime", &projPhys->explodeTime);
}

static void ED_Proj_AddToEdit(void *elem, lwlib::Edit *edit)
{
	ED_Proj_t *proj;

	proj = (ED_Proj_t *)elem;

	edit->addItem_string("name", proj->name, sizeof(proj->name));
	edit->addItem_string("startState", proj->startState,
		sizeof(proj->startState));
	edit->addItem_enum("damageType", (int *)&proj->damageType,
		ED_DamageType_Str, NULL, lwlib_CountArray(ED_DamageType_Str));
	edit->addItem_int("baseDamage", &proj->baseDamage);
	edit->addItem_int("rndDamageMultiplier", &proj->rndDamageMultiplier);
	edit->addItem_int("rndDamageDivisor", &proj->rndDamageDivisor);
	edit->addItem_string("explState", proj->explState, 
		sizeof(proj->explState));
	edit->addItem_bool("playHitSound", &proj->playHitSound);
	edit->addItem_string("shootSound", proj->shootSound, 
		sizeof(proj->shootSound));
	edit->addItem_string("hitSound", proj->hitSound, 
		sizeof(proj->hitSound));
	edit->addItem_bool("hitsPlayer", &proj->hitsPlayer);
	edit->addItem_bool("hitsEnemies", &proj->hitsEnemies);
	edit->addItem_string("smokeState", proj->smokeState, 
		sizeof(proj->smokeState));
	lwlib_Edit_AddGroup(edit, "spriteOffset", 
		ED_SpriteOffset_AddToEdit(&proj->spriteOffset, edit));
	edit->addItem_string("physName", proj->physName, 
		sizeof(proj->physName));
	edit->addItem_bool("destroysWalls", &proj->destroysWalls);
}

static void ED_Inject_AddToEdit(ED_Inject_t *inject, lwlib::Edit *edit)
{
	edit->addItem_string("actorName", 
		inject->actorName, sizeof(inject->actorName));
	edit->addItem_int("tilex", &inject->tilex);
	edit->addItem_int("tiley", &inject->tiley);
	edit->addItem_bool("enabled", &inject->enabled);
}

static void ED_Sound_AddToEdit(void *elem, lwlib::Edit *edit)
{
	ED_Sound_t *sound;

	sound = (ED_Sound_t *)elem;

	edit->addItem_string("name", 
		sound->name, sizeof(sound->name));
	edit->addItem_string("soundFileName", 
		sound->soundFileName, sizeof(sound->soundFileName));
	edit->addItem_int("maxTileDist", &sound->maxTileDist);
}

static void ED_AmmoBonus_AddToEdit(ED_AmmoBonus_t *ammoBonus,
	lwlib::Edit *edit)
{
	edit->addItem_enum("ammoType",
		(int *)&ammoBonus->ammoType, ED_AmmoType_Str, NULL,
		lwlib_CountArray(ED_AmmoType_Str));
	edit->addItem_int("ammo", 
		&ammoBonus->ammo);
}

static void ED_Bonus_AddToEdit(void *elem, lwlib::Edit *edit)
{
	int i;
	ED_Bonus_t *bonus;
	lwlib_String_t tmpStr;

	bonus = (ED_Bonus_t *)elem;

	edit->addItem_string("name", 
		bonus->name, sizeof(bonus->name));
	edit->addItem_int("healthPoints", 
		&bonus->healthPoints);
	edit->addItem_bool("isKey", &bonus->isKey);
	edit->addItem_int("key", &bonus->key);
	edit->addItem_string("message", 
		bonus->message, sizeof(bonus->message));
	edit->addItem_string("sound", 
		bonus->sound, sizeof(bonus->sound));
	edit->addItem_int("scorePoints", 
		&bonus->scorePoints);
	edit->addItem_bool("isTreasure",
		&bonus->isTreasure);
	for (i = 0; (size_t)i < lwlib_CountArray(bonus->ammoBonuses); i++)
	{
		edit->addItem_group(
			lwlib_VaString(tmpStr, "ammoBonus[%d]", i));
		ED_AmmoBonus_AddToEdit(&bonus->ammoBonuses[i], edit);
		edit->navigateBack();
	}
	edit->addItem_enum("weapon",
		(int *)&bonus->weapon, ED_WeaponType_Str, ED_WeaponType_Vals,
		lwlib_CountArray(ED_WeaponType_Str));
	edit->addItem_int("armorPoints", 
		&bonus->armorPoints);
	edit->addItem_bool("armorJumpTo",
		&bonus->armorJumpTo);
	edit->addItem_bool("regenHealth",
		&bonus->regenHealth);
	edit->addItem_bool("regenArmor",
		&bonus->regenArmor);
	edit->addItem_int("lives", &bonus->lives);
	edit->addItem_bool("usedSprite",
		&bonus->usedSprite);
	edit->addItem_string("usedSpriteDef", 
		bonus->usedSpriteDef, sizeof(bonus->usedSpriteDef));
	edit->addItem_int("damagePoints",
		&bonus->damagePoints);
	edit->addItem_bool("usePickup",
		&bonus->usePickup);
	edit->addItem_string("logChoice", 
		bonus->logChoice, sizeof(bonus->logChoice));
	edit->addItem_int("medkits", &bonus->medkits);
	edit->addItem_bool("increaseMedkitLimit",
		&bonus->increaseMedkitLimit);
	edit->addItem_bool("increaseAmmoLimit",
		&bonus->increaseAmmoLimit);
	edit->addItem_bool("giveBackpackAmmo",
		&bonus->giveBackpackAmmo);
}

static void ED_FloorTile_AddToEdit(void *elem, lwlib::Edit *edit)
{
	ED_FloorTile_t *floorTile;
	floorTile = (ED_FloorTile_t *)elem;
	edit->addItem_int("floorTile", floorTile);
}

static void ED_FootStep_AddToEdit(void *elem, lwlib::Edit *edit)
{
	ED_FootStep_t *footStep;

	footStep = (ED_FootStep_t *)elem;

	lwlib_Edit_AddVec(edit, "floorTiles", footStep->floorTiles, 
		NULL, NULL, ED_FloorTile_AddToEdit);
	edit->addItem_string("soundDef", footStep->soundDef, 
		sizeof(footStep->soundDef));
}

static void ED_AmbientActor_AddToEdit(void *elem, lwlib::Edit *edit)
{
	ED_AmbientActor_t *ambientActor;
	ambientActor = (ED_AmbientActor_t *)elem;
	edit->addItem_int("actor", ambientActor);
}

static void ED_Ambient_AddToEdit(void *elem, lwlib::Edit *edit)
{
	ED_Ambient_t *ambient;

	ambient = (ED_Ambient_t *)elem;

	edit->addItem_bool("looped", &ambient->looped);
	edit->addItem_string("soundDef", ambient->soundDef, 
		sizeof(ambient->soundDef));
	edit->addItem_int("actor", &ambient->actor);
	edit->addItem_int("minTimePeriod", &ambient->minTimePeriod);
	edit->addItem_int("maxTimePeriod", &ambient->maxTimePeriod);
	edit->addItem_int("fadeInTime", &ambient->fadeInTime);
	edit->addItem_int("fadeOutTime", &ambient->fadeOutTime);
	edit->addItem_bool("distanceScaled", &ambient->distanceScaled);
	edit->addItem_enum("type",
		(int *)&ambient->type, ED_AmbientType_Str, ED_AmbientType_Vals,
		lwlib_CountArray(ED_AmbientType_Str));
	lwlib_Edit_AddVec(edit, "actors", ambient->actors, 
		NULL, NULL, ED_AmbientActor_AddToEdit);
}

static void ED_Ricochet_AddToEdit(void *elem, lwlib::Edit *edit)
{
	ED_Ricochet_t *ricochet;

	ricochet = (ED_Ricochet_t *)elem;

	edit->addItem_string("name", ricochet->name, 
		sizeof(ricochet->name));
	edit->addItem_string("state", ricochet->state, 
		sizeof(ricochet->state));
	edit->addItem_string("soundDef", ricochet->soundDef, 
		sizeof(ricochet->soundDef));
}

static void ED_WallAttr_AddToEdit(void *elem, lwlib::Edit *edit)
{
	ED_WallAttr_t *wallAttr;

	wallAttr = (ED_WallAttr_t *)elem;

	edit->addItem_string("ricochetName", wallAttr->ricochetName, 
		sizeof(wallAttr->ricochetName));
	edit->addItem_bool("breakableWall", &wallAttr->breakableWall);
}

static void ED_AnimatedWallTile_AddToEdit(void *elem, lwlib::Edit *edit)
{
	ED_AnimatedWallTile_t *animatedWallTile;

	animatedWallTile = (ED_AnimatedWallTile_t *)elem;

	edit->addItem_int("animStartTile", 
		&animatedWallTile->animStartTile);
	edit->addItem_int("animNumTiles",
		&animatedWallTile->animNumTiles);
	edit->addItem_int("animTics",
		&animatedWallTile->animTics);
}

static void ED_VSWAPInfo_AddToEdit(ED_VSWAPInfo_t *vswapInfo,
	lwlib::Edit *edit)
{
	edit->addItem_int("skyStart", &vswapInfo->skyStart);
	edit->addItem_int("wallPatchStart", 
		&vswapInfo->wallPatchStart);
	edit->addItem_int("glowPatchStart", 
		&vswapInfo->glowPatchStart);
	edit->addItem_int("glowPatchEnd", 
		&vswapInfo->glowPatchEnd);
	lwlib_Edit_AddVec(edit, "animatedWallTiles", 
		vswapInfo->animatedWallTiles, 
		NULL, NULL, ED_AnimatedWallTile_AddToEdit);
}

static void ED_TileInfo_AddToEdit(ED_TileInfo_t *tileInfo,
	lwlib::Edit *edit)
{
	edit->addItem_int("firstWallPatchActor", 
		&tileInfo->firstWallPatchActor);
	edit->addItem_int("lastWallPatchActor", 
		&tileInfo->lastWallPatchActor);
	lwlib_Edit_AddVec(edit, "burrowFloorTiles", tileInfo->burrowFloorTiles, 
		NULL, NULL, ED_FloorTile_AddToEdit);
	edit->addItem_int("firstTruckTile", &tileInfo->firstTruckTile);
	edit->addItem_int("lastTruckTile", &tileInfo->lastTruckTile);
	edit->addItem_int("woodenDoorStartTile", &tileInfo->woodenDoorStartTile);
	edit->addItem_int("waterFloorTile", &tileInfo->waterFloorTile);
}

static void ED_AmbientSettings_AddToEdit(ED_AmbientSettings_t *ambientSettings,
	lwlib::Edit *edit)
{
	edit->addItem_int("radius", &ambientSettings->radius);
	edit->addItem_double("scale", &ambientSettings->scale);
	edit->addItem_double("hiddenScale", &ambientSettings->hiddenScale);
	edit->addItem_double("convPasses", &ambientSettings->convPasses);
	edit->addItem_double("doorVisFactor", &ambientSettings->doorVisFactor);
}

static void ED_AddToEdit(ED_t ed, lwlib::Edit *edit)
{
	int i;
	const char *toolName;
	lwlib::Tool *tool;
	lwlib_GetPrivAssert(ED_Priv_t, ed);

	edit->navigateTop();
	lwlib_Edit_AddVec(edit, "sprites", priv->sprites, 
		NULL, NULL, ED_Sprite_AddToEdit);
	edit->addPersistGroup(&priv->persist);
	lwlib_Edit_AddVec(edit, "statInfos", priv->statInfos, 
		NULL, NULL, ED_StatInfo_AddToEdit);
	lwlib_Edit_AddVec(edit, "states", priv->states,
		NULL, NULL, ED_State_AddToEdit);
	lwlib_Edit_AddVec(edit, "spawns", priv->spawns,
		NULL, NULL, ED_Spawn_AddToEdit);
	lwlib_Edit_AddVec(edit, "enemies", priv->enemies,
		NULL, NULL, ED_Enemy_AddToEdit);
	edit->addItem_enum("activeTool", (int *)&priv->activeTool,
		ED_ToolId_Str, NULL, lwlib_CountArray(ED_ToolId_Str));
	for (i = 0; (size_t)i < vec_size(priv->toolPtrs); i++)
	{
		tool = vec_at(priv->toolPtrs, i);
		toolName = ED_ToolId_Str[i];

		lwlib_Edit_AddGroup(edit, toolName, 
			tool->addToEdit(edit));
	}
	lwlib_Edit_AddGroup(edit, "enemyConfig", 
		ED_EnemyConfig_AddToEdit(&priv->enemyConfig, edit));
	lwlib_Edit_AddVec(edit, "projs", priv->projs,
		NULL, NULL, ED_Proj_AddToEdit);
	lwlib_Edit_AddVec(edit, "projPhysSeq", priv->projPhysSeq,
		NULL, NULL, ED_ProjPhys_AddToEdit);
	lwlib_Edit_AddGroup(edit, "inject", 
		ED_Inject_AddToEdit(&priv->inject, edit));
	lwlib_Edit_AddGroup(edit, "mouse", 
		priv->mouse->addToEdit(edit));
	lwlib_Edit_AddVec(edit, "sounds", priv->sounds, 
		NULL, NULL, ED_Sound_AddToEdit);
	lwlib_Edit_AddVec(edit, "randomSounds", priv->randomSounds, 
		NULL, NULL, ED_RandomSound_AddToEdit);
	lwlib_Edit_AddVec(edit, "bonuses", priv->bonuses, 
		NULL, NULL, ED_Bonus_AddToEdit);
	lwlib_Edit_AddVec(edit, "footSteps", priv->footSteps, 
		NULL, NULL, ED_FootStep_AddToEdit);
	
	lwlib_Edit_AddVec(edit, "chanceRanges", priv->chanceRanges, 
		NULL, NULL, ED_ChanceRange_AddToEdit);
	lwlib_Edit_AddGroup(edit, "vswapInfo", 
		ED_VSWAPInfo_AddToEdit(&priv->vswapInfo, edit));
	lwlib_Edit_AddVec(edit, "ambients", priv->ambients, 
		NULL, NULL, ED_Ambient_AddToEdit);
	lwlib_Edit_AddGroup(edit, "tileInfo", 
		ED_TileInfo_AddToEdit(&priv->tileInfo, edit));
	lwlib_Edit_AddVec(edit, "ricochets", priv->ricochets, 
		NULL, NULL, ED_Ricochet_AddToEdit);
	lwlib_Edit_AddVec(edit, "wallAttrs", priv->wallAttrs, 
		NULL, NULL, ED_WallAttr_AddToEdit);
	edit->addItem_int("lrPack", &priv->lrPack);
	lwlib_Edit_AddGroup(edit, "ambientSettings", 
		ED_AmbientSettings_AddToEdit(&priv->ambientSettings, edit));
}

static void ED_RefreshEditItems(ED_t ed)
{
	lwlib_GetPrivAssert(ED_Priv_t, ed);

	priv->edit->clearItems();
	ED_AddToEdit(ed, priv->edit);
}

static void ED_Load(ED_t ed)
{
	lwlib_GetPrivAssert(ED_Priv_t, ed);

	lwlib_StrCopy(priv->persist.fileName, ED_STATEFILE);
	priv->persist.fileMode = lwlib::Fs_FileMode::ReadText;
	priv->persist.commit = true;

	lwlib_Edit_CheckPersist(priv->persist, 
		ED_AddToEdit(ed, fsEdit));
}

static void ED_UpdateEdit(ED_t ed, lwlib::Edit *edit)
{
	int i;
	int flg;
	const char *toolName;
	lwlib::Tool *tool;
	lwlib_GetPrivAssert(ED_Priv_t, ed);

	for (i = 0; (size_t)i < vec_size(priv->toolPtrs); i++)
	{
		tool = vec_at(priv->toolPtrs, i);
		toolName = ED_ToolId_Str[i];

		edit->navigateItem(toolName);
		tool->updateEdit(edit);
		edit->navigateBack();

		flg = edit->getItemFlags(toolName);
		lwlib_SetBitLValue(
			flg, lwlib::Edit_ItemFlag::Disabled, 
			(priv->activeTool != i)
			);
		edit->setItemFlags(toolName, flg);
	}

	edit->updateFrame();

	lwlib_Edit_CheckPersist(priv->persist, 
		ED_AddToEdit(ed, fsEdit));
}

static void ED_UpdateFrame(ED_t ed)
{
	int i;
	lwlib::Tool *tool;
	lwlib_GetPrivAssert(ED_Priv_t, ed);

	ED_UpdateEdit(ed, priv->edit);

	for (i = 0; (size_t)i < vec_size(priv->toolPtrs); i++)
	{
		tool = vec_at(priv->toolPtrs, i);
		if (priv->activeTool == i)
		{
			tool->updateFrame();
			break;
		}
	}

	if (lwlib_Keys_CheckEvent(priv->keys, "esc", lwlib_Keys_Event(Pressed)))
	{
		priv->quit = true;
	}

	priv->mouse->pollState();
	priv->mouse->updateFrame();

	if (priv->mapTool->pickerPending())
	{
		priv->activeTool = ED_ToolId_Picker;
	}

	if (priv->picker->mapToolPending())
	{
		priv->activeTool = ED_ToolId_MapTool;
	}
}

static int ED_PickerDelegateGetDoorTile(void *obj, int lock)
{
	ED_Priv_t *priv = (ED_Priv_t *)obj;
	return priv->mapTool->getPickerDoorTile(lock);
}

ED_t ED_New(void)
{
	int i, j, k;
	int count;
	ED_Sprite_t *sprite;
	ED_StatInfo_t *statInfo;
	ED_t ed;
	lwlib::TileViewer *tileViewer;
	lwlib::PicViewer *picViewer;
	lwlib::ModelMaker *modelMaker;
	ED_StateHandler_t *stateHandler;
	ED_State_t *state;
	ED_Spawn_t *spawn;
	ED_Enemy_t *enemy;
	ED_Proj_t *proj;
	ED_ProjPhys_t *projPhys;
	ED_Sound_t *sound;
	ED_RandomSound_t *randomSound;
	ED_Bonus_t *bonus;
	int floorTile;
	ED_FootStep_t *footStep;
	ED_SpriteRepl_t *spriteRepl;
	ED_AnimatedWallTile_t *animatedWallTile;
	ED_Ambient_t *ambient;
	lwlib::Picker_DelegateAPI_t pickerDelegateAPI;
	ED_Ricochet_t *ricochet;
	ED_WallAttr_t *wallAttr;
	int ricochetNum;
	ED_Feature_t *feature;
	int bossIndex;
	ED_Boss_t *boss;

	lwlib_Zero(ed);
	ed.priv = lwlib_CallocSingle(ED_Priv_t);

	lwlib_GetPriv(ED_Priv_t, ed);

#ifdef SDL_EDITOR
	const bool sdlEmu = true;
#else
	const bool sdlEmu = false;
#endif
	priv->gl = new lwlib::Gl(sdlEmu);

	priv->keys = lwlib_Keys_New();
	priv->colorRemap = new lwlib::ColorRemap();

	priv->palette = gamepal;

	priv->font = new lwlib::Font(ED_FONTPATH, priv->palette);
	priv->font->setColorRemap(priv->colorRemap);
	priv->font->setGl(priv->gl);

	priv->edit = new lwlib::Edit();
	priv->edit->setKeys(priv->keys);
	priv->edit->setFont(priv->font);

	priv->mouse = new lwlib::Mouse(
		lwlib::vec2i(screenWidth, screenHeight));
	priv->mouse->setEdit(priv->edit);

	tileViewer = new lwlib::TileViewer();
	tileViewer->setGl(priv->gl);
	tileViewer->setKeys(priv->keys);
	tileViewer->setFont(priv->font);
	tileViewer->setPalette(priv->palette);
	tileViewer->setEdit(priv->edit);

	modelMaker = new lwlib::ModelMaker();

	picViewer = new lwlib::PicViewer();
	picViewer->setGl(priv->gl);
	picViewer->setKeys(priv->keys);
	picViewer->setFont(priv->font);
	picViewer->setPalette(priv->palette);
	picViewer->setEdit(priv->edit);
	priv->picViewer = picViewer;

	{
		lwlib::Picker_DelegateAPI_t *delegateAPI = &pickerDelegateAPI;
		lwlib_InitDelegate(getDoorTile, priv, 
			ED_PickerDelegateGetDoorTile);
	}

	priv->picker = new lwlib::Picker(
		priv->keys,
		priv->gl,
		priv->font,
		priv->edit,
		tileViewer,
		priv->mouse,
		priv->palette,
		pickerDelegateAPI
		);

	priv->mapTool = new lwlib::MapTool(
		priv->gl,
		tileViewer,
		priv->mouse,
		priv->keys,
		priv->edit,
		priv->font,
		priv->picker
		);

	// create generic vector of tools
	vec_push(priv->toolPtrs, lwlib::ToolPtr_t, priv->mapTool);
	vec_push(priv->toolPtrs, lwlib::ToolPtr_t, tileViewer);
	vec_push(priv->toolPtrs, lwlib::ToolPtr_t, priv->picViewer);
	vec_push(priv->toolPtrs, lwlib::ToolPtr_t, priv->picker);
	vec_push(priv->toolPtrs, lwlib::ToolPtr_t, modelMaker);

	ED_Load(ed);
	ED_RefreshEditItems(ed);

	for (i = 0; (size_t)i < vec_size(priv->sprites); i++)
	{
		sprite = &vec_at(priv->sprites, i);
		lwlib_StrMapInsert(&priv->spriteMap, sprite->name, (lwlib_Word_t)i);
	}

	for (i = 0; (size_t)i < vec_size(priv->statInfos); i++)
	{
		statInfo = &vec_at(priv->statInfos, i);
		if (statInfo->actor != 0)
		{
			lwlib_IntMapInsert(&priv->statInfoMap, statInfo->actor, (lwlib_Word_t)i);
		}
	}

	priv->stateHandlers = ED_StateHandlers;
	priv->numStateHandlers = lwlib_CountArray(ED_StateHandlers);

	for (i = 0; i < priv->numStateHandlers; i++)
	{
		stateHandler = &priv->stateHandlers[i];
		lwlib_StrMapInsert(&priv->stateHandlerMap, stateHandler->name, (lwlib_Word_t)i);
	}

	for (i = 0; (size_t)i < vec_size(priv->states); i++)
	{
		state = &vec_at(priv->states, i);
		lwlib_StrMapInsert(&priv->stateMap, state->name, (lwlib_Word_t)i);

		for (j = 0; (size_t)j < vec_size(state->features); j++)
		{
			feature = &vec_at(state->features, j);
			switch (feature->type)
			{
			case feat_resurrection:
			case feat_watermut:
			case feat_boss:
			case feat_music:
			case feat_damagedebris:
			case feat_skillprofile:
			case feat_melee:
			case feat_summoner:
			case feat_guarded:
				break;
			case feat_playvoice:
				state->playVoice = &feature->u.playVoice;
				break;
			}
		}
	}

	for (i = 0; (size_t)i < vec_size(priv->spawns); i++)
	{
		spawn = &vec_at(priv->spawns, i);
		lwlib_IntMapInsert(&priv->spawnMap, spawn->actor, (lwlib_Word_t)i);
	}

	bossIndex = 0;

	for (i = 0; (size_t)i < vec_size(priv->enemies); i++)
	{
		enemy = &vec_at(priv->enemies, i);
		lwlib_StrMapInsert(&priv->enemyMap, enemy->name,
			(lwlib_Word_t)i);

		for (j = 0; j < enemy->shootChanceConfig.numRanges; j++)
		{
			vec_push(
				enemy->shootChanceConfig.ranges, ED_ChanceRange_t, 
				vec_at(
					priv->chanceRanges,
					enemy->shootChanceConfig.firstRange + j
					)
				);
		}

		for (j = 0; (size_t)j < vec_size(enemy->spriteRepls); j++)
		{
			spriteRepl = &vec_at(enemy->spriteRepls, j);

			if (ED_SPR(spriteRepl->lastOldSprite) == -1)
			{
				count = 1;
			}
			else
			{
				count = ED_SPR(spriteRepl->lastOldSprite) -
					ED_SPR(spriteRepl->oldSprite) + 1;
			}

			for (k = 0; k < count; k++)
			{
				lwlib_IntMapInsert(&enemy->spriteReplMap,
					ED_SPR(spriteRepl->oldSprite) + k, 
					ED_SPR(spriteRepl->newSprite) + k);
			}
		}

		for (j = 0; (size_t)j < vec_size(enemy->features); j++)
		{
			feature = &vec_at(enemy->features, j);
			switch (feature->type)
			{
			case feat_resurrection:
				enemy->resurrection = &feature->u.resurrection;
				break;
			case feat_watermut:
				enemy->waterMut = &feature->u.waterMut;
				break;
			case feat_boss:
				boss = enemy->boss = &feature->u.boss;
				boss->index = bossIndex++;
				break;
			case feat_damagedebris:
				enemy->damageDebris = &feature->u.damageDebris;
				break;
			case feat_skillprofile:
				if (enemy->numSkillProfiles < ED_MAX_SKILLPROFILES)
				{
					enemy->skillProfiles[enemy->numSkillProfiles] =
						&feature->u.skillProfile;
					enemy->numSkillProfiles++;
				}
				break;
			case feat_playvoice:
				break;
			case feat_melee:
				enemy->melee = &feature->u.melee;
				break;
			case feat_summoner:
				enemy->summoner = &feature->u.summoner;
				break;
			case feat_guarded:
				enemy->guarded = &feature->u.guarded;
				break;
			case feat_music:
				enemy->music = &feature->u.music;
				break;
			}
		}
	}

	for (i = 0; (size_t)i < vec_size(priv->projs); i++)
	{
		proj = &vec_at(priv->projs, i);
		lwlib_StrMapInsert(&priv->projMap, proj->name, (lwlib_Word_t)i);
		if (strcmp(proj->name, "mechmut4noarms") == 0)
		{
			proj->hitsScenery = true;
		}
	}

	for (i = 0; (size_t)i < vec_size(priv->projPhysSeq); i++)
	{
		projPhys = &vec_at(priv->projPhysSeq, i);
		lwlib_StrMapInsert(&priv->projPhysMap, projPhys->name, 
			(lwlib_Word_t)i);
	}

	for (i = 0; (size_t)i < vec_size(priv->sounds); i++)
	{
		sound = &vec_at(priv->sounds, i);
		lwlib_StrMapInsert(&priv->soundMap, sound->name, (lwlib_Word_t)i);
	}

	for (i = 0; (size_t)i < vec_size(priv->randomSounds); i++)
	{
		randomSound = &vec_at(priv->randomSounds, i);
		lwlib_StrMapInsert(&priv->randomSoundMap, randomSound->name, 
			(lwlib_Word_t)i);
	}

	for (i = 0; (size_t)i < vec_size(priv->bonuses); i++)
	{
		bonus = &vec_at(priv->bonuses, i);
		lwlib_StrMapInsert(&priv->bonusMap, bonus->name, (lwlib_Word_t)i);
	}

	for (i = 0; (size_t)i < vec_size(priv->footSteps); i++)
	{
		footStep = &vec_at(priv->footSteps, i);
		for (j = 0; (size_t)j < vec_size(footStep->floorTiles); j++)
		{
			floorTile = vec_at(footStep->floorTiles, j);
			lwlib_IntMapInsert(&priv->footStepMap, floorTile,
				(lwlib_Word_t)i);
		}
	}

	for (i = 0; (size_t)i < vec_size(priv->vswapInfo.animatedWallTiles); i++)
	{
		animatedWallTile = &vec_at(priv->vswapInfo.animatedWallTiles, i);
		priv->vswapInfo.animatedWallTileMap[
			animatedWallTile->animStartTile] = i + 1;
	}

	for (i = 0; (size_t)i < vec_size(priv->ambients); i++)
	{
		ambient = &vec_at(priv->ambients, i);
		lwlib_IntMapInsert(&priv->ambientMap, ambient->actor,
			(lwlib_Word_t)i);
		lwlib_StrMapInsert(&priv->ambientStrMap, ambient->soundDef,
			(lwlib_Word_t)i);
	}

	for (i = 0; (size_t)i < vec_size(priv->ricochets); i++)
	{
		ricochet = &vec_at(priv->ricochets, i);
		lwlib_StrMapInsert(&priv->ricochetMap, ricochet->name,
			(lwlib_Word_t)i);
	}

	for (i = 0; (size_t)i < vec_size(priv->wallAttrs); i++)
	{
		wallAttr = &vec_at(priv->wallAttrs, i);
		ricochetNum = (int)lwlib_StrMapElem(&priv->ricochetMap, 
			wallAttr->ricochetName);
		if (ricochetNum >= 0 && (size_t)ricochetNum < vec_size(priv->ricochets))
		{
			wallAttr->ricochet = &vec_at(priv->ricochets, ricochetNum);
		}
	}

	if (priv->lrPack > MAX_LRPACK)
	{
		Quit("LRPack %d is larger than maximum of %d!\n", priv->lrPack,
			MAX_LRPACK);
	}

	priv->ambientSettings.radius = 
		lwlib_MIN(priv->ambientSettings.radius, 
		ED_VISFACTOR_MATRIX_SIZE / 2);

	return ed;
}

ED_t ED_Free(ED_t ed)
{
	int i;
	lwlib::Tool *tool;
	lwlib_GetPriv(ED_Priv_t, ed);
	ED_Enemy_t *enemy;
	ED_FootStep_t *footStep;
	ED_Ambient_t *ambient;

	if (priv)
	{
		vec_free(priv->ricochets, ED_Ricochet_t);
		lwlib_StrMapClear(&priv->ricochetMap);
		vec_free(priv->wallAttrs, ED_WallAttr_t);
		for (i = 0; (size_t)i < vec_size(priv->ambients); i++)
		{
			ambient = &vec_at(priv->ambients, i);
			vec_free(ambient->actors, ED_AmbientActor_t);
		}
		vec_free(priv->ambients, ED_Ambient_t);
		lwlib_IntMapClear(&priv->ambientMap);
		lwlib_StrMapClear(&priv->ambientStrMap);
		vec_free(priv->vswapInfo.animatedWallTiles, ED_AnimatedWallTile_t);
		vec_free(priv->chanceRanges, ED_ChanceRange_t);
		lwlib_IntMapClear(&priv->footStepMap);
		for (i = 0; (size_t)i < vec_size(priv->footSteps); i++)
		{
			footStep = &vec_at(priv->footSteps, i);
			vec_free(footStep->floorTiles, ED_FloorTile_t);
		}
		vec_free(priv->footSteps, ED_FootStep_t);
		lwlib_StrMapClear(&priv->bonusMap);
		vec_free(priv->bonuses, ED_Bonus_t);
		lwlib_StrMapClear(&priv->soundMap);
		vec_free(priv->sounds, ED_Sound_t);
		lwlib_StrMapClear(&priv->randomSoundMap);
		vec_free(priv->randomSounds, ED_RandomSound_t);
		lwlib_StrMapClear(&priv->projMap);
		vec_free(priv->projs, ED_Proj_t);
		lwlib_StrMapClear(&priv->projPhysMap);
		vec_free(priv->projPhysSeq, ED_ProjPhys_t);
		lwlib_StrMapClear(&priv->enemyMap);
		for (i = 0; (size_t)i < vec_size(priv->enemies); i++)
		{
			enemy = &vec_at(priv->enemies, i);
			vec_free(enemy->spriteRepls, ED_SpriteRepl_t);
			lwlib_IntMapClear(&enemy->spriteReplMap);
			vec_free(enemy->goalEnemyNames, lwlib_Str_t);
			vec_free(enemy->features, ED_Feature_t);
		}
		vec_free(priv->enemies, ED_Enemy_t);
		lwlib_IntMapClear(&priv->spawnMap);
		vec_free(priv->spawns, ED_Spawn_t);
		lwlib_StrMapClear(&priv->stateMap);
		for (i = 0; (size_t)i < vec_size(priv->states); i++)
		{
			vec_free(vec_at(priv->states, i).features, ED_Feature_t);
		}
		vec_free(priv->states, ED_State_t);
		lwlib_StrMapClear(&priv->stateHandlerMap);
		lwlib_IntMapClear(&priv->statInfoMap);
		vec_free(priv->statInfos, ED_StatInfo_t);
		lwlib_StrMapClear(&priv->spriteMap);
		vec_free(priv->sprites, ED_Sprite_t);
		for (i = 0; (size_t)i < vec_size(priv->toolPtrs); i++)
		{
			tool = vec_at(priv->toolPtrs, i);
			delete tool;
		}
		vec_free(priv->toolPtrs, lwlib::ToolPtr_t);
		delete priv->edit;
		delete priv->font;
		delete priv->colorRemap;
		delete priv->mouse;
		lwlib_Keys_Free(priv->keys);
		delete priv->signonImg;
		delete priv->gl;
		free(priv);
	}

	lwlib_Zero(ed);
	return ed;
}

static void ED_SendKeyEvents(ED_t ed)
{
	SDL_Event event;
	lwlib_GetPrivAssert(ED_Priv_t, ed);

	lwlib_Keys_ResetTransientEvents(priv->keys);

	while (SDL_PollEvent(&event))
	{
		switch (event.type)
		{
		case SDL_KEYDOWN:
		case SDL_KEYUP:
			lwlib_Keys_ProcessEvent(priv->keys, 
				event.key.keysym.sym, 
				(event.key.state == SDL_PRESSED));
			break;
		case SDL_ACTIVEEVENT:
			if (event.active.state & SDL_APPACTIVE)
			{
				if (event.active.gain)
				{
					priv->sdlApp.visible = true;
				}
				else
				{
					priv->sdlApp.visible = false;
				}
			}
			if (event.active.state & SDL_APPMOUSEFOCUS)
			{
				if (event.active.gain)
				{
					priv->sdlApp.mouseFocus = true;
					priv->mouse->setActiveFlagState(
						lwlib::Mouse_ActiveFlag::AppFocus, true);
				}
				else
				{
					priv->sdlApp.mouseFocus = false;
					priv->mouse->setActiveFlagState(
						lwlib::Mouse_ActiveFlag::AppFocus, false);
				}
			}
			if (event.active.state & SDL_APPINPUTFOCUS)
			{
				if (event.active.gain)
				{
					priv->sdlApp.keyboardFocus = true;
					lwlib_Keys_KeyboardFocusReturned(priv->keys);
				}
				else
				{
					priv->sdlApp.keyboardFocus = false;
				}
			}
			break;
		}
	}

	lwlib_Keys_ProcessEvent(priv->keys, SDLK_UNKNOWN, false);
}

static void ED_Draw(ED_t ed)
{
	lwlib::Tool *tool;
	lwlib_GetPrivAssert(ED_Priv_t, ed);

	priv->gl->lockScreen();

	priv->signonImg->drawScaled(priv->gl, 0, 0, 
		screenWidth, screenHeight, false);

	if (priv->activeTool >= 0 && (size_t)priv->activeTool < 
		vec_size(priv->toolPtrs))
	{
		tool = vec_at(priv->toolPtrs, priv->activeTool);
		tool->draw(priv->gl);
	}
	priv->edit->draw();

	priv->gl->unlockScreen();
}

static void ED_LoopMenu(ED_t ed)
{
	lwlib_GetPrivAssert(ED_Priv_t, ed);

	while (!priv->quit)
	{
		ED_SendKeyEvents(ed);

		ED_UpdateFrame(ed);
		lwlib::g_frameTimer = lwlib::Timer();

		priv->gl->clearScreen();

		ED_Draw(ed);

		priv->gl->updateScreen();
		SDL_Delay(8);
	}
}

static void ED_DarkenBackdrop(ED_t ed)
{
	lwlib_GetPrivAssert(ED_Priv_t, ed);

	if (priv->signonImg == NULL)
	{
		GenerateShadeTable(0, 0, 0, gamepal, 0);

		priv->signonImg = priv->picViewer->getPicImg(SIGNONPIC);
		priv->signonImg->scaleIntensity(192);
	}
}

void ED_LaunchMenu(ED_t ed)
{
	ED_DarkenBackdrop(ed);
	ED_LoopMenu(ed);
}

const char *ED_SpriteName(ED_t ed, int index)
{
	lwlib_GetPrivAssert(ED_Priv_t, ed);
	return vec_at(priv->sprites, index).name;
}

int ED_NumSprites(ED_t ed)
{
	lwlib_GetPrivAssert(ED_Priv_t, ed);
	return vec_size(priv->sprites);
}

short ED_SpriteNum(ED_t ed, const char *name)
{
	short spriteNum;
	lwlib_GetPrivAssert(ED_Priv_t, ed);

	if (lwlib_StrMapContains(&priv->spriteMap, name))
	{
		spriteNum = (short)lwlib_StrMapElem(&priv->spriteMap, name);
	}
	else
	{
		spriteNum = -1;
		sscanf(name, "%hd", &spriteNum);
	}

	return spriteNum;
}

ED_StatInfo_t *ED_GetStatInfos(ED_t ed)
{
	lwlib_GetPrivAssert(ED_Priv_t, ed);
	return &vec_at(priv->statInfos, 0);
}

int ED_ActorStatInfo(ED_t ed, int actor)
{
	int type;
	lwlib_GetPrivAssert(ED_Priv_t, ed);

	if (lwlib_IntMapContains(&priv->statInfoMap, actor))
	{
		type = (int)lwlib_IntMapElem(&priv->statInfoMap, actor);
	}
	else
	{
		type = -1;
	}

	return type;
}

ED_SpriteSeq_t *ED_Sprites(ED_t ed)
{
	lwlib_GetPrivAssert(ED_Priv_t, ed);
	return &priv->sprites;
}

ED_StateIndex_t ED_StateIndex(ED_t ed, const char *name)
{
	ED_StateIndex_t stateIndex;
	lwlib_GetPrivAssert(ED_Priv_t, ed);

	stateIndex = 0;
	if (lwlib_StrMapContains(&priv->stateMap, name))
	{
		stateIndex = (int)lwlib_StrMapElem(&priv->stateMap, name) + 1;
	}

	return stateIndex;
}

ED_State_t *ED_State(ED_t ed, const char *name)
{
	ED_State_t *state;
	ED_StateIndex_t stateIndex;
	lwlib_GetPrivAssert(ED_Priv_t, ed);

	state = NULL;

	stateIndex = ED_StateIndex(ed, name);
	if (stateIndex != 0)
	{
		state = &vec_at(priv->states, stateIndex - 1);
	}

	return state;
}

ED_State_t *ED_States(ED_t ed)
{
	lwlib_GetPrivAssert(ED_Priv_t, ed);
	return &vec_at(priv->states, 0);
}

ED_State_t *ED_StateByIndex(ED_t ed, ED_StateIndex_t stateIndex)
{
	static ED_State_t dummy = { { 0 } };
	lwlib_GetPrivAssert(ED_Priv_t, ed);

	if (stateIndex == 0)
	{
		return &dummy;
	}

	return &vec_at(priv->states, stateIndex - 1);
}

ED_ThinkFunc_t ED_ThinkFunc(ED_t ed, const char *name)
{
	int stateHandlerNum;
	ED_ThinkFunc_t thinkFunc;
	ED_StateHandler_t *stateHandler;
	lwlib_GetPrivAssert(ED_Priv_t, ed);

	thinkFunc = NULL;
	if (lwlib_StrMapContains(&priv->stateHandlerMap, name))
	{
		stateHandlerNum = (int)lwlib_StrMapElem(&priv->stateHandlerMap, name);
		stateHandler = &priv->stateHandlers[stateHandlerNum];
		thinkFunc = stateHandler->thinkFunc;
	}

	return thinkFunc;
}

ED_Spawn_t *ED_ActorSpawn(ED_t ed, int actor)
{
	ED_Spawn_t *spawn;
	int spawnNum;
	lwlib_GetPrivAssert(ED_Priv_t, ed);

	spawn = NULL;
	if (lwlib_IntMapContains(&priv->spawnMap, actor))
	{
		spawnNum = (int)lwlib_IntMapElem(&priv->spawnMap, actor);
		spawn = &vec_at(priv->spawns, spawnNum);
	}

	return spawn;
}

ED_Spawn_t *ED_SpawnByEnemyName(ED_t ed, const char *name, int dir)
{
	int i;
	ED_Spawn_t *spawn;
	lwlib_GetPrivAssert(ED_Priv_t, ed);

	for (i = 0; (size_t)i < vec_size(priv->spawns); i++)
	{
		spawn = &vec_at(priv->spawns, i);
		if (strcmp(spawn->enemyName, name) == 0 && 
			(dir == -1 || spawn->dir == dir))
		{
			return spawn;
		}
	}

	return NULL;
}

ED_EnemyIndex_t ED_EnemyIndex(ED_t ed, const char *name)
{
	ED_EnemyIndex_t enemyNum;
	lwlib_GetPrivAssert(ED_Priv_t, ed);

	enemyNum = 0;
	if (lwlib_StrMapContains(&priv->enemyMap, name))
	{
		enemyNum = (int)lwlib_StrMapElem(&priv->enemyMap, name) + 1;
	}

	return enemyNum;
}

ED_Enemy_t *ED_Enemy(ED_t ed, const char *name)
{
	ED_Enemy_t *enemy;
	ED_EnemyIndex_t enemyIndex;
	lwlib_GetPrivAssert(ED_Priv_t, ed);

	enemy = NULL;

	enemyIndex = ED_EnemyIndex(ed, name);
	if (enemyIndex != 0)
	{
		enemy = &vec_at(priv->enemies, enemyIndex - 1);
	}

	return enemy;
}

ED_Enemy_t *ED_Enemies(ED_t ed)
{
	lwlib_GetPrivAssert(ED_Priv_t, ed);
	return &vec_at(priv->enemies, 0);
}

ED_Enemy_t *ED_EnemyByIndex(ED_t ed, ED_EnemyIndex_t enemyIndex)
{
	static ED_Enemy_t dummy = { { 0 } };
	lwlib_GetPrivAssert(ED_Priv_t, ed);

	if (enemyIndex == 0)
	{
		return &dummy;
	}

	return &vec_at(priv->enemies, enemyIndex - 1);
}

void ED_SetVideoMode(ED_t ed)
{
	lwlib::Gl_Screen_t glScreen;
	lwlib_GetPrivAssert(ED_Priv_t, ed);

	lwlib_Zero(glScreen);
	glScreen.width = screenWidth;
	glScreen.height = screenHeight;
	glScreen.bits = screenBits;
#ifndef SDL_TRUE_COLOR
	glScreen.palette = gamepal;
#endif
	priv->gl->setVideoMode(&glScreen);
	//screenSurf = glScreen.surf;
	//usedoublebuffering = glScreen.isDoubleBuffered;
}

void ED_InitGl(ED_t ed)
{
	lwlib_GetPrivAssert(ED_Priv_t, ed);
	priv->gl->init();
}

ED_EnemyConfig_t *ED_EnemyConfig(ED_t ed)
{
	lwlib_GetPrivAssert(ED_Priv_t, ed);
	return &priv->enemyConfig;
}

ED_ProjIndex_t ED_ProjIndex(ED_t ed, const char *name)
{
	ED_ProjIndex_t projNum;
	lwlib_GetPrivAssert(ED_Priv_t, ed);

	projNum = 0;
	if (lwlib_StrMapContains(&priv->projMap, name))
	{
		projNum = (int)lwlib_StrMapElem(&priv->projMap, name) + 1;
	}

	return projNum;
}

const ED_Proj_t *ED_Proj(ED_t ed, const char *name)
{
	ED_Proj_t *proj;
	ED_ProjIndex_t projIndex;
	lwlib_GetPrivAssert(ED_Priv_t, ed);

	proj = NULL;

	projIndex = ED_ProjIndex(ed, name);
	if (projIndex != 0)
	{
		proj = &vec_at(priv->projs, projIndex - 1);
	}

	return proj;
}

const ED_Proj_t *ED_Projs(ED_t ed)
{
	lwlib_GetPrivAssert(ED_Priv_t, ed);
	return &vec_at(priv->projs, 0);
}

const ED_Proj_t *ED_ProjByIndex(ED_t ed, ED_ProjIndex_t projIndex)
{
	static const ED_Proj_t dummy = { { 0 } };
	lwlib_GetPrivAssert(ED_Priv_t, ed);

	if (projIndex == 0)
	{
		return &dummy;
	}

	return &vec_at(priv->projs, projIndex - 1);
}

int ED_ProjPhysIndex(ED_t ed, const char *name)
{
	int projPhysNum;
	lwlib_GetPrivAssert(ED_Priv_t, ed);

	projPhysNum = -1;
	if (lwlib_StrMapContains(&priv->projPhysMap, name))
	{
		projPhysNum = (int)lwlib_StrMapElem(&priv->projPhysMap, name);
	}

	return projPhysNum;
}

const ED_ProjPhys_t *ED_ProjPhys(ED_t ed, const char *name)
{
	int projPhysIndex;
	ED_ProjPhys_t *projPhys;
	lwlib_GetPrivAssert(ED_Priv_t, ed);

	projPhys = NULL;

	projPhysIndex = ED_ProjPhysIndex(ed, name);
	if (projPhysIndex != -1)
	{
		projPhys = &vec_at(priv->projPhysSeq, projPhysIndex);
	}

	return projPhys;
}

ED_Inject_t *ED_Inject(ED_t ed)
{
	lwlib_GetPrivAssert(ED_Priv_t, ed);
	return &priv->inject;
}

int ED_StatInfoByName(ED_t ed, const char *name)
{
	int i;
	ED_StatInfo_t *statInfo;
	lwlib_GetPrivAssert(ED_Priv_t, ed);

	for (i = 0; (size_t)i < vec_size(priv->statInfos); i++)
	{
		statInfo = &vec_at(priv->statInfos, i);
		if (strcmp(statInfo->name, name) == 0)
		{
			return i;
		}
	}

	return -1;
}

int ED_FindFreeActor(ED_t ed)
{
	int actor;
	int endActorCode;
	lwlib_GetPrivAssert(ED_Priv_t, ed);

	endActorCode = priv->tileInfo.firstWallPatchActor;

	for (actor = 23; actor < endActorCode; actor++)
	{
		if (actor == DEADGUARDTILE || (actor >= ICONARROWS && 
			actor <= KEY4ENEMYBLOCK))
		{
			continue;
		}

		if (!lwlib_IntMapContains(&priv->statInfoMap, actor) &&
			!lwlib_IntMapContains(&priv->spawnMap, actor))
		{
			break;
		}
	}

	if (actor == endActorCode)
	{
		return 0;
	}

	return actor;
}

bool ED_AddStatInfo(ED_t ed, ED_StatInfo_t statInfo)
{
	int actor, terminalIndex;
	lwlib_GetPrivAssert(ED_Priv_t, ed);

	actor = ED_FindFreeActor(ed);
	if (actor == 0)
	{
		return false;
	}

	assert(vec_size(priv->statInfos) > 0);
	terminalIndex = vec_size(priv->statInfos) - 1;

	vec_push(priv->statInfos, ED_StatInfo_t, statInfo);

	vec_top(priv->statInfos) = vec_at(priv->statInfos, 
		terminalIndex);

	statInfo.actor = actor;
	lwlib_IntMapInsert(&priv->statInfoMap, actor, 
		(lwlib_Word_t)terminalIndex);
	vec_at(priv->statInfos, terminalIndex) = statInfo;
	return true;
}

ED_SoundIndex_t ED_SoundIndex(ED_t ed, const char *name)
{
	int i;
	ED_SoundIndex_t soundIndex;
	ED_SoundIndex_t randomSoundIndex;
	ED_RandomSound_t *randomSound;
	lwlib_Str_t *soundDef;
	lwlib_GetPrivAssert(ED_Priv_t, ed);

	if (name == NULL)
	{
		return -1;
	}

	if (lwlib_StrMapContains(&priv->randomSoundMap, name))
	{
		randomSoundIndex = (int)lwlib_StrMapElem(
			&priv->randomSoundMap, name);
		randomSound = &vec_at(priv->randomSounds, randomSoundIndex);

		const int n = vec_size(randomSound->soundDefs);
		if (n > 0)
		{
			i = lwlib_RandInt(n);
			if (i >= n)
			{
				i = n - 1;
			}
			soundDef = &vec_at(randomSound->soundDefs, i);
			name = soundDef->v;
		}
	}

	soundIndex = -1;
	if (lwlib_StrMapContains(&priv->soundMap, name))
	{
		soundIndex = (int)lwlib_StrMapElem(&priv->soundMap, name);
	}

	return soundIndex;
}

ED_Sound_t *ED_Sound(ED_t ed, const char *name)
{
	ED_Sound_t *sound;
	ED_SoundIndex_t soundIndex;
	lwlib_GetPrivAssert(ED_Priv_t, ed);

	sound = NULL;

	soundIndex = ED_SoundIndex(ed, name);
	if (soundIndex != -1)
	{
		sound = &vec_at(priv->sounds, soundIndex);
	}

	return sound;
}

ED_Sound_t *ED_Sounds(ED_t ed)
{
	lwlib_GetPrivAssert(ED_Priv_t, ed);
	return &vec_at(priv->sounds, 0);
}

int ED_NumSounds(ED_t ed)
{
	lwlib_GetPrivAssert(ED_Priv_t, ed);
	return vec_size(priv->sounds);
}

ED_BonusIndex_t ED_BonusIndex(ED_t ed, const char *name)
{
	ED_BonusIndex_t bonusIndex;
	lwlib_GetPrivAssert(ED_Priv_t, ed);

	bonusIndex = -1;
	if (lwlib_StrMapContains(&priv->bonusMap, name))
	{
		bonusIndex = (int)lwlib_StrMapElem(&priv->bonusMap, name);
	}

	return bonusIndex;
}

ED_Bonus_t *ED_Bonus(ED_t ed, const char *name)
{
	ED_Bonus_t *bonus;
	ED_BonusIndex_t bonusIndex;
	lwlib_GetPrivAssert(ED_Priv_t, ed);

	bonus = NULL;

	bonusIndex = ED_BonusIndex(ed, name);
	if (bonusIndex != -1)
	{
		bonus = &vec_at(priv->bonuses, bonusIndex);
	}

	return bonus;
}

ED_Bonus_t *ED_Bonuses(ED_t ed)
{
	lwlib_GetPrivAssert(ED_Priv_t, ed);
	return &vec_at(priv->bonuses, 0);
}

int ED_NumBonuses(ED_t ed)
{
	lwlib_GetPrivAssert(ED_Priv_t, ed);
	return vec_size(priv->bonuses);
}

const char *ED_FootStepByFloorTile(ED_t ed, int floorTile)
{
	int footStepIndex;
	ED_FootStep_t *footStep;
	lwlib_GetPrivAssert(ED_Priv_t, ed);

	if (lwlib_IntMapContains(&priv->footStepMap, floorTile))
	{
		footStepIndex = (int)lwlib_IntMapElem(&priv->footStepMap,
			floorTile);
		footStep = &vec_at(priv->footSteps, footStepIndex);

		return footStep->soundDef;
	}

	return NULL;
}

int ED_ReplaceSprite(const ED_Enemy_t *enemy, int shapenum)
{
	if (lwlib_IntMapContains(&enemy->spriteReplMap, shapenum))
	{
		shapenum = (int)lwlib_IntMapElem(&enemy->spriteReplMap, shapenum);
	}
	return shapenum;
}

ED_VSWAPInfo_t *ED_GetVSWAPInfo(ED_t ed)
{
	lwlib_GetPrivAssert(ED_Priv_t, ed);
	return &priv->vswapInfo;
}

ED_TileInfo_t *ED_GetTileInfo(ED_t ed)
{
	lwlib_GetPrivAssert(ED_Priv_t, ed);
	return &priv->tileInfo;
}

int ED_FirstPatchTile(ED_t ed)
{
	return (ED_GetVSWAPInfo(ed)->wallPatchStart / 2) + 1;
}

int ED_LastPatchTile(ED_t ed)
{
	return ((ED_GetVSWAPInfo(ed)->skyStart + 1) / 2) - 1;
}

bool ED_PatchPicIsGlow(int patchpic)
{
	lwlib_GetPrivAssert(ED_Priv_t, ed);
	return (priv->vswapInfo.glowPatchStart != 0 &&
		patchpic >= priv->vswapInfo.glowPatchStart &&
		patchpic <= priv->vswapInfo.glowPatchEnd);
}

int ED_AnimateWallTile(ED_t ed, int tile)
{
	int index;
	ED_AnimatedWallTile_t *animatedWallTile;
	lwlib_GetPrivAssert(ED_Priv_t, ed);

	index = priv->vswapInfo.animatedWallTileMap[tile];
	if (index != 0)
	{
		index--;
		animatedWallTile = &vec_at(priv->vswapInfo.animatedWallTiles,
			index);
		tile = animatedWallTile->animStartTile + 
			((gamestate.TimeCount / animatedWallTile->animTics) % 
			animatedWallTile->animNumTiles);
	}

	return tile;
}

void ED_StatInfoReport(ED_t ed)
{
	int i;
	int spriteNum;
	ED_StatInfo_t *statInfo;
	const ED_Spawn_t *spawn;
	ED_State_t *startState;
	const ED_Enemy_t *enemy;
	lwlib_GetPrivAssert(ED_Priv_t, ed);

	printf("static objects\n");
	for (i = 0; (size_t)i < vec_size(priv->statInfos); i++)
	{
		statInfo = &vec_at(priv->statInfos, i);
		spriteNum = lwlib_StrMapElem(&priv->spriteMap,
			statInfo->picnum);
		printf("%d - %s (%s: %d)\n", statInfo->actor, statInfo->name,
			statInfo->picnum, spriteNum);
	}

	printf("actors\n");
	for (i = 0; (size_t)i < vec_size(priv->spawns); i++)
	{
		spawn = &vec_at(priv->spawns, i);
		printf("%d - %s", spawn->actor, spawn->enemyName);
		if (spawn->dir != 8)
		{
			printf(", dir %d", spawn->dir);
		}
		if (spawn->patrol)
		{
			printf(", patrol");
		}

		enemy = ED_Enemy(ed, spawn->enemyName);
		assert(enemy != NULL);

		startState = ED_State(ed, spawn->startState);
		assert(startState != NULL);

		spriteNum = ED_SpriteNum(ed, startState->shapenum);
		spriteNum += ED_SPR(enemy->spriteOffset.newStart) -
			ED_SPR(enemy->spriteOffset.oldStart);
		spriteNum = ED_ReplaceSprite(enemy, spriteNum);

		if (spriteNum != -1)
		{
			printf(" (%s: %d)", vec_at(priv->sprites, spriteNum).name, spriteNum);
		}
		printf("\n");
	}
}

int ED_StatInfoPickerTile(ED_t ed, int statInfoNum)
{
	int tile;
	ED_StatInfo_t *statInfo;
	lwlib_GetPrivAssert(ED_Priv_t, ed);

	if (statInfoNum < 0 || (size_t)statInfoNum >= vec_size(priv->statInfos))
	{
		return -1;
	}

	statInfo = &vec_at(priv->statInfos, statInfoNum);

	tile = ED_SPR(statInfo->picnum);
	if (tile == -1)
	{
		return -1;
	}

	tile += PMSpriteStart;
	return tile;
}

int ED_SpawnPickerTile(ED_t ed, int spawnNum)
{
	int spriteNum;
	ED_Spawn_t *spawn;
	ED_State_t *startState;
	const ED_Enemy_t *enemy;
	lwlib_GetPrivAssert(ED_Priv_t, ed);

	if (spawnNum < 0 || (size_t)spawnNum >= vec_size(priv->spawns))
	{
		return -1;
	}

	spawn = &vec_at(priv->spawns, spawnNum);

	enemy = ED_Enemy(ed, spawn->enemyName);
	assert(enemy != NULL);

	startState = ED_State(ed, spawn->startState);
	assert(startState != NULL);

	spriteNum = ED_SpriteNum(ed, startState->shapenum);
	if (spriteNum == -1)
	{
		return 0;
	}

	spriteNum += ED_SPR(enemy->spriteOffset.newStart) -
		ED_SPR(enemy->spriteOffset.oldStart);
	spriteNum = ED_ReplaceSprite(enemy, spriteNum);

	return spriteNum + PMSpriteStart;
}

lwlib_Str_t ED_SpawnPickerInfoString(ED_t ed, int spawnNum)
{
	ED_Spawn_t *spawn;
	lwlib_String_t infoStr;
	lwlib_String_t tmpStr[2];
	lwlib_GetPrivAssert(ED_Priv_t, ed);

	lwlib_Zero(infoStr);

	if (spawnNum < 0 || (size_t)spawnNum >= vec_size(priv->spawns))
	{
		return lwlib_Str("");
	}

	spawn = &vec_at(priv->spawns, spawnNum);

	lwlib_Snprintf(infoStr, "actor: %d, %s", spawnNum, spawn->enemyName);
	if (spawn->dir != 8)
	{
		lwlib_Copy(tmpStr[0], infoStr);
		lwlib_Snprintf(tmpStr[1], ", dir %d", spawn->dir);
		lwlib_Snprintf(infoStr, "%s%s", tmpStr[0], tmpStr[1]);
	}
	if (spawn->patrol)
	{
		lwlib_Copy(tmpStr[0], infoStr);
		lwlib_StrCopy(tmpStr[1], ", patrol");
		lwlib_Snprintf(infoStr, "%s%s", tmpStr[0], tmpStr[1]);
	}

	return lwlib_Str(infoStr);
}

int ED_MoveStatInfo(ED_t ed, int statInfoNum, int move)
{
	int i;
	lwlib_GetPrivAssert(ED_Priv_t, ed);

	assert(statInfoNum >= 0 && (size_t)statInfoNum < vec_size(priv->statInfos));
	
	for (i = 0; i < abs(move); i++)
	{
		if (move > 0)
		{
			if (vec_size(priv->statInfos) > 1 && 
				(size_t)statInfoNum < vec_size(priv->statInfos) - 1)
			{
				lwlib_IntMapInsert(&priv->statInfoMap, 
					vec_at(priv->statInfos, statInfoNum).actor,
					statInfoNum + 1);

				lwlib_IntMapInsert(&priv->statInfoMap, 
					vec_at(priv->statInfos, statInfoNum + 1).actor,
					statInfoNum);

				lwlib_Swap(vec_at(priv->statInfos, statInfoNum), 
					vec_at(priv->statInfos, statInfoNum + 1),
					ED_StatInfo_t);
				statInfoNum++;
			}
		}
		else
		{
			if (statInfoNum > 0)
			{
				lwlib_IntMapInsert(&priv->statInfoMap, 
					vec_at(priv->statInfos, statInfoNum).actor,
					statInfoNum - 1);

				lwlib_IntMapInsert(&priv->statInfoMap, 
					vec_at(priv->statInfos, statInfoNum - 1).actor,
					statInfoNum);

				lwlib_Swap(vec_at(priv->statInfos, statInfoNum), 
					vec_at(priv->statInfos, statInfoNum - 1),
					ED_StatInfo_t);
				statInfoNum--;
			}
		}
	}

	return statInfoNum;
}

ED_SpawnSeq_t ED_Spawns(ED_t ed)
{
	lwlib_GetPrivAssert(ED_Priv_t, ed);
	return priv->spawns;
}

void ED_InsertState(ED_t ed, ED_State_t state)
{
	lwlib_GetPrivAssert(ED_Priv_t, ed);

	lwlib_StrMapInsert(&priv->stateMap, state.name, (lwlib_Word_t)
		vec_size(priv->states));
	vec_push(priv->states, ED_State_t, state);
}

void ED_InsertSpawn(ED_t ed, ED_Spawn_t spawn)
{
	lwlib_GetPrivAssert(ED_Priv_t, ed);

	lwlib_IntMapInsert(&priv->spawnMap, spawn.actor, (lwlib_Word_t)
		vec_size(priv->spawns));
	vec_push(priv->spawns, ED_Spawn_t, spawn);
}

void ED_InsertEnemy(ED_t ed, ED_Enemy_t enemy)
{
	lwlib_GetPrivAssert(ED_Priv_t, ed);

	lwlib_StrMapInsert(&priv->enemyMap, enemy.name, (lwlib_Word_t)
		vec_size(priv->enemies));
	vec_push(priv->enemies, ED_Enemy_t, enemy);
}

void ED_InsertSound(ED_t ed, ED_Sound_t sound)
{
	lwlib_GetPrivAssert(ED_Priv_t, ed);

	lwlib_StrMapInsert(&priv->soundMap, sound.name, (lwlib_Word_t)
		vec_size(priv->sounds));
	vec_push(priv->sounds, ED_Sound_t, sound);

	ASM_Audiosegs = (sample *)realloc(ASM_Audiosegs, sizeof(sample) * 
		vec_size(priv->sounds));
	lwlib_Zero(ASM_Audiosegs[vec_size(priv->sounds) - 1]);
}

ED_Ambient_t *ED_ActorAmbient(ED_t ed, int tile)
{
	int ambientIndex;
	ED_Ambient_t *ambient;
	lwlib_GetPrivAssert(ED_Priv_t, ed);

	ambient = NULL;

	if (lwlib_IntMapContains(&priv->ambientMap, tile))
	{
		ambientIndex = (int)lwlib_IntMapElem(&priv->ambientMap, tile);
		ambient = &vec_at(priv->ambients, ambientIndex);
	}

	return ambient;
}

ED_Ambient_t *ED_Ambients(ED_t ed)
{
	lwlib_GetPrivAssert(ED_Priv_t, ed);
	return &vec_at(priv->ambients, 0);
}

int ED_NumAmbients(ED_t ed)
{
	lwlib_GetPrivAssert(ED_Priv_t, ed);
	return vec_size(priv->ambients);
}

void ED_ResetAmbients(ED_t ed)
{
	int i;
	ED_Ambient_t *ambient;
	lwlib_GetPrivAssert(ED_Priv_t, ed);

	for (i = 0; (size_t)i < vec_size(priv->ambients); i++)
	{
		ambient = &vec_at(priv->ambients, i);
		ambient->chan = 0;
		ED_RandomizeAmbientTimePeriod(ed, i);
	}
}

ED_Ambient_t *ED_AmbientByName(ED_t ed, const char *soundName)
{
	int ambientIndex;
	ED_Ambient_t *ambient;
	lwlib_GetPrivAssert(ED_Priv_t, ed);

	ambient = NULL;

	if (lwlib_StrMapContains(&priv->ambientStrMap, soundName))
	{
		ambientIndex = (int)lwlib_StrMapElem(&priv->ambientStrMap, 
			soundName);
		ambient = &vec_at(priv->ambients, ambientIndex);
	}

	return ambient;
}

void ED_RandomizeAmbientTimePeriod(ED_t ed, int ambientNum)
{
	ED_Ambient_t *ambient;
	lwlib_GetPrivAssert(ED_Priv_t, ed);

	ambient = &vec_at(priv->ambients, ambientNum);
	ambient->timer = 0;
	ambient->timePeriod = US_RndT_Range(
		ambient->minTimePeriod, ambient->maxTimePeriod);
}

bool ED_ActorIsWallPatch(ED_t ed, int actor)
{
	lwlib_GetPrivAssert(ED_Priv_t, ed);
	return actor >= priv->tileInfo.firstWallPatchActor &&
		actor <= priv->tileInfo.lastWallPatchActor;
}

int ED_ActorToWallPatch(ED_t ed, int actor)
{
	lwlib_GetPrivAssert(ED_Priv_t, ed);
	return ED_FirstPatchTile(ed) + (actor - 
		priv->tileInfo.firstWallPatchActor);
}

bool ED_CanBurrowFloor(ED_t ed, int floor)
{
	int i;
	lwlib_GetPrivAssert(ED_Priv_t, ed);
	
	for (i = 0; (size_t)i < vec_size(priv->tileInfo.burrowFloorTiles); i++)
	{
		if (floor == vec_at(priv->tileInfo.burrowFloorTiles, i))
		{
			return true;
		}
	}

	return false;
}

bool ED_FloorIsWater(ED_t ed, int floor)
{
	lwlib_GetPrivAssert(ED_Priv_t, ed);
	return floor == priv->tileInfo.waterFloorTile;
}

double ED_RandomConst_Eval(const ED_RandomConst_t *randomConst)
{
	if (randomConst->min == randomConst->max)
	{
		return randomConst->min;
	}
	return US_RndT() * (randomConst->max - randomConst->min) / 255.0 + 
		randomConst->min;
}

int ED_RandomInt_Eval(const ED_RandomInt_t *randomInt)
{
	if (randomInt->min == randomInt->max)
	{
		return randomInt->min;
	}
	return US_RndT() * (randomInt->max - randomInt->min) / 255 + 
		randomInt->min;
}

bool ED_TileIsTruckTile(ED_t ed, int tile)
{
	lwlib_GetPrivAssert(ED_Priv_t, ed);
	return tile >= priv->tileInfo.firstTruckTile && 
		tile <= priv->tileInfo.lastTruckTile;
}

ED_SoundIndex_t ED_GetThemedSound(ED_t ed, const ED_SoundDef_t soundDef)
{
	bool isWood;
	lwlib_GetPrivAssert(ED_Priv_t, ed);

	isWood = (DOORPAGE_START == priv->tileInfo.woodenDoorStartTile);

	if (strcmp(soundDef, "OPENDOORSND") == 0)
	{
		soundDef = isWood ? "OPENDOORWOODSND" : "OPENDOORMETALSND";
	}
	else if (strcmp(soundDef, "CLOSEDOORSND") == 0)
	{
		soundDef = isWood ? "CLOSEDOORWOODSND" : "CLOSEDOORMETALSND";
	}

	return ED_SND(soundDef);
}

ED_WallAttr_t *ED_WallAttr(ED_t ed, int tile)
{
	lwlib_GetPrivAssert(ED_Priv_t, ed);

	if (tile < 0 || (size_t)tile >= vec_size(priv->wallAttrs))
	{
		return NULL;
	}

	return &vec_at(priv->wallAttrs, tile);
}

ED_Ricochet_t *ED_WallRicochet(ED_t ed, int tile)
{
	ED_WallAttr_t *wallAttr;
	lwlib_GetPrivAssert(ED_Priv_t, ed);

	wallAttr = ED_WallAttr(ed, tile);
	if (wallAttr == NULL)
	{
		wallAttr = ED_WallAttr(ed, 0);
		if (wallAttr == NULL)
		{
			return NULL;
		}
	}

	return wallAttr->ricochet;
}

bool ED_ActorIsBreakableWall(ED_t ed, int actor)
{
	int wallAttrIndex;
	ED_TileInfo_t *tileInfo;
	ED_WallAttr_t *wallAttr;
	
	tileInfo = ED_GetTileInfo(ed);
	if (actor < tileInfo->firstWallPatchActor || 
		actor > tileInfo->lastWallPatchActor)
	{
		return false;
	}

	wallAttrIndex = actor - tileInfo->firstWallPatchActor;

	wallAttr = ED_WallAttr(ed, wallAttrIndex);
	if (wallAttr == NULL)
	{
		return false;
	}

	return wallAttr->breakableWall;
}

int ED_LRPack(ED_t ed)
{
	lwlib_GetPrivAssert(ED_Priv_t, ed);
	return priv->lrPack;
}

ED_AmbientSettings_t *ED_AmbientSettings(ED_t ed)
{
	lwlib_GetPrivAssert(ED_Priv_t, ed);
	return &priv->ambientSettings;
}
