//
//	WL Edit
//	WL_ED.h - Game Editor
//	By Linux Wolf - Team RayCAST
//

#ifndef WL_ED_H
#define WL_ED_H

#include "lw_misc.h"
#include "lw_intmap.h"
#include "lw_vec.h"

#define ED_SPR(name) ED_SpriteNum(ed, name)
#define ED_STATINFO ED_GetStatInfos(ed)
#define ED_STATE(name) ED_State(ed, name)
#define ED_THINK(name) ((void (*)(objtype *))ED_ThinkFunc(ed, name))
#define ED_SOUND(name) ED_Sound(ed, name)
#define ED_SND(name) ED_SoundIndex(ed, name)
#define ED_BONUS(name) ED_Bonus(ed, name)
#define ED_BONUSIDX(name) ED_BonusIndex(ed, name)
#define ED_AMB(name) ED_AmbientByName(ed, name)

#define ED_EnemyRangeAttackVariant(name) ED_RangeAttackVariant_ ## name

#define ED_MAX_SKILLPROFILES 4
#define ED_VISFACTOR_MATRIX_SIZE 16

typedef lwlib_String_t ED_SpriteDef_t;
typedef lwlib_String_t ED_BonusDef_t;
typedef lwlib_String_t ED_SoundDef_t;
typedef lwlib_String_t ED_RandomSoundDef_t;
typedef lwlib_String_t ED_EnemyKey_t;
typedef lwlib_String_t ED_ProjPhysKey_t;

typedef lwlib_StrSeq_t ED_EnemyKeySeq_t;

typedef enum
{
	none,
	block,
	bo_mg42_mounted,
	bo_any,
} ED_Stat_t;

lwlib_EnumStrExtern(ED_Stat);

typedef enum
{
    dam_none,
    dam_melee,
    dam_bullet,
    dam_mutagen,
    dam_rifle,
    dam_fire,
    dam_chaingun,
    dam_rocket,
    dam_selfdestruct,
    dam_infection,
} ED_DamageType_t;

lwlib_EnumStrExtern(ED_DamageType);

typedef enum
{
	FL_NONE             = 0x00000000,
	FL_SHOOTABLE        = 0x00000001,
	FL_BONUS            = 0x00000002,
	FL_NEVERMARK        = 0x00000004,
	FL_VISABLE          = 0x00000008,
	FL_ATTACKMODE       = 0x00000010,
	FL_FIRSTATTACK      = 0x00000020,
	FL_AMBUSH           = 0x00000040,
	FL_NONMARK          = 0x00000080,
	FL_FULLBRIGHT       = 0x00000100,
//#ifdef USE_DIR3DSPR
	// you can choose one of the following values in wl_act1.cpp
	// to make a static sprite a directional 3d sprite
	// (see example at the end of the statinfo array)
	FL_DIR_HORIZ_MID    = 0x00000200,
	FL_DIR_HORIZ_FW     = 0x00000400,
	FL_DIR_HORIZ_BW     = 0x00000600,
	FL_DIR_VERT_MID     = 0x00000a00,
	FL_DIR_VERT_FW      = 0x00000c00,
	FL_DIR_VERT_BW      = 0x00000e00,

	// these values are just used to improve readability of code
	FL_DIR_NONE         = 0x00000000,
	FL_DIR_POS_MID      = 0x00000200,
	FL_DIR_POS_FW       = 0x00000400,
	FL_DIR_POS_BW       = 0x00000600,
	FL_DIR_POS_MASK     = 0x00000600,
	FL_DIR_VERT_FLAG    = 0x00000800,
	FL_DIR_MASK         = 0x00000e00,
//#endif
	FL_BLOCK            = 0x00001000,
	FL_UNKILLABLE       = 0x00002000,
	FL_SUPPRESSPAIN     = 0x00004000,
	FL_FENCING          = 0x00008000,
	FL_UNDERGROUND      = 0x00010000,
	FL_WEAKTOROCKETS    = 0x00020000,
	FL_NODAMAGE         = 0x00040000,
	FL_SPAWNEDGHOST     = 0x00080000,
	FL_GLOWBRIGHT       = 0x00100000,
} ED_ObjFlag_t;

lwlib_EnumStrExtern(ED_ObjFlag);
lwlib_EnumValExtern(ED_ObjFlag);

typedef struct ED_Sprite_s
{
	lwlib_String_t name;
} ED_Sprite_t;

typedef vec_t(ED_Sprite_t) ED_SpriteSeq_t;

typedef struct ED_StatInfo_s
{
	lwlib_String_t name;
	ED_SpriteDef_t picnum;
	ED_Stat_t type;
	ED_BonusDef_t bonus;
	ED_ObjFlag_t specialFlags;
	int animframes;
	int actor;
} ED_StatInfo_t;

typedef vec_t(ED_StatInfo_t) ED_StatInfoSeq_t;

typedef lwlib_String_t ED_ThinkDef_t;
typedef void (*ED_ThinkFunc_t)(void *obj);

typedef lwlib_String_t ED_StateKey_t;

typedef struct ED_SpriteRepl_s
{
	ED_SpriteDef_t oldSprite;
	ED_SpriteDef_t lastOldSprite;
	ED_SpriteDef_t newSprite;
} ED_SpriteRepl_t;

typedef vec_t(ED_SpriteRepl_t) ED_SpriteReplSeq_t;

typedef struct ED_SpriteOffset_s
{
	ED_SpriteDef_t newStart;
	ED_SpriteDef_t oldStart;
} ED_SpriteOffset_t;

typedef struct ED_Spawn_s
{
	lwlib_String_t enemyName;
	ED_StateKey_t startState;
	int dir;
	bool patrol;
	int actor;
} ED_Spawn_t;

typedef vec_t(ED_Spawn_t) ED_SpawnSeq_t;

typedef int ED_EnemyIndex_t;
typedef int ED_ProjIndex_t;
typedef int ED_StateIndex_t;
typedef int ED_SoundIndex_t;
typedef int ED_RandomSoundIndex_t;
typedef int ED_BonusIndex_t;

typedef int ED_ProjPhysId_t;

typedef struct ED_ChanceRange_s
{
	int startDist;
	int endDist;
	int startChance;
	int endChance;
} ED_ChanceRange_t;

typedef vec_t(ED_ChanceRange_t) ED_ChanceRangeSeq_t;

typedef struct ED_RandomSound_s
{
	ED_RandomSoundDef_t name;
	lwlib_StrSeq_t soundDefs;
} ED_RandomSound_t;

typedef vec_t(ED_RandomSound_t) ED_RandomSoundSeq_t;

typedef enum ED_EnemyRangeAttackVariant_e
{
	ED_EnemyRangeAttackVariant(Normal),
	ED_EnemyRangeAttackVariant(Long),
} ED_EnemyRangeAttackVariant_t;

typedef struct ED_EnemyMg42_s
{
	bool canUse;
	bool canWalkThrough;
	ED_StateKey_t shootState;
	ED_StateKey_t waitState;
} ED_EnemyMg42_t;

typedef struct ED_EnemyMecha_s
{
	bool canUse;
} ED_EnemyMecha_t;

typedef struct ED_DropItem_s
{
	ED_BonusDef_t bonusDef;
	int chance;
} ED_DropItem_t;

typedef vec_t(ED_DropItem_t) ED_DropItemSeq_t;

typedef struct ED_EnemyDeath_s
{
	ED_StateKey_t burnState;
	ED_StateKey_t gibState;
	ED_StateKey_t headShotState;
	ED_StateKey_t dieState;
	ED_DropItemSeq_t dropItems;
	int burnHealth;
	ED_SoundDef_t screamSound;
	ED_SoundDef_t burnScreamSound;
	bool doVictory;
} ED_EnemyDeath_t;

typedef struct ED_EnemyReactionTime_s
{
	int baseTics;
	int rndMultiplier;
	int rndDivisor;
} ED_EnemyReactionTime_t;

typedef struct ED_EnemyBloodPuff_s
{
	ED_StateKey_t state;
	ED_SoundDef_t soundDef;
	ED_SpriteOffset_t spriteOffset;
} ED_EnemyBloodPuff_t;

typedef struct ED_EnemyBurrow_s
{
	ED_StateKey_t digInState;
	ED_StateKey_t digOutState;
} ED_EnemyBurrow_t;

typedef struct ED_EnemySpawnFlags_s
{
	bool override;
	ED_ObjFlag_t flags[3];
} ED_EnemySpawnFlags_t;

typedef struct ED_EnemySurpriseAttack_s
{
	bool noAmbushReset;
	ED_EnemyKey_t spawnEnemyName;
} ED_EnemySurpriseAttack_t;

typedef struct ED_EnemyChanceConfig_s
{
	int timeout;
	int firstRange;
	int numRanges;
	ED_ChanceRangeSeq_t ranges;
} ED_EnemyChanceConfig_t;

typedef enum ED_FeatureType_e
{
	feat_resurrection,
	feat_watermut,
	feat_boss,
	feat_damagedebris,
	feat_skillprofile,
	feat_playvoice,
	feat_melee,
	feat_summoner,
	feat_guarded,
	feat_music,
} ED_FeatureType_t;

lwlib_EnumStrExtern(ED_FeatureType);

typedef struct ED_Resurrection_s
{
	ED_EnemyKey_t undeadName;
	ED_StateKey_t resurrectState;
	ED_SoundDef_t resurrectSound;
	ED_StateKey_t enterBodyState;
	ED_SoundDef_t enterBodySound;
	ED_StateKey_t skinState;
} ED_Resurrection_t;

typedef struct ED_WaterMut_s
{
	ED_StateKey_t shootState;
	ED_StateKey_t dieState;
} ED_WaterMut_t;

typedef int ED_MapNum_t;

typedef vec_t(ED_MapNum_t) ED_MapNumSeq_t;

typedef struct ED_Boss_s
{
	int index;
	ED_MapNumSeq_t musicMaps;
	ED_MapNum_t lastMusicMap;
	bool ownsMusic;
} ED_Boss_t;

typedef struct ED_DamageDebris_s
{
	lwlib_String_t projName;
	int hitpoints;
} ED_DamageDebris_t;

typedef struct ED_SkillProfile_s
{
	int hitpoints;
} ED_SkillProfile_t;

typedef struct ED_PlayVoice_s
{
	ED_SoundDef_t sound;
} ED_PlayVoice_t;

typedef struct ED_Melee_s
{
	ED_SoundDef_t hitSound;
	ED_SoundDef_t missSound;
} ED_Melee_t;

typedef struct ED_Summoner_s
{
	int summonChance;
	ED_StateKey_t summonState;
	ED_SoundDef_t summonSound;
} ED_Summoner_t;

typedef struct ED_Guarded_s
{
} ED_Guarded_t;

typedef struct ED_Music_s
{
    int mapNum;
    int trackNum;
} ED_Music_t;

typedef struct ED_Feature_s
{
	ED_FeatureType_t type;
	union
	{
		ED_Resurrection_t resurrection;
		ED_WaterMut_t waterMut;
		ED_Boss_t boss;
		ED_DamageDebris_t damageDebris;
		ED_SkillProfile_t skillProfile;
		ED_PlayVoice_t playVoice;
		ED_Melee_t melee;
		ED_Summoner_t summoner;
		ED_Guarded_t guarded;
		ED_Music_t music;
	} u;
} ED_Feature_t;

typedef vec_t(ED_Feature_t) ED_FeatureSeq_t;

typedef struct ED_Enemy_s
{
	lwlib_String_t name;
	ED_SpriteOffset_t spriteOffset;
	ED_SpriteReplSeq_t spriteRepls;
	lwlib_IntMap_t spriteReplMap;
	short health;
	short killpoints;
	int speed;
	int chaseSpeed;
	ED_StateKey_t shootState;
	ED_StateKey_t painStateOdd;
	ED_StateKey_t painStateEven;
	ED_StateKey_t chaseState;
	ED_StateKey_t meleeState;
	ED_StateKey_t escapeState;
	ED_EnemyDeath_t death;
	ED_EnemyMg42_t mg42;
	ED_SoundDef_t alertSound;
	ED_EnemyRangeAttackVariant_t rangeAttackVariant;
	ED_EnemyReactionTime_t reactionTime;
	ED_EnemyBloodPuff_t bloodPuff;
	ED_EnemyBurrow_t burrow;
	bool cannotOpenDoors;
	ED_EnemySpawnFlags_t spawnFlags;
	ED_EnemySurpriseAttack_t surpriseAttack;
	ED_EnemyChanceConfig_t shootChanceConfig;
	int runAwayDist;
	int altShootChance;
	ED_StateKey_t altShootState;
	lwlib_String_t projName;
	int projSpeed;
	int meleeDamageBase;
	int meleeDamageDivisor;
	int meleeDamageChance;
	int defenseChance;
	ED_StateKey_t defenseState;
	ED_EnemyKeySeq_t goalEnemyNames;
	ED_ThinkDef_t reachedGoalThink;
	bool lightWeight;
	bool naziFriends;
	bool meleeHurtsMecha;
	int32_t animalSoundTimer;
	int32_t animalSoundPeriod;
	ED_FeatureSeq_t features;
	ED_Resurrection_t *resurrection;
	ED_WaterMut_t *waterMut;
	ED_Boss_t *boss;
	ED_DamageDebris_t *damageDebris;
	ED_SkillProfile_t *skillProfiles[ED_MAX_SKILLPROFILES];
	int numSkillProfiles;
	ED_Melee_t *melee;
	ED_Summoner_t *summoner;
	ED_Guarded_t *guarded;
	ED_Music_t *music;
} ED_Enemy_t;

typedef vec_t(ED_Enemy_t) ED_EnemySeq_t;

typedef struct ED_State_s
{
	ED_StateKey_t name;
	int rotate;
	ED_SpriteDef_t shapenum;
	short tictime;
	ED_ThinkDef_t think;
	ED_ThinkDef_t action;
	ED_StateKey_t next;
	bool fullbright;
	ED_FeatureSeq_t features;
	ED_PlayVoice_t *playVoice;
} ED_State_t;

typedef vec_t(ED_State_t) ED_StateSeq_t;

typedef struct ED_RandomConst_s
{
	double min;
	double max;
} ED_RandomConst_t;

typedef struct ED_RandomInt_s
{
	int min;
	int max;
} ED_RandomInt_t;

typedef struct ED_ProjPhys_s
{
	lwlib_String_t name;
	ED_RandomConst_t speed;
	ED_RandomConst_t pitch;
	ED_RandomInt_t angleOffset;
	double gravity;
	double bounceDamping;
	int bounceLimit;
	ED_SoundDef_t bounceSound;
	double explodeTime;
} ED_ProjPhys_t;

typedef vec_t(ED_ProjPhys_t) ED_ProjPhysSeq_t;

typedef struct ED_Proj_s
{
	lwlib_String_t name;
	ED_StateKey_t startState;
	ED_DamageType_t damageType;
	int baseDamage;
	int rndDamageMultiplier;
	int rndDamageDivisor;
	ED_StateKey_t explState;
	bool playHitSound;
	ED_SoundDef_t shootSound;
	ED_SoundDef_t hitSound;
	bool hitsPlayer;
	bool hitsEnemies;
	ED_StateKey_t smokeState;
	ED_SpriteOffset_t spriteOffset;
	ED_ProjPhysKey_t physName;
	bool destroysWalls;
	bool hitsScenery;
} ED_Proj_t;

typedef vec_t(ED_Proj_t) ED_ProjSeq_t;

typedef struct ED_EnemyConfigCheck_s
{
	int timeout;
	int delayTime;
	ED_ChanceRangeSeq_t chanceRanges;
} ED_EnemyConfigCheck_t;

typedef struct ED_EnemyConfig_s
{
	ED_EnemyConfigCheck_t wakeCheck;
	ED_EnemyConfigCheck_t escapeCheck;
} ED_EnemyConfig_t;

typedef struct ED_Inject_s
{
	lwlib_String_t actorName;
	int tilex;
	int tiley;
	bool enabled;
} ED_Inject_t;

typedef struct ED_Sound_s
{
	lwlib_String_t name;
	lwlib_String_t soundFileName;
	int maxTileDist;
} ED_Sound_t;

typedef vec_t(ED_Sound_t) ED_SoundSeq_t;

typedef enum ED_AmmoType_e
{
	ammo_none=-1,
	ammo_bullet,
	ammo_rifle,
	ammo_stg44,
	ammo_flame,
	ammo_chaingun,
	ammo_rocket,
	ammo_shotgun,
	ammo_fuel,
	ammo_antidote,
	NUMAMMOTYPES
} ED_AmmoType_t;

typedef struct ED_AmmoBonus_s
{
	ED_AmmoType_t ammoType;
	int ammo;
} ED_AmmoBonus_t;

typedef enum ED_WeaponType_e
{
	wp_none=-1,
	wp_cyringe,
	wp_knife,
	wp_pistol,
	wp_kar98k,
	wp_sniper,
	wp_mp40,
	wp_stg44,
	wp_shotgun,
	wp_flamethrower,
	wp_chaingun,
	wp_panzerschreck,
	wp_wirecutters,
	wp_mg42_mounted,
	wp_mecha,
	NUMWEAPONS
} ED_WeaponType_t;

typedef struct ED_Bonus_s
{
	lwlib_String_t name;
	int healthPoints;
	bool isKey;
	int key;
	lwlib_String_t message;
	lwlib_String_t sound;
	int scorePoints;
	bool isTreasure;
	ED_AmmoBonus_t ammoBonuses[3];
	ED_WeaponType_t weapon;
	int armorPoints;
	bool armorJumpTo;
	bool regenHealth;
	bool regenArmor;
	int lives;
	bool usedSprite;
	lwlib_String_t usedSpriteDef;
	int damagePoints;
	bool usePickup;
	lwlib_String_t logChoice;
	int medkits;
	bool increaseMedkitLimit;
	bool increaseAmmoLimit;
	bool giveBackpackAmmo;
} ED_Bonus_t;

typedef vec_t(ED_Bonus_t) ED_BonusSeq_t;

typedef int ED_FloorTile_t;

typedef vec_t(ED_FloorTile_t) ED_FloorTileSeq_t;

typedef struct ED_FootStep_s
{
	ED_FloorTileSeq_t floorTiles;
	ED_SoundDef_t soundDef;
} ED_FootStep_t;

typedef vec_t(ED_FootStep_t) ED_FootStepSeq_t;

typedef struct ED_AnimatedWallTile_s
{
	int animStartTile;
	int animNumTiles;
	int animTics;
} ED_AnimatedWallTile_t;

typedef vec_t(ED_AnimatedWallTile_t) ED_AnimatedWallTileSeq_t;

typedef struct ED_VSWAPInfo_s
{
	int skyStart;
	int wallPatchStart;
	int glowPatchStart, glowPatchEnd;
	ED_AnimatedWallTileSeq_t animatedWallTiles;
	int animatedWallTileMap[256]; // must match MAXWALLTILES!
} ED_VSWAPInfo_t;

typedef struct ED_TileInfo_s
{
	int firstWallPatchActor;
	int lastWallPatchActor;
	ED_FloorTileSeq_t burrowFloorTiles;
	int firstTruckTile;
	int lastTruckTile;
	int woodenDoorStartTile;
	int waterFloorTile;
} ED_TileInfo_t;

typedef enum ED_AmbientType_e
{
	amb_area,
	amb_rain,
	amb_raincover,
	amb_water,
	amb_scenery,
} ED_AmbientType_t;

typedef int ED_AmbientActor_t;

typedef vec_t(ED_AmbientActor_t) ED_AmbientActorSeq_t;

typedef struct ED_Ambient_s
{
	bool looped;
	ED_SoundDef_t soundDef;
	int actor;
	int minTimePeriod;
	int maxTimePeriod;
	int fadeInTime;
	int fadeOutTime;
	bool distanceScaled;
	ED_AmbientType_t type;
	ED_AmbientActorSeq_t actors;
	int chan;
	int32_t timer;
	int32_t timePeriod;
} ED_Ambient_t;

typedef vec_t(ED_Ambient_t) ED_AmbientSeq_t;

typedef lwlib_String_t ED_RicochetName_t;

typedef struct ED_Ricochet_s
{
	ED_RicochetName_t name;
	ED_StateKey_t state;
	ED_SoundDef_t soundDef;
} ED_Ricochet_t;

typedef vec_t(ED_Ricochet_t) ED_RicochetSeq_t;

typedef struct ED_WallAttr_s
{
	ED_RicochetName_t ricochetName;
	bool breakableWall;
	ED_Ricochet_t *ricochet;
} ED_WallAttr_t;

typedef vec_t(ED_WallAttr_t) ED_WallAttrSeq_t;

typedef lwlib::Mat<double, ED_VISFACTOR_MATRIX_SIZE> ED_VisFactorMat_t;

typedef struct ED_AmbientSettings_s
{
	int radius;
	double scale;
	double hiddenScale;
	double convPasses;
	double doorVisFactor;
	lwlib::Ray3f earRay;
	int ox, oy;
	double u, v;
} ED_AmbientSettings_t;

typedef struct ED_s
{
	void *priv;
} ED_t;

ED_t ED_New(void);

ED_t ED_Free(ED_t ed);

void ED_LaunchMenu(ED_t ed);

const char *ED_SpriteName(ED_t ed, int index);

int ED_NumSprites(ED_t ed);

short ED_SpriteNum(ED_t ed, const char *name);

ED_StatInfo_t *ED_GetStatInfos(ED_t ed);

int ED_ActorStatInfo(ED_t ed, int actor);

ED_SpriteSeq_t *ED_Sprites(ED_t ed);

ED_StateIndex_t ED_StateIndex(ED_t ed, const char *name);

ED_State_t *ED_State(ED_t ed, const char *name);

ED_State_t *ED_States(ED_t ed);

ED_State_t *ED_StateByIndex(ED_t ed, ED_StateIndex_t stateIndex);

ED_ThinkFunc_t ED_ThinkFunc(ED_t ed, const char *name);

ED_Spawn_t *ED_ActorSpawn(ED_t ed, int actor);

ED_Spawn_t *ED_SpawnByEnemyName(ED_t ed, const char *name, int dir);

ED_EnemyIndex_t ED_EnemyIndex(ED_t ed, const char *name);

ED_Enemy_t *ED_Enemy(ED_t ed, const char *name);

ED_Enemy_t *ED_Enemies(ED_t ed);

ED_Enemy_t *ED_EnemyByIndex(ED_t ed, ED_EnemyIndex_t enemyIndex);

void ED_SetVideoMode(ED_t ed);

void ED_InitGl(ED_t ed);

ED_EnemyConfig_t *ED_EnemyConfig(ED_t ed);

ED_ProjIndex_t ED_ProjIndex(ED_t ed, const char *name);

const ED_Proj_t *ED_Proj(ED_t ed, const char *name);

const ED_Proj_t *ED_Projs(ED_t ed);

const ED_Proj_t *ED_ProjByIndex(ED_t ed, ED_ProjIndex_t projIndex);

ED_Inject_t *ED_Inject(ED_t ed);

int ED_ProjPhysIndex(ED_t ed, const char *name);

const ED_ProjPhys_t *ED_ProjPhys(ED_t ed, const char *name);

int ED_StatInfoByName(ED_t ed, const char *name);

int ED_FindFreeActor(ED_t ed);

bool ED_AddStatInfo(ED_t ed, ED_StatInfo_t statInfo);

ED_SoundIndex_t ED_SoundIndex(ED_t ed, const char *name);

ED_Sound_t *ED_Sound(ED_t ed, const char *name);

ED_Sound_t *ED_Sounds(ED_t ed);

int ED_NumSounds(ED_t ed);

ED_BonusIndex_t ED_BonusIndex(ED_t ed, const char *name);

ED_Bonus_t *ED_Bonus(ED_t ed, const char *name);

ED_Bonus_t *ED_Bonuses(ED_t ed);

int ED_NumBonuses(ED_t ed);

const char *ED_FootStepByFloorTile(ED_t ed, int floorTile);

int ED_ReplaceSprite(const ED_Enemy_t *enemy, int shapenum);

ED_VSWAPInfo_t *ED_GetVSWAPInfo(ED_t ed);

ED_TileInfo_t *ED_GetTileInfo(ED_t ed);

int ED_FirstPatchTile(ED_t ed);

int ED_LastPatchTile(ED_t ed);

bool ED_PatchPicIsGlow(int patchpic);

int ED_AnimateWallTile(ED_t ed, int tile);

void ED_StatInfoReport(ED_t ed);

int ED_StatInfoPickerTile(ED_t ed, int statInfoNum);

int ED_SpawnPickerTile(ED_t ed, int spawnNum);

lwlib_Str_t ED_SpawnPickerInfoString(ED_t ed, int spawnNum);

int ED_MoveStatInfo(ED_t ed, int statInfoNum, int move);

ED_SpawnSeq_t ED_Spawns(ED_t ed);

void ED_InsertState(ED_t ed, ED_State_t state);

void ED_InsertSpawn(ED_t ed, ED_Spawn_t spawn);

void ED_InsertEnemy(ED_t ed, ED_Enemy_t enemy);

void ED_InsertSound(ED_t ed, ED_Sound_t sound);

ED_Ambient_t *ED_ActorAmbient(ED_t ed, int tile);

ED_Ambient_t *ED_Ambients(ED_t ed);

int ED_NumAmbients(ED_t ed);

void ED_ResetAmbients(ED_t ed);

ED_Ambient_t *ED_AmbientByName(ED_t ed, const char *soundName);

void ED_RandomizeAmbientTimePeriod(ED_t ed, int ambientNum);

bool ED_ActorIsWallPatch(ED_t ed, int actor);

int ED_ActorToWallPatch(ED_t ed, int actor);

bool ED_CanBurrowFloor(ED_t ed, int floor);

bool ED_FloorIsWater(ED_t ed, int floor);

double ED_RandomConst_Eval(const ED_RandomConst_t *randomConst);

int ED_RandomInt_Eval(const ED_RandomInt_t *randomInt);

bool ED_TileIsTruckTile(ED_t ed, int tile);

ED_SoundIndex_t ED_GetThemedSound(ED_t ed, const ED_SoundDef_t soundDef);

ED_WallAttr_t *ED_WallAttr(ED_t ed, int tile);

ED_Ricochet_t *ED_WallRicochet(ED_t ed, int tile);

bool ED_ActorIsBreakableWall(ED_t ed, int actor);

int ED_LRPack(ED_t ed);

ED_AmbientSettings_t *ED_AmbientSettings(ED_t ed);

#endif
