#include "version.h"

#ifdef USE_FLOORCEILINGTEX

#include "wl_def.h"
#include "wl_shade.h"
#include "fixedptc.h"

static inline byte GetFlatMapTile(int curx, int cury, fixed gu, fixed gv, int ceil)
{
    return (flatmap[curx][cury] >> (ceil ? 8 : 0)) & 0xff;
}

//
// WSJ: apply any animation if used
//
unsigned GetFlatTexture (unsigned texnum)
{
    unsigned flattex = texnum;

    switch (texnum)
    {
        case 185:
        case 221:
            flattex+=(gamestate.TimeCount/10)%3;
            break;
    }

    return flattex;
}

// Textured Floor and Ceiling by DarkOne
// With multi-textured floors and ceilings stored in lower and upper bytes of
// according tile in third mapplane, respectively.
void DrawFloorAndCeiling(byte *vbuf, unsigned vbufPitch, int min_wallheight)
{
    int x, i;
    fixed dist;                                // distance to row projection
    fixed tex_step;                            // global step per one screen pixel
    fixed gu, gv, du, dv;                      // global texture coordinates
    int u, v;                                  // local texture coordinates
    byte *tex;
    byte *toptex;
    int yd;
    int y;
    int offset;
    int bot, offsetadd;
    byte *curshades;

    int halfheight = viewheight >> 1;
    int y0 = min_wallheight >> 3;              // starting y value
    if(y0 > halfheight)
        return;                                // view obscured by walls
    if(!y0) y0 = 1;                            // don't let division by zero
    unsigned bot_offset0 = vbufPitch * (halfheight + WallMidY(y0, 1));
    unsigned top_offset0 = vbufPitch * (halfheight - WallMidY(y0, -1) - 1);
    fixed zoomoffset = focallength/0x5700;    // WSJ: zoom-in texture fix

    // draw horizontal lines
    for (i = 0; i < 2; i++)
    {
        bot = 1 - i * 2;
        for(
            // init
            y = WallMidY(y0, bot), 
            offset = i ? top_offset0 : bot_offset0; 
            // check
            y < halfheight; 
            // iterate
            y++, 
            offset += bot * vbufPitch)
        {
            dist = (heightnumerator / InvWallMidY(y, bot)) << 5;
            gu =  viewx + FixedMul(dist, viewcos);
            gv = -viewy + FixedMul(dist, viewsin);
            tex_step = (dist << 8) / viewwidth / 175;
            if (zoomoffset > 1) tex_step /= (fixed)(zoomoffset*0.645);    // WSJ: zoom-in texture fix
            du =  FixedMul(tex_step, viewsin);
            dv = -FixedMul(tex_step, viewcos);
            gu -= (viewwidth >> 1) * du;
            gv -= (viewwidth >> 1) * dv; // starting point (leftmost)

            curshades = shadetable[ShadeGlow::off][GetShade(InvWallMidY(y, bot) << 3)];
            for(x = 0, offsetadd = offset; x < viewwidth; 
                x++, offsetadd++)
            {
                yd = wallheight[x] >> 3;
                if(y >= WallMidY(yd, bot))
                {
                    const int curx = (gu >> TILESHIFT) & (MAPSIZE - 1);
                    const int cury = (-(gv >> TILESHIFT) - 1) & (MAPSIZE - 1);

                    unsigned curtex = GetFlatMapTile(curx, cury, gu, gv, i);
                    if(curtex)
                    {
                        curtex = GetFlatTexture(curtex);
                        tex = PM_GetTexture(curtex);

                        u = (gu >> (TILESHIFT - TEXTURESHIFT)) & 
                            (TEXTURESIZE - 1);
                        v = (gv >> (TILESHIFT - TEXTURESHIFT)) & 
                            (TEXTURESIZE - 1);
                        unsigned texoffs = (u << TEXTURESHIFT) + 
                            (TEXTURESIZE - 1) - v;
                        if (tex[texoffs] != 0xff)
                        {
                            if (i == 0 && lightningtics)
                            {
                                unsigned curtoptex = GetFlatMapTile(curx, 
                                    cury, gu, gv, 1);
                                if (curtoptex)
                                {
                                    curtoptex = GetFlatTexture(curtoptex);
                                    toptex = PM_GetTexture(curtoptex);
                                    if (toptex[texoffs] == 0xff)
                                    {
                                        vbuf[offsetadd] = tex[texoffs];
                                    }
                                    else
                                    {
                                        vbuf[offsetadd] = 
                                            curshades[tex[texoffs]];
                                    }
                                }
                                else
                                {
                                    vbuf[offsetadd] = tex[texoffs];
                                }
                            }
                            else
                            {
                                vbuf[offsetadd] = curshades[tex[texoffs]];
                            }
                        }
                    }
                }
                gu += du;
                gv += dv;
            }
        }
    }
}

//==========================================================================

static flat_t MapSpotToFlat(int x, int y)
{
    return (flat_t)MAPSPOT(x, y, 2);
}

flat_t GetGameLevelFlat(int x, int y)
{
    return MapSpotToFlat(x, y);
}

#endif
