// WL_GAME.C

#include <math.h>
#include "wl_def.h"
#include <SDL_mixer.h>
#pragma hdrstop

#ifdef MYPROFILE
#include <TIME.H>
#endif


/*
=============================================================================

                             LOCAL CONSTANTS

=============================================================================
*/


/*
=============================================================================

                             GLOBAL VARIABLES

=============================================================================
*/

boolean         ingame,fizzlein;
gametype        gamestate;
byte            bordercol=VIEWCOLOR;        // color of the Change View/Ingame border

byte            searchdata[SEARCHDATASZ];
byte*           sdqstart = searchdata + maparea;

#ifdef SPEAR
int32_t         spearx,speary;
unsigned        spearangle;
boolean         spearflag;
#endif

#ifdef USE_FEATUREFLAGS
int ffDataTopLeft, ffDataTopRight, ffDataBottomLeft, ffDataBottomRight,
    ffDataMusic, ffDataBossMusic, ffDataSecretLevel, ffDataThunderVol;
#endif

//
// ELEVATOR BACK MAPS - REMEMBER (-1)!!
//
// WSJ: no longer used
//
int ElevatorBackTo[]={1,1,7,3,5,3};

void SetupGameLevel (void);
void DrawPlayScreen (void);
void LoadLatchMem (void);
void GameLoop (void);

#ifdef DEBUGKEYS
boolean infinitelives;    // WSJ: used for tedlevel
#endif

lwlib_EnumStr(
    surpriseatktype,
    "surpriseatk_default",
    "surpriseatk_hit",
    "surpriseatk_delay",
    "surpriseatk_prox",
    "surpriseatk_prox_delay",
    "surpriseatk_out_of_sight",
    );

/*
=============================================================================

                             LOCAL VARIABLES

=============================================================================
*/



//===========================================================================
//===========================================================================


/*
==========================
=
= SetSoundLoc - Given the location of an object (in terms of global
=       coordinates, held in globalsoundx and globalsoundy), munges the values
=       for an approximate distance from the left and right ear, and puts
=       those values into leftchannel and rightchannel.
=
= JAB
=
==========================
*/

int leftchannel, rightchannel, channeldist;
#define ATABLEMAX 15
byte righttable[ATABLEMAX][ATABLEMAX * 2] = {
{ 8, 8, 8, 8, 8, 8, 8, 7, 7, 7, 7, 7, 7, 6, 0, 0, 0, 0, 0, 1, 3, 5, 8, 8, 8, 8, 8, 8, 8, 8},
{ 8, 8, 8, 8, 8, 8, 8, 7, 7, 7, 7, 7, 6, 4, 0, 0, 0, 0, 0, 2, 4, 6, 8, 8, 8, 8, 8, 8, 8, 8},
{ 8, 8, 8, 8, 8, 8, 8, 7, 7, 7, 7, 6, 6, 4, 1, 0, 0, 0, 1, 2, 4, 6, 8, 8, 8, 8, 8, 8, 8, 8},
{ 8, 8, 8, 8, 8, 8, 8, 7, 7, 7, 7, 6, 5, 4, 2, 1, 0, 1, 2, 3, 5, 7, 8, 8, 8, 8, 8, 8, 8, 8},
{ 8, 8, 8, 8, 8, 8, 8, 8, 7, 7, 7, 6, 5, 4, 3, 2, 2, 3, 3, 5, 6, 8, 8, 8, 8, 8, 8, 8, 8, 8},
{ 8, 8, 8, 8, 8, 8, 8, 8, 7, 7, 7, 6, 6, 5, 4, 4, 4, 4, 5, 6, 7, 8, 8, 8, 8, 8, 8, 8, 8, 8},
{ 8, 8, 8, 8, 8, 8, 8, 8, 8, 7, 7, 7, 6, 6, 5, 5, 5, 6, 6, 7, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8},
{ 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 7, 7, 7, 6, 6, 7, 7, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8},
{ 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8},
{ 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8},
{ 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8},
{ 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8},
{ 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8},
{ 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8},
{ 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8}
};
byte lefttable[ATABLEMAX][ATABLEMAX * 2] = {
{ 8, 8, 8, 8, 8, 8, 8, 8, 5, 3, 1, 0, 0, 0, 0, 0, 6, 7, 7, 7, 7, 7, 7, 8, 8, 8, 8, 8, 8, 8},
{ 8, 8, 8, 8, 8, 8, 8, 8, 6, 4, 2, 0, 0, 0, 0, 0, 4, 6, 7, 7, 7, 7, 7, 8, 8, 8, 8, 8, 8, 8},
{ 8, 8, 8, 8, 8, 8, 8, 8, 6, 4, 2, 1, 0, 0, 0, 1, 4, 6, 6, 7, 7, 7, 7, 8, 8, 8, 8, 8, 8, 8},
{ 8, 8, 8, 8, 8, 8, 8, 8, 7, 5, 3, 2, 1, 0, 1, 2, 4, 5, 6, 7, 7, 7, 7, 8, 8, 8, 8, 8, 8, 8},
{ 8, 8, 8, 8, 8, 8, 8, 8, 8, 6, 5, 3, 3, 2, 2, 3, 4, 5, 6, 7, 7, 7, 8, 8, 8, 8, 8, 8, 8, 8},
{ 8, 8, 8, 8, 8, 8, 8, 8, 8, 7, 6, 5, 4, 4, 4, 4, 5, 6, 6, 7, 7, 7, 8, 8, 8, 8, 8, 8, 8, 8},
{ 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 7, 6, 6, 5, 5, 5, 6, 6, 7, 7, 7, 8, 8, 8, 8, 8, 8, 8, 8, 8},
{ 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 7, 7, 6, 6, 7, 7, 7, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8},
{ 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8},
{ 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8},
{ 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8},
{ 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8},
{ 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8},
{ 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8},
{ 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8}
};

void
SetSoundLoc(fixed gx,fixed gy,int maxTileDist)
{
    fixed   xt,yt;
    int     x,y;
    fixed   dist;

//
// translate point to view centered coordinates
//
    gx -= viewx;
    gy -= viewy;

//
// calculate newx
//
    xt = FixedMul(gx,viewcos);
    yt = FixedMul(gy,viewsin);
    x = (xt - yt) >> TILESHIFT;

//
// calculate newy
//
    xt = FixedMul(gx,viewsin);
    yt = FixedMul(gy,viewcos);
    y = (yt + xt) >> TILESHIFT;

    if (y >= ATABLEMAX)
        y = ATABLEMAX - 1;
    else if (y <= -ATABLEMAX)
        y = -ATABLEMAX;
    if (x < 0)
        x = -x;
    if (x >= ATABLEMAX)
        x = ATABLEMAX - 1;
    leftchannel  =  lefttable[x][y + ATABLEMAX];
    rightchannel = righttable[x][y + ATABLEMAX];

    if (maxTileDist >= 0)
    {
        dist = fixedpt_sqrt(fixedpt_mul(gx, gx) + fixedpt_mul(gy, gy));
        if (maxTileDist == 0)
        {
            maxTileDist = 64;
            dist *= 2;
        }
        if (dist >= FP(maxTileDist))
        {
            dist = FP(maxTileDist);
        }
        channeldist = fixedpt_toint(dist * 255 / maxTileDist);
    }
    else
    {
        channeldist = 0;
    }

#if 0
    CenterWindow(8,1);
    US_PrintSigned(leftchannel);
    US_Print(",");
    US_PrintSigned(rightchannel);
    VW_UpdateScreen();
#endif
}

/*
==========================
=
= SetSoundLocGlobal - Sets up globalsoundx & globalsoundy and then calls
=       UpdateSoundLoc() to transform that into relative channel volumes. Those
=       values are then passed to the Sound Manager so that they'll be used for
=       the next sound played (if possible).
=
= JAB
=
==========================
*/
int PlaySoundLocGlobal(word s,fixed gx,fixed gy,int volume,
    bool foreground,bool noDistAtt)
{
    ASM_SoundPlay_t soundPlay;
    ED_Sound_t *edSound;

    lwlib_Zero(soundPlay);
    soundPlay.sound = s;
    soundPlay.posOn = true;
    soundPlay.posGx = gx;
    soundPlay.posGy = gy;
    soundPlay.vol = Vol_Mono(volume);
    soundPlay.foreground = foreground;

    int channel = SD_PlaySoundEx(soundPlay);
    if(channel != -1)
    {
        edSound = &ED_Sounds(ed)[soundPlay.sound];

        channelSoundPos[channel].globalsoundx = gx;
        channelSoundPos[channel].globalsoundy = gy;
        channelSoundPos[channel].maxTileDist = edSound->maxTileDist;
        channelSoundPos[channel].valid = 1;
    }

    return channel;
}

void UpdateSoundLoc(void)
{
#ifdef VIEASM
    for(int i = 0; i < ASM_CurChannels(); i++)
#else
    for(int i = 0; i < MIX_CHANNELS; i++)
#endif
    {
        if(channelSoundPos[i].valid)
        {
            SetSoundLoc(channelSoundPos[i].globalsoundx,
                channelSoundPos[i].globalsoundy,
                channelSoundPos[i].maxTileDist);
            SD_SetPosition(i, leftchannel, rightchannel, channeldist);
        }
    }

    //ASM_UpdateForeground();
}

/*
**      JAB End
*/

/*
==========================
=
= PlaceValidArea
=
==========================
*/

void PlaceValidArea (int tilex, int tiley)
{
    word *map,tile;
    map = mapsegs[0]+(tiley<<mapshift)+tilex;
    tile = *map;
    if (!VALIDAREATILE(tile))
    {
        tilemap[tilex][tiley] = 0;

        if (VALIDAREATILE(*(map+1)))
            tile = *(map+1);
        else if (VALIDAREATILE(*(map-mapwidth)))
            tile = *(map-mapwidth);
        else if (VALIDAREATILE(*(map+mapwidth)))
            tile = *(map+mapwidth);
        else if (VALIDAREATILE(*(map-1)))
            tile = *(map-1);
        else
            tile = AREATILE;    // WSJ: if all else fails, use the first floor code

        *map = tile;
    }
}

#ifdef DEBUG_INJECT
static void InjectTestObj(int x, int y, word *start)
{
	int type;
    const ED_Inject_t *inject;
    const ED_Spawn_t *spawn;

    inject = ED_Inject(ed);
    if (inject->tilex == x && inject->tiley == y && inject->enabled)
    {
        type = ED_StatInfoByName(ed, inject->actorName);
        spawn = ED_SpawnByEnemyName(ed, inject->actorName, -1);

        if (type != -1)
        {
            *start = ED_STATINFO[type].actor;
        }
        else if (spawn != NULL)
        {
            *start = spawn->actor;
        }
    }
}
#endif

static int RandomizeActor(int tile, int x, int y)
{
    // en_tablemut -> en_tablemutdamage
    // 30% chance
    if (tile == 227 && US_RndT() < 96)
    {
        tile = 512;
    }

    // Operating Table 1 -> en_clothmutdamage
    if (tile == 147)
    {
        tile = 513;
    }

    if (gamestate.mapon >= 8)
    {
        struct
        {
            int first[5];
            int numFirst;
            int second[5]; 
            int numSecond;
            int chanceTo;
            int chanceFrom;
        } maps[] =
        {
            // en_cagemut -> Cage 2 or Cage 3
            // 30% chance
            { { 225 }, 1, { 408, 409 }, 2, 30, 30 },

            // en_wellmut -> Well
            // 50% chance
            { { 232 }, 1, { 193 }, 1, 50, 50 },

            // en_tablemut -> en_tablemutdamage
            // 70% chance
            { { 227 }, 1, { 512 }, 1, 70, 70 },

            // en_hangedmut or en_hangedmut2 -> en_damagemut
            // 50% chance
            { { 238, 239 }, 2, { 285 }, 1, 50, 50 },

            // en_domemut -> Broken Dome
            // 30% chance
            { { 241 }, 1, { 514 }, 1, 30, 30 },

            // en_knightmut -> Knight
            // 30% chance
            { { 230 }, 1, { 334 }, 1, 30, 30 },

            // en_clothmut -> en_clothmutdamage
            // 70% chance
            { { 228 }, 1, { 513 }, 1, 70, 70 },
        };
        const int numMaps = sizeof(maps) / sizeof(maps[0]);

        for (int i = 0; i < numMaps; i++)
        {
            int *found, *start, *end;

            start = maps[i].first;
            end = start + maps[i].numFirst;
            found = std::find(start, end, tile);
            if (found != end && US_RndT_Range(0, 100) < maps[i].chanceTo)
            {
                tile = maps[i].second[US_RndT_Range(0, 
                    maps[i].numSecond - 1)];
                break;
            }

            start = maps[i].second;
            end = start + maps[i].numSecond;
            found = std::find(start, end, tile);
            if (found != end && US_RndT_Range(0, 100) < 
                maps[i].chanceFrom)
            {
                tile = maps[i].first[US_RndT_Range(0, 
                    maps[i].numFirst - 1)];
                MAPSPOT(x,y,0) = AREATILE + (US_RndT() % surpriseatk_max);
                break;
            }
        }
    }

    return tile;
}

/*
==========================
=
= ScanInfoPlane
=
= Spawn all actors and mark down special places
=
==========================
*/

static void ScanInfoPlane(void)
{
    unsigned x,y;
    int      tile;
    word     *start;

    start = mapsegs[1];
    for (y=0;y<mapheight;y++)
    {
        for (x=0;x<mapwidth;x++)
        {
            #ifdef DEBUG_INJECT
            InjectTestObj(x, y, start);
            #endif

            tile = *start++;
            if (!tile)
                continue;

            switch (tile)
            {
                case PLAYERSPAWNTILE:
                case PLAYERSPAWNTILE + 1:
                case PLAYERSPAWNTILE + 2:
                case PLAYERSPAWNTILE + 3:
                    SpawnPlayer(x, y, NORTH + tile - PLAYERSPAWNTILE);
                    break;

                case PUSHABLETILE: // P wall
                    if (!loadedgame)
                        gamestate.secrettotal++;
                    break;

                case DEADGUARDTILE:
                    SpawnSimpleStatic (x,y,ED_SPR("SPR_GRD_DEAD"));
                    break;

                default:
                    tile = RandomizeActor(tile, x, y);
                    SpawnActor(x, y, tile);
                    break;
            }
        }
    }
}

//==========================================================================

/*
==================
=
= SetupGameLevel
=
==================
*/

void SetupGameLevel (void)
{
    int  x,y;
    word *map;
    word tile;
    word patchspot;

    if (!loadedgame)
    {
        gamestate.TimeCount
            = gamestate.secrettotal
            = gamestate.killtotal
            = gamestate.treasuretotal
            = gamestate.secretcount
            = gamestate.killcount
            = gamestate.treasurecount
            = alertedarea = lastalertedarea
            = pwallstate = pwallpos = facetimes = 0;
        killerobj = NULL;
    }

    if (demoplayback || demorecord)
        US_InitRndT (false);
    else
        US_InitRndT (true);

//
// load the level
//
    CA_CacheMap (gamestate.mapon+10*gamestate.episode);
    mapon-=gamestate.episode*10;

#ifdef USE_FEATUREFLAGS
    // Temporary definition to make things clearer
    #define MXX MAPSIZE - 1

    // Read feature flags data from map corners and overwrite corners with adjacent tiles
    ffDataTopLeft     = MAPSPOT(0,   0,   0); MAPSPOT(0,   0,   0) = MAPSPOT(1,       0,       0);
    ffDataTopRight    = MAPSPOT(MXX, 0,   0); MAPSPOT(MXX, 0,   0) = MAPSPOT(MXX,     1,       0);
    ffDataBottomRight = MAPSPOT(MXX, MXX, 0); MAPSPOT(MXX, MXX, 0) = MAPSPOT(MXX - 1, MXX,     0);
    ffDataBottomLeft  = MAPSPOT(0,   MXX, 0); MAPSPOT(0,   MXX, 0) = MAPSPOT(0,       MXX - 1, 0);
    ffDataMusic       = MAPSPOT(0,   1,   0); MAPSPOT(0,   1,   0) = MAPSPOT(1,       1,       0);
    ffDataBossMusic   = MAPSPOT(0,   2,   0); MAPSPOT(0,   2,   0) = MAPSPOT(1,       2,       0);
    ffDataSecretLevel = MAPSPOT(0,   3,   0); MAPSPOT(0,   3,   0) = MAPSPOT(1,       3,       0);
    if (GetFeatureFlags() & FF_THUNDER)
    {
        ffDataThunderVol  = MAPSPOT(1, MXX, 0);
        MAPSPOT(1, MXX, 0) = MAPSPOT(1, MXX - 1, 0);
    }
    else
    {
        ffDataThunderVol = 0;
    }

    #undef MXX
#endif

//
// copy the wall data to a data segment array
//
    memset (tilemap,0,sizeof(tilemap));
    memset (wallpatch,0,sizeof(wallpatch));
    memset (actorat,0,sizeof(actorat));
    memset (flatmap,0,sizeof(flatmap));
    memset (ambiance,0,sizeof(ambiance));
    memset (tileflags,0,sizeof(tileflags));
    map = mapsegs[0];
    for (y=0;y<mapheight;y++)
    {
        for (x=0;x<mapwidth;x++)
        {
            flatmap[x][y] = GetGameLevelFlat(x, y);

            tile = *map++;
            if (tile<AREATILE)
            {
                // solid wall
                tilemap[x][y] = (byte) tile;
                actorat[x][y] = (objtype *)(uintptr_t) tile;

                patchspot = MAPSPOT(x,y,1);
                if (ED_ActorIsWallPatch(ed, patchspot))
                {
                    wallpatch[x][y] = ED_ActorToWallPatch(ed, patchspot);
                }
            }
            else
            {
                // area floor
                tilemap[x][y] = 0;
                actorat[x][y] = 0;
                wallpatch[x][y] = 0;
            }
        }
    }

//
// spawn doors
//
    InitActorList ();                       // start spawning things with a clean slate
    InitDoorList ();
    InitStaticList ();

    map = mapsegs[0];
    for (y=0;y<mapheight;y++)
    {
        for (x=0;x<mapwidth;x++)
        {
            tile = *map++;
            if (tile >= DOORTILE_FIRST && tile <= DOORTILE_LAST)
            {
                if (tile % 2 == 0)
                {
                    SpawnDoor (x, y, 1, (tile - DOORTILE_FIRST) / 2);
                }
                else
                {
                    SpawnDoor (x, y, 0,
                        (tile - DOORTILE_FIRST - 1) / 2);
                }
            }
        }
    }

//
// spawn actors
//
    ScanInfoPlane ();


//
// take out the ambush markers
//
    map = mapsegs[0];
    for (y=0;y<mapheight;y++)
    {
        for (x=0;x<mapwidth;x++)
        {
            if (!tilemap[x][y] && MAPSPOT(x, y, 0) == AMBUSHTILE)
                PlaceValidArea(x,y);
        }
    }


//
// have the caching manager load and purge stuff to make sure all marks
// are in memory
//
    CA_LoadAllSounds ();

// make sure all ambients are reset since no ambient sounds should be
// playing when level is loaded or started
    ED_ResetAmbients(ed);
}


//==========================================================================


/*
===================
=
= DrawPlayBorderSides
=
= To fix window overwrites
=
===================
*/
void DrawPlayBorderSides(void)
{
    if(viewsize == 21) return;

	const int sw = screenWidth;
	const int sh = screenHeight;
	const int vw = viewwidth;
	const int vh = viewheight;
	const int px = scaleFactor; // size of one "pixel"

	const int h  = sh - px * STATUSLINES;
	const int xl = sw / 2 - vw / 2;
	const int yl = (h - vh) / 2;

    if(xl != 0)
    {
	    VWB_BarScaledCoord(0,            0, xl - px,     h, bordercol);                 // left side
	    VWB_BarScaledCoord(xl + vw + px, 0, xl - px * 2, h, bordercol);                 // right side
    }

    if(yl != 0)
    {
	    VWB_BarScaledCoord(0, 0,            sw, yl - px, bordercol);                    // upper side
	    VWB_BarScaledCoord(0, yl + vh + px, sw, yl - px, bordercol);                    // lower side
    }

    if(xl != 0)
    {
        // Paint game view border lines
	    VWB_BarScaledCoord(xl - px, yl - px, vw + px, px,          0);                      // upper border
	    VWB_BarScaledCoord(xl,      yl + vh, vw + px, px,          bordercol - 2);          // lower border
	    VWB_BarScaledCoord(xl - px, yl - px, px,      vh + px,     0);                      // left border
	    VWB_BarScaledCoord(xl + vw, yl - px, px,      vh + px * 2, bordercol - 2);          // right border
	    VWB_BarScaledCoord(xl - px, yl + vh, px,      px,          bordercol - 3);          // lower left highlight
    }
    else
    {
        // Just paint a lower border line
        VWB_BarScaledCoord(0, yl+vh, vw, px, bordercol-2);       // lower border
    }
}


/*
===================
=
= DrawStatusBorder
=
===================
*/

void DrawStatusBorder (byte color)
{
    int statusborderw = SCREEN_XOFF;

    VWB_BarScaledCoord (0,0,screenWidth,screenHeight-scaleFactor*(STATUSLINES-3),color);
    VWB_BarScaledCoord (0,screenHeight-scaleFactor*(STATUSLINES-3),
        statusborderw+scaleFactor*8,scaleFactor*(STATUSLINES-4),color);
    VWB_BarScaledCoord (0,screenHeight-scaleFactor*2,screenWidth,scaleFactor*2,color);
    VWB_BarScaledCoord (screenWidth-statusborderw-scaleFactor*8, screenHeight-scaleFactor*(STATUSLINES-3),
        statusborderw+scaleFactor*8,scaleFactor*(STATUSLINES-4),color);

    VWB_BarScaledCoord (statusborderw+scaleFactor*9, screenHeight-scaleFactor*3,
        scaleFactor*97, scaleFactor*1, color-1);
    VWB_BarScaledCoord (statusborderw+scaleFactor*106, screenHeight-scaleFactor*3,
        scaleFactor*161, scaleFactor*1, color-2);
    VWB_BarScaledCoord (statusborderw+scaleFactor*267, screenHeight-scaleFactor*3,
        scaleFactor*44, scaleFactor*1, color-3);
    VWB_BarScaledCoord (screenWidth-statusborderw-scaleFactor*9, screenHeight-scaleFactor*(STATUSLINES-4),
        scaleFactor*1, scaleFactor*20, color-2);
    VWB_BarScaledCoord (screenWidth-statusborderw-scaleFactor*9, screenHeight-scaleFactor*(STATUSLINES/2-4),
        scaleFactor*1, scaleFactor*14, color-3);
}


/*
===================
=
= DrawPlayBorder
=
===================
*/

void DrawPlayBorder (void)
{
	const int px = scaleFactor; // size of one "pixel"

    if (bordercol != VIEWCOLOR)
        DrawStatusBorder(bordercol);
    else
    {
        const int statusborderw = (screenWidth-px*320)/2;
        VWB_BarScaledCoord (0, screenHeight-px*STATUSLINES,
            statusborderw+px*8, px*STATUSLINES, bordercol);
        VWB_BarScaledCoord (screenWidth-statusborderw-px*8, screenHeight-px*STATUSLINES,
            statusborderw+px*8, px*STATUSLINES, bordercol);
    }

    if((unsigned) viewheight == screenHeight) return;

    VWB_BarScaledCoord (0,0,screenWidth,screenHeight-px*STATUSLINES,bordercol);

    const int xl = screenWidth/2-viewwidth/2;
    const int yl = (screenHeight-px*STATUSLINES-viewheight)/2;
    VWB_BarScaledCoord (xl,yl,viewwidth,viewheight,0);

    if(xl != 0)
    {
        // Paint game view border lines
        VWB_BarScaledCoord(xl-px, yl-px, viewwidth+px, px, 0);                      // upper border
        VWB_BarScaledCoord(xl, yl+viewheight, viewwidth+px, px, bordercol-2);       // lower border
        VWB_BarScaledCoord(xl-px, yl-px, px, viewheight+px, 0);                     // left border
        VWB_BarScaledCoord(xl+viewwidth, yl-px, px, viewheight+2*px, bordercol-2);  // right border
        VWB_BarScaledCoord(xl-px, yl+viewheight, px, px, bordercol-3);              // lower left highlight
    }
    else
    {
        // Just paint a lower border line
        VWB_BarScaledCoord(0, yl+viewheight, viewwidth, px, bordercol-2);       // lower border
    }
}


/*
===================
=
= DrawPlayScreen
=
===================
*/

void DrawPlayScreen (void)
{
    if (Mecha_Equipped())
    {
        LatchDrawPicScaledCoord (SCREEN_XOFF, screenHeight - scaleFactor * 
            STATUSLINES, MECHASTATUSBARPIC, 1);
    }
    else
    {
        VWB_DrawPicScaledCoord (SCREEN_XOFF, screenHeight-scaleFactor * 
            STATUSLINES, STATUSBARPIC);
    }
    DrawPlayBorder ();

    DrawFace ();
    DrawHealth ();
    DrawLives ();
    DrawLevel ();
    DrawAmmo ();
    DrawKeys ();
    DrawWeapon ();
    DrawScore ();
}

void ShowActStatus()
{
    // Draw status bar without borders
    byte *source = grsegs[STATUSBARPIC];
    int	picnum = STATUSBARPIC - STARTPICS;
    int width = pictable[picnum].width;
    int height = pictable[picnum].height;
    int destx = SCREEN_XOFF + 9 * scaleFactor;
    int desty = screenHeight - (height - 4) * scaleFactor;
    VL_MemToScreenScaledCoord(source, width, height, 9, 4, destx, desty, width - 18, height - 7);

    ingame = false;
    DrawFace ();
    DrawHealth ();
    DrawLives ();
    DrawLevel ();
    DrawAmmo ();
    DrawKeys ();
    DrawWeapon ();
    DrawScore ();
    ingame = true;
}


//==========================================================================

/*
==================
=
= StartDemoRecord
=
==================
*/

char    demoname[13] = "DEMO?.";

#ifndef REMDEBUG
#define MAXDEMOSIZE     8192

void StartDemoRecord (int levelnumber)
{
    demobuffer=malloc(MAXDEMOSIZE);
    CHECKMALLOCRESULT(demobuffer);
    demoptr = (int8_t *) demobuffer;
    lastdemoptr = demoptr+MAXDEMOSIZE;

    *demoptr = levelnumber;
    demoptr += 4;                           // leave space for length
    demorecord = true;
}


/*
==================
=
= FinishDemoRecord
=
==================
*/

void FinishDemoRecord (void)
{
    int32_t    length,level;

    demorecord = false;

    length = (int32_t) (demoptr - (int8_t *)demobuffer);

    demoptr = ((int8_t *)demobuffer)+1;
    demoptr[0] = (int8_t) length;
    demoptr[1] = (int8_t) (length >> 8);
    demoptr[2] = 0;

    VW_FadeIn();
    CenterWindow(24,3);
    PrintY+=6;
    fontnumber=0;
    SETFONTCOLOR(0,15);
    US_Print(" Demo number (0-9): ");
    VW_UpdateScreen();

    if (US_LineInput (px,py,str,NULL,true,1,0))
    {
        level = atoi (str);
        if (level>=0 && level<=9)
        {
            demoname[4] = (char)('0'+level);
            CA_WriteFile (demoname,demobuffer,length);
        }
    }

    free(demobuffer);
}

//==========================================================================

/*
==================
=
= RecordDemo
=
= Fades the screen out, then starts a demo.  Exits with the screen faded
=
==================
*/

void RecordDemo (void)
{
    int level,esc,maps;

    CenterWindow(26,3);
    PrintY+=6;
    CA_CacheGrChunk(STARTFONT);
    fontnumber=0;
    SETFONTCOLOR(0,15);
#ifndef SPEAR
#ifdef UPLOAD
    US_Print("  Demo which level(1-10): "); maps = 10;
#else
    US_Print("  Demo which level(1-60): "); maps = 60;
#endif
#else
    US_Print("  Demo which level(1-21): "); maps = 21;
#endif
    VW_UpdateScreen();
    VW_FadeIn ();
    esc = !US_LineInput (px,py,str,NULL,true,2,0);
    if (esc)
        return;

    level = atoi (str);
    level--;

    if (level >= maps || level < 0)
        return;

    VW_FadeOut ();

#ifndef SEAMLESSLEVELS
    NewGame (gd_hard,level/10);
    gamestate.mapon = level%10;
#else
    NewGame (gd_hard,0);
    gamestate.mapon = level;
#endif

    StartDemoRecord (level);

    DrawPlayScreen ();
    VW_FadeIn ();

    startgame = false;
    demorecord = true;

    SetupGameLevel ();
    StartMusic ();

    fizzlein = true;

    PlayLoop ();

    demoplayback = false;

    StopMusic ();
    VW_FadeOut ();
    ClearMemory ();

    FinishDemoRecord ();
}
#else
void FinishDemoRecord (void) {return;}
void RecordDemo (void) {return;}
#endif



//==========================================================================

/*
==================
=
= PlayDemo
=
= Fades the screen out, then starts a demo.  Exits with the screen unfaded
=
==================
*/

void PlayDemo (int demonumber)
{
    int length;
#ifdef DEMOSEXTERN
// debug: load chunk
#ifndef SPEARDEMO
    int dems[4]={T_DEMO0,T_DEMO1,T_DEMO2,T_DEMO3};
#else
    int dems[1]={T_DEMO0};
#endif

    CA_CacheGrChunk(dems[demonumber]);
    demoptr = (int8_t *) grsegs[dems[demonumber]];
#else
    demoname[4] = '0'+demonumber;
    CA_LoadFile (demoname,&demobuffer);
    demoptr = (int8_t *)demobuffer;
#endif

    NewGame (1,0);
    gamestate.mapon = *demoptr++;
    gamestate.difficulty = gd_hard;
    length = READWORD(*(uint8_t **)&demoptr);
    // TODO: Seems like the original demo format supports 16 MB demos
    //       But T_DEM00 and T_DEM01 of Wolf have a 0xd8 as third length size...
    demoptr++;
    lastdemoptr = demoptr-4+length;

    VW_FadeOut ();

    SETFONTCOLOR(0,15);
    DrawPlayScreen ();

    startgame = false;
    demoplayback = true;

    SetupGameLevel ();
    StartMusic ();

    PlayLoop ();

#ifdef DEMOSEXTERN
    UNCACHEGRCHUNK(dems[demonumber]);
#else
    MM_FreePtr (&demobuffer);
#endif

    demoplayback = false;

    StopMusic ();
    ClearMemory ();
}

//==========================================================================

/*
==================
=
= Died
=
==================
*/

#define DEATHROTATE 2

void Died (void)
{
    float   fangle;
    int32_t dx,dy;
    int     iangle,curangle,clockwise,counter,change;
    int     fizzlecolor;

    if (screenfaded)
    {
        ThreeDRefresh ();
        VW_FadeIn ();
    }

    fizzlecolor = gamestate.infected ? GAMEPAL_BLOODGREEN : 
        GAMEPAL_BLOODRED;

    DiscardWeaponsAtEnd();
    DiscardAmmoAtEnd();
    gamestate.weapon = (weapontype) -1;                     // take away weapon
    gamestate.activebosses = 0;
    gamestate.bosskeydrop = 0;
    gamestate.bossmusicowner = 0;
    gamestate.guardedObjNum = 0;
    messagetime = 0;
    Player_ClearInfection();
    SD_PlaySound (ED_SND("PLAYERDEATHSND"));

    //
    // swing around to face attacker
    //
    if(killerobj)
    {
        dx = OB_GetX(killerobj) - OB_GetX(player);
        dy = OB_GetY(player) - OB_GetY(killerobj);

        fangle = (float) atan2((float) dy, (float) dx);     // returns -pi to pi
        if (fangle<0)
            fangle = (float) (M_PI*2+fangle);

        iangle = (int) (fangle/(M_PI*2)*ANGLES);
    }
    else
    {
        iangle = player->angle + ANGLES / 2;
        if(iangle >= ANGLES) iangle -= ANGLES;
    }

    if (player->angle > iangle)
    {
        counter = player->angle - iangle;
        clockwise = ANGLES-player->angle + iangle;
    }
    else
    {
        clockwise = iangle - player->angle;
        counter = player->angle + ANGLES-iangle;
    }

    curangle = player->angle;

    if (clockwise<counter)
    {
        //
        // rotate clockwise
        //
        if (curangle>iangle)
            curangle -= ANGLES;
        do
        {
            change = tics*DEATHROTATE;
            if (curangle + change > iangle)
                change = iangle-curangle;

            curangle += change;
            player->angle += change;
            if (player->angle >= ANGLES)
                player->angle -= ANGLES;
            player->fineangle = player->angle * FINEANGLES / ANGLES;

            ThreeDRefresh ();
            CalcTics ();
        } while (curangle != iangle);
    }
    else
    {
        //
        // rotate counterclockwise
        //
        if (curangle<iangle)
            curangle += ANGLES;
        do
        {
            change = -(int)tics*DEATHROTATE;
            if (curangle + change < iangle)
                change = iangle-curangle;

            curangle += change;
            player->angle += change;
            if (player->angle < 0)
                player->angle += ANGLES;
            player->fineangle = player->angle * FINEANGLES / ANGLES;

            ThreeDRefresh ();
            CalcTics ();
        } while (curangle != iangle);
    }

    //
    // fade to red
    //
    FinishPaletteShifts ();

    VL_BarScaledCoord (viewscreenx,viewscreeny,viewwidth,viewheight,
        fizzlecolor);

    IN_ClearKeysDown ();

    FizzleFade(screenBuffer,screenSurf,viewscreenx,viewscreeny,viewwidth,viewheight,70,false);

    IN_UserInput(100);
    SD_WaitSoundDone ();
    ClearMemory();

#ifdef USE_LIVES
#ifdef DEBUGKEYS
    if (!infinitelives)
#endif
        gamestate.lives--;
#endif

#ifdef USE_LIVES
    if (gamestate.lives > -1)
    {
#endif
        DiedReset();
        memset(gamestate.ambiance, 0, sizeof(gamestate.ambiance));

        if(viewsize != 21)
        {
            DrawPlayScreen();
        }
#ifdef USE_LIVES
    }
#endif
}

//==========================================================================

/*
===================
=
= GameLoop
=
===================
*/

void GameLoop (void)
{
    boolean died;
#ifdef MYPROFILE
    clock_t start,end;
#endif

restartgame:
    ClearMemory ();
    SETFONTCOLOR(0,15);
    VW_FadeOut();
    if (startgame)
    {
        IntermissionScreens();
    }
    DrawPlayScreen ();
    died = false;
    do
    {
        if (!loadedgame)
            gamestate.score = gamestate.oldscore;
        if(!died || viewsize != 21) DrawScore();

        startgame = false;
        if (!loadedgame)
            SetupGameLevel ();

#ifdef SPEAR
        if (gamestate.mapon == 20)      // give them the key allways
        {
            gamestate.keys |= 1;
            DrawKeys ();
        }
#endif

        ingame = true;
        if(loadedgame)
        {
            ContinueMusic(lastgamemusicoffset);
            loadedgame = false;
        }
        else StartMusic ();

        if (!died)
            PreloadGraphics ();             // TODO: Let this do something useful!
        else
            died = false;

        fizzlein = true;

        DrawLevel ();

#ifdef SPEAR
startplayloop:
#endif
        PlayLoop ();

#ifdef SPEAR
        if (spearflag)
        {
            SD_StopSound();
            SD_PlaySound(ED_SND("GETSPEARSND"));
            if (DigiMode != sds_Off)
            {
                Delay(150);
            }
            else
                SD_WaitSoundDone();

            ClearMemory ();
            gamestate.oldscore = gamestate.score;
            gamestate.mapon = 20;
            SetupGameLevel ();
            StartMusic ();
            OB_SetX(player, spearx);
            OB_SetY(player, speary);
            player->angle = (short)spearangle;
            player->fineangle = player->angle * FINEANGLES / ANGLES;
            spearflag = false;
            Thrust (0,0);
            goto startplayloop;
        }
#endif

        StopMusic ();
        ingame = false;

        if (demorecord && playstate != ex_warped)
            FinishDemoRecord ();

        if (startgame || loadedgame)
            goto restartgame;

        switch (playstate)
        {
            case ex_completed:
            case ex_secretlevel:
                if(viewsize == 21) DrawPlayScreen();
                gamestate.keys = 0;
                gamestate.activebosses = 0;
                gamestate.bosskeydrop = 0;
                gamestate.bossmusicowner = 0;
                gamestate.guardedObjNum = 0;
                messagetime = 0;
                Player_ClearInfection();
                DrawKeys ();
                memset(gamestate.ambiance, 0, sizeof(gamestate.ambiance));
                VW_FadeOut ();

                ClearMemory ();

                LevelCompleted ();              // do the intermission
                DiscardWeaponsAtEnd();
                DiscardAmmoAtEnd();
                if ((GetFeatureFlags() & FF_WEAPONRESET) != 0)
                {
                    DiedReset();
                }
                if(viewsize == 21) DrawPlayScreen();

#ifdef SPEARDEMO
                if (gamestate.mapon == 1)
                {
                    died = true;                    // don't "get psyched!"

                    VW_FadeOut ();

                    ClearMemory ();

                    CheckHighScore (gamestate.score,gamestate.mapon+1);
#ifndef JAPAN
                    strcpy(MainMenu[viewscores].string,STR_VS);
#endif
                    MainMenu[viewscores].routine = CP_ViewScores;
                    return;
                }
#endif

#ifdef JAPDEMO
                if (gamestate.mapon == 3)
                {
                    died = true;                    // don't "get psyched!"

                    VW_FadeOut ();

                    ClearMemory ();

                    CheckHighScore (gamestate.score,gamestate.mapon+1);
#ifndef JAPAN
                    strcpy(MainMenu[viewscores].string,STR_VS);
#endif
                    MainMenu[viewscores].routine = CP_ViewScores;
                    return;
                }
#endif

                gamestate.oldscore = gamestate.score;

                //
                // GOING TO SECRET LEVEL
                //
                if (playstate == ex_secretlevel && 
                    ffDataSecretLevel > 0)
                {
                    gamestate.altmapon = gamestate.mapon+1;
                    gamestate.mapon = ffDataSecretLevel - 1;
                }
                else
                    //
                    // COMING BACK FROM SECRET LEVEL
                    //
                    if (gamestate.altmapon > 0)
                    {
                        gamestate.mapon = gamestate.altmapon;    // back from secret
                        gamestate.altmapon = 0;
                    }
                    else
                        //
                        // GOING TO NEXT LEVEL
                        //
                        gamestate.mapon++;
#ifdef AUTOINTER
                IntermissionScreens();                   // WSJ: INTART screen if file exists
                ClearMemory();                           // WSJ: "
                DrawPlayScreen();                        // WSJ: "
#endif
                break;

            case ex_died:
                Died ();
                died = true;                    // don't "get psyched!"

#ifdef USE_LIVES
                if (gamestate.lives > -1)
#endif
                    break;                          // more lives left

                VW_FadeOut ();
                if(screenHeight % 200 != 0)
                    VL_ClearScreen(0);

                ClearMemory ();

                CheckHighScore (gamestate.score,gamestate.mapon+1);
#ifndef JAPAN
                strcpy(MainMenu[viewscores].string,STR_VS);
#endif
                MainMenu[viewscores].routine = CP_ViewScores;
                return;

            case ex_victorious:
                if(viewsize == 21) DrawPlayScreen();
#ifndef SEAMLESSLEVELS
                VW_FadeOut ();
#else
                //VL_FadeOut (0,255,0,17,17,300);
                VL_FadeOut (0,255,0,0,0,200);
#endif
                ClearMemory ();

                Victory ();

                ClearMemory ();

                CheckHighScore (gamestate.score,gamestate.mapon+1);
#ifndef JAPAN
                strcpy(MainMenu[viewscores].string,STR_VS);
#endif
                MainMenu[viewscores].routine = CP_ViewScores;
                return;

            default:
                if(viewsize == 21) DrawPlayScreen();
                ClearMemory ();
                break;
        }
    } while (1);
}
