#include <set>
#include "wl_def.h"

namespace SnowMutants
{
    void defineTable(byte *table)
    {
        // identity
        for (int i = 0; i < 256; i++)
        {
            table[i] = i;
        }
        // clothes
        table[0x1E] = 0x1B;
        table[0x1F] = 0x1C;
        table[0x6E] = 0x11;
        table[0x6F] = 0x15;
        table[0xEE] = 0x18;
        table[0xA0] = 0x1B;
        table[0xDF] = 0x1F;
        table[0xED] = 0x17;
        // face
        table[0x11] = 0x72;
        table[0x13] = 0x77;
        table[0x14] = 0x77;
        table[0x15] = 0x78;
        table[0x16] = 0x7A;
        table[0x17] = 0x79;
        table[0x18] = 0x7B;
        table[0x19] = 0x7B;
        table[0x1A] = 0x7B;
        table[0x1B] = 0x7D;
        table[0x1D] = 0x1A;
    }

    static void convertSprite(int shapenum, byte *table)
    {
        int x, y, i;
        word *cmdptr;
        t_compshape *shape;
        byte *sprite;
        short *linecmds;

        shape = (t_compshape *)PM_GetSprite(shapenum);
        sprite = (byte *)shape;

        cmdptr = shape->dataofs;
        for (x = shape->leftpix; x <= shape->rightpix; x++)
        {
            linecmds = (short *)(sprite + *cmdptr++);
            for (; *linecmds; linecmds += 3)
            {
                i = linecmds[2] / 2 + linecmds[1];
                for (y = linecmds[2] / 2; y < linecmds[0] / 2; y++, i++)
                {
                    sprite[i] = table[sprite[i]];
                }
            }
        }
    }

    static void addRange(std::set<int> &sprites, int first, int last)
    {
        for (int i = first; i <= last; i++)
        {
            sprites.insert(sprites.begin(), i);
        }
    }

    static void defineSprites(std::set<int> &sprites)
    {
        addRange(sprites, 306, 372);
        addRange(sprites, 930, 958);
        addRange(sprites, 959, 978);
        addRange(sprites, 1018, 1037);
        addRange(sprites, 1087, 1092);
        addRange(sprites, 1097, 1102);
        addRange(sprites, 1598, 1605);
        addRange(sprites, 1610, 1613);
        addRange(sprites, 2307, 2335);
    }

    void convertSprites(void)
    {
        byte table[256];
        defineTable(table);

        std::set<int> sprites;
        defineSprites(sprites);

        for (int i = PMSpriteStart; i < PMSoundStart; i++)
        {
            const int shapenum = i - PMSpriteStart;
            if (sprites.count(shapenum))
            {
                convertSprite(shapenum, table);
            }
        }
    }
}
