// WL_STATE.C

#include "wl_def.h"
#include "lw.h"
#pragma hdrstop

/*
=============================================================================

                            LOCAL CONSTANTS

=============================================================================
*/

#define SKELETOVHEALTH SKILLAMOUNT(50)

/*
=============================================================================

                            GLOBAL VARIABLES

=============================================================================
*/


static const dirtype opposite[9] =
    {west,southwest,south,southeast,east,northeast,north,northwest,nodir};

static const dirtype diagonal[9][9] =
{
    /* east */  {nodir,nodir,northeast,nodir,nodir,nodir,southeast,nodir,nodir},
                {nodir,nodir,nodir,nodir,nodir,nodir,nodir,nodir,nodir},
    /* north */ {northeast,nodir,nodir,nodir,northwest,nodir,nodir,nodir,nodir},
                {nodir,nodir,nodir,nodir,nodir,nodir,nodir,nodir,nodir},
    /* west */  {nodir,nodir,northwest,nodir,nodir,nodir,southwest,nodir,nodir},
                {nodir,nodir,nodir,nodir,nodir,nodir,nodir,nodir,nodir},
    /* south */ {southeast,nodir,nodir,nodir,southwest,nodir,nodir,nodir,nodir},
                {nodir,nodir,nodir,nodir,nodir,nodir,nodir,nodir,nodir},
                {nodir,nodir,nodir,nodir,nodir,nodir,nodir,nodir,nodir}
};



void    SpawnNewObj (unsigned tilex, unsigned tiley, statetype *state);
void    NewState (objtype *ob, statetype *state);

boolean TryWalk (objtype *ob);
void    MoveObj (objtype *ob, int32_t move);

void    KillActor (objtype *ob, objtype *killer, damage_t damagetype);
boolean DamageActor (objtype *ob, objtype *attacker, unsigned damage, damage_t damagetype);

boolean CheckLine (objtype *ob);
void    FirstSighting (objtype *ob);
boolean CheckSight (objtype *ob);

/*
=============================================================================

                                LOCAL VARIABLES

=============================================================================
*/



//===========================================================================

//#define DEBUG_SPAWNOBJ

/*
===================
=
= SpawnNewObj
=
= Spaws a new actor at the given TILE coordinates, with the given state, and
= the given size in GLOBAL units.
=
= newobj = a pointer to an initialized new actor
=
===================
*/

void SpawnNewObj (unsigned tilex, unsigned tiley, statetype *state)
{
    #ifdef DEBUG_SPAWNOBJ
    fprintf(stderr, "spawning obj at (%d,%d) with state %s\n", (int)tilex, (int)tiley, state->name);
    #endif
    GetNewActor ();
    OB_SetState(newobj, state);
    if (state->tictime)
        newobj->ticcount = DEMOCHOOSE_ORIG_SDL(
                US_RndT () % state->tictime,
                US_RndT () % state->tictime + 1);     // Chris' moonwalk bugfix ;D
    else
        newobj->ticcount = 0;

    newobj->tilex = (short) tilex;
    newobj->tiley = (short) tiley;
    OB_SetX(newobj, ((int32_t)tilex<<TILESHIFT)+TILEGLOBAL/2);
    OB_SetY(newobj, ((int32_t)tiley<<TILESHIFT)+TILEGLOBAL/2);
    newobj->dir = nodir;

    actorat[tilex][tiley] = newobj;
    PlaceValidArea(newobj->tilex,newobj->tiley);
    newobj->areanumber = AreaAt(newobj->tilex,newobj->tiley);
}



/*
===================
=
= NewState
=
= Changes ob to a new state, setting ticcount to the max for that state
=
===================
*/

void NewState (objtype *ob, statetype *state)
{
    OB_SetState(ob, state);
    if (state != NULL)
    {
        ob->ticcount = state->tictime;
    }
}



/*
=============================================================================

                        ENEMY TILE WORLD MOVEMENT CODE

=============================================================================
*/


/*
==================================
=
= TryWalk
=
= Attempts to move ob in its current (ob->dir) direction.
=
= If blocked by either a wall or an actor returns FALSE
=
= If move is either clear or blocked only by a door, returns TRUE and sets
=
= ob->tilex         = new destination
= ob->tiley
= ob->areanumber    = the floor tile number (0-(NUMAREAS-1)) of destination
= ob->distance      = TILEGLOBAl, or -doornumber if a door is blocking the way
=
= If a door is in the way, an OpenDoor call is made to start it opening.
= The actor code should wait until
=       doorobjlist[-ob->distance].action = dr_open, meaning the door has been
=       fully opened
=
==================================
*/

boolean CheckEnemyBlock (objtype *ob, int x, int y)
{
    word tile = MAPSPOT(x,y,1);

    if ((ob->hidden == 2 || (ob->flags & FL_UNDERGROUND)) &&
        !ValidGround(ob,x,y))
    {
        return true;
    }

    switch (tile)
    {
        case MG42TILE:
            if (!CanUseMG42(ob) && !ob->enemy->mg42.canWalkThrough)
                return true;
            break;
        case ENEMYBLOCK:
            return true;
        case KEY1ENEMYBLOCK:
        case KEY2ENEMYBLOCK:
        case KEY3ENEMYBLOCK:
        case KEY4ENEMYBLOCK:
            if ( !(gamestate.keys & (1<<(tile-KEY1ENEMYBLOCK))) )
                return true;
            break;
    }

    if (OB_GuardedObjBlocksMove(ob, x, y))
    {
        return true;
    }

    return false;
}

static bool DoorCheck(objtype *ob, int x, int y, int *pdoornum)
{
    uintptr_t temp;
    int doornum;

    temp = (uintptr_t)actorat[x][y];

    doornum = (int)temp & 127;
    if (doorobjlist[doornum].lock >= dr_lock1 && 
        doorobjlist[doornum].lock <= dr_lock4 && 
        !(gamestate.keys & (1 << (doorobjlist[doornum].lock - dr_lock1))))
    {
        *pdoornum = doornum;
        return false;
    }
    OpenDoor(doornum);
    ob->waitdoornum = doornum + 1;

    *pdoornum = doornum;
    return true;
}

static bool CheckSide(objtype *ob, int x, int y, int *pdoornum)
{
    uintptr_t temp;
    objtype *check;

    if (CheckEnemyBlock(ob,x,y))
    {
        return false;
    }

    temp = (uintptr_t)actorat[x][y];
    if (temp)
    {
        if (pdoornum != NULL)
        {
            if (temp < 128)
            {
                return false;
            }
        }
        if (temp < 256)
        {
            if (pdoornum == NULL ||
                !DoorCheck(ob, x, y, pdoornum))
            {
                return false;
            }
        }
        else
        {
            check = (objtype *)temp;
            if ((check->flags & FL_BLOCK) != 0 && 
                !OB_CanUseGoalObj(ob, x, y))
            {
                return false;
            }
        }
    }

    return true;
}


boolean TryWalk (objtype *ob)
{
    int doornum = -1;

    switch (ob->dir)
    {
        case north:
            if (ob->enemy->cannotOpenDoors || ob->hidden == 2)
            {
                if (!CheckSide(ob, ob->tilex, ob->tiley-1, NULL))
                {
                    return false;
                }
            }
            else
            {
                if (!CheckSide(ob, ob->tilex,ob->tiley-1,&doornum))
                {
                    return false;
                }
            }
            ob->tiley--;
            break;

        case northeast:
            if (!CheckSide(ob, ob->tilex+1,ob->tiley-1, NULL))
            {
                return false;
            }
            if (!CheckSide(ob, ob->tilex+1,ob->tiley, NULL))
            {
                return false;
            }
            if (!CheckSide(ob, ob->tilex,ob->tiley-1, NULL))
            {
                return false;
            }
            ob->tilex++;
            ob->tiley--;
            break;

        case east:
            if (ob->enemy->cannotOpenDoors || ob->hidden == 2)
            {
                if (!CheckSide(ob, ob->tilex+1,ob->tiley, NULL))
                {
                    return false;
                }
            }
            else
            {
                if (!CheckSide(ob, ob->tilex+1,ob->tiley,&doornum))
                {
                    return false;
                }
            }
            ob->tilex++;
            break;

        case southeast:
            if (!CheckSide(ob, ob->tilex+1,ob->tiley+1, NULL))
            {
                return false;
            }
            if (!CheckSide(ob, ob->tilex+1,ob->tiley, NULL))
            {
                return false;
            }
            if (!CheckSide(ob, ob->tilex,ob->tiley+1, NULL))
            {
                return false;
            }
            ob->tilex++;
            ob->tiley++;
            break;

        case south:
            if (ob->enemy->cannotOpenDoors || ob->hidden == 2)
            {
                if (!CheckSide(ob, ob->tilex,ob->tiley+1, NULL))
                {
                    return false;
                }
            }
            else
            {
                if (!CheckSide(ob, ob->tilex,ob->tiley+1,&doornum))
                {
                    return false;
                }
            }
            ob->tiley++;
            break;

        case southwest:
            if (!CheckSide(ob, ob->tilex-1,ob->tiley+1, NULL))
            {
                return false;
            }
            if (!CheckSide(ob, ob->tilex-1,ob->tiley, NULL))
            {
                return false;
            }
            if (!CheckSide(ob, ob->tilex,ob->tiley+1, NULL))
            {
                return false;
            }
            ob->tilex--;
            ob->tiley++;
            break;

        case west:
            if (ob->enemy->cannotOpenDoors || ob->hidden == 2)
            {
                if (!CheckSide(ob, ob->tilex-1,ob->tiley, NULL))
                {
                    return false;
                }
            }
            else
            {
                if (!CheckSide(ob, ob->tilex-1,ob->tiley,&doornum))
                {
                    return false;
                }
            }
            ob->tilex--;
            break;

        case northwest:
            if (!CheckSide(ob, ob->tilex-1,ob->tiley-1, NULL))
            {
                return false;
            }
            if (!CheckSide(ob, ob->tilex-1,ob->tiley, NULL))
            {
                return false;
            }
            if (!CheckSide(ob, ob->tilex,ob->tiley-1, NULL))
            {
                return false;
            }
            ob->tilex--;
            ob->tiley--;
            break;

        case nodir:
            return false;

        default:
            Quit ("Walk: Bad dir");
    }

    if (VALIDAREA(ob->tilex,ob->tiley))
    {
        ob->areanumber = AreaAt(ob->tilex,ob->tiley);
    }

    ob->distance = TILEGLOBAL;
    return true;
}

static boolean CheckTile1(objtype *ob, int x, int y)
{
    uintptr_t t;

    t = (uintptr_t)actorat[x][y];
    return
        (
            !CheckEnemyBlock(ob,x,y) && 
            (
                t == 0 || 
                ((t >= 128) && (t < 256)) || 
                (
                    ISPOINTER(actorat[x][y]) && 
                    !(
                        (actorat[x][y]->flags & FL_BLOCK) != 0 &&
                        !OB_CanUseGoalObj(ob, x, y)
                    )
                )
            )
        );
}

static boolean CheckTile2(objtype *ob, int x, int y)
{
    uintptr_t t;
    t = (uintptr_t)actorat[x][y];
    return 
        (
            !CheckEnemyBlock(ob,x,y) && 
            (
                (t == 0) || 
                (
                    ISPOINTER(actorat[x][y]) && 
                    !(actorat[x][y]->flags&FL_BLOCK)
                )
            )
        );
}

namespace SearchMemChecker
{
    class Exception
    {
    };

    class Pointer
    {
        byte *p;

    public:
        Pointer() : p(NULL)
        {
        }

        Pointer(byte *p_) : p(p_)
        {
        }

        byte *operator->()
        {
            if (p - searchdata < 0 || p - searchdata >= SEARCHDATASZ)
            {
                throw Exception();
            }
            return p;
        }

        byte &operator*()
        {
            if (p - searchdata < 0 || p - searchdata >= SEARCHDATASZ)
            {
                throw Exception();
            }
            return *p;
        }

        const byte &operator*() const
        {
            if (p - searchdata < 0 || p - searchdata >= SEARCHDATASZ)
            {
                throw Exception();
            }
            return *p;
        }

        Pointer &operator++()
        {
            ++p;
            return *this;
        }

        Pointer operator++(int)
        {
            Pointer copy(p);
            ++p;
            return copy;
        }

        bool operator<(const Pointer &x) const
        {
            return p < x.p;
        }

        Pointer operator+(int len) const
        {
            return Pointer(p + len);
        }

        Pointer operator-(int len) const
        {
            return Pointer(p - len);
        }
    };

    boolean selectNextDir(objtype *ob, int dtilex, int dtiley)
    {
        try
        {
            return SelectNextDir(ob, dtilex, dtiley);
        }
        catch (Exception e)
        {
            memset(searchdata, 0, SEARCHDATASZ);
            ob->dir = nodir;
            return false;
        }
    }
}

/*
==================================
=
= SelectNextDir
=
= Uses a parallel search algorithm to find and select the next
= direction in as close a path as is possible to the given tile
=
= WSJ: changed to boolean
=
==================================
*/

boolean SelectNextDir(objtype *ob, int dtilex, int dtiley)
{
    SearchMemChecker::Pointer p1, p2, p3;
    byte tx, ty;
    byte tdir;
    int l, t, r, b;

    l = ob->tilex - 1;
    t = ob->tiley - 1;
    r = ob->tilex + 1;
    b = ob->tiley + 1;
    if (ob->tilex == dtilex)
    {
        if (ob->tiley == dtiley)
        {
            ob->dir = nodir;
            return false;
        }
        else if (t == dtiley)
        {
            ob->dir = north;
            if (!TryWalk(ob))
            {
                ob->dir = nodir;
                return false;
            }
            return true;
        }
        else if (b == dtiley)
        {
            ob->dir = south;
            if (!TryWalk(ob))
            {
                ob->dir = nodir;
                return false;
            }
            return true;
        }
    }
    else if (ob->tiley == dtiley)
    {
        if (l == dtilex)
        {
            ob->dir = west;
            if (!TryWalk(ob))
            {
                ob->dir = nodir;
                return false;
            }
            return true;
        }
        else if (r == dtilex)
        {
            ob->dir = east;
            if (!TryWalk(ob))
            {
                ob->dir = nodir;
                return false;
            }
            return true;
        }
    }
    p1 = searchdata;
    p3 = searchdata + ((ob->tiley << mapshift) + ob->tilex);
    for (; p1 < p3; p1++)
        *p1 = 1;
    *(p1++) = 0;
    for (; p1 < sdqstart; p1++)
        *p1 = 1;
    p2 = p1;
    if (ob->enemy->cannotOpenDoors || ob->hidden == 2)
    {
        if (ob->tiley > 0)
        {
            tx = (byte)ob->tilex;
            ty = (byte)t;
            *(p3 - mapwidth) = 0;
            if (CheckTile2(ob, tx, ty))
            {
                *(p2++) = tx;
                *(p2++) = ty;
                *(p2++) = 2;
            }
        }
        if (ob->tilex < 63)
        {
            tx = (byte)r;
            ty = (byte)ob->tiley;
            *(p3 + 1) = 0;
            if (CheckTile2(ob, tx, ty))
            {
                *(p2++) = tx;
                *(p2++) = ty;
                *(p2++) = 0;
            }
        }
        if (ob->tiley < 63)
        {
            tx = (byte)ob->tilex;
            ty = (byte)b;
            *(p3 + mapwidth) = 0;
            if (CheckTile2(ob, tx, ty))
            {
                *(p2++) = tx;
                *(p2++) = ty;
                *(p2++) = 6;
            }
        }
        if (ob->tilex > 0)
        {
            tx = (byte)l;
            ty = (byte)ob->tiley;
            *(p3 - 1) = 0;
            if (CheckTile2(ob, tx, ty))
            {
                *(p2++) = tx;
                *(p2++) = ty;
                *(p2++) = 4;
            }
        }
        for (; p1 < p2; )
        {
            tx = *(p1++);
            ty = *(p1++);
            l = tx - 1;
            t = ty - 1;
            r = tx + 1;
            b = ty + 1;
            if (tx == dtilex)
            {
                if ((t == dtiley) || (b == dtiley))
                {
                    ob->dir = (dirtype)*p1;
                    if (!TryWalk(ob))
                    {
                        ob->dir = nodir;
                        return false;
                    }
                    return true;
                }
            }
            else if (ty == dtiley)
            {
                if ((l == dtilex) || (r == dtilex))
                {
                    ob->dir = (dirtype)*p1;
                    if (!TryWalk(ob))
                    {
                        ob->dir = nodir;
                        return false;
                    }
                    return true;
                }
            }
            tdir = *(p1++);
            if (ty > 0)
            {
                p3 = searchdata + ((t << mapshift) + (int)tx);
                if (*p3)
                {
                    *p3 = 0;
                    if (CheckTile2(ob, tx, t))
                    {
                        *(p2++) = tx;
                        *(p2++) = (byte)t;
                        *(p2++) = tdir;
                    }
                }
            }
            if (tx < 63)
            {
                p3 = searchdata + (((int)ty << mapshift) + r);
                if (*p3)
                {
                    *p3 = 0;
                    if (CheckTile2(ob, r, ty))
                    {
                        *(p2++) = (byte)r;
                        *(p2++) = ty;
                        *(p2++) = tdir;
                    }
                }
            }
            if (ty < 63)
            {
                p3 = searchdata + ((b << mapshift) + (int)tx);
                if (*p3)
                {
                    if (CheckTile2(ob, tx, b))
                    {
                        *(p2++) = tx;
                        *(p2++) = (byte)b;
                        *(p2++) = tdir;
                    }
                }
            }
            if (tx > 0)
            {
                p3 = searchdata + (((int)ty << mapshift) + l);
                if (*p3)
                {
                    *p3 = 0;
                    if (CheckTile2(ob, l, ty))
                    {
                        *(p2++) = (byte)l;
                        *(p2++) = ty;
                        *(p2++) = tdir;
                    }
                }
            }
        }
    }
    else
    {
        if (ob->tiley > 0)
        {
            tx = (byte)ob->tilex;
            ty = (byte)t;
            *(p3 - mapwidth) = 0;
            if (CheckTile1(ob, tx, ty))
            {
                *(p2++) = tx;
                *(p2++) = ty;
                *(p2++) = 2;
            }
        }
        if (ob->tilex < 63)
        {
            tx = (byte)r;
            ty = (byte)ob->tiley;
            *(p3 + 1) = 0;
            if (CheckTile1(ob, tx, ty))
            {
                *(p2++) = tx;
                *(p2++) = ty;
                *(p2++) = 0;
            }
        }
        if (ob->tiley < 63)
        {
            tx = (byte)ob->tilex;
            ty = (byte)b;
            *(p3 + mapwidth) = 0;
            if (CheckTile1(ob, tx, ty))
            {
                *(p2++) = tx;
                *(p2++) = ty;
                *(p2++) = 6;
            }
        }
        if (ob->tilex > 0)
        {
            tx = (byte)l;
            ty = (byte)ob->tiley;
            *(p3 - 1) = 0;
            if (CheckTile1(ob, tx, ty))
            {
                *(p2++) = tx;
                *(p2++) = ty;
                *(p2++) = 4;
            }
        }
        for (; p1 < p2; )
        {
            tx = *(p1++);
            ty = *(p1++);
            l = tx - 1;
            t = ty - 1;
            r = tx + 1;
            b = ty + 1;
            if (tx == dtilex)
            {
                if ((t == dtiley) || (b == dtiley))
                {
                    ob->dir = (dirtype)*p1;
                    if (!TryWalk(ob))
                    {
                        ob->dir = nodir;
                        return false;
                    }
                    return true;
                }
            }
            else if (ty == dtiley)
            {
                if ((l == dtilex) || (r == dtilex))
                {
                    ob->dir = (dirtype)*p1;
                    if (!TryWalk(ob))
                    {
                        ob->dir = nodir;
                        return false;
                    }
                    return true;
                }
            }
            tdir = *(p1++);
            if (ty > 0)
            {
                p3 = searchdata + ((t << mapshift) + (int)tx);
                if (*p3)
                {
                    *p3 = 0;
                    if (CheckTile1(ob, tx, t))
                    {
                        *(p2++) = tx;
                        *(p2++) = (byte)t;
                        *(p2++) = tdir;
                    }
                }
            }
            if (tx < 63)
            {
                p3 = searchdata + (((int)ty << mapshift) + r);
                if (*p3)
                {
                    *p3 = 0;
                    if (CheckTile1(ob, r, ty))
                    {
                        *(p2++) = (byte)r;
                        *(p2++) = ty;
                        *(p2++) = tdir;
                    }
                }
            }
            if (ty < 63)
            {
                p3 = searchdata + ((b << mapshift) + (int)tx);
                if (*p3)
                {
                    if (CheckTile1(ob, tx, b))
                    {
                        *(p2++) = tx;
                        *(p2++) = (byte)b;
                        *(p2++) = tdir;
                    }
                }
            }
            if (tx > 0)
            {
                p3 = searchdata + (((int)ty << mapshift) + l);
                if (*p3)
                {
                    *p3 = 0;
                    if (CheckTile1(ob, l, ty))
                    {
                        *(p2++) = (byte)l;
                        *(p2++) = ty;
                        *(p2++) = tdir;
                    }
                }
            }
        }
    }
    ob->dir = nodir;
    return false;
}


/*
=================
=
= CheckGoal
=
=================
*/

void ResetGoal (objtype *ob)
{
    int goalObjIndex;
    objtype *goalObj;
    struct statstruct *goalitem;

    if (ob->reaction > react_dead)
    {
        if (ob->goalitemnum != 0)
        {
            goalitem = &statobjlist[ob->goalitemnum - 1];
            if (goalitem->flags & FL_FIRSTATTACK)
                goalitem->flags &= ~FL_FIRSTATTACK;
        }
        else
        {
            if (ob->objIndexArr[oit_goaltarget] != 0)
            {
                goalObjIndex = ob->objIndexArr[oit_goaltarget];
                goalObj = &objlist[goalObjIndex];

                assert(goalObj->objIndexArr[oit_goalseeker] == 
                    ob - objlist);

                ob->objIndexArr[oit_goaltarget] = 
                    goalObj->objIndexArr[oit_goalseeker] = 0;
            }
        }
    }
    ob->reaction = react_none;
}

#define MINUSEDIST 0x2000

static void CheckMG42Goal(objtype *ob)
{
    ED_State_t *mg42WaitState;
    struct statstruct *goalitem;

    assert(ob->reaction == react_findmg42 && ob->goalitemnum != 0);

    goalitem = &statobjlist[ob->goalitemnum - 1];

    boolean reachedgoal = (
        LABS(OB_GetX(ob) - TILECENTER(goalitem->tilex)) < MINUSEDIST && 
        LABS(OB_GetY(ob) - TILECENTER(goalitem->tiley)) < MINUSEDIST
        );

    if (reachedgoal)
    {
        ob->tilex = goalitem->tilex;
        ob->tiley = goalitem->tiley;
        OB_SetX(ob, TILECENTER(goalitem->tilex));
        OB_SetY(ob, TILECENTER(goalitem->tiley));

        PlaySoundLocActor(ED_SND("USEMG42SND"),ob);
        goalitem->flags |= FL_ATTACKMODE;

        mg42WaitState = ED_STATE(ob->enemy->mg42.waitState);
        if (mg42WaitState != NULL)
        {
            NewState (ob, mg42WaitState);
        }

        ResetGoal (ob);
        ob->reaction = react_usingmg42;
    }
    else if (player->tilex == goalitem->tilex && player->tiley == goalitem->tiley)
        ResetGoal (ob);
}

static void CheckGoalObj(objtype *ob)
{
    int goalObjIndex;
    objtype *goalObj;
    boolean reachedgoal;
    void (*think)(objtype *);

    assert(ob->reaction == react_findmg42 && ob->goalitemnum == 0);

    goalObjIndex = ob->objIndexArr[oit_goaltarget];
    assert(goalObjIndex != 0);

    goalObj = &objlist[goalObjIndex];
    if (!OB_CanResurrect(ob))
    {
        if (goalObj->hitpoints <= 0)
        {
            ResetGoal(ob);
            return;
        }
    }

    if (OB_TileCentred(goalObj, MINUSEDIST))
    {
        reachedgoal = OB_CheckCloserThan(ob, goalObj, MINUSEDIST);
    }
    else
    {
        reachedgoal = OB_CheckCloserThan(ob, goalObj, 0x8000);
    }
    if (reachedgoal)
    {
        OB_SetX(ob, OB_GetX(goalObj));
        OB_SetY(ob, OB_GetY(goalObj));

        think = ED_THINK(ob->enemy->reachedGoalThink);
        if (think != NULL)
        {
            think(ob);
        }
    }
}

void CheckGoal (objtype *ob)
{
    if (ob->reaction != react_findmg42)
        return;

    if (ob->goalitemnum != 0)
    {
        CheckMG42Goal(ob);
    }
    else
    {
        CheckGoalObj(ob);
    }
}


/*
==================================
=
= SelectDodgeDir
=
= Attempts to choose and initiate a movement for ob that sends it towards
= the player while dodging
=
= If there is no possible move (ob is totally surrounded)
=
= ob->dir           =       nodir
=
= Otherwise
=
= ob->dir           = new direction to follow
= ob->distance      = TILEGLOBAL or -doornumber
= ob->tilex         = new destination
= ob->tiley
= ob->areanumber    = the floor tile number (0-(NUMAREAS-1)) of destination
=
==================================
*/

void SelectDodgeDir (objtype *ob)
{
    int         deltax,deltay,i;
    unsigned    absdx,absdy;
    dirtype     dirtry[5];
    dirtype     turnaround,tdir;


    if (ob->flags & FL_FIRSTATTACK)
    {
        //
        // turning around is only ok the very first time after noticing the
        // player
        //
        turnaround = nodir;
        ob->flags &= ~FL_FIRSTATTACK;
    }
    else
        turnaround=opposite[ob->dir];

    deltax = player->tilex - ob->tilex;
    deltay = player->tiley - ob->tiley;

    //
    // arange 5 direction choices in order of preference
    // the four cardinal directions plus the diagonal straight towards
    // the player
    //

    if (deltax>0)
    {
        dirtry[1]= east;
        dirtry[3]= west;
    }
    else
    {
        dirtry[1]= west;
        dirtry[3]= east;
    }

    if (deltay>0)
    {
        dirtry[2]= south;
        dirtry[4]= north;
    }
    else
    {
        dirtry[2]= north;
        dirtry[4]= south;
    }

    if (ob->runAway)
    {
        SWAP_VALS(dirtry[1], dirtry[3], tdir);
        SWAP_VALS(dirtry[2], dirtry[4], tdir);
    }

    //
    // randomize a bit for dodging
    //
    absdx = abs(deltax);
    absdy = abs(deltay);

    if (absdx > absdy)
    {
        tdir = dirtry[1];
        dirtry[1] = dirtry[2];
        dirtry[2] = tdir;
        tdir = dirtry[3];
        dirtry[3] = dirtry[4];
        dirtry[4] = tdir;
    }

    if (US_RndT() < 128)
    {
        tdir = dirtry[1];
        dirtry[1] = dirtry[2];
        dirtry[2] = tdir;
        tdir = dirtry[3];
        dirtry[3] = dirtry[4];
        dirtry[4] = tdir;
    }

    dirtry[0] = diagonal [ dirtry[1] ] [ dirtry[2] ];

    //
    // try the directions util one works
    //
    for (i=0;i<5;i++)
    {
        if ( dirtry[i] == nodir || dirtry[i] == turnaround)
            continue;

        ob->dir = dirtry[i];
        if (TryWalk(ob))
            return;
    }

    //
    // turn around only as a last resort
    //
    if (turnaround != nodir)
    {
        ob->dir = turnaround;

        if (TryWalk(ob))
            return;
    }

    ob->dir = nodir;
}


/*
============================
=
= SelectChaseDir
=
= As SelectDodgeDir, but doesn't try to dodge
=
============================
*/

void SelectChaseDir (objtype *ob)
{
    int     deltax,deltay;
    dirtype d[3];
    dirtype tdir, olddir, turnaround;
    int goalObjIndex;
    objtype *goalObj;
    struct statstruct *goalitem;
    using namespace SearchMemChecker;

    if (ob->reaction == react_findmg42)
    {
        if (ob->goalitemnum != 0)
        {
            goalitem = &statobjlist[ob->goalitemnum - 1];
            if (selectNextDir(ob, goalitem->tilex, goalitem->tiley))
            {
                return;
            }
        }
        else
        {
            goalObjIndex = ob->objIndexArr[oit_goaltarget];
            assert(goalObjIndex != 0);

            goalObj = &objlist[goalObjIndex];

            if (selectNextDir(ob, goalObj->tilex, goalObj->tiley))
            {
                return;
            }
        }
    }
    else if (OB_CanDefendGuardedObj(ob))
    {
        objtype *guardedObj = objlist + gamestate.guardedObjNum;
        if (selectNextDir(ob, guardedObj->tilex, guardedObj->tiley))
        {
            return;
        }
        if (selectNextDir(ob, player->tilex, player->tiley))
        {
            return;
        }
    }
    else if (selectNextDir(ob, player->tilex, player->tiley))
    {
        return;
    }

    olddir=ob->dir;
    turnaround=opposite[olddir];

    deltax=player->tilex - ob->tilex;
    deltay=player->tiley - ob->tiley;

    d[1]=nodir;
    d[2]=nodir;

    if (deltax>0)
        d[1]= east;
    else if (deltax<0)
        d[1]= west;
    if (deltay>0)
        d[2]=south;
    else if (deltay<0)
        d[2]=north;

    if (abs(deltay)>abs(deltax))
    {
        tdir=d[1];
        d[1]=d[2];
        d[2]=tdir;
    }

    if (d[1]==turnaround)
        d[1]=nodir;
    if (d[2]==turnaround)
        d[2]=nodir;


    if (d[1]!=nodir)
    {
        ob->dir=d[1];
        if (TryWalk(ob))
            return;     /*either moved forward or attacked*/
    }

    if (d[2]!=nodir)
    {
        ob->dir=d[2];
        if (TryWalk(ob))
            return;
    }

    /* there is no direct path to the player, so pick another direction */

    if (olddir!=nodir)
    {
        ob->dir=olddir;
        if (TryWalk(ob))
            return;
    }

    if (US_RndT()>128)      /*randomly determine direction of search*/
    {
        for (tdir=north; tdir<=west; tdir=(dirtype)(tdir+1))
        {
            if (tdir!=turnaround)
            {
                ob->dir=tdir;
                if ( TryWalk(ob) )
                    return;
            }
        }
    }
    else
    {
        for (tdir=west; tdir>=north; tdir=(dirtype)(tdir-1))
        {
            if (tdir!=turnaround)
            {
                ob->dir=tdir;
                if ( TryWalk(ob) )
                    return;
            }
        }
    }

    if (turnaround !=  nodir)
    {
        ob->dir=turnaround;
        if (ob->dir != nodir)
        {
            if ( TryWalk(ob) )
                return;
        }
    }

    ob->dir = nodir;                // can't move
}

/*
============================
=
= SelectRunDir
=
= Run Away from player
=
============================
*/

void SelectRunDir (objtype *ob)
{
    int deltax,deltay;
    dirtype d[3];
    dirtype tdir;


    deltax=player->tilex - ob->tilex;
    deltay=player->tiley - ob->tiley;

    if (deltax<0)
        d[1]= east;
    else
        d[1]= west;
    if (deltay<0)
        d[2]=south;
    else
        d[2]=north;

    if (abs(deltay)>abs(deltax))
    {
        tdir=d[1];
        d[1]=d[2];
        d[2]=tdir;
    }

    ob->dir=d[1];
    if (TryWalk(ob))
        return;     /*either moved forward or attacked*/

    ob->dir=d[2];
    if (TryWalk(ob))
        return;

    /* there is no direct path to the player, so pick another direction */

    if (US_RndT()>128)      /*randomly determine direction of search*/
    {
        for (tdir=north; tdir<=west; tdir=(dirtype)(tdir+1))
        {
            ob->dir=tdir;
            if ( TryWalk(ob) )
                return;
        }
    }
    else
    {
        for (tdir=west; tdir>=north; tdir=(dirtype)(tdir-1))
        {
            ob->dir=tdir;
            if ( TryWalk(ob) )
                return;
        }
    }

    ob->dir = nodir;                // can't move
}


/*
=================
=
= MoveObj
=
= Moves ob be move global units in ob->dir direction
= Actors are not allowed to move inside the player
= Does NOT check to see if the move is tile map valid
=
= OB_GetX(ob)                 = adjusted for new position
= OB_GetY(ob)
=
=================
*/

void MoveObj (objtype *ob, int32_t move)
{
    MoveObjInDir (ob,move,ob->dir);
}

void MoveObjInDir (objtype *ob, int32_t move, dirtype dir)
{
    int32_t    deltax,deltay;

    switch (dir)
    {
        case north:
            OB_SetY(ob, OB_GetY(ob) - move);
            break;
        case northeast:
            OB_SetX(ob, OB_GetX(ob) + move);
            OB_SetY(ob, OB_GetY(ob) - move);
            break;
        case east:
            OB_SetX(ob, OB_GetX(ob) + move);
            break;
        case southeast:
            OB_SetX(ob, OB_GetX(ob) + move);
            OB_SetY(ob, OB_GetY(ob) + move);
            break;
        case south:
            OB_SetY(ob, OB_GetY(ob) + move);
            break;
        case southwest:
            OB_SetX(ob, OB_GetX(ob) - move);
            OB_SetY(ob, OB_GetY(ob) + move);
            break;
        case west:
            OB_SetX(ob, OB_GetX(ob) - move);
            break;
        case northwest:
            OB_SetX(ob, OB_GetX(ob) - move);
            OB_SetY(ob, OB_GetY(ob) - move);
            break;

        case nodir:
            return;

        default:
            Quit ("MoveObj: bad dir!");
    }

    //
    // check to make sure it's not on top of player
    //
    if (areabyplayer[ob->areanumber])
    {
        deltax = OB_GetX(ob) - OB_GetX(player);
        if (deltax < -MINACTORDIST || deltax > MINACTORDIST)
            goto moveok;
        deltay = OB_GetY(ob) - OB_GetY(player);
        if (deltay < -MINACTORDIST || deltay > MINACTORDIST)
            goto moveok;

        if (ob->hidden == 1)          // move closer until he meets CheckLine
            goto moveok;

        //
        // back up
        //
        switch (dir)
        {
            case north:
                OB_SetY(ob, OB_GetY(ob) + move);
                break;
            case northeast:
                OB_SetX(ob, OB_GetX(ob) - move);
                OB_SetY(ob, OB_GetY(ob) + move);
                break;
            case east:
                OB_SetX(ob, OB_GetX(ob) - move);
                break;
            case southeast:
                OB_SetX(ob, OB_GetX(ob) - move);
                OB_SetY(ob, OB_GetY(ob) - move);
                break;
            case south:
                OB_SetY(ob, OB_GetY(ob) - move);
                break;
            case southwest:
                OB_SetX(ob, OB_GetX(ob) + move);
                OB_SetY(ob, OB_GetY(ob) - move);
                break;
            case west:
                OB_SetX(ob, OB_GetX(ob) + move);
                break;
            case northwest:
                OB_SetX(ob, OB_GetX(ob) + move);
                OB_SetY(ob, OB_GetY(ob) + move);
                break;

            case nodir:
                return;
        }
        return;
    }
moveok:
    ob->distance -=move;
}

/*
=============================================================================

                                STUFF

=============================================================================
*/

bool CheckPlaysMusic(objtype *ob)
{
    int i;
    bool playBossMusic;
    ED_Boss_t *boss;

    if (gamestate.bossmusicowner != 0)
    {
        playBossMusic = (ob - objlist == gamestate.bossmusicowner);
    }
    else
    {
        playBossMusic = false;

        boss = ob->enemy->boss;
        if (boss != NULL)
        {
            if (vec_size(boss->musicMaps) == 0)
            {
                playBossMusic = (boss->lastMusicMap == 0 ||
                    gamestate.mapon <= boss->lastMusicMap);
            }
            else
            {
                for (i = 0; (size_t)i < vec_size(boss->musicMaps); i++)
                {
                    if (gamestate.mapon == vec_at(boss->musicMaps, i))
                    {
                        playBossMusic = true;
                        break;
                    }
                }
            }
        }
    }

    ED_Music_t *music = ob->enemy->music;
    if (music != NULL && music->mapNum == gamestate.mapon)
    {
        playBossMusic = true;
    }

    return playBossMusic;
}

/*
===============
=
= KillActor
=
===============
*/

void KillActor (objtype *ob, objtype *killer, damage_t damagetype)
{
    int     i;
    int     tilex,tiley;
    boolean gibdeath,burndeath,headshot,selfdestruct;
    const ED_EnemyDeath_t *death;
    const ED_DropItem_t *dropItem;
    int     dropChance, headShotChance;
    watermutobjtype *watermut;
    ED_Boss_t *boss;
    ED_Bonus_t *bonus;
    struct statstruct *goalitem;

    boss = ob->enemy->boss;
    death = &ob->enemy->death;

    headshot = gibdeath = burndeath = selfdestruct = false;

    watermut = &ob->u.watermut;
    if (ob->enemy->waterMut != NULL &&
        watermut->waterbound)
    {
        damagetype = dam_bullet;
    }

    switch (damagetype)
    {
        case dam_chaingun:
        case dam_rocket:
            if (OBJTILEDIST(ob,killer) < 3)
                gibdeath = true;
            break;
        case dam_fire:
            burndeath = true;
            break;
        case dam_rifle:
            headShotChance = US_RndT_Range(0, 100);
            if 
                (
                    killer == player &&
                    (
                        (
                            OB_InShootDistDelta(ob, 3) && headShotChance < 30
                        ) ||
                        (
                            OB_InShootDistDelta(ob, 6) && headShotChance >= 30 && 
                            headShotChance < 75
                        )
                    )
                )
            {
                headshot = true;
            }
            break;
        case dam_selfdestruct:
            selfdestruct = true;
            break;
        default:
            break;
    }

    tilex = ob->tilex = (word)(OB_GetX(ob) >> TILESHIFT);         // drop item on center
    tiley = ob->tiley = (word)(OB_GetY(ob) >> TILESHIFT);

    if (ob->reaction == react_usingmg42 && ob->goalitemnum != 0)
    {
        goalitem = &statobjlist[ob->goalitemnum - 1];
        goalitem->flags &= ~FL_ATTACKMODE;
    }
    ResetGoal (ob);

    ob->deathtyp = death_normal;
    if (burndeath && ED_STATE(death->burnState))
    {
        ob->hitpoints = ob->enemy->death.burnHealth;
        ob->speed *= 3;
        NewState (ob,ED_STATE(death->burnState));
        ob->deathtyp = death_burned;
    }
    else if (gibdeath && ED_STATE(death->gibState))
    {
        NewState (ob,ED_STATE(death->gibState));
        ob->deathtyp = death_gibbed;
    }
    else if (headshot && ED_STATE(death->headShotState))
    {
        NewState (ob,ED_STATE(death->headShotState));
        ob->deathtyp = death_headshot;
    }
    else if (selfdestruct && ED_STATE("s_mechmut4expl1"))
    {
        NewState (ob,ED_STATE("s_mechmut4expl1"));
    }
    else if (ob->enemy->waterMut != NULL &&
        watermut->waterbound)
    {
        NewState (ob,ED_STATE(ob->enemy->waterMut->dieState));
    }
    else
    {
        NewState (ob,ED_STATE(ob->enemy->death.dieState));
    }

    dropChance = US_RndT();
    for (i = 0; (size_t)i < vec_size(death->dropItems); i++)
    {
        dropItem = &vec_at(death->dropItems, i);

        // NYE key drop on level 1
        if (gamestate.mapon == 0)
        {
            bonus = ED_BONUS(dropItem->bonusDef);
            if (bonus != NULL && boss != NULL && dropItem->chance == -1)
            {
                PlaceItemType (dropItem->bonusDef, tilex, tiley);
                break;
            }
        }

        if (dropItem->bonusDef[0] != '\0' && (dropItem->chance == 0 ||
            dropChance <= dropItem->chance * 255 / 100))
        {
            bonus = ED_BONUS(dropItem->bonusDef);
            if (bonus != NULL && bonus->isKey && boss != NULL)
            {
                if (!lwlib_BitOn(gamestate.bosskeydrop, boss->index))
                {
                    PlaceItemType (dropItem->bonusDef, tilex, tiley);
                    lwlib_SetBitLValue(gamestate.bosskeydrop, 
                        boss->index, 1);
                }
            }
            else
            {
                PlaceItemType (dropItem->bonusDef, tilex, tiley);
            }
            break;
        }
    }

    if (ob->killpoints)
    {
        gamestate.killcount++;
        GivePoints (ob->killpoints);
        ob->killpoints = 0;
    }

    ob->reaction = react_dead;
    ob->flags &= ~FL_SHOOTABLE;
    ob->flags &= ~FL_ATTACKMODE;
    actorat[ob->tilex][ob->tiley] = NULL;
    ob->flags |= FL_NONMARK;

    if (CheckPlaysMusic(ob))
    {
        UpdateBossMusic(ob, false);
    }

    OB_KillActorAlert(ob);
}

/*
===================
=
= DamageDebris
=
= This will spawn debris projectiles around enemy when damaged.
=
===================
*/
void DamageDebris(objtype *ob, unsigned damage, damage_t damagetype)
{
    int i;
    int numDebris;
    int debrisHp;
    int startPuff, endPuff;
    ED_DamageDebris_t *damageDebris;

    damageDebris = ob->enemy->damageDebris;
    if (damageDebris != NULL && (damagetype == dam_bullet || 
        damagetype == dam_chaingun || damagetype == dam_rocket))
    {
        debrisHp = damageDebris->hitpoints;
        startPuff = (ob->hitpoints + debrisHp) / debrisHp;
        endPuff = (ob->hitpoints + debrisHp - (short)damage) / debrisHp;
        if (startPuff > endPuff)
        {
            numDebris = startPuff - endPuff;
            for (i = 0; i < numDebris; i++)
            {
                ProjFire(ob, 0, ED_Proj(ed, damageDebris->projName),
                    NULL, 0, 0);
            }
        }
    }
}


/*
===================
=
= DamageActor
=
= Called when the player succesfully hits an enemy.
=
= Does damage points to enemy ob, either putting it into a stun frame or
= killing it.
=
= WSJ: changed to a boolean in case the enemy was immune to the damage
=
===================
*/

boolean DamageActor (objtype *ob, objtype *attacker, unsigned damage, damage_t damagetype)
{
    ED_State_t *painState;

    madenoise = true;

    //if (damagetype == dam_rocket)
        //damage *= 2;

    //
    // do double damage if shooting a non attack mode actor
    //
    if ( !(ob->flags & FL_ATTACKMODE) )
    {
        if (damagetype == dam_melee)
        {
            damage = lwlib_MIN(ob->hitpoints, 80);
        }
        else
        {
            damage <<= 1;
        }
    }

    DamageDebris(ob, damage, damagetype);

    if ((ob->flags & FL_NODAMAGE) == 0)
    {
        ob->hitpoints -= (short)damage;
    }

    if (ob->flags & FL_UNKILLABLE)
    {
        ob->hitpoints = lwlib_MAX(ob->hitpoints, 1);
    }

    if (ob->hitpoints<=0)
    {
        KillActor (ob,attacker,damagetype);
    }
    else if (ob->flags & FL_SUPPRESSPAIN)
    {
        // no action
    }
    else
    {
        if (! (ob->flags & FL_ATTACKMODE) )
            FirstSighting (ob);             // put into combat mode

        if (ob->reaction == react_usingmg42)
        {
            NewState (ob,ED_STATE(ob->enemy->mg42.waitState));
        }
        else
        {
            if (ob->hitpoints&1)
            {
                painState = ED_STATE(ob->enemy->painStateOdd);
            }
            else
            {
                painState = ED_STATE(ob->enemy->painStateEven);
            }
            if (painState != NULL)
            {
                NewState (ob, painState);
            }

            if (ob->hitpoints < ob->starthitpoints / 2 &&
                ob->enemy == ED_Enemy(ed, "en_boss"))
            {
                OB_SetEnemy(ob, ED_Enemy(ed, "en_uberskel"));
                NewState(ob, ED_STATE("s_bosstoskel1"));
                PlayActorVoice(ED_SND("HANS2UBERSKELSND"),ob);
            }
        }
    }

    return true;
}

/*
=============================================================================

                                CHECKSIGHT

=============================================================================
*/


/*
=====================
=
= CheckLine
=
= Returns true if a straight line between the player and ob is unobstructed
=
= WSJ: edited to check between specific coordinates
=
=====================
*/

boolean CheckCoordLine (fixed fx1, fixed fy1, fixed fx2, fixed fy2,
    boolean binarydoor)
{
    int         x1,y1,xt1,yt1,x2,y2,xt2,yt2;
    int         x,y;
    int         xdist,ydist,xstep,ystep;
    int         partial,delta;
    int32_t     ltemp;
    int         xfrac,yfrac,deltafrac;
    unsigned    value,intercept;


    x1 = fx1 >> UNSIGNEDSHIFT;            // 1/256 tile precision
    y1 = fy1 >> UNSIGNEDSHIFT;
    xt1 = x1 >> 8;
    yt1 = y1 >> 8;

    x2 = fx2 >> UNSIGNEDSHIFT;
    y2 = fy2 >> UNSIGNEDSHIFT;
    xt2 = x2 >> 8;
    yt2 = y2 >> 8;

    xdist = abs(xt2-xt1);

    if (xdist > 0)
    {
        if (xt2 > xt1)
        {
            partial = 256-(x1&0xff);
            xstep = 1;
        }
        else
        {
            partial = x1&0xff;
            xstep = -1;
        }

        deltafrac = abs(x2-x1);
        delta = y2-y1;
        ltemp = ((int32_t)delta<<8)/deltafrac;
        if (ltemp > 0x7fffl)
            ystep = 0x7fff;
        else if (ltemp < -0x7fffl)
            ystep = -0x7fff;
        else
            ystep = ltemp;
        yfrac = y1 + (((int32_t)ystep*partial) >>8);

        x = xt1+xstep;
        xt2 += xstep;
        do
        {
            y = yfrac>>8;
            yfrac += ystep;

            value = (unsigned)tilemap[x][y];
            x += xstep;

            if (!value)
                continue;

            if (value<128 || value>256)
                return false;

            //
            // see if the door is open enough
            //
            value &= ~0x80;
            intercept = yfrac-ystep/2;

            if (binarydoor)
            {
                if (doorposition[value] == 0)
                {
                    return false;
                }
            }
            else
            {
                if (intercept>doorposition[value])
                {
                    return false;
                }
            }

        } while (x != xt2);
    }

    ydist = abs(yt2-yt1);

    if (ydist > 0)
    {
        if (yt2 > yt1)
        {
            partial = 256-(y1&0xff);
            ystep = 1;
        }
        else
        {
            partial = y1&0xff;
            ystep = -1;
        }

        deltafrac = abs(y2-y1);
        delta = x2-x1;
        ltemp = ((int32_t)delta<<8)/deltafrac;
        if (ltemp > 0x7fffl)
            xstep = 0x7fff;
        else if (ltemp < -0x7fffl)
            xstep = -0x7fff;
        else
            xstep = ltemp;
        xfrac = x1 + (((int32_t)xstep*partial) >>8);

        y = yt1 + ystep;
        yt2 += ystep;
        do
        {
            x = xfrac>>8;
            xfrac += xstep;

            value = (unsigned)tilemap[x][y];
            y += ystep;

            if (!value)
                continue;

            if (value<128 || value>256)
                return false;

            //
            // see if the door is open enough
            //
            value &= ~0x80;
            intercept = xfrac-xstep/2;

            if (binarydoor)
            {
                if (doorposition[value] == 0)
                {
                    return false;
                }
            }
            else
            {
                if (intercept>doorposition[value])
                {
                    return false;
                }
            }
        } while (y != yt2);
    }

    return true;
}

boolean CheckLine (objtype *ob)
{
    return CheckCoordLine (OB_GetX(ob),OB_GetY(ob),
        player->x,player->y,false);
}

boolean CheckTileLine (unsigned tx1, unsigned ty1, unsigned tx2, unsigned ty2,
    boolean binarydoor)
{
    int32_t x1,y1,x2,y2;

    x1 = ((int32_t)tx1<<TILESHIFT)+TILEGLOBAL/2;
    y1 = ((int32_t)ty1<<TILESHIFT)+TILEGLOBAL/2;
    x2 = ((int32_t)tx2<<TILESHIFT)+TILEGLOBAL/2;
    y2 = ((int32_t)ty2<<TILESHIFT)+TILEGLOBAL/2;

    return CheckCoordLine (x1,y1,x2,y2,binarydoor);
}

boolean CheckLinePlayerBody (objtype *ob)
{
    const lwlib::Ray3f ray = OB_GetViewRay(player);

    const lwlib::Vector3f u = vec_scale(vec_rot90(ray.p[1]), 0.25);

    const lwlib::Point3f p0 = ray.p[0];
    if (CheckCoordLine(ob->x, ob->y, fixedpt_rconst(p0.v[0]), 
        fixedpt_rconst(p0.v[1]), false))
    {
        return true;
    }

    const lwlib::Point3f p1 = vec_add(p0, u);
    if (CheckCoordLine(ob->x, ob->y, fixedpt_rconst(p1.v[0]), 
        fixedpt_rconst(p1.v[1]), false))
    {
        return true;
    }

    const lwlib::Point3f p2 = vec_sub(p0, u);
    if (CheckCoordLine(ob->x, ob->y, fixedpt_rconst(p2.v[0]), 
        fixedpt_rconst(p2.v[1]), false))
    {
        return true;
    }

    return false;
}

/*
================
=
= CheckSight
=
= Checks a straight line between player and current object
=
= If the sight is ok, check alertness and angle to see if they notice
=
= returns true if the player has been spoted
=
================
*/

#define MINSIGHT        0x18000l

boolean CheckObjSight (objtype *ob, objtype *check)
{
    int32_t deltax,deltay;

    //
    // don't bother tracing a line if the area isn't connected to the player's
    //
    if (!areabyplayer[ob->areanumber])
        return false;

    //
    // if the player is real close, sight is automatic
    //
    deltax = OB_GetX(check) - OB_GetX(ob);
    deltay = OB_GetY(check) - OB_GetY(ob);

    if (deltax > -MINSIGHT && deltax < MINSIGHT
        && deltay > -MINSIGHT && deltay < MINSIGHT)
        return true;

    //
    // see if they are looking in the right direction
    //
    switch (ob->dir)
    {
        case north:
            if (deltay > 0)
                return false;
            break;

        case east:
            if (deltax < 0)
                return false;
            break;

        case south:
            if (deltay < 0)
                return false;
            break;

        case west:
            if (deltax > 0)
                return false;
            break;

        // check diagonal moving guards fix

        case northwest:
            if (DEMOCOND_SDL && deltay > -deltax)
                return false;
            break;

        case northeast:
            if (DEMOCOND_SDL && deltay > deltax)
                return false;
            break;

        case southwest:
            if (DEMOCOND_SDL && deltax > deltay)
                return false;
            break;

        case southeast:
            if (DEMOCOND_SDL && -deltax > deltay)
                return false;
            break;

        default:
            break;
    }

    //
    // trace a line to check for blocking tiles (corners)
    //
    return CheckObjLine(ob,check);
}

boolean CheckSight (objtype *ob)
{
    objtype *check;

    if (ob->flags & FL_AMBUSH)
        return CheckObjSight(ob,player);

    // WSJ: check for dead bodies and alerted guards
    for (ITEROBJ(check, 0))
    {
        if (check != player && !(check->flags & FL_ATTACKMODE) && check->reaction != react_dead)
            continue;

        if (CheckObjSight(ob,check))
        {
            if (check == player)
                ob->reaction = react_sight;
            else if (check->reaction == react_dead)
                ob->reaction = react_findbody;

            return true;
        }
    }

    return false;
}

/*
===============
=
= FirstSighting
=
= Puts an actor into attack mode and possibly reverses the direction
= if the player is behind it
=
===============
*/

void FirstSighting (objtype *ob)
{
    if (ED_STATE(ob->enemy->chaseState))
    {
        NewState (ob,ED_STATE(ob->enemy->chaseState));
    }
    ob->speed = ob->enemy->chaseSpeed;        // go faster when chasing player

    if (ob->distance < 0)
        ob->distance = 0;       // ignore the door opening command

    ob->flags |= FL_ATTACKMODE|FL_FIRSTATTACK;

    if (CheckPlaysMusic(ob))
    {
        UpdateBossMusic(ob, true);
    }
}


/*
===============
=
= SightPlayer
=
= Called by actors that ARE NOT chasing the player.  If the player
= is detected (by sight, noise, or proximity), the actor is put into
= it's combat frame and true is returned.
=
= Incorporates a random reaction delay
=
= WSJ: added "state" for special animations. If NULL, then go to
= FirstSighting instead.
=
===============
*/

boolean SightPlayer (objtype *ob, statetype *state)
{
    if (ob->flags & FL_ATTACKMODE)
        Quit ("An actor in ATTACKMODE called SightPlayer!");

    if (invisiblemode)
        return false;

    if (player->tilex==ob->tilex && player->tiley==ob->tiley)
        return false;    // WSJ: don't get up if player is right on top of object

    if (ob->temp2)
    {
        //
        // count down reaction time
        //
        ob->temp2 -= (short) tics;
        if (ob->temp2 > 0)
            return false;
        ob->temp2 = 0;                                  // time to react
    }
    else
    {
        if (!CheckSight (ob))
        {
            if (ob->flags & FL_AMBUSH)
                return false;

            if (!AreaByActor(ob,lastalertedarea))
                return false;

            if (madenoise)
                ob->reaction = react_hear;
        }

        if (ob->flags & FL_AMBUSH)
        {
            ob->temp2 = 1;    // WSJ: instant reaction by default
        }
        else
        {
            if (ob->enemy->reactionTime.rndDivisor != 0)
            {
                ob->temp2 = 1 + ob->enemy->reactionTime.baseTics + 
                    (US_RndT() * 
                    ob->enemy->reactionTime.rndMultiplier / 
                    ob->enemy->reactionTime.rndDivisor);
            }
            else
            {
                ob->temp2 = 1;
            }
        }
        return false;
    }

    if (state == NULL)
    {
        A_AlertSound (ob);
        FirstSighting (ob);
    }
    else
    {
        NewState (ob,(statetype *)state);
    }

    return true;
}
